#include <src/analyseur/analyseur.h>
#include <src/donnees/instructions/instructions.h>
#include <src/donnees/expressions/expressions.h>
#define YYLTYPE Traceur::Donnees::Source::Position
#include <src/analyseur/analyseur.syn.h>
#include <src/donnees/global/cacheobjets.h>

#include <sys/stat.h>
#include <unistd.h>
#define FICHIER_STDIN "-"
#define PREFIXE_STANDARD "="
#define VARIABLE_FICHIER "TRACEUR"
#define UTILISATEUR_FICHIER "/.traceur"
#include <config.h>
#define SYSTEME_FICHIER SYSTEM_INCLUDE

using namespace Traceur::Analyseur;

int parseurparse(Traceur::Donnees::Instruction::BlocSP&, Traceur::Donnees::Global::CacheObjetGenerique& cache_objets);
int parseurlex_destroy();
extern FILE* parseurin;

Analyseur::~Analyseur()
{
	parseurlex_destroy();
}

Traceur::Donnees::Instruction::BlocSP Analyseur::analyse(ListeFichiers sources)
{
	AnalyseurSP analyseur = Analyseur::analyseur();
	analyseur->_parseur_source=sources;
	if(analyseur->_parseur_source.empty())
		analyseur->_parseur_source.push_back(FICHIER_STDIN);
	parseurin = analyseur->premier_fichier(parseurlloc);
	Traceur::Donnees::Instruction::BlocSP bloc;
	Traceur::Donnees::Global::CacheObjetGenerique cache_objets;
	if(parseurparse(bloc,cache_objets))
	{
		return Traceur::Donnees::Instruction::BlocSP();
	}
	analyseur.reset();
	return bloc;
}

bool Analyseur::teste_fichier(std::string fichier)
{
#ifdef TRACEUR_DEBUG
	std::cerr << "teste fichier " << fichier << std::endl;
#endif
	if(::access(fichier.c_str(),R_OK))
		return false;
#ifdef TRACEUR_DEBUG
	std::cerr << "access " << fichier << std::endl;
#endif
	struct stat buf;
	if(::stat(fichier.c_str(),&buf))
		return false;
#ifdef TRACEUR_DEBUG
	std::cerr << "stat " << fichier << std::endl;
#endif
	if(S_ISDIR(buf.st_mode))
		return false;
#ifdef TRACEUR_DEBUG
	std::cerr << "OK " << fichier << std::endl;
#endif
	return true;
}

FILE *Analyseur::ouvrir(std::string fichier)
{
	TRACEUR_TEST(not fichier.empty());
	if(fichier==FICHIER_STDIN)
		return stdin;
	if(fichier.substr(0,std::string(PREFIXE_STANDARD).size())==PREFIXE_STANDARD)
	{
		std::string suffixe = fichier.substr(std::string(PREFIXE_STANDARD).size());
		std::string prefixe;
		bool substitution = false;
		if(::getenv(VARIABLE_FICHIER))
		{
			prefixe = std::string(::getenv(VARIABLE_FICHIER));
			if(teste_fichier(prefixe+suffixe)) 
			{
				fichier=prefixe+suffixe;
				substitution = true;
#ifdef TRACEUR_DEBUG
				std::cerr << "var\n";
#endif
			}
		}
		if(not substitution and ::getenv("HOME"))
		{
			prefixe = std::string(::getenv("HOME"))+std::string(UTILISATEUR_FICHIER);
#ifdef TRACEUR_DEBUG
			std::cerr << prefixe+suffixe << std::endl;
#endif
			if(teste_fichier(prefixe+suffixe)) 
			{
				fichier=prefixe+suffixe;
				substitution = true;
#ifdef TRACEUR_DEBUG
				std::cerr << "usr\n";
#endif
			}
		}
		if(not substitution)
		{
			prefixe = std::string(SYSTEME_FICHIER);
			if(teste_fichier(prefixe+suffixe)) 
			{
				fichier=prefixe+suffixe;
				substitution = true;
#ifdef TRACEUR_DEBUG
				std::cerr << "sys\n";
#endif
			}
		}
	}

	if(not teste_fichier(fichier)) 
		throw Traceur::Exception::PasDeFichierSource(fichier);

	FILE *fic = ::fopen(fichier.c_str(),"r");
	return fic;
}

FILE *Analyseur::source_suivant(Traceur::Donnees::Source::Position& position)
{
	++_fichier_courant;
	if(_fichier_courant==_parseur_source.end())
		return 0;
	FILE *fd;
	fd=Analyseur::ouvrir(*_fichier_courant);
	if(not fd)
		throw Traceur::Exception::PasDeFichierSource(*_fichier_courant);
	position.nouveau_fichier(*_fichier_courant);
	return fd;
}
