#ifndef __ANALYSEUR_ANALYSEUR__
#define __ANALYSEUR_ANALYSEUR__
#include <cstdio>
#include <string>
#include <deque>

#include <src/global/global.h>
namespace Traceur { namespace Analyseur { DECL_SHARED_PTR(Analyseur); } }
#include <src/exceptions/exceptions.h>
#include <src/donnees/sources/position.h>
#include <src/donnees/instructions/bloc.h>

namespace Traceur
{
	namespace Analyseur
	{
		class Analyseur
		{
			public:
				typedef std::deque<std::string> ListeFichiers;
				static AnalyseurSP& analyseur() { static AnalyseurSP analyseur(new Analyseur()); return analyseur; };
				static Traceur::Donnees::Instruction::BlocSP analyse(ListeFichiers sources);
				FILE *source_suivant(Traceur::Donnees::Source::Position& position);
				~Analyseur();
			private:
				Analyseur() {};
				ListeFichiers _parseur_source;
				ListeFichiers::iterator _fichier_courant;
				FILE *premier_fichier(Traceur::Donnees::Source::Position& position)
				{
					_fichier_courant = _parseur_source.begin();
					position.nouveau_fichier(*_fichier_courant);
					return ouvrir(*_fichier_courant);
				};
				static FILE *ouvrir(std::string fichier);
				static bool teste_fichier(std::string fichier);
		};
	}
}

namespace Traceur
{
	namespace Exception
	{
		class PasDeFichierSource : public Compilation
		{
			public:
				PasDeFichierSource(std::string fichier_source)
				 :Compilation(Traceur::Donnees::Source::PositionSP(),std::string("Source file ")+fichier_source+std::string(" is invalid.")) {};
		};
	}
}
#endif
