#ifndef _DONNEES_ASSEMBLEUR_ARBREBITMAP_H_
#define _DONNEES_ASSEMBLEUR_ARBREBITMAP_H_

#include <iostream>

#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Assembleur {
	template<typename Couleur> class ArbreBitmap;
	template<typename Couleur> class ArbreBitmapNoeud;
	template<typename Couleur> class ArbreBitmapFeuille;
} } }

#include <src/donnees/assembleur/systemecoordonnees.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Assembleur
		{
			template<typename Couleur>
				class CacheBitmap;
			class Point;
			template<typename Couleur>
			class ArbreBitmap : public std::enable_shared_from_this<ArbreBitmap<Couleur> >
			{
				friend class ArbreBitmapFeuille<Couleur>;
				friend class ArbreBitmapNoeud<Couleur>;
				public:
					ArbreBitmap() = default;
					static SHARED_PTR(ArbreBitmap<Couleur>) instancie(Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, Couleur couleur_fond);
					virtual SHARED_PTR(CacheBitmap<Couleur>) bitmap(Point point, Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, Couleur couleur_fond, SHARED_PTR(ArbreBitmap<Couleur>)& arbre, bool instancie_si_necessaire = true) = 0;
					virtual void format(std::ostream& os) const = 0;
					virtual ~ArbreBitmap() {};
				private:
					virtual bool feuille() const = 0;
					virtual void copie(SHARED_PTR(ArbreBitmapFeuille<Couleur>) feuille, Coordonnee x, Coordonnee y) = 0;
				protected:
					virtual SHARED_PTR(ArbreBitmapFeuille<Couleur>) fusionne() = 0; // renvoie un SP faux s'il n'y a pas eu de fusion
					Coordonnee separation(Coordonnee min, Coordonnee max) { return (min+max)/2+1; }
			};

			template<typename Couleur>
			class ArbreBitmapFeuille : public ArbreBitmap<Couleur>
			{
				public:
					ArbreBitmapFeuille(Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, Couleur couleur_fond);
					virtual SHARED_PTR(CacheBitmap<Couleur>) bitmap(Point point, Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, Couleur couleur_fond, SHARED_PTR(ArbreBitmap<Couleur>)& arbre, bool instancie_si_necessaire = true);
					virtual SHARED_PTR(ArbreBitmapFeuille<Couleur>) fusionne() { return SHARED_PTR(ArbreBitmapFeuille<Couleur>)(); };
					virtual bool feuille() const { return true; };
					virtual void copie(SHARED_PTR(ArbreBitmapFeuille<Couleur>) feuille, Coordonnee x, Coordonnee y);
					void trace(Coordonnee x, Coordonnee y, Couleur couleur, Traceur::Donnees::Expression::Entier::TypeInterne alpha);
					void trace(Coordonnee x, Coordonnee y, Couleur couleur);
					void remplace(Coordonnee x, Coordonnee y, Couleur couleur);
					Couleur couleur(Coordonnee x, Coordonnee y) const;
					virtual void format(std::ostream& os) const;
					virtual ~ArbreBitmapFeuille() {};
				private:
					typedef std::vector<Couleur> Colonne;
					typedef std::vector<Colonne> Pixels;
					Pixels _pixels;
			};

			template<typename Couleur>
			class ArbreBitmapNoeud : public ArbreBitmap<Couleur>
			{
				public:
					ArbreBitmapNoeud(Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y);
					virtual SHARED_PTR(CacheBitmap<Couleur>) bitmap(Point point, Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, Couleur couleur_fond, SHARED_PTR(ArbreBitmap<Couleur>)& arbre, bool instancie_si_necessaire = true);
					virtual SHARED_PTR(ArbreBitmapFeuille<Couleur>) fusionne();
					virtual bool feuille() const { return false; };
					virtual void copie(SHARED_PTR(ArbreBitmapFeuille<Couleur>) feuille, Coordonnee x, Coordonnee y) { throw; };
					virtual void format(std::ostream& os) const;
					virtual ~ArbreBitmapNoeud() {};
				private:
					// la bitmap est [_min_x,_max_x) et pareil sur y
					Coordonnee _min_x;
					Coordonnee _min_y;
					Coordonnee _max_x;
					Coordonnee _max_y;
					SHARED_PTR(ArbreBitmap<Couleur>) _sud_est;
					SHARED_PTR(ArbreBitmap<Couleur>) _sud_ouest;
					SHARED_PTR(ArbreBitmap<Couleur>) _nord_est;
					SHARED_PTR(ArbreBitmap<Couleur>) _nord_ouest;
			};
		}
	}
}
#include <src/donnees/assembleur/point.h>
#include <src/donnees/assembleur/cachebitmap.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Assembleur
		{
	template<typename Couleur>
ArbreBitmapNoeud<Couleur>::ArbreBitmapNoeud(Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y)
	:_min_x(min_x), _min_y(min_y), _max_x(max_x), _max_y(max_y)
{
}

	template<typename Couleur>
ArbreBitmapFeuille<Couleur>::ArbreBitmapFeuille(Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, Couleur couleur_fond)
{
	Colonne colonne(max_y-min_y,couleur_fond);
	_pixels = Pixels(max_x-min_x,colonne);
}

	template<typename Couleur>
void ArbreBitmapNoeud<Couleur>::format(std::ostream& os) const
{
	os << "// noeud [" << _min_x << "," << _max_x << ")x[" << _min_y << "," << _max_y << ")" << std::endl;
	os << "// SO" << std::endl;
	if(_sud_ouest)
	{
		_sud_ouest->format(os);
	}
	os << "// SE" << std::endl;
	if(_sud_est)
	{
		_sud_est->format(os);
	}
	os << "// NO" << std::endl;
	if(_nord_ouest)
	{
		_nord_ouest->format(os);
	}
	os << "// NE" << std::endl;
	if(_nord_est)
	{
		_nord_est->format(os);
	}
	os << "// fin noeud" << std::endl;
}

	template<typename Couleur>
void ArbreBitmapFeuille<Couleur>::format(std::ostream& os) const
{
	os << "// feuille " << _pixels.size() << ", " << _pixels.begin()->size() << std::endl;
	for(auto itx=_pixels.begin() ; itx!=_pixels.end() ; ++itx)
	{
		os << "//" ;
		for(auto ity=itx->begin() ; ity!=itx->end() ; ++ity)
		{
			Couleur couleur=*ity;
			couleur.format(os);
		}
		os << std::endl;
	}
}

	template<typename Couleur>
SHARED_PTR(ArbreBitmap<Couleur>) ArbreBitmap<Couleur>::instancie(Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, Couleur couleur_fond)
{
#ifdef TRACEUR_DEBUG
	std::cerr << "INSTANCIE " << min_x << " " << min_y << " " << max_x << " " << max_y << ": ";
#endif
	if((max_x-min_x)*(max_y-min_y)<64)
	{
#ifdef TRACEUR_DEBUG
		std::cerr << "feuille" << std::endl;
#endif
		SHARED_PTR(ArbreBitmap<Couleur>) feuille(new ArbreBitmapFeuille<Couleur>(min_x,min_y,max_x,max_y,couleur_fond));
		return feuille;
	}
#ifdef TRACEUR_DEBUG
	std::cerr << "noeud" << std::endl;
#endif
	SHARED_PTR(ArbreBitmap<Couleur>) noeud(new ArbreBitmapNoeud<Couleur>(min_x,min_y,max_x,max_y));
	return noeud;
}

	template<typename Couleur>
SHARED_PTR(ArbreBitmapFeuille<Couleur>) ArbreBitmapNoeud<Couleur>::fusionne()
{
	if(not (_sud_ouest and _sud_est and _nord_ouest and _nord_est))
		return SHARED_PTR(ArbreBitmapFeuille<Couleur>)();
	if(not _sud_ouest->feuille())
		return SHARED_PTR(ArbreBitmapFeuille<Couleur>)();
	if(not _sud_est->feuille())
		return SHARED_PTR(ArbreBitmapFeuille<Couleur>)();
	if(not _nord_ouest->feuille())
		return SHARED_PTR(ArbreBitmapFeuille<Couleur>)();
	if(not _nord_est->feuille())
		return SHARED_PTR(ArbreBitmapFeuille<Couleur>)();
#ifdef TRACEUR_DEBUG
	std::cerr << "FUSION: " << _min_x << " " << _min_y << " " << _max_x << " " << _max_y << std::endl;
#endif
	SHARED_PTR(ArbreBitmapFeuille<Couleur>) feuille(new ArbreBitmapFeuille<Couleur>(_min_x,_min_y,_max_x,_max_y,Couleur()));
	_sud_ouest->copie(feuille,0,0);
	_sud_est->copie(feuille,this->separation(_min_x,_max_x)-_min_x,0);
	_nord_ouest->copie(feuille,0,this->separation(_min_y,_max_y)-_min_y);
	_nord_est->copie(feuille,this->separation(_min_x,_max_x)-_min_x,this->separation(_min_y,_max_y)-_min_y);
	return feuille;
}

	template<typename Couleur>
void ArbreBitmapFeuille<Couleur>::copie(SHARED_PTR(ArbreBitmapFeuille<Couleur>) feuille, Coordonnee x, Coordonnee y)
{
#ifdef TRACEUR_DEBUG
	std::cerr << "COPIE " << x <<" " << y << ": " << _pixels.size() << " " << _pixels.begin()->size() << 
		" => " << feuille->_pixels.size() << " " << feuille->_pixels.begin()->size() << std::endl;
#endif
	Coordonnee ax=x;
	for(auto itx=_pixels.begin() ; itx!=_pixels.end() ; ++itx, ++ax)
	{
		Coordonnee ay=y;
		for(auto ity=itx->begin() ; ity!=itx->end() ; ++ity, ++ay)
		{
#ifdef TRACEUR_DEBUG
			std::cerr << ax << " " << ay << std::endl;
#endif
			feuille->_pixels[ax][ay]=*ity;
		}
	}
}

	template<typename Couleur>
SHARED_PTR(CacheBitmap<Couleur>) ArbreBitmapFeuille<Couleur>::bitmap(Point point, Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, Couleur couleur_fond, SHARED_PTR(ArbreBitmap<Couleur>)& arbre, bool instancie_si_necessaire)
{
	SHARED_PTR(ArbreBitmapFeuille<Couleur>) feuille = std::static_pointer_cast<ArbreBitmapFeuille>(this->shared_from_this());
	SHARED_PTR(CacheBitmap<Couleur>) cache(new CacheBitmap<Couleur>(min_x,min_y,max_x,max_y,feuille));
	return cache;
}

	template<typename Couleur>
SHARED_PTR(CacheBitmap<Couleur>) ArbreBitmapNoeud<Couleur>::bitmap(Point point, Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, Couleur couleur_fond, SHARED_PTR(ArbreBitmap<Couleur>)& arbre, bool instancie_si_necessaire)
{
	SHARED_PTR(ArbreBitmap<Couleur>) scope;
	Coordonnee x=this->separation(_min_x,_max_x);
	Coordonnee y=this->separation(_min_y,_max_y);
	SHARED_PTR(CacheBitmap<Couleur>) cache;
	if((point._x<x) and (point._y<y))
	{
		if(not _sud_ouest)
		{
			if(not instancie_si_necessaire)
				return cache;
			_sud_ouest = this->instancie(_min_x,_min_y,x,y,couleur_fond);
		}
		cache = _sud_ouest->bitmap(point,_min_x,_min_y,x,y,couleur_fond,_sud_ouest);
	}
	else if((point._x>=x) and (point._y<y))
	{
		if(not _sud_est)
		{
			if(not instancie_si_necessaire)
				return cache;
			_sud_est = this->instancie(x,_min_y,_max_x,y,couleur_fond);
		}
		cache = _sud_est->bitmap(point,x,_min_y,_max_x,y,couleur_fond,_sud_est);
	}
	else if((point._x<x) and (point._y>=y))
	{
		if(not _nord_ouest)
		{
			if(not instancie_si_necessaire)
				return cache;
			_nord_ouest = this->instancie(_min_x,y,x,_max_y,couleur_fond);
		}
		cache = _nord_ouest->bitmap(point,_min_x,y,x,_max_y,couleur_fond,_nord_ouest);
	}
	else if((point._x>=x) and (point._y>=y))
	{
		if(not _nord_est)
		{
			if(not instancie_si_necessaire)
				return cache;
			_nord_est = this->instancie(x,y,_max_x,_max_y,couleur_fond);
		}
		cache = _nord_est->bitmap(point,x,y,_max_x,_max_y,couleur_fond,_nord_est);
	}
	TRACEUR_TEST(cache);
	SHARED_PTR(ArbreBitmapFeuille<Couleur>) fusion = fusionne();
	if(fusion)
	{
		cache = SHARED_PTR(CacheBitmap<Couleur>)(new CacheBitmap<Couleur>(_min_x,_min_y,_max_x,_max_y,fusion));
		scope= this->shared_from_this(); // pour maintenir this jusqu'a la fin de la methode
		arbre=fusion;
	}
	return cache;
}

	template<typename Couleur>
void ArbreBitmapFeuille<Couleur>::trace(Coordonnee x, Coordonnee y, Couleur couleur, Traceur::Donnees::Expression::Entier::TypeInterne alpha)
{
	_pixels[x][y].integre(couleur,alpha);
}

	template<typename Couleur>
void ArbreBitmapFeuille<Couleur>::trace(Coordonnee x, Coordonnee y, Couleur couleur)
{
	_pixels[x][y].integre(couleur);
}

	template<typename Couleur>
void ArbreBitmapFeuille<Couleur>::remplace(Coordonnee x, Coordonnee y, Couleur couleur)
{
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE ; _pixels[x][y].format(std::cerr); std::cerr << " <= "; couleur.format(std::cerr); std::cerr << std::endl;
#endif
	_pixels[x][y] = couleur;
}

	template<typename Couleur>
Couleur ArbreBitmapFeuille<Couleur>::couleur(Coordonnee x, Coordonnee y) const
{
	return _pixels[x][y];
}
		}
	}
}

#endif
