#include <src/donnees/assembleur/assembleur.h>
#include <src/donnees/assembleur/drawpoint.h>
#include <src/donnees/assembleur/drawat.h>
#include <src/donnees/assembleur/conv.h>
#include <src/donnees/assembleur/with.h>
#include <src/donnees/assembleur/image.h>
#include <src/donnees/fonctions/declarations.h>
#include <src/donnees/expressions/expressionimageevaluee.h>
#include <src/donnees/assembleur/write.h>
#include <src/donnees/assembleur/read.h>
#include <iostream>
using namespace Traceur::Donnees::Assembleur;

void Assembleur::debut_bloc(Traceur::Donnees::Fonction::DeclarationsSP& declarations)
{
}

void Assembleur::bloc_interrompu()
{
}

void Assembleur::fin_bloc()
{
} 

void Assembleur::disp()
{
}

void Assembleur::disp(Traceur::Donnees::Expression::ExpressionSP& expression, Traceur::Donnees::Expression::ExpressionSP& valeur)
{
	std::cout << "disp ";
	expression->format(std::cout);
	std::cout << ": ";
	valeur->format(std::cout);
	std::cout << std::endl;
}

void Assembleur::draw(Traceur::Donnees::Expression::ExpressionSP& valeur)
{
	if(_pile.empty())
		return;

	InstructionSP instruction = valeur->draw(*this);
	if(instruction)
	{
		ajoute(instruction);
	}
}

void Assembleur::drawat(Traceur::Donnees::Expression::ExpressionSP& image, Traceur::Donnees::Expression::ExpressionSP& point, Traceur::Donnees::Expression::ExpressionSP& alpha)
{
	if(_pile.empty())
		return;
	Traceur::Donnees::Expression::ImageEvalueeSP expr_image = std::dynamic_pointer_cast<Traceur::Donnees::Expression::ImageEvaluee>(image);
	Image::ImageSP image_evaluee = expr_image->_image;
	if(image_evaluee->type()!=_pile.top()->type())
		throw Traceur::Donnees::Instruction::Valeur::type_image_incorrect();
	Traceur::Donnees::Expression::PointSP expr_point = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Point>(point);
	Point point_evalue(expr_point,_pile.top()->_systeme_coordonnees);
	Traceur::Donnees::Expression::NombreSP expr_alpha = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(alpha);
	Traceur::Donnees::Expression::Entier::TypeInterne alpha_evalue = expr_alpha->entier()._entier ;
	DrawAtSP drawat(new DrawAt(image_evaluee,point_evalue,alpha_evalue));
	ajoute(drawat);
}

void Assembleur::conv(Traceur::Donnees::Expression::ExpressionSP& noyau)
{
	if(_pile.empty())
		return;
	Traceur::Donnees::Expression::ImageEvalueeSP expr_noyau = std::dynamic_pointer_cast<Traceur::Donnees::Expression::ImageEvaluee>(noyau);
	Image::ImageSP noyau_evaluee = expr_noyau->_image;
	if((_pile.top()->type()!=Image::Image::Type::REELLE) and (_pile.top()->type()!=Image::Image::Type::NOYAU))
		throw Traceur::Donnees::Instruction::Valeur::type_image_incorrect();
	ConvSP conv(new Conv(noyau_evaluee));
	ajoute(conv);
}

void Assembleur::with(Traceur::Donnees::Expression::ExpressionSP& valeur)
{
	if(_pile.empty())
		return;
	Traceur::Donnees::Expression::CouleurRVBASP expr_rvba = std::dynamic_pointer_cast<Traceur::Donnees::Expression::CouleurRVBA>(valeur);
	if(expr_rvba)
	{
		Couleur::Couleur couleur(expr_rvba);
		with(couleur);
		return;
	}
	Traceur::Donnees::Expression::CouleurHSVASP expr_hsva = std::dynamic_pointer_cast<Traceur::Donnees::Expression::CouleurHSVA>(valeur);
	if(expr_hsva)
	{
		Traceur::Donnees::Expression::CouleurRVBASP expr_rvba(new Traceur::Donnees::Expression::CouleurRVBA(*expr_hsva));
		Couleur::Couleur couleur(expr_rvba);
		with(couleur);
		return;
	}
}

Couleur::Couleur Assembleur::couleur_courante()
{
	if(_pile.empty())
		return Couleur::Couleur();
	return _pile.top()->_couleur_courante;
}

void Assembleur::with(Couleur::Couleur couleur)
{
	if(_pile.empty())
		return;
	_pile.top()->_couleur_courante=couleur;
	WithSP with(new With(couleur));
	_pile.top()->ajoute(with);
}

void Assembleur::ajoute(const InstructionSP& instruction)
{
	if(_pile.empty())
	{
		_code.push_back(instruction);
		return;
	}
	_pile.top()->ajoute(instruction);
}

void Assembleur::write(Traceur::Donnees::Expression::ExpressionSP& image, Traceur::Donnees::Expression::ExpressionSP& nom_fichier)
{
	Traceur::Donnees::Expression::ImageEvalueeSP expr_image = std::dynamic_pointer_cast<Traceur::Donnees::Expression::ImageEvaluee>(image);
	Image::ImageSP image_evaluee = expr_image->_image;
	if(image_evaluee->type()!=Image::Image::Type::REELLE)
		throw Traceur::Donnees::Instruction::Valeur::type_image_incorrect();
	Traceur::Donnees::Expression::ChaineSP expr_nom_fichier = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Chaine>(nom_fichier);
	NomFichier nom_fichier_evalue(expr_nom_fichier);
	WriteSP write(new Write(image_evaluee,nom_fichier_evalue));
	write->teste();
	_code.push_back(write);
}

void Assembleur::read(Traceur::Donnees::Expression::ExpressionSP& nom_fichier)
{
	if(not _pile.empty() and _pile.top()->type()!=Image::Image::Type::REELLE)
		throw Traceur::Donnees::Instruction::Valeur::type_image_incorrect();
	Traceur::Donnees::Expression::ChaineSP expr_nom_fichier = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Chaine>(nom_fichier);
	NomFichier nom_fichier_evalue(expr_nom_fichier);
	ReadSP read(new Read(nom_fichier_evalue));
	read->teste();
	ajoute(read);
}

void Assembleur::dpt()
{
	if(_pile.empty())
		return;
	if(_pile.top()->type()!=Image::Image::Type::REELLE)
		throw Traceur::Donnees::Instruction::Valeur::type_image_incorrect();
	_pile.top()->trace();
	Image::ReelleSP reelle = std::static_pointer_cast<Image::Reelle>(_pile.top());
	Image::ImageSP frequencielle(new Image::Frequencielle(*reelle));
	_pile.pop();
	_pile.push(frequencielle);
}

void Assembleur::idpt()
{
	if(_pile.empty())
		return;
	if(_pile.top()->type()!=Image::Image::Type::FREQUENCIELLE)
		throw Traceur::Donnees::Instruction::Valeur::type_image_incorrect();
	_pile.top()->trace();
	Image::FrequencielleSP frequencielle = std::static_pointer_cast<Image::Frequencielle>(_pile.top());
	Image::ImageSP reelle(new Image::Reelle(*frequencielle));
	_pile.pop();
	_pile.push(reelle);
}

void Assembleur::image(Traceur::Donnees::Expression::ExpressionSP& taille_x, Traceur::Donnees::Expression::ExpressionSP& taille_y,
		Traceur::Donnees::Expression::SystemeCoordonneesSP& coordonnees, Traceur::Donnees::Expression::ExpressionSP& couleur_fond)
{
	Traceur::Donnees::Expression::CouleurRVBASP rvba = std::dynamic_pointer_cast<Traceur::Donnees::Expression::CouleurRVBA>(couleur_fond);
	if(not rvba)
	{
		Traceur::Donnees::Expression::CouleurHSVASP hsva = std::dynamic_pointer_cast<Traceur::Donnees::Expression::CouleurHSVA>(couleur_fond);
		TRACEUR_TEST(hsva);
		rvba = Traceur::Donnees::Expression::CouleurRVBASP(new Traceur::Donnees::Expression::CouleurRVBA(*hsva));
	}
	Image::ImageSP image(new Image::Reelle(taille_x,taille_y,coordonnees,rvba));
	_pile.push(image);
}

void Assembleur::noyau(Traceur::Donnees::Expression::ExpressionSP& taille_x, Traceur::Donnees::Expression::ExpressionSP& taille_y,
						Traceur::Donnees::Expression::SystemeCoordonneesSP& coordonnees)
{
	Image::ImageSP image(new Image::Noyau(taille_x,taille_y,coordonnees));
	_pile.push(image);
}

Image::ImageSP Assembleur::image()
{
	TRACEUR_TEST(not _pile.empty());
	Image::ImageSP image = _pile.top();
	_pile.pop();
	image->trace();
	return image;
}

Image::ImageSP Assembleur::trace_image()
{
	if(_pile.empty())
		throw Traceur::Donnees::Instruction::Valeur::coordonnee_hors_image();
	Image::ImageSP image = _pile.top();
	image->trace();
	return image;
}

SystemeCoordonnees Assembleur::coordonnees(Traceur::Donnees::Instruction::ValeurSP erreur)
{
	if(_pile.empty())
		throw erreur;
	return _pile.top()->_systeme_coordonnees;
}

void Assembleur::execute()
{
	std::for_each(_code.begin(),_code.end(),bind(&Instruction::execute,std::placeholders::_1));
}
