#ifndef _DONNEES_ASSEMBLEUR_ASSEMBLEUR_H_
#define _DONNEES_ASSEMBLEUR_ASSEMBLEUR_H_

#include <deque>
#include <stack>
#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Assembleur { DECL_SHARED_PTR(Assembleur); } } }

#include <src/donnees/expressions/expression.h>
#include <src/donnees/expressions/expressionimage.h>
#include <src/donnees/assembleur/couleur.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Fonction { DECL_SHARED_PTR(Declarations); }
		namespace Instruction { class With; }
		namespace Expression { class Point; class Cercle; class Ligne; class Solide; class Gomme; class Remplissage; class Triangle; class Rectangle; class Ellipse; }
		namespace Assembleur
		{
			DECL_SHARED_PTR(Instruction);
			namespace Image { DECL_SHARED_PTR(Image); }
			class SystemeCoordonnees;
			class Assembleur
			{
				friend class Traceur::Donnees::Instruction::With;
				friend class Traceur::Donnees::Expression::Point;
				friend class Traceur::Donnees::Expression::Cercle;
				friend class Traceur::Donnees::Expression::Ligne;
				friend class Traceur::Donnees::Expression::Solide;
				friend class Traceur::Donnees::Expression::Gomme;
				friend class Traceur::Donnees::Expression::Remplissage;
				friend class Traceur::Donnees::Expression::Triangle;
				friend class Traceur::Donnees::Expression::Rectangle;
				friend class Traceur::Donnees::Expression::Ellipse;
				public:
					typedef std::deque<InstructionSP> ListeInstructions;
					typedef std::stack<Image::ImageSP> PileImages;
					Assembleur() = default;
					~Assembleur() {};
					
					void debut_bloc(Traceur::Donnees::Fonction::DeclarationsSP& declarations);
					void bloc_interrompu();
					void fin_bloc();

					void disp();
					void disp(Traceur::Donnees::Expression::ExpressionSP& expression, Traceur::Donnees::Expression::ExpressionSP& valeur);
					void draw(Traceur::Donnees::Expression::ExpressionSP& valeur);
					void drawat(Traceur::Donnees::Expression::ExpressionSP& image, Traceur::Donnees::Expression::ExpressionSP& point, Traceur::Donnees::Expression::ExpressionSP& alpha);
					void conv(Traceur::Donnees::Expression::ExpressionSP& noyau);
					void dpt();
					void idpt();
					void with(Traceur::Donnees::Expression::ExpressionSP& valeur);
					void write(Traceur::Donnees::Expression::ExpressionSP& image, Traceur::Donnees::Expression::ExpressionSP& nom_fichier);
					void read(Traceur::Donnees::Expression::ExpressionSP& nom_fichier);
					void image(Traceur::Donnees::Expression::ExpressionSP& taille_x, Traceur::Donnees::Expression::ExpressionSP& taille_y,
						Traceur::Donnees::Expression::SystemeCoordonneesSP& coordonnees, Traceur::Donnees::Expression::ExpressionSP& couleur_fond);
					void noyau(Traceur::Donnees::Expression::ExpressionSP& taille_x, Traceur::Donnees::Expression::ExpressionSP& taille_y,
						Traceur::Donnees::Expression::SystemeCoordonneesSP& coordonnees);
					Image::ImageSP image();
					Image::ImageSP trace_image();
					SystemeCoordonnees coordonnees(Traceur::Donnees::Instruction::ValeurSP erreur);
					void execute();
				private:
					Couleur::Couleur couleur_courante();
					void with(Couleur::Couleur couleur);
					void ajoute(const InstructionSP& instruction);
					ListeInstructions _code;
					PileImages _pile;
			};
		}
	}
}
#endif
