#ifndef _DONNEES_ASSEMBLEUR_BITMAP_H_
#define _DONNEES_ASSEMBLEUR_BITMAP_H_

#include <iostream>
#include <vector>

#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Assembleur {
	template<typename Couleur> class Bitmap;
} } }

#include <src/donnees/assembleur/systemecoordonnees.h>
#include <src/donnees/assembleur/cachebitmap.h>
#include <src/donnees/assembleur/arbrebitmap.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Assembleur
		{
			class Point;
			namespace Image { class Reelle; class Noyau; }
			template<typename Couleur>
			class Bitmap
			{
				friend class Image::Reelle;
				friend class Image::Noyau;
				public:
					Bitmap(Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, Couleur couleur_fond);
					void trace(Point point, Couleur couleur, Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, Couleur couleur_fond);
					void remplace(Point point, Couleur couleur, Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, Couleur couleur_fond);
					void trace(Point point, Couleur couleur, Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, Couleur couleur_fond, Traceur::Donnees::Expression::Entier::TypeInterne alpha);
					Couleur couleur(Point point, Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, Couleur couleur_fond);
					virtual void format(std::ostream& os) const;
					virtual ~Bitmap() {};
				private:
					Bitmap() {};
					SHARED_PTR(CacheBitmap<Couleur>) _cache;
					SHARED_PTR(ArbreBitmap<Couleur>) _arbre;
			};
		}
	}
}

#include <src/donnees/assembleur/point.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Assembleur
		{
template <typename Couleur>
Bitmap<Couleur>::Bitmap(Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, Couleur couleur_fond)
{
	_arbre = ArbreBitmap<Couleur>::instancie(min_x,min_y,max_x,max_y,couleur_fond);
}

template <typename Couleur>
void Bitmap<Couleur>::trace(Point point, Couleur couleur, Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, Couleur couleur_fond, Traceur::Donnees::Expression::Entier::TypeInterne alpha)
{
	if((not _cache) or (_cache and not _cache->cible(point)))
		_cache = _arbre->bitmap(point,min_x,min_y,max_x,max_y,couleur_fond,_arbre);
	_cache->trace(point,couleur,alpha);
}

template <typename Couleur>
void Bitmap<Couleur>::trace(Point point, Couleur couleur, Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, Couleur couleur_fond)
{
	if((not _cache) or (_cache and not _cache->cible(point)))
		_cache = _arbre->bitmap(point,min_x,min_y,max_x,max_y,couleur_fond,_arbre);
	_cache->trace(point,couleur);
}

template <typename Couleur>
void Bitmap<Couleur>::remplace(Point point, Couleur couleur, Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, Couleur couleur_fond)
{
	if((not _cache) or (_cache and not _cache->cible(point)))
		_cache = _arbre->bitmap(point,min_x,min_y,max_x,max_y,couleur_fond,_arbre);
	_cache->remplace(point,couleur);
}

template <typename Couleur>
Couleur Bitmap<Couleur>::couleur(Point point, Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, Couleur couleur_fond)
{
	if(_cache and _cache->cible(point))
		return _cache->couleur(point);
	SHARED_PTR(CacheBitmap<Couleur>) cache = _arbre->bitmap(point,min_x,min_y,max_x,max_y,couleur_fond,_arbre,false);
	if(cache)
		return cache->couleur(point);
	return couleur_fond;
}

template <typename Couleur>
void Bitmap<Couleur>::format(std::ostream& os) const
{
	os << "// bitmap:" << std::endl;
	if(_cache)
	{
		_cache->format(os);
	}
	_arbre->format(os);
}
		}
	}
}
#endif
