#ifndef _DONNEES_ASSEMBLEUR_CACHEBITMAP_H_
#define _DONNEES_ASSEMBLEUR_CACHEBITMAP_H_

#include <iostream>
#include <vector>

#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Assembleur {
	template<typename Couleur> class CacheBitmap;
} } }

#include <src/donnees/assembleur/systemecoordonnees.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Assembleur
		{
			template <typename Couleur>
				class ArbreBitmapFeuille;
			class Point;
			template<typename Couleur>
			class CacheBitmap
			{
				public:
					typedef unsigned char Canal;
					CacheBitmap(Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, SHARED_PTR(ArbreBitmapFeuille<Couleur>) bitmap);
					bool cible(Point point);
					void trace(Point point, Couleur couleur);
					void remplace(Point point, Couleur couleur);
					void trace(Point point, Couleur couleur, Traceur::Donnees::Expression::Entier::TypeInterne alpha);
					Couleur couleur(Point point);
					virtual void format(std::ostream& os) const;
					virtual ~CacheBitmap() {};
				private:
					Coordonnee _min_x;
					Coordonnee _min_y;
					Coordonnee _max_x;
					Coordonnee _max_y;
					SHARED_PTR(ArbreBitmapFeuille<Couleur>) _bitmap;
			};
		}
	}
}

#include <src/donnees/assembleur/point.h>
#include <src/donnees/assembleur/couleur.h>
#include <src/donnees/assembleur/arbrebitmap.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Assembleur
		{
template <typename Couleur>
CacheBitmap<Couleur>::CacheBitmap(Coordonnee min_x, Coordonnee min_y, Coordonnee max_x, Coordonnee max_y, SHARED_PTR(ArbreBitmapFeuille<Couleur>) bitmap)
	:_min_x(min_x), _min_y(min_y), _max_x(max_x), _max_y(max_y), _bitmap(bitmap)
{
}

template <typename Couleur>
bool CacheBitmap<Couleur>::cible(Point point)
{
	return (point._x>=_min_x) and (point._x<_max_x) and (point._y>=_min_y) and (point._y<_max_y);
}

template <typename Couleur>
void CacheBitmap<Couleur>::trace(Point point, Couleur couleur, Traceur::Donnees::Expression::Entier::TypeInterne alpha)
{
	_bitmap->trace(point._x-_min_x,point._y-_min_y,couleur,alpha);
}

template <typename Couleur>
void CacheBitmap<Couleur>::trace(Point point, Couleur couleur)
{
	_bitmap->trace(point._x-_min_x,point._y-_min_y,couleur);	
}

template <typename Couleur>
void CacheBitmap<Couleur>::remplace(Point point, Couleur couleur)
{
	_bitmap->remplace(point._x-_min_x,point._y-_min_y,couleur);
}

template <typename Couleur>
Couleur CacheBitmap<Couleur>::couleur(Point point)
{
	return _bitmap->couleur(point._x-_min_x,point._y-_min_y);	
}

template <typename Couleur>
void CacheBitmap<Couleur>::format(std::ostream& os) const
{
	os << "// cachebitmap (" << _min_x << "," << _min_y << ") (" << _max_x << "," << _max_y << ")" << std::endl;
}
		}
	}
}
#endif
