#include <src/donnees/assembleur/couleur.h>
#include <cmath>
using namespace Traceur::Donnees::Assembleur::Couleur;

Reelle::Reelle(const Traceur::Donnees::Expression::CouleurRVBASP& couleur)
{
	Traceur::Donnees::Expression::NombreSP rouge = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(couleur->rouge());
	TRACEUR_TEST(rouge);
	_rouge = static_cast<Canal>(rouge->entier_borne(*Traceur::Donnees::Expression::Entier::zero(),*Traceur::Donnees::Expression::Entier::max_canal_couleur())._entier);
	Traceur::Donnees::Expression::NombreSP vert = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(couleur->vert());
	TRACEUR_TEST(vert);
	_vert = static_cast<Canal>(vert->entier_borne(*Traceur::Donnees::Expression::Entier::zero(),*Traceur::Donnees::Expression::Entier::max_canal_couleur())._entier);
	Traceur::Donnees::Expression::NombreSP bleu = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(couleur->bleu());
	TRACEUR_TEST(bleu);
	_bleu = static_cast<Canal>(bleu->entier_borne(*Traceur::Donnees::Expression::Entier::zero(),*Traceur::Donnees::Expression::Entier::max_canal_couleur())._entier);
	Traceur::Donnees::Expression::NombreSP alpha = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(couleur->alpha());
	TRACEUR_TEST(alpha);
	_alpha = static_cast<Canal>(alpha->entier_borne(*Traceur::Donnees::Expression::Entier::zero(),*Traceur::Donnees::Expression::Entier::max_canal_couleur())._entier);
}

Reelle::Reelle(const Couleur& couleur)
{
	Traceur::Donnees::Expression::Entier rouge(couleur._rouge);
	_rouge = static_cast<Canal>(rouge.entier_borne(*Traceur::Donnees::Expression::Entier::zero(),*Traceur::Donnees::Expression::Entier::max_canal_couleur())._entier);
	Traceur::Donnees::Expression::Entier vert (couleur._vert );
	_vert  = static_cast<Canal>(vert .entier_borne(*Traceur::Donnees::Expression::Entier::zero(),*Traceur::Donnees::Expression::Entier::max_canal_couleur())._entier);
	Traceur::Donnees::Expression::Entier bleu (couleur._bleu );
	_bleu  = static_cast<Canal>(bleu .entier_borne(*Traceur::Donnees::Expression::Entier::zero(),*Traceur::Donnees::Expression::Entier::max_canal_couleur())._entier);
	Traceur::Donnees::Expression::Entier alpha(couleur._alpha);
	_alpha = static_cast<Canal>(alpha.entier_borne(*Traceur::Donnees::Expression::Entier::zero(),*Traceur::Donnees::Expression::Entier::max_canal_couleur())._entier);
}

void Reelle::integre(Reelle couleur, Traceur::Donnees::Expression::Entier::TypeInterne alpha)
{
	if(alpha == 0)
		return;
	if(couleur._alpha==0)
		return;
	Traceur::Donnees::Expression::Entier::TypeInterne couleur_alpha = (couleur._alpha*alpha)/255;
	Traceur::Donnees::Expression::Entier::TypeInterne intermediaire_alpha = (((_alpha+couleur_alpha)*255)-_alpha*couleur_alpha)/255;
	_rouge = ((_rouge*255)+(couleur._rouge-_rouge)*couleur_alpha)/255;
	_vert  = ((_vert *255)+(couleur._vert -_vert )*couleur_alpha)/255;
	_bleu  = ((_bleu *255)+(couleur._bleu -_bleu )*couleur_alpha)/255;
	_alpha = intermediaire_alpha;
}

void Reelle::integre(Reelle couleur)
{
	if(couleur._alpha==0)
		return;
	Traceur::Donnees::Expression::Entier::TypeInterne intermediaire_alpha = (((_alpha+couleur._alpha)*255)-_alpha*couleur._alpha)/255;
	_rouge = ((_rouge*255)+(couleur._rouge-_rouge)*couleur._alpha)/255;
	_vert  = ((_vert *255)+(couleur._vert -_vert )*couleur._alpha)/255;
	_bleu  = ((_bleu *255)+(couleur._bleu -_bleu )*couleur._alpha)/255;
	_alpha = intermediaire_alpha;
}

void Reelle::remplace(Reelle couleur)
{
	_rouge = couleur._rouge;
	_vert  = couleur._vert ;
	_bleu  = couleur._bleu ;
	_alpha = couleur._alpha;
}

void Reelle::format(std::ostream& os)
{
	os << "COLOR[" << static_cast<unsigned int>(_rouge) << ";" << static_cast<unsigned int>(_vert) << ";" << static_cast<unsigned int>(_bleu) << ";" << static_cast<unsigned int>(_alpha) << "]" ;
}

Frequencielle::Frequencielle(const Traceur::Donnees::Expression::CouleurRVBASP& couleur)
{
	Traceur::Donnees::Expression::NombreSP rouge = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(couleur->rouge());
	TRACEUR_TEST(rouge);
	_rouge = rouge->reel()._reel;
	Traceur::Donnees::Expression::NombreSP vert = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(couleur->vert());
	TRACEUR_TEST(vert);
	_vert = vert->reel()._reel;
	Traceur::Donnees::Expression::NombreSP bleu = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(couleur->bleu());
	TRACEUR_TEST(bleu);
	_bleu = bleu->reel()._reel;
	Traceur::Donnees::Expression::NombreSP alpha = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(couleur->alpha());
	TRACEUR_TEST(alpha);
	_alpha = alpha->reel()._reel;
	_phase_rouge=0;
	_phase_vert =0;
	_phase_bleu =0;
	_phase_alpha=0;
}

Frequencielle::Frequencielle(const Couleur& couleur)
{
	_rouge = couleur._rouge;
	_vert  = couleur._vert ;
	_bleu  = couleur._bleu ;
	_alpha = couleur._alpha;
	_phase_rouge=0;
	_phase_vert =0;
	_phase_bleu =0;
	_phase_alpha=0;
}

void Frequencielle::integre(Frequencielle couleur, Traceur::Donnees::Expression::Entier::TypeInterne alpha)
{
	_rouge += (couleur._rouge*alpha)/255;
	_vert  += (couleur._vert *alpha)/255;
	_bleu  += (couleur._bleu *alpha)/255;
	_alpha += (couleur._alpha*alpha)/255;
}

void Frequencielle::integre(Frequencielle couleur)
{
	_rouge += couleur._rouge;
	_vert  += couleur._vert ;
	_bleu  += couleur._bleu ;
	_alpha += couleur._alpha;
}

void Frequencielle::remplace(Frequencielle couleur)
{
	_rouge = couleur._rouge;
	_vert  = couleur._vert ;
	_bleu  = couleur._bleu ;
	_alpha = couleur._alpha;
}

void Frequencielle::format(std::ostream& os)
{
	os << "COLOR[" << _rouge << ";" << _vert << ";" << _bleu << ";" << _alpha << "] PHASE[" 
		<< _phase_rouge << ";" << _phase_vert << ";" << _phase_bleu << ";" << _phase_alpha << "] ";
}

Couleur::Couleur(const Traceur::Donnees::Expression::CouleurRVBASP& couleur)
{
	Traceur::Donnees::Expression::NombreSP rouge = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(couleur->rouge());
	TRACEUR_TEST(rouge);
	_rouge = rouge->reel()._reel;
	Traceur::Donnees::Expression::NombreSP vert = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(couleur->vert());
	TRACEUR_TEST(vert);
	_vert = vert->reel()._reel;
	Traceur::Donnees::Expression::NombreSP bleu = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(couleur->bleu());
	TRACEUR_TEST(bleu);
	_bleu = bleu->reel()._reel;
	Traceur::Donnees::Expression::NombreSP alpha = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(couleur->alpha());
	TRACEUR_TEST(alpha);
	_alpha = alpha->reel()._reel;
}

Couleur::Couleur(const Reelle& couleur)
{
	Traceur::Donnees::Expression::Reel rouge(couleur._rouge);
	_rouge = rouge._reel;
	Traceur::Donnees::Expression::Reel vert (couleur._vert );
	_vert  = vert ._reel;
	Traceur::Donnees::Expression::Reel bleu (couleur._bleu );
	_bleu  = bleu ._reel;
	Traceur::Donnees::Expression::Reel alpha(couleur._alpha);
	_alpha = alpha._reel;
}

Couleur::Couleur(const Frequencielle& couleur)
{
	Traceur::Donnees::Expression::Reel rouge(couleur._rouge);
	_rouge = rouge._reel;
	Traceur::Donnees::Expression::Reel vert (couleur._vert );
	_vert  = vert ._reel;
	Traceur::Donnees::Expression::Reel bleu (couleur._bleu );
	_bleu  = bleu ._reel;
	Traceur::Donnees::Expression::Reel alpha(couleur._alpha);
	_alpha = alpha._reel;
}

void Couleur::format(std::ostream& os)
{
	os << "COLOR[" << _rouge << ";" << _vert << ";" << _bleu << ";" << _alpha << "]" ;
}

void Couleur::integre(Couleur couleur, Traceur::Donnees::Expression::Entier::TypeInterne alpha)
{
	_rouge += (couleur._rouge*alpha)/255;
	_vert  += (couleur._vert *alpha)/255;
	_bleu  += (couleur._bleu *alpha)/255;
	_alpha += (couleur._alpha*alpha)/255;
}

void Couleur::integre(Couleur couleur)
{
	_rouge += couleur._rouge;
	_vert  += couleur._vert ;
	_bleu  += couleur._bleu ;
	_alpha += couleur._alpha;
}

void Couleur::remplace(Couleur couleur)
{
	_rouge = couleur._rouge;
	_vert  = couleur._vert ;
	_bleu  = couleur._bleu ;
	_alpha = couleur._alpha;
}

