#include <src/donnees/assembleur/drawfill.h>
#include <stack>
using namespace Traceur::Donnees::Assembleur;

/*
//The scanline floodfill algorithm using our own stack routines, faster
void floodFillScanlineStack(int x, int y, int newColor, int oldColor)
{
	if(oldColor == newColor) return;
	emptyStack();

	int y1; 
	bool spanLeft, spanRight;

	if(!push(x, y)) return;

	while(pop(x, y))
	{    
		y1 = y;
		while(y1 >= 0 && screenBuffer[x][y1] == oldColor) y1--;
		y1++;
		spanLeft = spanRight = 0;
		while(y1 < h && screenBuffer[x][y1] == oldColor )
		{
			screenBuffer[x][y1] = newColor;
			if(!spanLeft && x > 0 && screenBuffer[x - 1][y1] == oldColor) 
			{
				if(!push(x - 1, y1)) return;
				spanLeft = 1;
			}
			else if(spanLeft && x > 0 && screenBuffer[x - 1][y1] != oldColor)
			{
				spanLeft = 0;
			}
			if(!spanRight && x < w - 1 && screenBuffer[x + 1][y1] == oldColor) 
			{
				if(!push(x + 1, y1)) return;
				spanRight = 1;
			}
			else if(spanRight && x < w - 1 && screenBuffer[x + 1][y1] != oldColor)
			{
				spanRight = 0;
			} 
			y1++;
		}
	}
}
*/

template<typename ModeTrace>
struct RemplissageNormal
{
	RemplissageNormal()
	:_mode_trace(ModeTrace()) {};
	bool correct(Image::Image& image, Couleur::Couleur couleur)
	{
		return _mode_trace.correct(image,couleur);
	};
	bool operator() (Image::Image& image, Couleur::Couleur initiale, Couleur::Couleur courante, Couleur::Couleur precedente)
	{
		return courante==initiale;
	};
	ModeTrace* operator->()
	{
		return &_mode_trace;
	};
	private:
	ModeTrace _mode_trace;
};

template<typename ModeTrace>
struct RemplissageAvance
{
	RemplissageAvance()
	:_mode_trace(ModeTrace()) {};
	bool correct(Image::Image& image, Couleur::Couleur couleur)
	{
		return true;
	};
	bool operator() (Image::Image& image, Couleur::Couleur initiale, Couleur::Couleur courante, Couleur::Couleur precedente)
	{
		Couleur::Couleur attendue = _mode_trace.simule(image,precedente);
#ifdef TRACEUR_DEBUG
		attendue.format(std::cerr);
		std::cerr << "!=";
		courante.format(std::cerr);
		std::cerr << std::endl;
#endif
		return attendue!=courante;
	};
	ModeTrace* operator->()
	{
		return &_mode_trace;
	};
	private:
	ModeTrace _mode_trace;
};

	template <typename ModeRemplissage>
void DrawFill::trace(Image::Image& image, ModeRemplissage mode_remplissage)
{
	std::stack<Point> germes;
	if(not image.sur_image(_point))
		return;
	Couleur::Couleur initiale = image.couleur(_point);
	Couleur::Couleur precedente = initiale;
	if(not mode_remplissage.correct(image, initiale))
		return;
	germes.push(_point);
	while(not germes.empty())
	{
		Point point = germes.top();
		germes.pop();
		while(point._y>=0)
		{  	
			Couleur::Couleur courante = image.couleur(point);
#ifdef TRACEUR_DEBUG
			point.format(std::cerr); std::cerr << std::endl;
#endif
			if(not mode_remplissage(image,initiale,courante,precedente))
				break;
			precedente = courante;
			point._y--;
		}
		point._y++;

		bool spangauche = false;
		bool spandroite = false;
		while(point._y<image._taille_y)
		{
#ifdef TRACEUR_DEBUG
			point.format(std::cerr); std::cerr << std::endl;
#endif
			Couleur::Couleur courante = image.couleur(point);
			if(not mode_remplissage(image,initiale,courante,precedente))
				break;
			precedente = courante;
			mode_remplissage->trace(image,point);
			// pour aller plus vite que la translation
			Point pointgauche = point;
			pointgauche._x--;
			Point pointdroite = point;
			pointdroite._x++;
			if((not spangauche) and (pointgauche._x>=0) and mode_remplissage(image,initiale,image.couleur(pointgauche),precedente))
			{
				germes.push(pointgauche);
				spangauche=true;
			}
			else if(spangauche and (pointgauche._x>=0) and (not mode_remplissage(image,initiale,image.couleur(pointgauche),precedente)))
			{
				spangauche=false;
			}

			if((not spandroite) and (pointdroite._x<(image._taille_x)) and mode_remplissage(image,initiale,image.couleur(pointdroite),precedente))
			{
				germes.push(pointdroite);
				spandroite=true;
			}
			else if(spandroite and (pointdroite._x<(image._taille_x)) and (not mode_remplissage(image,initiale,image.couleur(pointdroite),precedente)))
			{
				spandroite=false;
			}
			point._y++;
		}
	}

}

void DrawFill::trace(Image::Image& image)
{
	if(_options._remplissage)
	{
		switch(_options)
		{
			case Image::ModeTrace::NORMAL:
				{
					RemplissageAvance<TraceSurImage<Image::ModeTrace::NORMAL> > remplissage;
					trace(image,remplissage);
					break;
				}
			case Image::ModeTrace::SOLIDE:
				{
					RemplissageAvance<TraceSurImage<Image::ModeTrace::SOLIDE> > remplissage;
					trace(image,remplissage);
					break;
				}
			case Image::ModeTrace::GOMME:
				{
					RemplissageAvance<TraceSurImage<Image::ModeTrace::GOMME> > remplissage;
					trace(image,remplissage);
					break;
				}
		}
	}
	else
	{
		switch(_options)
		{
			case Image::ModeTrace::NORMAL:
				{
					RemplissageNormal<TraceSurImage<Image::ModeTrace::NORMAL> > remplissage;
					trace(image,remplissage);
					break;
				}
			case Image::ModeTrace::SOLIDE:
				{
					RemplissageNormal<TraceSurImage<Image::ModeTrace::SOLIDE> > remplissage;
					trace(image,remplissage);
					break;
				}
			case Image::ModeTrace::GOMME:
				{
					RemplissageNormal<TraceSurImage<Image::ModeTrace::GOMME> > remplissage;
					trace(image,remplissage);
					break;
				}
		}
	}
}

void DrawFill::format(std::ostream& os)
{
	os << "draw FILL[";
	_point.format(os);
	_options.format(os);
	os << "] // asm" << std::endl;
}

