#include <src/donnees/assembleur/image.h>
#include <src/donnees/assembleur/instruction.h>
#include <src/donnees/assembleur/draw.h>
#include <src/donnees/assembleur/conv.h>
#include <src/donnees/assembleur/with.h>
#include <src/donnees/assembleur/read.h>
#include <algorithm>
#include <functional>
#include <cmath>
using namespace Traceur::Donnees::Assembleur::Image;

Reelle::Reelle(const Traceur::Donnees::Expression::ExpressionSP& taille_x, const Traceur::Donnees::Expression::ExpressionSP& taille_y,
		const Traceur::Donnees::Expression::SystemeCoordonneesSP& coordonnees, const Traceur::Donnees::Expression::CouleurRVBASP& couleur_fond)
:Image(couleur_fond)
{
	Traceur::Donnees::Expression::NombreSP entier_x = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(taille_x);
	TRACEUR_TEST(entier_x);
	_taille_x = entier_x->entier()._entier;
	Traceur::Donnees::Expression::NombreSP entier_y = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(taille_y);
	TRACEUR_TEST(entier_y);
	_taille_y = entier_y->entier()._entier;
	if(_taille_x<0 or _taille_y<0)
		throw Traceur::Donnees::Instruction::Valeur::taille_image_incorrecte();
	_systeme_coordonnees = SystemeCoordonnees(entier_x->entier(),entier_y->entier(),coordonnees);
	_couleur_courante = _couleur_fond;
	_bitmap = Bitmap<Traceur::Donnees::Assembleur::Couleur::Reelle>(0,0,_taille_x,_taille_y,couleur_fond);
}

#define FOURIER_TRAFIQUE
Frequencielle::Frequencielle(Reelle& reelle)
:Image(reelle._taille_x,reelle._taille_y,reelle._systeme_coordonnees,reelle._couleur_fond,reelle._couleur_courante,reelle._instructions),_bitmap(0,0,reelle._taille_x,reelle._taille_y,Traceur::Donnees::Assembleur::Couleur::Frequencielle())
{
	TRACEUR_TEST(reelle._instructions.empty());
#ifdef FOURIER_TRAFIQUE
	// Transformee de Fourier discrete bidouillee
	Traceur::Donnees::Expression::Reel::TypeInterne normalisation = 1./(_taille_x*_taille_y);
	for(Traceur::Donnees::Assembleur::Coordonnee abscisse_freq=0 ; abscisse_freq<_taille_x ; ++abscisse_freq)
	{
		Traceur::Donnees::Expression::Entier::TypeInterne ppcm_x = Traceur::Donnees::Expression::Entier::ppcm(Traceur::Donnees::Expression::Entier(abscisse_freq+1),Traceur::Donnees::Expression::Entier(_taille_x))._entier;

		for(Traceur::Donnees::Assembleur::Coordonnee ordonnee_freq=0 ; ordonnee_freq<_taille_y ; ++ordonnee_freq)
		{
			Traceur::Donnees::Expression::Entier::TypeInterne ppcm_y = Traceur::Donnees::Expression::Entier::ppcm(Traceur::Donnees::Expression::Entier(ordonnee_freq+1),Traceur::Donnees::Expression::Entier(_taille_y))._entier;

			Traceur::Donnees::Assembleur::Couleur::Complexe couleur_frequencielle;

			for(Traceur::Donnees::Assembleur::Coordonnee abscisse_reel=0 ; abscisse_reel<ppcm_x ; ++abscisse_reel)
			{
				for(Traceur::Donnees::Assembleur::Coordonnee ordonnee_reel=0 ; ordonnee_reel<ppcm_y ; ++ordonnee_reel)
				{
					Traceur::Donnees::Assembleur::Point point_reel(abscisse_reel%_taille_x,ordonnee_reel%_taille_y);
					Traceur::Donnees::Assembleur::Couleur::Couleur couleur_reelle = reelle.couleur(point_reel);

couleur_frequencielle += Traceur::Donnees::Assembleur::Couleur::Complexe::exp_2i_pi(-(abscisse_reel/(abscisse_freq+1.)+ordonnee_reel/(ordonnee_freq+1.)))*Traceur::Donnees::Assembleur::Couleur::Couleur(couleur_reelle);
				}
			}
			Traceur::Donnees::Assembleur::Point point_freq(abscisse_freq,ordonnee_freq);
#ifdef TRACEUR_DEBUG
			Traceur::Donnees::Assembleur::Couleur::Frequencielle coul = couleur_frequencielle;
			coul.format(std::cerr);
#endif
			remplace(point_freq,couleur_frequencielle*normalisation);
		}
	}
#else
	Traceur::Donnees::Expression::Reel::TypeInterne normalisation = 1./(_taille_x*_taille_y);
	for(Traceur::Donnees::Assembleur::Coordonnee abscisse_freq=0 ; abscisse_freq<_taille_x ; ++abscisse_freq)
	{
		for(Traceur::Donnees::Assembleur::Coordonnee ordonnee_freq=0 ; ordonnee_freq<_taille_y ; ++ordonnee_freq)
		{
			Traceur::Donnees::Assembleur::Couleur::Complexe couleur_frequencielle;

			for(Traceur::Donnees::Assembleur::Coordonnee abscisse_reel=0 ; abscisse_reel<_taille_x ; ++abscisse_reel)
			{
				for(Traceur::Donnees::Assembleur::Coordonnee ordonnee_reel=0 ; ordonnee_reel<_taille_y ; ++ordonnee_reel)
				{
					Traceur::Donnees::Assembleur::Point point_reel(abscisse_reel,ordonnee_reel);
					Traceur::Donnees::Assembleur::Couleur::Couleur couleur_reelle = reelle.couleur(point_reel);

couleur_frequencielle += Traceur::Donnees::Assembleur::Couleur::Complexe::exp_2i_pi(-((static_cast<Traceur::Donnees::Assembleur::Couleur::Complexe::Canal>(abscisse_reel*abscisse_freq))/_taille_x+(static_cast<Traceur::Donnees::Assembleur::Couleur::Complexe::Canal>(ordonnee_reel*ordonnee_freq))/_taille_y))*Traceur::Donnees::Assembleur::Couleur::Couleur(couleur_reelle);
				}
			}
			Traceur::Donnees::Assembleur::Point point_freq(abscisse_freq,ordonnee_freq);
#ifdef TRACEUR_DEBUG
			Traceur::Donnees::Assembleur::Couleur::Frequencielle coul = couleur_frequencielle;
			coul.format(std::cerr);
#endif
			remplace(point_freq,couleur_frequencielle*normalisation);
		}
	}
#endif
}

Reelle::Reelle(Frequencielle& frequencielle)
:Image(frequencielle._taille_x,frequencielle._taille_y,frequencielle._systeme_coordonnees,frequencielle._couleur_fond,frequencielle._couleur_courante,frequencielle._instructions),_bitmap(0,0,frequencielle._taille_x,frequencielle._taille_y,Traceur::Donnees::Assembleur::Couleur::Reelle())
{
	TRACEUR_TEST(frequencielle._instructions.empty());
	// Transformee de Fourier discrete bidouillee inverse
	for(Traceur::Donnees::Assembleur::Coordonnee abscisse_reel=0 ; abscisse_reel<_taille_x ; ++abscisse_reel)
	{
		for(Traceur::Donnees::Assembleur::Coordonnee ordonnee_reel=0 ; ordonnee_reel<_taille_y ; ++ordonnee_reel)
		{
			Traceur::Donnees::Assembleur::Couleur::Complexe couleur_reelle;

			for(Traceur::Donnees::Assembleur::Coordonnee abscisse_freq=0 ; abscisse_freq<_taille_x ; ++abscisse_freq)
			{
				for(Traceur::Donnees::Assembleur::Coordonnee ordonnee_freq=0 ; ordonnee_freq<_taille_y ; ++ordonnee_freq)
				{
					Traceur::Donnees::Assembleur::Point point_freq(abscisse_freq,ordonnee_freq);
					Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur_frequencielle = frequencielle.couleur_frequencielle(point_freq);

#ifdef TRACEUR_DEBUG
					std::cout << abscisse_reel << "," << ordonnee_reel << " - " << abscisse_freq << "," << ordonnee_freq << ": ";
					Traceur::Donnees::Assembleur::Couleur::Complexe::exp_2i_pi((abscisse_reel/(abscisse_freq+1.)+ordonnee_reel/(ordonnee_freq+1.))).format(std::cout); std::cout << " * " ; couleur_frequencielle.format(std::cout); std::cout << std::endl;
#endif
#ifdef FOURIER_TRAFIQUE
couleur_reelle += Traceur::Donnees::Assembleur::Couleur::Complexe::exp_2i_pi((abscisse_reel/(abscisse_freq+1.)+ordonnee_reel/(ordonnee_freq+1.)))*Traceur::Donnees::Assembleur::Couleur::Complexe(couleur_frequencielle);
#else
couleur_reelle += Traceur::Donnees::Assembleur::Couleur::Complexe::exp_2i_pi(((static_cast<Traceur::Donnees::Assembleur::Couleur::Complexe::Canal>(abscisse_reel*abscisse_freq))/_taille_x+(static_cast<Traceur::Donnees::Assembleur::Couleur::Complexe::Canal>(ordonnee_reel*ordonnee_freq))/_taille_y))*Traceur::Donnees::Assembleur::Couleur::Complexe(couleur_frequencielle);
#endif
				}
			}
			Traceur::Donnees::Assembleur::Point point_reel(abscisse_reel,ordonnee_reel);
			remplace(point_reel,couleur_reelle);
		}
	}
}

void Image::trace()
{
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE << "debut" << std::endl;
	std::for_each(_instructions.begin(),_instructions.end(),bind(&Instruction::format,std::placeholders::_1,std::ref(std::cerr)));
	TRACEUR_TRACE << "fin" << std::endl;
#endif
	std::for_each(_instructions.begin(),_instructions.end(),bind(&Instruction::trace,std::placeholders::_1,std::ref(*this)));
	_instructions.clear();
}

void Reelle::format(std::ostream& os)
{
	os << "IMAGE[" << _taille_x << ";" << _taille_y << "] ";
	_systeme_coordonnees.format(os);
	os << " " ;
	_couleur_fond.format(os);
	os << " // asm/normal" << std::endl;
	_bitmap.format(os);
	os << "{" << std::endl;
	std::for_each(_instructions.begin(),_instructions.end(),bind(&Instruction::format,std::placeholders::_1,std::ref(os)));
	os << "}" << std::endl;
}

void Frequencielle::format(std::ostream& os)
{
	os << "IMAGE[" << _taille_x << ";" << _taille_y << "] ";
	_systeme_coordonnees.format(os);
	os << " " ;
	_couleur_fond.format(os);
	os << " // asm/frequencies" << std::endl;
	_bitmap.format(os);
	os << "{" << std::endl;
	std::for_each(_instructions.begin(),_instructions.end(),bind(&Instruction::format,std::placeholders::_1,std::ref(os)));
	os << "}" << std::endl;
}

void Noyau::format(std::ostream& os)
{
	os << "KERNEL[" << _taille_x << ";" << _taille_y << "] ";
	_systeme_coordonnees.format(os);
	os << " // asm" << std::endl;
	_bitmap.format(os);
	os << "{" << std::endl;
	std::for_each(_instructions.begin(),_instructions.end(),bind(&Instruction::format,std::placeholders::_1,std::ref(os)));
	os << "}" << std::endl;
}

void Image::ajoute(const InstructionSP& instruction)
{
	_instructions.push_back(instruction);
}

void Reelle::trace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur, Traceur::Donnees::Expression::Entier::TypeInterne alpha)
{
	if(not sur_image(point))
		return;
	if(alpha==0)
		return;
	if(couleur._alpha==0)
		return;
	_bitmap.trace(point,Couleur::Couleur(couleur),0,0,_taille_x,_taille_y,_couleur_fond,alpha);
}

void Reelle::trace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur)
{
	if(not sur_image(point))
		return;
	if(couleur._alpha==0)
		return;
	_bitmap.trace(point,Couleur::Couleur(couleur),0,0,_taille_x,_taille_y,_couleur_fond);
}

void Reelle::remplace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur)
{
	if(not sur_image(point))
		return;
	_bitmap.remplace(point,Couleur::Couleur(couleur),0,0,_taille_x,_taille_y,_couleur_fond);
}

Traceur::Donnees::Assembleur::Couleur::Couleur Reelle::simule(Traceur::Donnees::Assembleur::Couleur::Couleur couleur)
{
	Traceur::Donnees::Assembleur::Couleur::Reelle reelle(couleur);
	reelle.integre(_couleur_courante);
	return Traceur::Donnees::Assembleur::Couleur::Couleur(reelle);
}

Traceur::Donnees::Assembleur::Couleur::Couleur Reelle::couleur(Point point)
{
	if(not sur_image(point))
		throw Traceur::Donnees::Instruction::Valeur::coordonnee_hors_image();
	return _bitmap.couleur(point,0,0,_taille_x,_taille_y,_couleur_fond);
}

void Frequencielle::trace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur, Traceur::Donnees::Expression::Entier::TypeInterne alpha)
{
	if(not sur_image(point))
		return;
	_bitmap.trace(point,couleur,0,0,_taille_x,_taille_y,_couleur_fond,alpha);
}

void Frequencielle::trace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur)
{
	if(not sur_image(point))
		return;
	_bitmap.trace(point,couleur,0,0,_taille_x,_taille_y,_couleur_fond);
}

void Frequencielle::remplace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur)
{
	if(not sur_image(point))
		return;
	_bitmap.remplace(point,couleur,0,0,_taille_x,_taille_y,_couleur_fond);
}

Traceur::Donnees::Assembleur::Couleur::Couleur Frequencielle::simule(Traceur::Donnees::Assembleur::Couleur::Couleur couleur)
{
	Traceur::Donnees::Assembleur::Couleur::Frequencielle frequencielle(couleur);
	frequencielle.integre(_couleur_courante);
	return Traceur::Donnees::Assembleur::Couleur::Couleur(frequencielle);
}

Traceur::Donnees::Assembleur::Couleur::Couleur Frequencielle::couleur(Point point)
{
	if(not sur_image(point))
		throw Traceur::Donnees::Instruction::Valeur::coordonnee_hors_image();
	return _bitmap.couleur(point,0,0,_taille_x,_taille_y,_couleur_fond);
}

Traceur::Donnees::Assembleur::Couleur::Frequencielle Frequencielle::couleur_frequencielle(Point point)
{
	return _bitmap.couleur(point,0,0,_taille_x,_taille_y,_couleur_fond);
}

Traceur::Donnees::Assembleur::Coordonnee Reelle::externe_lignes()
{
	if(_systeme_coordonnees._type==SystemeCoordonnees::Type::STANDARD)
		return _taille_x;
	return _taille_y;
}

Traceur::Donnees::Assembleur::Coordonnee Reelle::externe_colonnes()
{
	if(_systeme_coordonnees._type==SystemeCoordonnees::Type::STANDARD)
		return _taille_y;
	return _taille_x;
}

Traceur::Donnees::Assembleur::Couleur::Reelle Reelle::externe_pixel(Traceur::Donnees::Assembleur::Coordonnee ligne, Traceur::Donnees::Assembleur::Coordonnee colonne)
{
	Point point(colonne,_taille_y-1-ligne);
	if(_systeme_coordonnees._type==SystemeCoordonnees::Type::STANDARD)
		point = Point(ligne,colonne);
	return couleur(point);
}

void Reelle::externe_trace(Traceur::Donnees::Assembleur::Coordonnee ligne, Traceur::Donnees::Assembleur::Coordonnee colonne, Traceur::Donnees::Assembleur::Couleur::Reelle couleur)
{
	Point point(colonne,_taille_y-1-ligne);
	if(_systeme_coordonnees._type==SystemeCoordonnees::Type::STANDARD)
		point = Point(ligne,colonne);
	trace(point,Couleur::Couleur(couleur));
}

Noyau::Noyau(const Traceur::Donnees::Expression::ExpressionSP& taille_x, const Traceur::Donnees::Expression::ExpressionSP& taille_y,
		const Traceur::Donnees::Expression::SystemeCoordonneesSP& coordonnees)
:Image(Couleur::Couleur())
{
	Traceur::Donnees::Expression::NombreSP entier_x = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(taille_x);
	TRACEUR_TEST(entier_x);
	_taille_x = entier_x->entier()._entier;
	Traceur::Donnees::Expression::NombreSP entier_y = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(taille_y);
	TRACEUR_TEST(entier_y);
	_taille_y = entier_y->entier()._entier;
	if(_taille_x<0 or _taille_y<0)
		throw Traceur::Donnees::Instruction::Valeur::taille_noyau_incorrecte();
	if(not (_taille_x%2==1 and _taille_y%2==1))
		throw Traceur::Donnees::Instruction::Valeur::taille_noyau_incorrecte();
	_systeme_coordonnees = SystemeCoordonnees(entier_x->entier(),entier_y->entier(),coordonnees,SystemeCoordonnees::Origine::CENTRE);
	_couleur_courante = _couleur_fond;
	_bitmap = Bitmap<Traceur::Donnees::Assembleur::Couleur::Couleur>(0,0,_taille_x,_taille_y,_couleur_fond);
}

void Noyau::trace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur, Traceur::Donnees::Expression::Entier::TypeInterne alpha)
{
	if(not sur_image(point))
		return;
	_bitmap.trace(point,couleur,0,0,_taille_x,_taille_y,_couleur_fond,alpha);
}

void Noyau::trace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur)
{
	if(not sur_image(point))
		return;
	_bitmap.trace(point,couleur,0,0,_taille_x,_taille_y,_couleur_fond);
}

void Noyau::remplace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur)
{
	if(not sur_image(point))
		return;
	_bitmap.remplace(point,couleur,0,0,_taille_x,_taille_y,_couleur_fond);
}

Traceur::Donnees::Assembleur::Couleur::Couleur Noyau::simule(Traceur::Donnees::Assembleur::Couleur::Couleur couleur)
{
	couleur.integre(_couleur_courante);
	return couleur;
}
Traceur::Donnees::Assembleur::Couleur::Couleur Noyau::couleur(Point point)
{
	if(not sur_image(point))
		throw Traceur::Donnees::Instruction::Valeur::coordonnee_hors_image();
	return _bitmap.couleur(point,0,0,_taille_x,_taille_y,_couleur_fond);
}

Traceur::Donnees::Assembleur::Couleur::Couleur Noyau::couleur_noyau(Point point)
{
	if(not sur_image(point))
		return _couleur_fond;
	return _bitmap.couleur(point,0,0,_taille_x,_taille_y,_couleur_fond);
}

Traceur::Donnees::Assembleur::Couleur::Couleur Reelle::couleur_noyau(Point point)
{
	if(not sur_image(point))
		return _couleur_fond;
	return _bitmap.couleur(point,0,0,_taille_x,_taille_y,_couleur_fond);
}

void Reelle::convolution(NoyauSP noyau)
{
	Noyau::convolution_interne<Traceur::Donnees::Assembleur::Couleur::Reelle>(*this,noyau);
}

void Noyau::convolution(NoyauSP noyau)
{
	Noyau::convolution_interne<Traceur::Donnees::Assembleur::Couleur::Couleur>(*this,noyau);
}
