#ifndef _DONNEES_ASSEMBLEUR_IMAGE_H_
#define _DONNEES_ASSEMBLEUR_IMAGE_H_

#include <iostream>
#include <set>
#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Assembleur { namespace Image { DECL_SHARED_PTR(Image); DECL_SHARED_PTR(Reelle); DECL_SHARED_PTR(Frequencielle); DECL_SHARED_PTR(Noyau); } } } }

#include <src/donnees/assembleur/assembleur.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/assembleur/couleur.h>
#include <src/donnees/assembleur/systemecoordonnees.h>
#include <src/donnees/assembleur/point.h>
#include <src/donnees/assembleur/bitmap.h>

namespace Traceur
{
	namespace Png { class Image; }
	namespace Donnees
	{
		namespace Expression { class Point; class Cercle; class Ligne; class Solide; class Gomme; class Remplissage; class Triangle; class Rectangle; class Ellipse; }
		namespace Assembleur
		{
			DECL_SHARED_PTR(Instruction);
			class DrawPoint;
			class DrawSolid;
			class DrawEraser;
			class DrawFill;
			class DrawCircle;
			class DrawAt;
			class With;
			namespace Image { enum class ModeTrace { NORMAL, SOLIDE, GOMME}; }
			template<Image::ModeTrace mode>
				struct TraceSurImage;
			namespace Image
			{
				class Image
				{
					friend class Traceur::Donnees::Assembleur::Assembleur;
					friend class Traceur::Donnees::Assembleur::DrawPoint;
					friend class Traceur::Donnees::Assembleur::DrawSolid;
					friend class Traceur::Donnees::Assembleur::DrawEraser;
					friend class Traceur::Donnees::Assembleur::DrawFill;
					friend class Traceur::Donnees::Assembleur::DrawCircle;
					friend class Traceur::Donnees::Assembleur::DrawAt;
					friend class Traceur::Donnees::Assembleur::With;
					friend class Traceur::Donnees::Assembleur::Image::Reelle;
					friend class Traceur::Donnees::Assembleur::Image::Frequencielle;
					friend class Traceur::Donnees::Assembleur::Image::Noyau;
					friend class Traceur::Donnees::Expression::Point;
					friend class Traceur::Donnees::Expression::Cercle;
					friend class Traceur::Donnees::Expression::Ligne;
					friend class Traceur::Donnees::Expression::Solide;
					friend class Traceur::Donnees::Expression::Gomme;
					friend class Traceur::Donnees::Expression::Remplissage;
					friend class Traceur::Donnees::Expression::Triangle;
					friend class Traceur::Donnees::Expression::Rectangle;
					friend class Traceur::Donnees::Expression::Ellipse;
					template<ModeTrace mode>
						friend struct Traceur::Donnees::Assembleur::TraceSurImage;
					public:
					enum class Type { REELLE, FREQUENCIELLE, NOYAU };
					Image(Traceur::Donnees::Assembleur::Couleur::Couleur couleur_fond) :_couleur_fond(couleur_fond) {};
					virtual Type type() const = 0;
					void ajoute(const InstructionSP& instruction);
					void trace();
					virtual void convolution(NoyauSP noyau) {};
					virtual void format(std::ostream& os) = 0;
					virtual ~Image() {};
					protected:
					Image(Traceur::Donnees::Assembleur::Coordonnee taille_x, Traceur::Donnees::Assembleur::Coordonnee taille_y,
						SystemeCoordonnees systeme_coordonnees,
						Traceur::Donnees::Assembleur::Couleur::Couleur couleur_fond,
						Traceur::Donnees::Assembleur::Couleur::Couleur couleur_courante,
						Assembleur::ListeInstructions instructions)
						:_taille_x(taille_x), _taille_y(taille_y), _systeme_coordonnees(systeme_coordonnees),
						_couleur_fond(couleur_fond), _couleur_courante(couleur_courante),
						_instructions(instructions) {};
					private:
					Traceur::Donnees::Assembleur::Coordonnee _taille_x;
					Traceur::Donnees::Assembleur::Coordonnee _taille_y;
					SystemeCoordonnees _systeme_coordonnees;
					Traceur::Donnees::Assembleur::Couleur::Couleur _couleur_fond;
					Traceur::Donnees::Assembleur::Couleur::Couleur _couleur_courante;
					Assembleur::ListeInstructions _instructions;
					virtual Traceur::Donnees::Assembleur::Couleur::Couleur couleur(Point point) = 0;
					virtual void trace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur, Traceur::Donnees::Expression::Entier::TypeInterne alpha) = 0;
					virtual void trace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur) = 0;
					virtual void remplace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur) = 0;
					virtual Traceur::Donnees::Assembleur::Couleur::Couleur simule(Traceur::Donnees::Assembleur::Couleur::Couleur couleur) = 0;
					void trace(Point point) { trace(point,_couleur_courante); };
					void remplace(Point point) { remplace(point,_couleur_courante); };
					bool sur_image(const Point& point) const
					{
						if(point._x<0)
						       return false;
						if(point._x>=_taille_x)
						       return false;
						if(point._y<0)
						       return false;
						if(point._y>=_taille_y)
							return false;
						return true;
					};
				};

				class Reelle : public Image
				{
					friend class Traceur::Donnees::Assembleur::Assembleur;
					friend class Traceur::Donnees::Assembleur::DrawPoint;
					friend class Traceur::Donnees::Assembleur::DrawAt;
					friend class Traceur::Donnees::Assembleur::Image::Frequencielle;
					friend class Traceur::Donnees::Assembleur::Image::Noyau;
					friend class Traceur::Donnees::Expression::Point;
					friend class Traceur::Png::Image;
					public:
					Reelle(const Traceur::Donnees::Expression::ExpressionSP& taille_x, const Traceur::Donnees::Expression::ExpressionSP& taille_y,
							const Traceur::Donnees::Expression::SystemeCoordonneesSP& coordonnees, const Traceur::Donnees::Expression::CouleurRVBASP& couleur_fond);
					Reelle(Frequencielle& frequencielle);
					virtual Type type() const { return Image::Type::REELLE; };
					virtual void convolution(NoyauSP noyau);
					virtual void format(std::ostream& os);
					virtual ~Reelle() {};
					private:
					virtual void trace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur, Traceur::Donnees::Expression::Entier::TypeInterne alpha);
					virtual void trace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur);
					virtual void remplace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur);
					virtual Traceur::Donnees::Assembleur::Couleur::Couleur simule(Traceur::Donnees::Assembleur::Couleur::Couleur couleur);
					virtual Traceur::Donnees::Assembleur::Couleur::Couleur couleur(Point point);
					Traceur::Donnees::Assembleur::Couleur::Couleur couleur_noyau(Point point);
					Bitmap<Traceur::Donnees::Assembleur::Couleur::Reelle> _bitmap;
					// export fichier
					Traceur::Donnees::Assembleur::Coordonnee externe_lignes();
					Traceur::Donnees::Assembleur::Coordonnee externe_colonnes();
					Traceur::Donnees::Assembleur::Couleur::Reelle externe_pixel(Traceur::Donnees::Assembleur::Coordonnee ligne, Traceur::Donnees::Assembleur::Coordonnee colonne);
					void externe_trace(Traceur::Donnees::Assembleur::Coordonnee ligne, Traceur::Donnees::Assembleur::Coordonnee colonne, Traceur::Donnees::Assembleur::Couleur::Reelle couleur);
				};

				class Frequencielle : public Image
				{
					friend class Traceur::Donnees::Assembleur::Assembleur;
					friend class Traceur::Donnees::Assembleur::DrawPoint;
					friend class Traceur::Donnees::Assembleur::DrawAt;
					friend class Traceur::Donnees::Assembleur::Image::Reelle;
					friend class Traceur::Donnees::Expression::Point;
					friend class Traceur::Png::Image;
					public:
					Frequencielle(Reelle& reelle);
					virtual Type type() const { return Image::Type::FREQUENCIELLE; };
					virtual void format(std::ostream& os);
					virtual ~Frequencielle() {};
					private:
					virtual void trace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur, Traceur::Donnees::Expression::Entier::TypeInterne alpha);
					virtual void trace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur);
					virtual void remplace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur);
					virtual Traceur::Donnees::Assembleur::Couleur::Couleur simule(Traceur::Donnees::Assembleur::Couleur::Couleur couleur);
					virtual Traceur::Donnees::Assembleur::Couleur::Couleur couleur(Point point);
					Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur_frequencielle(Point point); // usage interne
					Bitmap<Traceur::Donnees::Assembleur::Couleur::Frequencielle> _bitmap;
				};

				class Noyau : public Image
				{
					friend class Traceur::Donnees::Assembleur::Assembleur;
					friend class Traceur::Donnees::Assembleur::DrawPoint;
					friend class Traceur::Donnees::Assembleur::DrawAt;
					friend class Traceur::Donnees::Assembleur::Image::Reelle;
					friend class Traceur::Donnees::Expression::Point;
					friend class Traceur::Png::Image;
					public:
					Noyau(const Traceur::Donnees::Expression::ExpressionSP& taille_x, const Traceur::Donnees::Expression::ExpressionSP& taille_y,
						const Traceur::Donnees::Expression::SystemeCoordonneesSP& coordonnees);
					virtual Type type() const { return Image::Type::NOYAU; };
					virtual void convolution(NoyauSP noyau);
					virtual void format(std::ostream& os);
					virtual ~Noyau() {};
					private:
					virtual void trace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur, Traceur::Donnees::Expression::Entier::TypeInterne alpha);
					virtual void trace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur);
					virtual void remplace(Point point, Traceur::Donnees::Assembleur::Couleur::Frequencielle couleur);
					virtual Traceur::Donnees::Assembleur::Couleur::Couleur simule(Traceur::Donnees::Assembleur::Couleur::Couleur couleur);
					virtual Traceur::Donnees::Assembleur::Couleur::Couleur couleur(Point point);
					Traceur::Donnees::Assembleur::Couleur::Couleur couleur_noyau(Point point);
					Bitmap<Traceur::Donnees::Assembleur::Couleur::Couleur> _bitmap;
					template<typename Couleur, typename Image>
						static void convolution_interne(Image& image, NoyauSP noyau)
					{
						Traceur::Donnees::Assembleur::Bitmap<Couleur> resultat(0,0,image._taille_x,image._taille_y,image._couleur_fond);
						
						Traceur::Donnees::Expression::Entier::TypeInterne rang_x = (noyau->_taille_x-1)/2;
						Traceur::Donnees::Expression::Entier::TypeInterne rang_y = (noyau->_taille_y-1)/2;

						Point centre_noyau(rang_x,rang_y);

						for(Traceur::Donnees::Assembleur::Coordonnee x=0 ; x<image._taille_x ; ++x)
						{
							for(Traceur::Donnees::Assembleur::Coordonnee y=0 ; y<image._taille_y ; ++y)
							{
								Traceur::Donnees::Assembleur::Couleur::Couleur resultante;
								Traceur::Donnees::Assembleur::Point point(x,y);

								for(Traceur::Donnees::Assembleur::Coordonnee noyau_x = -rang_x ; noyau_x<=rang_x ; ++noyau_x)
								{
									for(Traceur::Donnees::Assembleur::Coordonnee noyau_y = -rang_y ; noyau_y<=rang_y ; ++noyau_y)
									{
										Traceur::Donnees::Assembleur::Point point_noyau(noyau_x,noyau_y);
										Traceur::Donnees::Assembleur::Point point_oppose_noyau(rang_x-noyau_x,rang_y-noyau_y);
										resultante += image.couleur_noyau(point.translate(point_noyau))*noyau->couleur_noyau(point_oppose_noyau);
									}
								}
								resultat.remplace(point,resultante,0,0,image._taille_x,image._taille_y,image._couleur_fond);
							}
						}
						image._bitmap=resultat;
					}
				};
			}

			template<Image::ModeTrace mode>
				struct TraceSurImage
				{
					void trace(Image::Image& image, Point point) { throw; };
					Couleur::Couleur simule(Image::Image& image, Couleur::Couleur couleur) { throw; };
					bool correct(Image::Image& image, Couleur::Couleur couleur) { throw; };
				};

			template<>
				struct TraceSurImage<Image::ModeTrace::NORMAL>
				{
					void trace(Image::Image& image, Point point)
					{
						image.trace(point);
					};
					Couleur::Couleur simule(Image::Image& image, Couleur::Couleur couleur)
					{
						return image.simule(couleur);
					};
					bool correct(Image::Image& image, Couleur::Couleur couleur)
					{
						Couleur::Couleur tranformee = couleur;
						tranformee.integre(image._couleur_courante);
						return couleur!=tranformee;
					};
				};

			template<>
				struct TraceSurImage<Image::ModeTrace::SOLIDE>
				{
					void trace(Image::Image& image, Point point)
					{
						image.remplace(point,image._couleur_courante);
					};
					Couleur::Couleur simule(Image::Image& image, Couleur::Couleur couleur)
					{
						return image._couleur_courante;
					};
					bool correct(Image::Image& image, Couleur::Couleur couleur)
					{
						return couleur!=image._couleur_courante;
					};
				};

			template<>
				struct TraceSurImage<Image::ModeTrace::GOMME>
				{
					void trace(Image::Image& image, Point point)
					{
						image.remplace(point,image._couleur_fond);
					};
					Couleur::Couleur simule(Image::Image& image, Couleur::Couleur couleur)
					{
						return image._couleur_fond;
					};
					bool correct(Image::Image& image, Couleur::Couleur couleur)
					{
						return couleur!=image._couleur_fond;
					};
				};
		}
	}
}

#endif
