#ifndef _DONNEES_ASSEMBLEUR_SYSTEME_COORDONNEES_H_
#define _DONNEES_ASSEMBLEUR_SYSTEME_COORDONNEES_H_

#include <iostream>
#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Assembleur { DECL_SHARED_PTR(SystemeCoordonnees); } } }

#include <src/donnees/assembleur/assembleur.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/instructions/valeur.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression { class Xmin; class Xmax; class Xscl; class Xsiz; class Ymin; class Ymax; class Yscl; class Ysiz; }
		namespace Assembleur
		{
			typedef Traceur::Donnees::Expression::Entier::TypeInterne Coordonnee;
			namespace Image { class Reelle; }
			class SystemeCoordonnees
			{
				friend class Traceur::Donnees::Expression::Xmin;
				friend class Traceur::Donnees::Expression::Xmax;
				friend class Traceur::Donnees::Expression::Xscl;
				friend class Traceur::Donnees::Expression::Xsiz;
				friend class Traceur::Donnees::Expression::Ymin;
				friend class Traceur::Donnees::Expression::Ymax;
				friend class Traceur::Donnees::Expression::Yscl;
				friend class Traceur::Donnees::Expression::Ysiz;
				friend class Image::Reelle;
				public:
					enum class Type {GEOMETRIE, STANDARD, MATHEMATIQUE};
					enum class Origine {NORMAL, CENTRE};
					SystemeCoordonnees(Traceur::Donnees::Expression::Entier taille_x, Traceur::Donnees::Expression::Entier taille_y,
						const Traceur::Donnees::Expression::SystemeCoordonneesSP& coordonnees, Origine origine = Origine::NORMAL)
						:_type(Type::GEOMETRIE),_origine(origine)
					{
						static Traceur::Donnees::Expression::Entier _1(1);
						static Traceur::Donnees::Expression::Entier _2(2);
						_xsiz = Traceur::Donnees::Expression::EntierSP(new Traceur::Donnees::Expression::Entier(taille_x));
						_ysiz = Traceur::Donnees::Expression::EntierSP(new Traceur::Donnees::Expression::Entier(taille_y));
						Traceur::Donnees::Expression::SystemeMathematiqueSP  maths = std::dynamic_pointer_cast<Traceur::Donnees::Expression::SystemeMathematique>(coordonnees);
						if(maths)
						{
							if((taille_x<=_1) or (taille_y<=_1))
							{
								throw Traceur::Donnees::Instruction::Valeur::image_mathematique_vide();
							}
							Traceur::Donnees::Expression::PointSP min = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Point>(maths->_point_min);
							TRACEUR_TEST(min);
							Traceur::Donnees::Expression::PointSP max = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Point>(maths->_point_max);
							TRACEUR_TEST(max);
							_xmin = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(min->gauche()); TRACEUR_TEST(_xmin);
							_xmax = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(max->gauche()); TRACEUR_TEST(_xmax);
							_xscl = Traceur::Donnees::Expression::ReelSP(new Traceur::Donnees::Expression::Reel((_xmax->reel()-_xmin->reel())/(taille_x-_1)));
							_ymin = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(min->droite()); TRACEUR_TEST(_ymin);
							_ymax = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(max->droite()); TRACEUR_TEST(_ymax);
							_yscl = Traceur::Donnees::Expression::ReelSP(new Traceur::Donnees::Expression::Reel((_ymax->reel()-_ymin->reel())/(taille_y-_1)));
							if((_xmin==_xmax) or (_ymin==_ymax))
								throw Traceur::Donnees::Instruction::Valeur::image_mathematique_vide();
							_type=Type::MATHEMATIQUE;
							return;
						}
						if(std::dynamic_pointer_cast<Traceur::Donnees::Expression::SystemeStandard>(coordonnees))
							_type=Type::STANDARD;
						if(_origine==Origine::NORMAL)
						{
							_xmin = Traceur::Donnees::Expression::Nombre::zero();
							_xmax = Traceur::Donnees::Expression::NombreSP(new Traceur::Donnees::Expression::Entier(taille_x-_1));
							_ymin = Traceur::Donnees::Expression::Nombre::zero();
							_ymax = Traceur::Donnees::Expression::NombreSP(new Traceur::Donnees::Expression::Entier(taille_y-_1));
						}
						else
						{
							_xmin = Traceur::Donnees::Expression::NombreSP(new Traceur::Donnees::Expression::Entier(-(taille_x-_1)/_2));
							_xmax = Traceur::Donnees::Expression::NombreSP(new Traceur::Donnees::Expression::Entier( (taille_x-_1)/_2));
							_ymin = Traceur::Donnees::Expression::NombreSP(new Traceur::Donnees::Expression::Entier(-(taille_y-_1)/_2));
							_ymax = Traceur::Donnees::Expression::NombreSP(new Traceur::Donnees::Expression::Entier( (taille_y-_1)/_2));
						}
						_xscl = Traceur::Donnees::Expression::Nombre::un();
						_yscl = Traceur::Donnees::Expression::Nombre::un();
					};
					SystemeCoordonnees() = default;
					~SystemeCoordonnees() {};
					void format(std::ostream& os) const
					{
						if(_type==Type::STANDARD)
							os << "STANDARD";
						else
							os << "GEOMETRY";
					};
					void transforme_point(const Traceur::Donnees::Expression::PointSP& point, Coordonnee& x, Coordonnee& y) const
					{
						Traceur::Donnees::Expression::NombreSP abscisse = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(point->gauche()); TRACEUR_TEST(abscisse);
						Traceur::Donnees::Expression::NombreSP ordonnee = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(point->droite()); TRACEUR_TEST(abscisse);
						if(_type==Type::MATHEMATIQUE)
						{
							Traceur::Donnees::Expression::Entier abs((abscisse->reel()-_xmin->reel())/_xscl->reel());
							Traceur::Donnees::Expression::Entier ord((ordonnee->reel()-_ymin->reel())/_yscl->reel());
							x=abs._entier;
							y=ord._entier;
						}
						else
						{
							if(_origine==Origine::NORMAL)
							{
								x=abscisse->entier()._entier;
								y=ordonnee->entier()._entier;
							}
							else
							{
								x=abscisse->entier()._entier-_xmin->entier()._entier;
								y=ordonnee->entier()._entier-_ymin->entier()._entier;
							}
						}
					};
					bool egalite_points(const Traceur::Donnees::Expression::PointSP& point1, const Traceur::Donnees::Expression::PointSP& point2)
					{
						Coordonnee point1_x;
						Coordonnee point1_y;
						transforme_point(point1,point1_x,point1_y);
						Coordonnee point2_x;
						Coordonnee point2_y;
						transforme_point(point2,point2_x,point2_y);
						return (point1_x==point2_x) and (point1_y==point2_y);
					}
				private:
					Type _type;
					Origine _origine;
					Traceur::Donnees::Expression::EntierSP _xsiz;
					Traceur::Donnees::Expression::NombreSP _xmin;
					Traceur::Donnees::Expression::NombreSP _xmax;
					Traceur::Donnees::Expression::NombreSP _xscl;
					Traceur::Donnees::Expression::EntierSP _ysiz;
					Traceur::Donnees::Expression::NombreSP _ymin;
					Traceur::Donnees::Expression::NombreSP _ymax;
					Traceur::Donnees::Expression::NombreSP _yscl;
			};

		}
	}
}

#endif

