#ifndef _DONNEES_EXPRESSIONS_EXPRESSION_H_
#define _DONNEES_EXPRESSIONS_EXPRESSION_H_

#include <iostream>
#include <sstream>
#include <type_traits>

#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Expression { DECL_SHARED_PTR(Expression); DECL_WEAK_PTR(Expression); } } }

#include <src/exceptions/exceptions.h>
#include <src/donnees/sources/position.h>
#include <src/donnees/fonctions/fonctionimaginaire.h>
#include <src/donnees/instructions/valeur.h>
#include <src/donnees/types/type.h>
#include <src/donnees/types/typeatomique.h>
#include <src/donnees/types/typefonction.h>
#include <src/donnees/instructions/bloc.h>
#include <src/donnees/global/clonesinstances.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Assembleur { DECL_SHARED_PTR(Assembleur); DECL_SHARED_PTR(Instruction); }
		namespace Expression
		{
			class Expression: public std::enable_shared_from_this<Expression>
			{
				friend class Valeur;
				friend class Traceur::Donnees::Fonction::FonctionImaginaire;
				public:
					virtual ~Expression() {};
					template<typename TestType>
						bool est_de_type()
						{
							static_assert(std::is_base_of<Traceur::Donnees::Type::Atomique,TestType>::value,
									"le type TestType ne derive pas de Traceur::Donnees::Type::Atomique");
							return static_cast<bool>(std::dynamic_pointer_cast<TestType>(_type));
						}
					bool est_une_fonction()
					{
						return static_cast<bool>(std::dynamic_pointer_cast<Traceur::Donnees::Type::Fonction>(_type));
					};
					const Traceur::Donnees::Type::TypeSP type() const { return _type; }
					template<typename TestNature>
						bool est_de_nature()
						{
							return dynamic_cast<TestNature*>(this);
						}

					virtual void edition_liens(Traceur::Donnees::Instruction::InstructionSP instruction) = 0 ;
					virtual void definition_fonction() = 0 ;
					virtual ExpressionSP simplification()
					{
#ifdef TRACEUR_DEBUG
						std::cerr << __FILE__ << " "; format(std::cerr); std::cerr << std::endl;
#endif
						return this->shared_from_this();
					};
					virtual bool est_constante() { return false; };
					virtual ExpressionSP clone(Traceur::Donnees::Global::ClonesInstances& clones_instances) = 0;
					virtual Traceur::Donnees::Instruction::ValeurSP evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur) = 0;

					virtual Traceur::Donnees::Assembleur::InstructionSP draw(Traceur::Donnees::Assembleur::Assembleur& assembleur)
					{
						return Traceur::Donnees::Assembleur::InstructionSP();
					};
					virtual void format(std::ostream& os) const =0;
					operator std::string() const
					{
						std::ostringstream oss;
						format(oss);
						return oss.str();
					}
				protected:
					Traceur::Donnees::Type::TypeSP _type;
			};
		}
	}
}

namespace Traceur
{
	namespace Exception
	{
		class TypeIncorrect : public Compilation
		{
			public:
				TypeIncorrect(Traceur::Donnees::Source::PositionSP position, Traceur::Donnees::Expression::ExpressionSP expression,Traceur::Donnees::Type::TypeSP type = Traceur::Donnees::Type::TypeSP())
					:Compilation(position,std::string(*expression)+" has an incorrect type "+std::string(*(expression->type())))
					{
						if(type)
							_message += (std::string(" instead of ")+std::string(*type)); 
					};
			protected:
				TypeIncorrect(Traceur::Donnees::Source::PositionSP position, std::string message)
					:Compilation(position,message) {};
		};
	}
}
#endif
