#include <src/donnees/expressions/expressionappelfonctionutilisateur.h>
#include <src/donnees/fonctions/fonctionprototype.h>
#include <src/donnees/fonctions/fonctionimaginaire.h>
#ifdef TRACEUR_DEBUG
#include <src/donnees/expressions/expressionfonctionimaginaire.h>
#endif
#include <src/donnees/instructions/instructioninterruption.h>
#include <algorithm>
#include <functional>
using namespace Traceur::Donnees::Expression;

Traceur::Donnees::Instruction::ValeurSP AppelFonctionUtilisateur::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	Traceur::Donnees::Fonction::FonctionSP fonction = _valeur_dynamique;
	if(not fonction)
	{
		Traceur::Donnees::Fonction::InstanceSP instance = _instance/*.lock()*/;
		TRACEUR_TEST(instance);
		fonction=instance->valeur_appel();
	}
	TRACEUR_TEST(fonction);
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE << this->shared_from_this() << std::endl;
	TRACEUR_TRACE << fonction << std::endl;
#endif

	Traceur::Donnees::Fonction::PrototypeSP prototype = fonction->prototype();

	Traceur::Donnees::Fonction::Fonction::ListeParametres parametres_gauche;
	auto ittypes_gauche = prototype->_parametres_gauche.begin();
	for(auto iter=_parametres_gauche.begin() ; iter!=_parametres_gauche.end() ; ++iter, ++ittypes_gauche)
	{
		Traceur::Donnees::Instruction::ValeurSP retour = (*iter)->evaluation(assembleur);
		if(not retour->a_une_valeur_retour())
		{
			return retour;
		}
		Traceur::Donnees::Type::TypeSP type_attendu(new Traceur::Donnees::Type::Fonction(*ittypes_gauche));
		Traceur::Donnees::Fonction::FonctionImaginaireSP fonction_parametre_gauche(new Traceur::Donnees::Fonction::FonctionImaginaire(*retour,_position,type_attendu));
		parametres_gauche.push_back(fonction_parametre_gauche);
	}

	Traceur::Donnees::Fonction::Fonction::ListeParametres parametres_droite;
	auto ittypes_droite = prototype->_parametres_droite.begin();
	for(auto iter=_parametres_droite.begin() ; iter!=_parametres_droite.end() ; ++iter,++ittypes_droite)
	{
		Traceur::Donnees::Instruction::ValeurSP retour = (*iter)->evaluation(assembleur);
		if(not retour->a_une_valeur_retour())
		{
			return retour;
		}
		Traceur::Donnees::Type::TypeSP type_attendu(new Traceur::Donnees::Type::Fonction(*ittypes_droite));
		Traceur::Donnees::Fonction::FonctionImaginaireSP fonction_parametre_droite(new Traceur::Donnees::Fonction::FonctionImaginaire(*retour,_position,type_attendu));
		parametres_droite.push_back(fonction_parametre_droite);
	}

	AppelFonctionUtilisateurSP appel = std::static_pointer_cast<AppelFonctionUtilisateur>(this->shared_from_this());
	Traceur::Donnees::Instruction::InstructionInterruptionSP retour = fonction->appel(parametres_gauche,parametres_droite,assembleur,appel);
	Traceur::Donnees::Instruction::ValeurSP valeur(new Traceur::Donnees::Instruction::Valeur());
	valeur->_interruption=retour;
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE << retour << std::endl;
	ExpressionSP expr = *valeur;
	Traceur::Donnees::Expression::FonctionImaginaireSP fonc = std::dynamic_pointer_cast<Traceur::Donnees::Expression::FonctionImaginaire>(expr);
	if(fonc)
	{
		TRACEUR_TRACE << fonc << std::endl;
	}
	TRACEUR_TRACE << fonction << std::endl;
#endif
	return valeur;
}

void AppelFonctionUtilisateur::disp(std::ostream& os) const
{
	_instruction.lock()->disp(os);
	os << std::endl << std::string(*_position) << "function call to" << std::endl;
}

void AppelFonctionUtilisateur::format(std::ostream& os) const
{
	_type->format(os);
	os << " ";
	if(_parametres_gauche.size()>0)
	{
		os << "(" ;
		auto iter = _parametres_gauche.cbegin();
		(*iter)->format(os);
		for(++iter ; iter!=_parametres_gauche.cend() ; ++iter)
		{
			os << ",";
			(*iter)->format(os);
		}
		os << ")" ;
	}
	os << _nom_fonction;
	if(_parametres_droite.size()>0)
	{
		os << "(" ;
		auto iter = _parametres_droite.cbegin();
		(*iter)->format(os);
		for(++iter ; iter!=_parametres_droite.cend() ; ++iter)
		{
			os << ",";
			(*iter)->format(os);
		}
		os << ")" ;
	}
#ifdef TRACEUR_DEBUG
	Traceur::Donnees::Fonction::InstanceSP instance=_instance/*.lock()*/;
	if(instance)
	{
		os << ": ";
		instance->format(os);
	}
	if(_valeur_dynamique)
	{
		os << " directly linked to fonction " << _valeur_dynamique.get() << ": ";
		_valeur_dynamique->format(os);
	}
#endif
}

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			struct TransformationPrototype
			{
				Traceur::Donnees::Fonction::PrototypeSP operator()(ExpressionSP expression)
				{
					Traceur::Donnees::Fonction::FonctionImaginaireSP fonction(new Traceur::Donnees::Fonction::FonctionImaginaire(expression,Traceur::Donnees::Source::PositionSP()));
					Traceur::Donnees::Fonction::PrototypeSP prototype = fonction->_prototype;
					return prototype;
				}
			};
		}
	}
}

void AppelFonctionUtilisateur::edition_liens(Traceur::Donnees::Instruction::InstructionSP instruction)
{
	_instruction = instruction;
	std::for_each(_parametres_gauche.cbegin(),_parametres_gauche.cend(),bind(&Traceur::Donnees::Expression::Expression::edition_liens,std::placeholders::_1,instruction));
	std::for_each(_parametres_droite.cbegin(),_parametres_droite.cend(),bind(&Traceur::Donnees::Expression::Expression::edition_liens,std::placeholders::_1,instruction));

	Traceur::Donnees::Fonction::Prototype::ListeParametres gauche;
	std::transform(_parametres_gauche.cbegin(),_parametres_gauche.cend(),std::back_inserter(gauche),TransformationPrototype());

	Traceur::Donnees::Fonction::Prototype::ListeParametres droite;
	std::transform(_parametres_droite.cbegin(),_parametres_droite.cend(),std::back_inserter(droite),TransformationPrototype());

	Traceur::Donnees::Fonction::PrototypeSP prototype(new Traceur::Donnees::Fonction::Prototype(_type,gauche,_nom_fonction,droite));
	
	_instance = instruction->parent()->_declarations->appel(prototype,_position);
}

void AppelFonctionUtilisateur::definition_fonction()
{
	std::for_each(_parametres_gauche.cbegin(),_parametres_gauche.cend(),bind(&Traceur::Donnees::Expression::Expression::definition_fonction,std::placeholders::_1));
	std::for_each(_parametres_droite.cbegin(),_parametres_droite.cend(),bind(&Traceur::Donnees::Expression::Expression::definition_fonction,std::placeholders::_1));
	_valeur_dynamique=_instance/*.lock()*/->valeur_dynamique();
}

ExpressionSP AppelFonctionUtilisateur::simplification()
{
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE << this->shared_from_this() << std::endl;
#endif
	Traceur::Donnees::Fonction::FonctionSP fonction = _valeur_dynamique;
	if(not fonction)
	{
		Traceur::Donnees::Fonction::InstanceSP instance = _instance/*.lock()*/;
		TRACEUR_TEST(instance);
		fonction = instance->valeur_appel();
	}
	if(not fonction)
		return this->shared_from_this();

#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE << "On substitue" << std::endl;
#endif
	Traceur::Donnees::Fonction::Fonction::ListeParametres parametres_gauche;
	for(auto iter=_parametres_gauche.begin() ; iter!=_parametres_gauche.end() ; ++iter)
	{
		Traceur::Donnees::Global::ClonesInstances clones_instances;
		Traceur::Donnees::Expression::ExpressionSP clone = (*iter)->clone(clones_instances);
		Traceur::Donnees::Fonction::FonctionImaginaireSP fonction(new Traceur::Donnees::Fonction::FonctionImaginaire(clone,_position));
		parametres_gauche.push_back(fonction);
	}

	Traceur::Donnees::Fonction::Fonction::ListeParametres parametres_droite;
	for(auto iter=_parametres_droite.begin() ; iter!=_parametres_droite.end() ; ++iter)
	{
		Traceur::Donnees::Global::ClonesInstances clones_instances;
		Traceur::Donnees::Expression::ExpressionSP clone = (*iter)->clone(clones_instances);
		Traceur::Donnees::Fonction::FonctionImaginaireSP fonction(new Traceur::Donnees::Fonction::FonctionImaginaire(clone,_position));
		parametres_droite.push_back(fonction);
	}

	ExpressionSP expression = fonction->substitution(parametres_gauche,parametres_droite);
	if(expression)
		return expression;
	return this->shared_from_this();
}

ExpressionSP AppelFonctionUtilisateur::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	Traceur::Donnees::Instruction::InstructionSP instruction = clones_instances[_instruction.lock().get()];
	if(not instruction)
		instruction = _instruction.lock();

	AppelFonctionUtilisateurSP clone(new AppelFonctionUtilisateur(_position,instruction));
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	std::transform(_parametres_gauche.begin(),_parametres_gauche.end(),std::back_inserter(clone->_parametres_gauche),bind(&Traceur::Donnees::Expression::Expression::clone,std::placeholders::_1,std::ref(clones_instances)));
	std::transform(_parametres_droite.begin(),_parametres_droite.end(),std::back_inserter(clone->_parametres_droite),bind(&Traceur::Donnees::Expression::Expression::clone,std::placeholders::_1,std::ref(clones_instances)));
	clone->_nom_fonction=_nom_fonction;
	if(_instance/*.lock()*/)
	{
		Traceur::Donnees::Fonction::InstanceSP instance = clones_instances[_instance/*.lock()*/.get()];
		if(instance)
			clone->_instance=instance;
		else
			clone->_instance=_instance;
	}
	if(_valeur_dynamique)
	{
		Traceur::Donnees::Fonction::FonctionSP valeur = _valeur_dynamique->clone(clones_instances);
		clone->_valeur_dynamique=valeur;
	}
	return clone;
}
