#include <src/donnees/expressions/expressionarccosinus.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_ARCCOSINUS(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ArccosinusSP(new EvaluateurOperateurUnaire<Arccosinus,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<Arccosinus,Entier> : public Arccosinus
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							EntierSP valeur_valeur=std::static_pointer_cast<Entier>(expr_valeur);
							if((*valeur_valeur)<Entier(-1) or (*valeur_valeur)>Entier(1))
								return Valeur::hors_domaine();
							else
							{ SECURISE_VALEUR(ReelSP(new Reel(Entier::arccosinus(*valeur_valeur)))); }
						}
				};
			template<>
				class EvaluateurOperateurUnaire<Arccosinus,Reel> : public Arccosinus
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ReelSP valeur_valeur=std::static_pointer_cast<Reel>(expr_valeur);
							if((*valeur_valeur)<Reel(-1) or (*valeur_valeur)>Reel(1))
								return Valeur::hors_domaine();
							else
							{ SECURISE_VALEUR(ReelSP(new Reel(Reel::arccosinus(*valeur_valeur)))); }
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Arccosinus::Arccosinus(ExpressionSP valeur,Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ArccosinusSP(new EvaluateurOperateurUnaire<Arccosinus,Expression>()))
{

	EVALUATEUR_ARCCOSINUS(Entier);
	EVALUATEUR_ARCCOSINUS(Reel  );
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionPrefixe(position,"acos",valeur);
}

ValeurSP Arccosinus::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	Arccosinus resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP Arccosinus::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void Arccosinus::format(std::ostream& os) const
{
	os << "acos (";
	_valeur->format(os);
	os << ")";
}

ExpressionSP Arccosinus::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ArccosinusSP clone(new Arccosinus(valeur));
	return clone;
}
