#include <src/donnees/expressions/expressionarcsinus.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_ARCSINUS(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ArcsinusSP(new EvaluateurOperateurUnaire<Arcsinus,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<Arcsinus,Entier> : public Arcsinus
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							EntierSP valeur_valeur=std::static_pointer_cast<Entier>(expr_valeur);
							if((*valeur_valeur)<Entier(-1) or (*valeur_valeur)>Entier(1))
								return Valeur::hors_domaine();
							else
							{ SECURISE_VALEUR(ReelSP(new Reel(Entier::arcsinus(*valeur_valeur)))); }
						}
				};
			template<>
				class EvaluateurOperateurUnaire<Arcsinus,Reel> : public Arcsinus
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ReelSP valeur_valeur=std::static_pointer_cast<Reel>(expr_valeur);
							if((*valeur_valeur)<Reel(-1) or (*valeur_valeur)>Reel(1))
								return Valeur::hors_domaine();
							else
							{ SECURISE_VALEUR(ReelSP(new Reel(Reel::arcsinus(*valeur_valeur)))); }
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Arcsinus::Arcsinus(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ArcsinusSP(new EvaluateurOperateurUnaire<Arcsinus,Expression>()))
{

	EVALUATEUR_ARCSINUS(Entier);
	EVALUATEUR_ARCSINUS(Reel  );
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionPrefixe(position,"asin",valeur);
}

ValeurSP Arcsinus::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	Arcsinus resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP Arcsinus::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void Arcsinus::format(std::ostream& os) const
{
	os << "asin (";
	_valeur->format(os);
	os << ")";
}

ExpressionSP Arcsinus::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ArcsinusSP clone(new Arcsinus(valeur));
	return clone;
}
