#include <src/donnees/expressions/expressionarctangente.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteur.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_ARCTANGENTE(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ArctangenteSP(new EvaluateurOperateurUnaire<Arctangente,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<Arctangente,Entier> : public Arctangente
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							EntierSP valeur_valeur=std::static_pointer_cast<Entier>(expr_valeur);
							SECURISE_VALEUR(ReelSP(new Reel(Entier::arctangente(*valeur_valeur))));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<Arctangente,Reel> : public Arctangente
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ReelSP valeur_valeur=std::static_pointer_cast<Reel>(expr_valeur);
							SECURISE_VALEUR(ReelSP(new Reel(Reel::arctangente(*valeur_valeur))));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<Arctangente,Point> : public Arctangente
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							PointSP valeur_valeur=std::static_pointer_cast<Point>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->argument());
						}
				};
			template<>
				class EvaluateurOperateurUnaire<Arctangente,Vecteur> : public Arctangente
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							VecteurSP valeur_valeur=std::static_pointer_cast<Vecteur>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->argument());
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Arctangente::Arctangente(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ArctangenteSP(new EvaluateurOperateurUnaire<Arctangente,Expression>()))
{

	EVALUATEUR_ARCTANGENTE(Entier );
	EVALUATEUR_ARCTANGENTE(Reel   );
	EVALUATEUR_ARCTANGENTE(Point  );
	EVALUATEUR_ARCTANGENTE(Vecteur);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Point>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Vector>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionPrefixe(position,"atan",valeur);
}

ValeurSP Arctangente::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	Arctangente resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP Arctangente::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void Arctangente::format(std::ostream& os) const
{
	os << "atan (";
	_valeur->format(os);
	os << ")";
}

ExpressionSP Arctangente::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ArctangenteSP clone(new Arctangente(valeur));
	return clone;
}
