#include <src/donnees/expressions/expressioncatch.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressions.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_CATCH(type1,type2) \
	if(_essai->est_de_nature<type1>() and _clause_catch->est_de_nature<type2>()) { _evaluateur = CatchSP(new EvaluateurOperateurCatch<Catch,type1,type2>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<typename OperateurCatch, typename TypeGauche, typename TypeDroite>
				class EvaluateurOperateurCatch : public OperateurCatch
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP essai, const Traceur::Donnees::Fonction::Identifiant nom_erreur, const ValeurSP clause_catch) const
						{
							return essai;
						}
				};
			template<>
				class EvaluateurOperateurCatch<Catch,Erreur,Expression>: public Catch
				{
						virtual ValeurSP evaluation(const ValeurSP essai, const Traceur::Donnees::Fonction::Identifiant nom_erreur, const ValeurSP clause_catch) const
						{
							ExpressionSP expr_essai(*essai);
							ErreurSP erreur = std::dynamic_pointer_cast<Erreur>(expr_essai);
							TRACEUR_TEST(erreur);
							if((nom_erreur.empty()) or (erreur->_erreur==nom_erreur))
							{
								return clause_catch;
							}
							return essai;
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Catch::Catch(ExpressionSP essai, Traceur::Donnees::Fonction::Identifiant nom_erreur, ExpressionSP clause_catch, Traceur::Donnees::Source::PositionSP position)
:_essai(essai),_nom_erreur(nom_erreur),_clause_catch(clause_catch), _position(position), _evaluateur(CatchSP(new EvaluateurOperateurCatch<Catch,Expression,Expression>()))
{
	aide_construction();
}

	Catch::Catch(ExpressionSP essai, ExpressionSP clause_catch, Traceur::Donnees::Source::PositionSP position)
:_essai(essai),_nom_erreur(""),_clause_catch(clause_catch), _position(position), _evaluateur(CatchSP(new EvaluateurOperateurCatch<Catch,Expression,Expression>()))
{
	aide_construction();
}

void Catch::aide_construction()
{
	EVALUATEUR_CATCH(Erreur,Expression);

	// calcul du type de retour
	if(_essai->type()->compatible(_clause_catch->type()))
	{
		_type = _essai->type();
		return;
	}
	if(_essai->est_de_type<Traceur::Donnees::Type::Error>())
	{
		_type = _clause_catch->type();
		return;
	}
	throw Traceur::Exception::TypeIncorrectCatch(_position,_essai,_clause_catch);
}

ValeurSP Catch::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_essai = _essai->evaluation(assembleur);
	if(not valeur_essai->est_une_erreur())
		return valeur_essai;
	ValeurSP valeur_clause_catch = _clause_catch->evaluation(assembleur);
	Catch resultat(*valeur_essai,_nom_erreur,*valeur_clause_catch,_position);
	return resultat.evaluation(valeur_essai,_nom_erreur,valeur_clause_catch);
}

ValeurSP Catch::evaluation(const ValeurSP essai, const Traceur::Donnees::Fonction::Identifiant nom_erreur, const ValeurSP clause_catch) const
{
	return _evaluateur->evaluation(essai,nom_erreur,clause_catch);
}

void Catch::format(std::ostream& os) const
{
	//_type->format(os);
	os << "CATCH[";
	_essai->format(os);
	if(not _nom_erreur.empty())
		os << ";" << _nom_erreur;
	os << ";";
	_clause_catch->format(os);
	os << "]";
}

void Catch::edition_liens(Traceur::Donnees::Instruction::InstructionSP instruction)
{
	_essai->edition_liens(instruction);
	_clause_catch->edition_liens(instruction);
}

void Catch::definition_fonction()
{
	_essai->definition_fonction();
	_clause_catch->definition_fonction();
}

ExpressionSP Catch::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP essai = _essai->clone(clones_instances);
	ExpressionSP clause_catch = _clause_catch->clone(clones_instances);
	CatchSP clone(new Catch());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_essai=essai;
	clone->_nom_erreur=_nom_erreur;
	clone->_clause_catch=clause_catch;
	return clone;
}

