#include <src/donnees/expressions/expressioncercle.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionnombre.h>
#include <src/donnees/expressions/expressionvecteur.h>
#include <src/donnees/expressions/expressionvecteurpolaire.h>
#include <src/donnees/assembleur/point.h>
#include <src/donnees/assembleur/drawcircle.h>
#include <src/donnees/assembleur/image.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_CERCLE(type1,type2) \
	if(centre->est_de_nature<type1>() and rayon->est_de_nature<type2>()) { _evaluateur = CercleSP(new EvaluateurOperateurBinaire<type1,Cercle,type2>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurBinaire<Point,Cercle,Nombre> : public Cercle
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							PointSP valeur_gauche=std::static_pointer_cast<Point>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							NombreSP valeur_droite=std::static_pointer_cast<Nombre>(expr_droite);
							if(valeur_droite<Nombre::zero())
								return Traceur::Donnees::Instruction::Valeur::hors_domaine();
							SECURISE_VALEUR(CercleSP(new Cercle(valeur_gauche,valeur_droite,_options)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Point,Cercle,Vecteur> : public Cercle
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							PointSP valeur_gauche=std::static_pointer_cast<Point>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							VecteurSP valeur_droite=std::static_pointer_cast<Vecteur>(expr_droite);
							SECURISE_VALEUR(CercleSP(new Cercle(valeur_gauche,valeur_droite,_options)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Point,Cercle,VecteurPolaire> : public Cercle
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							PointSP valeur_gauche=std::static_pointer_cast<Point>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							VecteurPolaireSP valeur_droite=std::static_pointer_cast<VecteurPolaire>(expr_droite);
							SECURISE_VALEUR(CercleSP(new Cercle(valeur_gauche,valeur_droite,_options)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Point,Cercle,Point> : public Cercle
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							PointSP valeur_gauche=std::static_pointer_cast<Point>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							PointSP valeur_droite=std::static_pointer_cast<Point>(expr_droite);
							SECURISE_VALEUR(CercleSP(new Cercle(valeur_gauche,valeur_droite,_options)));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Cercle::Cercle(ExpressionSP centre, ExpressionSP rayon, OptionsSurfaceSP options, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneBinaire(centre,rayon), _evaluateur(CercleSP(new EvaluateurOperateurBinaire<Expression,Cercle,Expression>())), _options(options)
{
	EVALUATEUR_CERCLE(Point,Nombre        );
	EVALUATEUR_CERCLE(Point,Vecteur       );
	EVALUATEUR_CERCLE(Point,VecteurPolaire);
	EVALUATEUR_CERCLE(Point,Point         );
	// calcul du type de retour
	if(_type)
		return;
	if(centre->est_de_type<Traceur::Donnees::Type::Point>())
	{
		if(rayon->est_de_type<Traceur::Donnees::Type::Number>())
		{
			_type = Traceur::Donnees::Type::Circle::SP();
			return;
		}
		if(rayon->est_de_type<Traceur::Donnees::Type::Vector>())
		{
			_type = Traceur::Donnees::Type::Circle::SP();
			return;
		}
		if(rayon->est_de_type<Traceur::Donnees::Type::Point>())
		{
			_type = Traceur::Donnees::Type::Circle::SP();
			return;
		}
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionDeux(position,"CIRCLE",centre,rayon);
}

ValeurSP Cercle::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_gauche = _gauche->evaluation(assembleur); if(not valeur_gauche->a_une_valeur_retour()) return valeur_gauche;
	ValeurSP valeur_droite = _droite->evaluation(assembleur); if(not valeur_droite->a_une_valeur_retour()) return valeur_droite;
	try
	{
		OptionsSurfaceSP options = _options->evaluation(assembleur);
		Cercle resultat(*valeur_gauche,*valeur_droite,options);
		return resultat.evaluation(valeur_gauche,valeur_droite,options);
	} catch(ValeurSP valeur) { return valeur; }
}

ValeurSP Cercle::evaluation(const ValeurSP gauche, const ValeurSP droite, const OptionsSurfaceSP options) const
{
	_evaluateur->_options=options;
	return _evaluateur->evaluation(gauche,droite);
}

ValeurSP Cercle::evaluation(const ValeurSP gauche, const ValeurSP droite) const
{
	return _evaluateur->evaluation(gauche,droite);
}

void Cercle::format(std::ostream& os) const
{
	//_type->format(os);
	os << "CIRCLE[";
	_gauche->format(os);
	os << ";";
	_droite->format(os);
	_options->format(os);
	os << "]";
}

ExpressionSP Cercle::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP gauche = _gauche->clone(clones_instances);
	ExpressionSP droite = _droite->clone(clones_instances);
	CercleSP clone(new Cercle());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_gauche=gauche;
	clone->_droite=droite;
	clone->_options = _options->clone(clones_instances);
	return clone;
}

Traceur::Donnees::Assembleur::InstructionSP Cercle::draw(Traceur::Donnees::Assembleur::Assembleur& assembleur)
{
	PointSP expr_centre = std::static_pointer_cast<Point>(_gauche);
	Traceur::Donnees::Assembleur::Point centre(expr_centre,assembleur._pile.top()->_systeme_coordonnees);
	Traceur::Donnees::Assembleur::Coordonnee rayon = 0; // pour eviter un warning
	if(_droite->est_de_nature<Nombre>()) // tout ce barda pour assurer au mieux quand le systeme de coordonnees n'est pas orthonorme
	{
		NombreSP expr_rayon = std::static_pointer_cast<Nombre>(_droite);
		Vecteur vecteur(expr_rayon,Nombre::zero());
		PointSP expr_point(new Point(expr_centre->translation(vecteur)));
		Traceur::Donnees::Assembleur::Point point(expr_point,assembleur._pile.top()->_systeme_coordonnees);
		rayon = centre.distance(point);
	}
	if(_droite->est_de_nature<Vecteur>())
	{
		VecteurSP expr_rayon = std::dynamic_pointer_cast<Vecteur>(_droite);
		PointSP expr_point(new Point(expr_centre->translation(*expr_rayon)));
		Traceur::Donnees::Assembleur::Point point(expr_point,assembleur._pile.top()->_systeme_coordonnees);
		rayon = centre.distance(point);
	}
	if(_droite->est_de_nature<Point>())
	{
		PointSP expr_rayon = std::static_pointer_cast<Point>(_droite);
		Traceur::Donnees::Assembleur::Point point(expr_rayon,assembleur._pile.top()->_systeme_coordonnees);
		rayon = centre.distance(point);
	}
	Traceur::Donnees::Assembleur::InstructionSP drawcircle(new Traceur::Donnees::Assembleur::DrawCircle(centre,rayon,*_options));
	return drawcircle;
}
