#include <src/donnees/expressions/expressionchaine.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <regex.h> // regex non supporte in g++4.4
#include <cstdio>
using namespace Traceur::Donnees::Expression;

Chaine Chaine::operator+(Chaine chaine)
{
	return Chaine(_chaine+chaine._chaine);
}
bool Chaine::operator==(Chaine chaine)
{
	return (_chaine==chaine._chaine);
}
bool Chaine::operator!=(Chaine chaine)
{
	return (_chaine!=chaine._chaine);
}
bool Chaine::operator<(Chaine chaine)
{
	return (_chaine<chaine._chaine);
}
bool Chaine::operator>(Chaine chaine)
{
	return (_chaine>chaine._chaine);
}
bool Chaine::operator<=(Chaine chaine)
{
	return (_chaine<=chaine._chaine);
}
bool Chaine::operator>=(Chaine chaine)
{
	return (_chaine>=chaine._chaine);
}
Entier Chaine::Chaine::operator/(Entier entier)
{
	return Entier(_chaine[entier._entier]);
}

Chaine Chaine::operator/(Chaine motif)
{
	std::string resultat;
	regex_t re;
	regmatch_t match;
	::regcomp(&re,motif._chaine.c_str(),REG_EXTENDED);
	if(::regexec(&re,_chaine.c_str(),1,&match,0)==0)
	{
		resultat = std::string(_chaine,match.rm_so,match.rm_eo-match.rm_so);
	}
	::regfree(&re);
	return Chaine(resultat);
}

Chaine Chaine::operator%(Chaine motif)
{
	std::string resultat = _chaine;
	regex_t re;
	regmatch_t match;
	::regcomp(&re,motif._chaine.c_str(),REG_EXTENDED);
	if(::regexec(&re,_chaine.c_str(),1,&match,0)==0)
	{
		resultat = std::string(_chaine,match.rm_eo);
	}
	::regfree(&re);
	return Chaine(resultat);
}

Entier Chaine::longueur()
{
	return Entier(_chaine.size());
}

bool Chaine::represente_un_entier()
{
	std::string resultat = _chaine;
	regex_t re;
	regmatch_t match;
	::regcomp(&re,"^[+-]?[0-9]+$",REG_EXTENDED);
	if(::regexec(&re,_chaine.c_str(),1,&match,0)==0)
	{
		::regfree(&re);
		return match.rm_so==0 and (match.rm_eo)==static_cast<int>(_chaine.size());
	}
	::regfree(&re);
	return false;
}

bool Chaine::represente_un_reel()
{
	std::string resultat = _chaine;
	regex_t re;
	regmatch_t match;
	::regcomp(&re,"^[+-]?[0-9]+\\.[0-9]*([eE][+-]?[0-9]+)?$",REG_EXTENDED);
	if(::regexec(&re,_chaine.c_str(),1,&match,0)==0)
	{
		::regfree(&re);
		return match.rm_so==0 and (match.rm_eo)==static_cast<int>(_chaine.size());
	}
	::regfree(&re);
	return false;
}

bool Chaine::est_un_format_entier(const std::string& format)
{
	if(format.empty())
		return false;
	switch(format[format.size()-1])
	{
		case 'd': case 'i': case 'o': case 'x': case 'X': return true;
	}
	return false;
}

bool Chaine::est_un_format_reel(const std::string& format)
{
	if(format.empty())
		return false;
	switch(format[format.size()-1])
	{
		case 'e': case 'E': case 'f': case 'F': return true;
	}
	return false;
}

bool Chaine::est_un_format_chaine(const std::string& format)
{
	if(format.empty())
		return false;
	switch(format[format.size()-1])
	{
		case 's': return true;
	}
	return false;
}

Chaine::Chaine(const Entier& entier, const std::string& format)
{
	_type = Traceur::Donnees::Type::String::SP();
	char buffer[256];
	::snprintf(buffer,255,(std::string("%")+format).c_str(),entier._entier);
	_chaine = TypeInterne(buffer);
}

Chaine::Chaine(const Reel& reel, const std::string& format)
{
	_type = Traceur::Donnees::Type::String::SP();
	char buffer[256];
	::snprintf(buffer,255,(std::string("%")+format).c_str(),reel._reel);
	_chaine = TypeInterne(buffer);
}


Traceur::Donnees::Instruction::ValeurSP Chaine::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	return Traceur::Donnees::Instruction::ValeurSP(new Traceur::Donnees::Instruction::Valeur(ChaineSP(new Chaine(*this))));
}

void Chaine::format(std::ostream& os) const
{
	os << "\"" << _chaine << "\"";
}
