#ifndef _DONNEES_EXPRESSIONS_CHAINE_H_
#define _DONNEES_EXPRESSIONS_CHAINE_H_

#include <string>

#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Expression { DECL_SHARED_PTR(Chaine); } } }

#include <src/donnees/expressions/expressionconstante.h>
#include <src/donnees/expressions/expressionbooleenvrai.h>
#include <src/donnees/expressions/expressionbooleenfaux.h>
#include <src/donnees/types/typestring.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Assembleur { class NomFichier; }
		namespace Expression
		{
			class Entier;
			class Reel;
			class Chaine: public Constante
			{
				friend class Entier;
				friend class Reel;
				friend class Traceur::Donnees::Assembleur::NomFichier;
				public:
					typedef std::string TypeInterne;
					Chaine(Chaine::TypeInterne chaine)
						:_chaine(chaine) { _type=Traceur::Donnees::Type::String::SP(); };
					Chaine(const Entier& entier, const std::string& format);
					Chaine(const Reel& reel, const std::string& format);
					virtual Traceur::Donnees::Instruction::ValeurSP evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur);
					virtual void format(std::ostream& os) const;
					virtual ~Chaine() {};
					Chaine operator+(Chaine chaine);
					Entier operator/(Entier entier);
					Chaine operator/(Chaine motif);
					Chaine operator%(Chaine motif);
					Entier longueur();
					bool operator==(Chaine chaine);
					bool operator!=(Chaine chaine);
					bool operator<(Chaine chaine);
					bool operator>(Chaine chaine);
					bool operator<=(Chaine chaine);
					bool operator>=(Chaine chaine);
					bool represente_un_entier();
					bool represente_un_reel();
					static bool est_un_format_entier(const std::string& format);
					static bool est_un_format_reel(const std::string& format);
					static bool est_un_format_chaine(const std::string& format);
				private:
					Chaine::TypeInterne _chaine;
			};
		}
	}
}

#endif
