#include <src/donnees/expressions/expressionconversionalpha.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/expressions/expressioncouleurhsva.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_CONVERSIONALPHA(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ConversionAlphaSP(new EvaluateurOperateurUnaire<ConversionAlpha,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<ConversionAlpha,CouleurRVBA> : public ConversionAlpha
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							CouleurRVBASP valeur_valeur=std::static_pointer_cast<CouleurRVBA>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->alpha());
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionAlpha,CouleurHSVA> : public ConversionAlpha
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							CouleurHSVASP valeur_valeur=std::static_pointer_cast<CouleurHSVA>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->alpha());
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	ConversionAlpha::ConversionAlpha(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ConversionAlphaSP(new EvaluateurOperateurUnaire<ConversionAlpha,Expression>()))
{

	EVALUATEUR_CONVERSIONALPHA(CouleurRVBA);
	EVALUATEUR_CONVERSIONALPHA(CouleurHSVA);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Color>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionUn(position,"ALPHA",valeur);
}

ValeurSP ConversionAlpha::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	ConversionAlpha resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP ConversionAlpha::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void ConversionAlpha::format(std::ostream& os) const
{
	os << "ALPHA[";
	_valeur->format(os);
	os << "]";
}

ExpressionSP ConversionAlpha::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ConversionAlphaSP clone(new ConversionAlpha(valeur));
	return clone;
}
