#include <src/donnees/expressions/expressionconversionbleu.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/expressions/expressioncouleurhsva.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_CONVERSIONBLEU(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ConversionBleuSP(new EvaluateurOperateurUnaire<ConversionBleu,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<ConversionBleu,CouleurRVBA> : public ConversionBleu
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							CouleurRVBASP valeur_valeur=std::static_pointer_cast<CouleurRVBA>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->bleu());
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionBleu,CouleurHSVA> : public ConversionBleu
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							CouleurHSVASP valeur_valeur=std::static_pointer_cast<CouleurHSVA>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->bleu());
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	ConversionBleu::ConversionBleu(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ConversionBleuSP(new EvaluateurOperateurUnaire<ConversionBleu,Expression>()))
{

	EVALUATEUR_CONVERSIONBLEU(CouleurRVBA);
	EVALUATEUR_CONVERSIONBLEU(CouleurHSVA);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Color>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionUn(position,"BLUE",valeur);
}

ValeurSP ConversionBleu::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	ConversionBleu resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP ConversionBleu::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void ConversionBleu::format(std::ostream& os) const
{
	os << "BLUE[";
	_valeur->format(os);
	os << "]";
}

ExpressionSP ConversionBleu::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ConversionBleuSP clone(new ConversionBleu(valeur));
	return clone;
}
