#include <src/donnees/expressions/expressionconversionchaine.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionchaine.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_CONVERSIONCHAINE(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ConversionChaineSP(new EvaluateurOperateurUnaire<ConversionChaine,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<ConversionChaine,Entier> : public ConversionChaine
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							EntierSP valeur_valeur=std::static_pointer_cast<Entier>(expr_valeur);
							if(Chaine::est_un_format_entier(_format))
							{ SECURISE_VALEUR(ChaineSP(new Chaine(*valeur_valeur,_format))); }
							else
								return Valeur::format_invalide();
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionChaine,Reel> : public ConversionChaine
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ReelSP valeur_valeur=std::static_pointer_cast<Reel>(expr_valeur);
							if(Chaine::est_un_format_reel(_format))
							{ SECURISE_VALEUR(ChaineSP(new Chaine(*valeur_valeur,_format))); }
							else
								return Valeur::format_invalide();
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionChaine,Chaine> : public ConversionChaine
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							if(Chaine::est_un_format_chaine(_format))
								return valeur;
							else
								return Valeur::format_invalide();
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	ConversionChaine::ConversionChaine(ExpressionSP valeur, std::string format, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ConversionChaineSP(new EvaluateurOperateurUnaire<ConversionChaine,Expression>())), _format(format)
{

	EVALUATEUR_CONVERSIONCHAINE(Entier);
	EVALUATEUR_CONVERSIONCHAINE(Reel  );
	EVALUATEUR_CONVERSIONCHAINE(Chaine);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::String::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::String>())
	{
		_type = Traceur::Donnees::Type::String::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionUnPlusChaine(position,"STRING",valeur,format);
}

ValeurSP ConversionChaine::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	ConversionChaine resultat(*valeur_valeur,_format);
	return resultat.evaluation(valeur_valeur,_format);
}

ValeurSP ConversionChaine::evaluation(const ValeurSP valeur, const std::string format) const
{
	_evaluateur->_format=format;
	return _evaluateur->evaluation(valeur);
}

Traceur::Donnees::Instruction::ValeurSP ConversionChaine::evaluation(const Traceur::Donnees::Instruction::ValeurSP) const
{
	throw;
}

void ConversionChaine::format(std::ostream& os) const
{
	os << "STRING[";
	_valeur->format(os);
	if(not _format.empty())
	{
		os << ";~" << _format;
	}
	os << "]";
}

ExpressionSP ConversionChaine::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ConversionChaineSP clone(new ConversionChaine(valeur,_format));
	return clone;
}
