#include <src/donnees/expressions/expressionconversioncouleur.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/expressions/expressioncouleurhsva.h>
#include <src/donnees/expressions/expressionpoint.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_CONVERSIONCOULEUR(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ConversionCouleurSP(new EvaluateurOperateurUnaire<ConversionCouleur,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<ConversionCouleur,CouleurRVBA> : public ConversionCouleur
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur)
						{
							return valeur;
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionCouleur,CouleurHSVA> : public ConversionCouleur
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur)
						{
							ExpressionSP expr_valeur(*valeur);
							CouleurHSVASP valeur_valeur=std::static_pointer_cast<CouleurHSVA>(expr_valeur);
							SECURISE_VALEUR(CouleurRVBASP(new CouleurRVBA(*valeur_valeur)));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionCouleur,Point> : public ConversionCouleur
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur)
						{
							ExpressionSP expr_valeur(*valeur);
							PointSP valeur_valeur=std::static_pointer_cast<Point>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->couleur(_assembleur));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	ConversionCouleur::ConversionCouleur(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ConversionCouleurSP(new EvaluateurOperateurUnaire<ConversionCouleur,Expression>()))
{

	EVALUATEUR_CONVERSIONCOULEUR(CouleurRVBA);
	EVALUATEUR_CONVERSIONCOULEUR(CouleurHSVA);
	EVALUATEUR_CONVERSIONCOULEUR(Point      );
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Color>())
	{
		_type = Traceur::Donnees::Type::Color::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Point>())
	{
		_type = Traceur::Donnees::Type::Color::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionUn(position,"COLOR",valeur);
}

ValeurSP ConversionCouleur::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	ConversionCouleur resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur,assembleur);
}

ValeurSP ConversionCouleur::evaluation(const ValeurSP valeur, Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	_evaluateur->_assembleur=assembleur;
	return _evaluateur->evaluation(valeur);
}

void ConversionCouleur::format(std::ostream& os) const
{
	os << "COLOR[";
	_valeur->format(os);
	os << "]";
}

ExpressionSP ConversionCouleur::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ConversionCouleurSP clone(new ConversionCouleur(valeur));
	return clone;
}
