#include <src/donnees/expressions/expressionconversioncouleuralpha.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/expressions/expressioncouleurhsva.h>
#include <src/donnees/expressions/expressionnombre.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_CONVERSIONCOULEURALPHA(type1,type2) \
	if((couleur->est_de_nature<type1>() and (alpha->est_de_nature<type2>()))) { _evaluateur = ConversionCouleurAlphaSP(new EvaluateurOperateurBinaire<type1,ConversionCouleurAlpha,type2>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurBinaire<CouleurRVBA,ConversionCouleurAlpha,Nombre> : public ConversionCouleurAlpha
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurRVBASP valeur_gauche=std::static_pointer_cast<CouleurRVBA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							NombreSP valeur_droite=std::static_pointer_cast<Nombre>(expr_droite);
							SECURISE_VALEUR(CouleurRVBASP(new CouleurRVBA(valeur_gauche,valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurHSVA,ConversionCouleurAlpha,Nombre> : public ConversionCouleurAlpha
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurHSVASP valeur_gauche=std::static_pointer_cast<CouleurHSVA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							NombreSP valeur_droite=std::static_pointer_cast<Nombre>(expr_droite);
							SECURISE_VALEUR(CouleurHSVASP(new CouleurHSVA(valeur_gauche,valeur_droite)));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	ConversionCouleurAlpha::ConversionCouleurAlpha(ExpressionSP couleur, ExpressionSP alpha, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneBinaire(couleur,alpha), _evaluateur(ConversionCouleurAlphaSP(new EvaluateurOperateurBinaire<Expression,ConversionCouleurAlpha,Expression>()))
{
	EVALUATEUR_CONVERSIONCOULEURALPHA(CouleurRVBA,Nombre);
	EVALUATEUR_CONVERSIONCOULEURALPHA(CouleurHSVA,Nombre);
	// calcul du type de retour
	if(_type)
		return;
	if((couleur->est_de_type<Traceur::Donnees::Type::Color>()) and (alpha->est_de_type<Traceur::Donnees::Type::Number>()))
	{
		_type = Traceur::Donnees::Type::Color::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionDeux(position,"COLOR",couleur,alpha);
}

ValeurSP ConversionCouleurAlpha::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_gauche = _gauche->evaluation(assembleur); if(not valeur_gauche->a_une_valeur_retour()) return valeur_gauche;
	ValeurSP valeur_droite = _droite->evaluation(assembleur); if(not valeur_droite->a_une_valeur_retour()) return valeur_droite;
	ConversionCouleurAlpha resultat(*valeur_gauche,*valeur_droite);
	return resultat.evaluation(valeur_gauche,valeur_droite);
}

ValeurSP ConversionCouleurAlpha::evaluation(const ValeurSP gauche, const ValeurSP droite) const
{
	return _evaluateur->evaluation(gauche,droite);
}

void ConversionCouleurAlpha::format(std::ostream& os) const
{
	os << "COLOR[";
	_gauche->format(os);
	os << ";";
	_droite->format(os);
	os << "]";
}

ExpressionSP ConversionCouleurAlpha::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP gauche = _gauche->clone(clones_instances);
	ExpressionSP droite = _droite->clone(clones_instances);
	ConversionCouleurAlphaSP clone(new ConversionCouleurAlpha(gauche,droite));
	return clone;
}
