#include <src/donnees/expressions/expressionconversionentier.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionchaine.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_CONVERSIONENTIER(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ConversionEntierSP(new EvaluateurOperateurUnaire<ConversionEntier,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<ConversionEntier,Entier> : public ConversionEntier
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							return valeur;
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionEntier,Reel> : public ConversionEntier
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ReelSP valeur_valeur=std::static_pointer_cast<Reel>(expr_valeur);
							SECURISE_VALEUR(EntierSP(new Entier(*valeur_valeur)));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionEntier,Chaine> : public ConversionEntier
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ChaineSP valeur_valeur=std::static_pointer_cast<Chaine>(expr_valeur);
							if(valeur_valeur->represente_un_entier())
							{ SECURISE_VALEUR(EntierSP(new Entier(*valeur_valeur))); }
							else
								return Valeur::conversion_entiere_invalide();
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	ConversionEntier::ConversionEntier(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ConversionEntierSP(new EvaluateurOperateurUnaire<ConversionEntier,Expression>()))
{

	EVALUATEUR_CONVERSIONENTIER(Entier);
	EVALUATEUR_CONVERSIONENTIER(Reel  );
	EVALUATEUR_CONVERSIONENTIER(Chaine);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Integer::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::String>())
	{
		_type = Traceur::Donnees::Type::Integer::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionUn(position,"INTEGER",valeur);
}

ValeurSP ConversionEntier::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	ConversionEntier resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP ConversionEntier::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void ConversionEntier::format(std::ostream& os) const
{
	os << "INTEGER[";
	_valeur->format(os);
	os << "]";
}

ExpressionSP ConversionEntier::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ConversionEntierSP clone(new ConversionEntier(valeur));
	return clone;
}
