#include <src/donnees/expressions/expressionconversionhsv.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/expressions/expressioncouleurhsva.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_CONVERSIONENTIER(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ConversionHSVSP(new EvaluateurOperateurUnaire<ConversionHSV,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<ConversionHSV,CouleurRVBA> : public ConversionHSV
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							CouleurRVBASP valeur_valeur=std::static_pointer_cast<CouleurRVBA>(expr_valeur);
							SECURISE_VALEUR(CouleurHSVASP(new CouleurHSVA(*valeur_valeur)));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionHSV,CouleurHSVA> : public ConversionHSV
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							return valeur;
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	ConversionHSV::ConversionHSV(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ConversionHSVSP(new EvaluateurOperateurUnaire<ConversionHSV,Expression>()))
{

	EVALUATEUR_CONVERSIONENTIER(CouleurRVBA);
	EVALUATEUR_CONVERSIONENTIER(CouleurHSVA);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Color>())
	{
		_type = Traceur::Donnees::Type::Color::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionUn(position,"HSV",valeur);
}

ValeurSP ConversionHSV::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	ConversionHSV resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP ConversionHSV::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void ConversionHSV::format(std::ostream& os) const
{
	os << "HSV[";
	_valeur->format(os);
	os << "]";
}

ExpressionSP ConversionHSV::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ConversionHSVSP clone(new ConversionHSV(valeur));
	return clone;
}
