#include <src/donnees/expressions/expressionconversionpoint.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteur.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_CONVERSIONPOINT(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ConversionPointSP(new EvaluateurOperateurUnaire<ConversionPoint,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<ConversionPoint,Point> : public ConversionPoint
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							return valeur;
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionPoint,Vecteur> : public ConversionPoint
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							VecteurSP valeur_valeur=std::static_pointer_cast<Vecteur>(expr_valeur);
							SECURISE_VALEUR(PointSP(new Point(*valeur_valeur)));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	ConversionPoint::ConversionPoint(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ConversionPointSP(new EvaluateurOperateurUnaire<ConversionPoint,Expression>()))
{

	EVALUATEUR_CONVERSIONPOINT(Point  );
	EVALUATEUR_CONVERSIONPOINT(Vecteur);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Point>())
	{
		_type = Traceur::Donnees::Type::Point::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Vector>())
	{
		_type = Traceur::Donnees::Type::Point::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionUn(position,"POINT",valeur);
}

ValeurSP ConversionPoint::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	ConversionPoint resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP ConversionPoint::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void ConversionPoint::format(std::ostream& os) const
{
	os << "POINT[";
	_valeur->format(os);
	os << "]";
}

ExpressionSP ConversionPoint::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ConversionPointSP clone(new ConversionPoint(valeur));
	return clone;
}
