#include <src/donnees/expressions/expressionconversionreel.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionchaine.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_CONVERSIONREEL(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ConversionReelSP(new EvaluateurOperateurUnaire<ConversionReel,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<ConversionReel,Entier> : public ConversionReel
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							EntierSP valeur_valeur=std::static_pointer_cast<Entier>(expr_valeur);
							SECURISE_VALEUR(ReelSP(new Reel(*valeur_valeur)));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionReel,Reel> : public ConversionReel
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							return valeur;
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionReel,Chaine> : public ConversionReel
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ChaineSP valeur_valeur=std::static_pointer_cast<Chaine>(expr_valeur);
							if(valeur_valeur->represente_un_reel())
							{ SECURISE_VALEUR(ReelSP(new Reel(*valeur_valeur))); }
							else
								return Valeur::conversion_reelle_invalide();
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	ConversionReel::ConversionReel(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ConversionReelSP(new EvaluateurOperateurUnaire<ConversionReel,Expression>()))
{

	EVALUATEUR_CONVERSIONREEL(Entier);
	EVALUATEUR_CONVERSIONREEL(Reel  );
	EVALUATEUR_CONVERSIONREEL(Chaine);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::String>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionUn(position,"REAL",valeur);
}

ValeurSP ConversionReel::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	ConversionReel resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP ConversionReel::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void ConversionReel::format(std::ostream& os) const
{
	os << "REAL[";
	_valeur->format(os);
	os << "]";
}

ExpressionSP ConversionReel::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ConversionReelSP clone(new ConversionReel(valeur));
	return clone;
}
