#include <src/donnees/expressions/expressionconversionrouge.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/expressions/expressioncouleurhsva.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_CONVERSIONROUGE(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ConversionRougeSP(new EvaluateurOperateurUnaire<ConversionRouge,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<ConversionRouge,CouleurRVBA> : public ConversionRouge
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							CouleurRVBASP valeur_valeur=std::static_pointer_cast<CouleurRVBA>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->rouge());
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionRouge,CouleurHSVA> : public ConversionRouge
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							CouleurHSVASP valeur_valeur=std::static_pointer_cast<CouleurHSVA>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->rouge());
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	ConversionRouge::ConversionRouge(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ConversionRougeSP(new EvaluateurOperateurUnaire<ConversionRouge,Expression>()))
{

	EVALUATEUR_CONVERSIONROUGE(CouleurRVBA);
	EVALUATEUR_CONVERSIONROUGE(CouleurHSVA);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Color>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionUn(position,"RED",valeur);
}

ValeurSP ConversionRouge::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	ConversionRouge resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP ConversionRouge::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void ConversionRouge::format(std::ostream& os) const
{
	os << "RED[";
	_valeur->format(os);
	os << "]";
}

ExpressionSP ConversionRouge::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ConversionRougeSP clone(new ConversionRouge(valeur));
	return clone;
}
