#include <src/donnees/expressions/expressionconversionsaturation.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/expressions/expressioncouleurhsva.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_CONVERSIONSATURATION(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ConversionSaturationSP(new EvaluateurOperateurUnaire<ConversionSaturation,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<ConversionSaturation,CouleurRVBA> : public ConversionSaturation
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							CouleurRVBASP valeur_valeur=std::static_pointer_cast<CouleurRVBA>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->saturation());
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionSaturation,CouleurHSVA> : public ConversionSaturation
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							CouleurHSVASP valeur_valeur=std::static_pointer_cast<CouleurHSVA>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->saturation());
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	ConversionSaturation::ConversionSaturation(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ConversionSaturationSP(new EvaluateurOperateurUnaire<ConversionSaturation,Expression>()))
{

	EVALUATEUR_CONVERSIONSATURATION(CouleurRVBA);
	EVALUATEUR_CONVERSIONSATURATION(CouleurHSVA);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Color>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionUn(position,"SATURATION",valeur);
}

ValeurSP ConversionSaturation::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	ConversionSaturation resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP ConversionSaturation::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void ConversionSaturation::format(std::ostream& os) const
{
	os << "SATURATION[";
	_valeur->format(os);
	os << "]";
}

ExpressionSP ConversionSaturation::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ConversionSaturationSP clone(new ConversionSaturation(valeur));
	return clone;
}
