#include <src/donnees/expressions/expressionconversionteinte.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/expressions/expressioncouleurhsva.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_CONVERSIONTEINTE(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ConversionTeinteSP(new EvaluateurOperateurUnaire<ConversionTeinte,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<ConversionTeinte,CouleurRVBA> : public ConversionTeinte
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							CouleurRVBASP valeur_valeur=std::static_pointer_cast<CouleurRVBA>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->teinte());
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionTeinte,CouleurHSVA> : public ConversionTeinte
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							CouleurHSVASP valeur_valeur=std::static_pointer_cast<CouleurHSVA>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->teinte());
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	ConversionTeinte::ConversionTeinte(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ConversionTeinteSP(new EvaluateurOperateurUnaire<ConversionTeinte,Expression>()))
{

	EVALUATEUR_CONVERSIONTEINTE(CouleurRVBA);
	EVALUATEUR_CONVERSIONTEINTE(CouleurHSVA);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Color>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	// TODO Point => Color
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionUn(position,"HUE",valeur);
}

ValeurSP ConversionTeinte::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	ConversionTeinte resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP ConversionTeinte::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void ConversionTeinte::format(std::ostream& os) const
{
	os << "HUE[";
	_valeur->format(os);
	os << "]";
}

ExpressionSP ConversionTeinte::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ConversionTeinteSP clone(new ConversionTeinte(valeur));
	return clone;
}
