#include <src/donnees/expressions/expressionconversionvaleur.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/expressions/expressioncouleurhsva.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_CONVERSIONVALEUR(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ConversionValeurSP(new EvaluateurOperateurUnaire<ConversionValeur,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<ConversionValeur,CouleurRVBA> : public ConversionValeur
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							CouleurRVBASP valeur_valeur=std::static_pointer_cast<CouleurRVBA>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->valeur());
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionValeur,CouleurHSVA> : public ConversionValeur
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							CouleurHSVASP valeur_valeur=std::static_pointer_cast<CouleurHSVA>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->valeur());
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	ConversionValeur::ConversionValeur(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ConversionValeurSP(new EvaluateurOperateurUnaire<ConversionValeur,Expression>()))
{

	EVALUATEUR_CONVERSIONVALEUR(CouleurRVBA);
	EVALUATEUR_CONVERSIONVALEUR(CouleurHSVA);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Color>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionUn(position,"VALUE",valeur);
}

ValeurSP ConversionValeur::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	ConversionValeur resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP ConversionValeur::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void ConversionValeur::format(std::ostream& os) const
{
	os << "VALUE[";
	_valeur->format(os);
	os << "]";
}

ExpressionSP ConversionValeur::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ConversionValeurSP clone(new ConversionValeur(valeur));
	return clone;
}
