#include <src/donnees/expressions/expressionconversionvecteur.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteur.h>
#include <src/donnees/expressions/expressionvecteurpolaire.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_CONVERSIONVECTEUR(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ConversionVecteurSP(new EvaluateurOperateurUnaire<ConversionVecteur,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<ConversionVecteur,Point> : public ConversionVecteur
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							PointSP valeur_valeur=std::static_pointer_cast<Point>(expr_valeur);
							SECURISE_VALEUR(VecteurSP(new Vecteur(*valeur_valeur)));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionVecteur,Vecteur> : public ConversionVecteur
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							VecteurSP valeur_valeur=std::static_pointer_cast<Vecteur>(expr_valeur);
							SECURISE_VALEUR(VecteurSP(new Vecteur(*valeur_valeur)));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionVecteur,VecteurPolaire> : public ConversionVecteur
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							VecteurPolaireSP valeur_valeur=std::static_pointer_cast<VecteurPolaire>(expr_valeur);
							SECURISE_VALEUR(VecteurSP(new Vecteur(*valeur_valeur)));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	ConversionVecteur::ConversionVecteur(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ConversionVecteurSP(new EvaluateurOperateurUnaire<ConversionVecteur,Expression>()))
{

	EVALUATEUR_CONVERSIONVECTEUR(Point         );
	EVALUATEUR_CONVERSIONVECTEUR(Vecteur       );
	EVALUATEUR_CONVERSIONVECTEUR(VecteurPolaire);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Point>())
	{
		_type = Traceur::Donnees::Type::Vector::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Vector>())
	{
		_type = Traceur::Donnees::Type::Vector::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionUn(position,"VECTOR",valeur);
}

ValeurSP ConversionVecteur::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	ConversionVecteur resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP ConversionVecteur::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void ConversionVecteur::format(std::ostream& os) const
{
	os << "VECTOR[";
	_valeur->format(os);
	os << "]";
}

ExpressionSP ConversionVecteur::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ConversionVecteurSP clone(new ConversionVecteur(valeur));
	return clone;
}
