#include <src/donnees/expressions/expressionconversionvecteurpolaire.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteur.h>
#include <src/donnees/expressions/expressionvecteurpolaire.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_CONVERSIONVECTEURPOLAIRE(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ConversionVecteurPolaireSP(new EvaluateurOperateurUnaire<ConversionVecteurPolaire,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<ConversionVecteurPolaire,Point> : public ConversionVecteurPolaire
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							PointSP valeur_valeur=std::static_pointer_cast<Point>(expr_valeur);
							SECURISE_VALEUR(VecteurPolaireSP(new VecteurPolaire(*valeur_valeur)));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionVecteurPolaire,Vecteur> : public ConversionVecteurPolaire
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							VecteurSP valeur_valeur=std::static_pointer_cast<Vecteur>(expr_valeur);
							SECURISE_VALEUR(VecteurPolaireSP(new VecteurPolaire(*valeur_valeur)));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionVecteurPolaire,VecteurPolaire> : public ConversionVecteurPolaire
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							VecteurPolaireSP valeur_valeur=std::static_pointer_cast<VecteurPolaire>(expr_valeur);
							SECURISE_VALEUR(VecteurPolaireSP(new VecteurPolaire(*valeur_valeur)));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	ConversionVecteurPolaire::ConversionVecteurPolaire(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ConversionVecteurPolaireSP(new EvaluateurOperateurUnaire<ConversionVecteurPolaire,Expression>()))
{

	EVALUATEUR_CONVERSIONVECTEURPOLAIRE(Point         );
	EVALUATEUR_CONVERSIONVECTEURPOLAIRE(Vecteur       );
	EVALUATEUR_CONVERSIONVECTEURPOLAIRE(VecteurPolaire);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Point>())
	{
		_type = Traceur::Donnees::Type::Vector::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Vector>())
	{
		_type = Traceur::Donnees::Type::Vector::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionUn(position,"POLAR",valeur);
}

ValeurSP ConversionVecteurPolaire::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	ConversionVecteurPolaire resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP ConversionVecteurPolaire::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void ConversionVecteurPolaire::format(std::ostream& os) const
{
	os << "POLAR[";
	_valeur->format(os);
	os << "]";
}

ExpressionSP ConversionVecteurPolaire::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ConversionVecteurPolaireSP clone(new ConversionVecteurPolaire(valeur));
	return clone;
}
