#include <src/donnees/expressions/expressionconversionvert.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/expressions/expressioncouleurhsva.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_CONVERSIONVERT(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ConversionVertSP(new EvaluateurOperateurUnaire<ConversionVert,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<ConversionVert,CouleurRVBA> : public ConversionVert
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							CouleurRVBASP valeur_valeur=std::static_pointer_cast<CouleurRVBA>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->vert());
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ConversionVert,CouleurHSVA> : public ConversionVert
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							CouleurHSVASP valeur_valeur=std::static_pointer_cast<CouleurHSVA>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->vert());
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	ConversionVert::ConversionVert(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ConversionVertSP(new EvaluateurOperateurUnaire<ConversionVert,Expression>()))
{

	EVALUATEUR_CONVERSIONVERT(CouleurRVBA);
	EVALUATEUR_CONVERSIONVERT(CouleurHSVA);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Color>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionUn(position,"GREEN",valeur);
}

ValeurSP ConversionVert::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	ConversionVert resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP ConversionVert::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void ConversionVert::format(std::ostream& os) const
{
	os << "GREEN[";
	_valeur->format(os);
	os << "]";
}

ExpressionSP ConversionVert::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ConversionVertSP clone(new ConversionVert(valeur));
	return clone;
}
