#include <src/donnees/expressions/expressioncosinus.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteur.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_COSINUS(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = CosinusSP(new EvaluateurOperateurUnaire<Cosinus,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<Cosinus,Entier> : public Cosinus
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							EntierSP valeur_valeur=std::static_pointer_cast<Entier>(expr_valeur);
							SECURISE_VALEUR(ReelSP(new Reel(Entier::cosinus(*valeur_valeur))));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<Cosinus,Reel> : public Cosinus
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ReelSP valeur_valeur=std::static_pointer_cast<Reel>(expr_valeur);
							SECURISE_VALEUR(ReelSP(new Reel(Reel::cosinus(*valeur_valeur))));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<Cosinus,Point> : public Cosinus
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							PointSP valeur_valeur=std::static_pointer_cast<Point>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->abscisse());
						}
				};
			template<>
				class EvaluateurOperateurUnaire<Cosinus,Vecteur> : public Cosinus
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							VecteurSP valeur_valeur=std::static_pointer_cast<Vecteur>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->abscisse());
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Cosinus::Cosinus(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(CosinusSP(new EvaluateurOperateurUnaire<Cosinus,Expression>()))
{

	EVALUATEUR_COSINUS(Entier );
	EVALUATEUR_COSINUS(Reel   );
	EVALUATEUR_COSINUS(Point  );
	EVALUATEUR_COSINUS(Vecteur);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Point>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Vector>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionPrefixe(position,"cos",valeur);
}

ValeurSP Cosinus::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	Cosinus resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP Cosinus::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void Cosinus::format(std::ostream& os) const
{
	os << "cos (";
	_valeur->format(os);
	os << ")";
}

ExpressionSP Cosinus::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	CosinusSP clone(new Cosinus(valeur));
	return clone;
}
