#include <src/donnees/expressions/expressioncouleurhsva.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionnombre.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_COULEURHSVA(type1,type2,type3,type4) \
	if(teinte->est_de_nature<type1>() and saturation->est_de_nature<type2>() and valeur->est_de_nature<type3>() and alpha->est_de_nature<type4>()) { _evaluateur = CouleurHSVASP(new EvaluateurOperateurQuaternaire<CouleurHSVA,type1,type2,type3,type4>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurQuaternaire<CouleurHSVA,Nombre,Nombre,Nombre,Nombre> : public CouleurHSVA
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP rouge, const ValeurSP vert, const ValeurSP bleu, const ValeurSP alpha) const
						{
							ExpressionSP expr_rouge(*rouge);
							NombreSP valeur_rouge=std::static_pointer_cast<Nombre>(expr_rouge);
							ExpressionSP expr_vert(*vert);
							NombreSP valeur_vert=std::static_pointer_cast<Nombre>(expr_vert);
							ExpressionSP expr_bleu(*bleu);
							NombreSP valeur_bleu=std::static_pointer_cast<Nombre>(expr_bleu);
							ExpressionSP expr_alpha(*alpha);
							NombreSP valeur_alpha=std::static_pointer_cast<Nombre>(expr_alpha);
							SECURISE_VALEUR(CouleurHSVASP(new CouleurHSVA(valeur_rouge,valeur_vert,valeur_bleu,valeur_alpha)));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	CouleurHSVA::CouleurHSVA(ExpressionSP teinte, ExpressionSP saturation, ExpressionSP valeur, ExpressionSP alpha, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneQuaternaire(teinte,saturation,valeur,alpha), _evaluateur(CouleurHSVASP(new EvaluateurOperateurQuaternaire<CouleurHSVA,Expression,Expression,Expression,Expression>()))
{
	EVALUATEUR_COULEURHSVA(Nombre,Nombre,Nombre,Nombre);
	// calcul du type de retour
	if(_type)
		return;
	if(teinte->est_de_type<Traceur::Donnees::Type::Number>() and saturation->est_de_type<Traceur::Donnees::Type::Number>() and valeur->est_de_type<Traceur::Donnees::Type::Number>() and alpha->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Color::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionQuatre(position,"HSV",teinte,saturation,valeur,alpha);
}

CouleurHSVA::CouleurHSVA(CouleurHSVASP couleur, NombreSP alpha)
{
	_type = couleur->_type;
	_valeur1 = couleur->_valeur1;
	_valeur2 = couleur->_valeur2;
	_valeur3 = couleur->_valeur3;
	_valeur4 = alpha;
}

CouleurHSVA::CouleurHSVA(const CouleurRVBA& couleur)
{
	static Reel _0(0);
	static Reel _1(1);
	static Reel _60(60);
	static Reel _100(100);
	static Reel _120(120);
	static Reel _240(240);
	static Reel _255(255);
	static Reel _360(360);
	NombreSP rouge = std::dynamic_pointer_cast<Nombre>(couleur._valeur1);
	TRACEUR_TEST(rouge);
	NombreSP vert = std::dynamic_pointer_cast<Nombre>(couleur._valeur2);
	TRACEUR_TEST(vert);
	NombreSP bleu = std::dynamic_pointer_cast<Nombre>(couleur._valeur3);
	TRACEUR_TEST(bleu);
	
	Reel rouge_ = rouge->reel_borne(_0,_255);

	Reel vert_ = vert->reel_borne(_0,_255);

	Reel bleu_ = bleu->reel_borne(_0,_255);
	
	Reel min_ = rouge_;
	Reel max_ = rouge_;
	if(vert_<min_)
		min_ = vert_;
	if(vert_>max_)
		max_ = vert_;
	if(bleu_<min_)
		min_ = bleu_;
	if(bleu_>max_)
		max_ = bleu_;
	
	if(min_==max_)
		_valeur1 = NombreSP(new Reel(0));
	else if(max_ == rouge_)
	{
		Reel teinte_ = _60*(vert_-bleu_)/(max_-min_)+_360;
		if(teinte_>=_360)
			teinte_ = teinte_ - _360;
		_valeur1 = NombreSP(new Reel(teinte_));
	}
	else if(max_ == vert_)
		_valeur1 = NombreSP(new Reel(_60*(bleu_-rouge_)/(max_-min_)+_120));
	else
		_valeur1 = NombreSP(new Reel(_60*(rouge_-vert_)/(max_-min_)+_240));
	
	if(max_ == _0)
		_valeur2 = NombreSP(new Reel(0));
	else
		_valeur2 = NombreSP(new Reel(_100*(_1-min_/max_)));

	_valeur3 = NombreSP(new Reel((_100*max_)/_255));

	_valeur4 = couleur._valeur4;
	_type = Traceur::Donnees::Type::Color::SP();
}

bool CouleurHSVA::operator==(const CouleurRVBA& couleur)
{
	CouleurRVBA couleur_rvba(*this);
	return couleur_rvba==couleur;
}

bool CouleurHSVA::operator==(const CouleurHSVA& couleur)
{
	CouleurRVBA couleur_rvba1(*this);
	CouleurRVBA couleur_rvba2(couleur);
	return couleur_rvba1==couleur_rvba2;
}

bool CouleurHSVA::operator!=(const CouleurRVBA& couleur)
{
	CouleurRVBA couleur_rvba(*this);
	return couleur_rvba!=couleur;
}

bool CouleurHSVA::operator!=(const CouleurHSVA& couleur)
{
	CouleurRVBA couleur_rvba1(*this);
	CouleurRVBA couleur_rvba2(couleur);
	return couleur_rvba1!=couleur_rvba2;
}

ValeurSP CouleurHSVA::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_rouge = _valeur1->evaluation(assembleur); if(not valeur_rouge->a_une_valeur_retour()) return valeur_rouge;
	ValeurSP valeur_vert = _valeur2->evaluation(assembleur); if(not valeur_vert->a_une_valeur_retour()) return valeur_vert;
	ValeurSP valeur_bleu = _valeur3->evaluation(assembleur); if(not valeur_bleu->a_une_valeur_retour()) return valeur_bleu;
	ValeurSP valeur_alpha = _valeur4->evaluation(assembleur); if(not valeur_alpha->a_une_valeur_retour()) return valeur_alpha;
	CouleurHSVA resultat(*valeur_rouge,*valeur_vert,*valeur_bleu,*valeur_alpha);
	return resultat.evaluation(valeur_rouge,valeur_vert,valeur_bleu,valeur_alpha);
}

ValeurSP CouleurHSVA::evaluation(const ValeurSP rouge, const ValeurSP vert, const ValeurSP bleu, const ValeurSP alpha) const
{
	return _evaluateur->evaluation(rouge,vert,bleu,alpha);
}

void CouleurHSVA::format(std::ostream& os) const
{
	//_type->format(os);
	os << "HSV[";
	_valeur1->format(os);
	os << ";";
	_valeur2->format(os);
	os << ";";
	_valeur3->format(os);
	os << ";";
	_valeur4->format(os);
	os << "]";
}

ExpressionSP CouleurHSVA::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP rouge = _valeur1->clone(clones_instances);
	ExpressionSP vert = _valeur2->clone(clones_instances);
	ExpressionSP bleu = _valeur3->clone(clones_instances);
	ExpressionSP alpha = _valeur4->clone(clones_instances);
	CouleurHSVASP clone(new CouleurHSVA());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_valeur1=rouge;
	clone->_valeur2=vert;
	clone->_valeur3=bleu;
	clone->_valeur4=alpha;
	return clone;
}

CouleurHSVASP CouleurHSVA::augmente_teinte(NombreSP valeur)
{
	CouleurHSVASP hsva(new CouleurHSVA(*this));
	NombreSP teinte = std::dynamic_pointer_cast<Nombre>(_valeur1);
	TRACEUR_TEST(teinte);
	hsva->_valeur1 = teinte+valeur;
	return hsva;
}

CouleurHSVASP CouleurHSVA::augmente_saturation(NombreSP valeur)
{
	CouleurHSVASP hsva(new CouleurHSVA(*this));
	NombreSP saturation = std::dynamic_pointer_cast<Nombre>(_valeur2);
	TRACEUR_TEST(saturation);
	hsva->_valeur2 = saturation+valeur;
	return hsva;
}

CouleurHSVASP CouleurHSVA::diminue_saturation(NombreSP valeur)
{
	CouleurHSVASP hsva(new CouleurHSVA(*this));
	NombreSP saturation = std::dynamic_pointer_cast<Nombre>(_valeur2);
	TRACEUR_TEST(saturation);
	hsva->_valeur2 = saturation-valeur;
	return hsva;
}

CouleurHSVASP CouleurHSVA::augmente_valeur(NombreSP valeur)
{
	CouleurHSVASP hsva(new CouleurHSVA(*this));
	NombreSP valeur_couleur = std::dynamic_pointer_cast<Nombre>(_valeur3);
	TRACEUR_TEST(valeur_couleur);
	hsva->_valeur3 = valeur_couleur+valeur;
	return hsva;
}

CouleurHSVASP CouleurHSVA::diminue_valeur(NombreSP valeur)
{
	CouleurHSVASP hsva(new CouleurHSVA(*this));
	NombreSP valeur_couleur = std::dynamic_pointer_cast<Nombre>(_valeur3);
	TRACEUR_TEST(valeur_couleur);
	hsva->_valeur3 = valeur_couleur-valeur;
	return hsva;
}

CouleurHSVASP CouleurHSVA::augmente_alpha(NombreSP valeur)
{
	CouleurHSVASP hsva(new CouleurHSVA(*this));
	NombreSP alpha = std::dynamic_pointer_cast<Nombre>(_valeur4);
	TRACEUR_TEST(alpha);
	hsva->_valeur4 = alpha+valeur;
	return hsva;
}

CouleurRVBASP CouleurHSVA::sur(const CouleurRVBASP couleur)
{
	CouleurRVBA rvba(*this);
	return rvba.sur(couleur);
}

CouleurRVBASP CouleurHSVA::sur(const CouleurHSVASP couleur)
{
	CouleurRVBA rvba(*this);
	CouleurRVBASP couleur_rvba(new CouleurRVBA(*couleur));
	return rvba.sur(couleur_rvba);
}

CouleurHSVASP CouleurHSVA::gris_equivalent()
{
	CouleurHSVASP resultat(new CouleurHSVA(_valeur1,Nombre::zero(),_valeur3,_valeur4));
	return resultat;
}

ExpressionSP CouleurHSVA::rouge()
{
	CouleurRVBA couleur(*this);
	return couleur._valeur1;
}

ExpressionSP CouleurHSVA::vert()
{
	CouleurRVBA couleur(*this);
	return couleur._valeur2;
}

ExpressionSP CouleurHSVA::bleu()
{
	CouleurRVBA couleur(*this);
	return couleur._valeur3;
}

ExpressionSP CouleurHSVA::alpha()
{
	return _valeur4;
}

ExpressionSP CouleurHSVA::teinte()
{
	return _valeur1;
}

ExpressionSP CouleurHSVA::saturation()
{
	return _valeur2;
}

ExpressionSP CouleurHSVA::valeur()
{
	return _valeur3;
}

NombreSP CouleurHSVA::operator-(CouleurHSVA& couleur)
{
	static Reel _0(0);
	static Reel _2(2);
	static Reel _90(90);
	static Reel _100(100);
	static Reel _180(180);
	static Reel _360(360);

	NombreSP teinte = std::dynamic_pointer_cast<Nombre>(_valeur1);
	TRACEUR_TEST(teinte);
	NombreSP couleur_teinte = std::dynamic_pointer_cast<Nombre>(couleur._valeur1);
	TRACEUR_TEST(couleur_teinte);
	NombreSP saturation = std::dynamic_pointer_cast<Nombre>(_valeur2);
	TRACEUR_TEST(saturation);
	NombreSP couleur_saturation = std::dynamic_pointer_cast<Nombre>(couleur._valeur2);
	TRACEUR_TEST(couleur_saturation);
	NombreSP valeur = std::dynamic_pointer_cast<Nombre>(_valeur3);
	TRACEUR_TEST(valeur);
	NombreSP couleur_valeur = std::dynamic_pointer_cast<Nombre>(couleur._valeur3);
	TRACEUR_TEST(couleur_valeur);

	Reel teinte_ = (teinte->reel()%_360)/_2;
	Reel couleur_teinte_ = (couleur_teinte->reel()%_360)/_2;

	if(couleur_teinte_<teinte_-_90)
		couleur_teinte_ = couleur_teinte_+_180;
	if(couleur_teinte_>teinte_+_90)
		couleur_teinte_ = couleur_teinte_-_180;

	teinte_ = Reel::radians(teinte_);
	couleur_teinte_ = Reel::radians(couleur_teinte_);

	Reel valeur_ = valeur->reel_borne(_0,_100);
	Reel couleur_valeur_ = couleur_valeur->reel_borne(_0,_100);

	Reel saturation_ = (saturation->reel_borne(_0,_100)*valeur_)/_100;
	Reel couleur_saturation_ = (couleur_saturation->reel_borne(_0,_100)*couleur_valeur_)/_100;

	Reel x_ = saturation_*Reel::cosinus(teinte_)-couleur_saturation_*Reel::cosinus(couleur_teinte_);
	Reel y_ = saturation_*Reel::  sinus(teinte_)-couleur_saturation_*Reel::  sinus(couleur_teinte_);
	Reel z_ = valeur_-couleur_valeur_;

	Reel intermediaire = x_*x_+y_*y_+z_*z_;

	NombreSP distance(new Reel(Reel::racine_carree(intermediaire)));
	return distance;
}

NombreSP CouleurHSVA::operator-(CouleurRVBA& couleur)
{
	CouleurHSVA hsva(couleur);
	return (*this)-hsva;
}
