#include <src/donnees/expressions/expressioncouleurrvba.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionnombre.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressioncouleurhsva.h>
#include <src/donnees/assembleur/couleur.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_COULEURRVBA(type1,type2,type3,type4) \
	if(rouge->est_de_nature<type1>() and vert->est_de_nature<type2>() and bleu->est_de_nature<type3>() and alpha->est_de_nature<type4>()) { _evaluateur = CouleurRVBASP(new EvaluateurOperateurQuaternaire<CouleurRVBA,type1,type2,type3,type4>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurQuaternaire<CouleurRVBA,Nombre,Nombre,Nombre,Nombre> : public CouleurRVBA
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP rouge, const ValeurSP vert, const ValeurSP bleu, const ValeurSP alpha) const
						{
							ExpressionSP expr_rouge(*rouge);
							NombreSP valeur_rouge=std::static_pointer_cast<Nombre>(expr_rouge);
							ExpressionSP expr_vert(*vert);
							NombreSP valeur_vert=std::static_pointer_cast<Nombre>(expr_vert);
							ExpressionSP expr_bleu(*bleu);
							NombreSP valeur_bleu=std::static_pointer_cast<Nombre>(expr_bleu);
							ExpressionSP expr_alpha(*alpha);
							NombreSP valeur_alpha=std::static_pointer_cast<Nombre>(expr_alpha);
							SECURISE_VALEUR(CouleurRVBASP(new CouleurRVBA(valeur_rouge,valeur_vert,valeur_bleu,valeur_alpha)));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	CouleurRVBA::CouleurRVBA(ExpressionSP rouge, ExpressionSP vert, ExpressionSP bleu, ExpressionSP alpha, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneQuaternaire(rouge,vert,bleu,alpha), _evaluateur(CouleurRVBASP(new EvaluateurOperateurQuaternaire<CouleurRVBA,Expression,Expression,Expression,Expression>()))
{
	EVALUATEUR_COULEURRVBA(Nombre,Nombre,Nombre,Nombre);
	// calcul du type de retour
	if(_type)
		return;
	if(rouge->est_de_type<Traceur::Donnees::Type::Number>() and vert->est_de_type<Traceur::Donnees::Type::Number>() and bleu->est_de_type<Traceur::Donnees::Type::Number>() and alpha->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Color::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionQuatre(position,"COLOR",rouge,vert,bleu,alpha);
}

CouleurRVBA::CouleurRVBA(CouleurRVBASP couleur, NombreSP alpha)
{
	_type = couleur->_type;
	_valeur1 = couleur->_valeur1;
	_valeur2 = couleur->_valeur2;
	_valeur3 = couleur->_valeur3;
	_valeur4 = alpha;
}

CouleurRVBA::CouleurRVBA(const CouleurHSVA& couleur)
{
	static Reel _0(0);
	static Reel _1(1);
	static Reel _60(60);
	static Reel _100(100);
	static Reel _255(255);
	static Reel _360(360);
	NombreSP teinte = std::dynamic_pointer_cast<Nombre>(couleur._valeur1);
	TRACEUR_TEST(teinte);
	NombreSP saturation = std::dynamic_pointer_cast<Nombre>(couleur._valeur2);
	TRACEUR_TEST(saturation);
	NombreSP valeur = std::dynamic_pointer_cast<Nombre>(couleur._valeur3);
	TRACEUR_TEST(valeur);
	Reel teinte_ = teinte->reel()/_360;
	teinte_ = _360*Reel::partie_decimale(teinte_);

	if(teinte_<_0)
		teinte_ = teinte_ + _360;
	if(teinte_>=_360)
		teinte_ = teinte_ - _360;

	Reel saturation_ = saturation->reel_borne(_0,_100);
	saturation_ = saturation_/_100;

	Reel valeur_ = valeur->reel_borne(_0,_100);

	valeur_ = (_255*valeur_) / _100;

	Reel preindice = teinte_/_60;
	Entier indice_teinte_(Reel::partie_entiere(preindice));
	Reel f(Reel::partie_decimale(preindice));
	Reel l(valeur_*(_1-saturation_));
	Reel m(valeur_*(_1-f*saturation_));
	Reel n(valeur_*(_1-(_1-f)*saturation_));
	switch(indice_teinte_._entier)
	{
		case 0:
			_valeur1 = NombreSP(new Reel(valeur_));
			_valeur2 = NombreSP(new Reel(n));
			_valeur3 = NombreSP(new Reel(l));
			break;
		case 1:
			_valeur1 = NombreSP(new Reel(m));
			_valeur2 = NombreSP(new Reel(valeur_));
			_valeur3 = NombreSP(new Reel(l));
			break;
		case 2:
			_valeur1 = NombreSP(new Reel(l));
			_valeur2 = NombreSP(new Reel(valeur_));
			_valeur3 = NombreSP(new Reel(n));
			break;
		case 3:
			_valeur1 = NombreSP(new Reel(l));
			_valeur2 = NombreSP(new Reel(m));
			_valeur3 = NombreSP(new Reel(valeur_));
			break;
		case 4:
			_valeur1 = NombreSP(new Reel(n));
			_valeur2 = NombreSP(new Reel(l));
			_valeur3 = NombreSP(new Reel(valeur_));
			break;
		case 5:
			_valeur1 = NombreSP(new Reel(valeur_));
			_valeur2 = NombreSP(new Reel(l));
			_valeur3 = NombreSP(new Reel(m));
			break;
		default: throw;
	}
	_valeur4 = couleur._valeur4;
	_type = Traceur::Donnees::Type::Color::SP();
}

CouleurRVBA::CouleurRVBA(const Traceur::Donnees::Assembleur::Couleur::Couleur& couleur)
{
	_type = Traceur::Donnees::Type::Color::SP();
	_valeur1 = ReelSP(new Reel(couleur._rouge));
	_valeur2 = ReelSP(new Reel(couleur._vert ));
	_valeur3 = ReelSP(new Reel(couleur._bleu ));
	_valeur4 = ReelSP(new Reel(couleur._alpha));
}

ValeurSP CouleurRVBA::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_rouge = _valeur1->evaluation(assembleur); if(not valeur_rouge->a_une_valeur_retour()) return valeur_rouge;
	ValeurSP valeur_vert = _valeur2->evaluation(assembleur); if(not valeur_vert->a_une_valeur_retour()) return valeur_vert;
	ValeurSP valeur_bleu = _valeur3->evaluation(assembleur); if(not valeur_bleu->a_une_valeur_retour()) return valeur_bleu;
	ValeurSP valeur_alpha = _valeur4->evaluation(assembleur); if(not valeur_alpha->a_une_valeur_retour()) return valeur_alpha;
	CouleurRVBA resultat(*valeur_rouge,*valeur_vert,*valeur_bleu,*valeur_alpha);
	return resultat.evaluation(valeur_rouge,valeur_vert,valeur_bleu,valeur_alpha);
}

ValeurSP CouleurRVBA::evaluation(const ValeurSP rouge, const ValeurSP vert, const ValeurSP bleu, const ValeurSP alpha) const
{
	return _evaluateur->evaluation(rouge,vert,bleu,alpha);
}

void CouleurRVBA::format(std::ostream& os) const
{
	//_type->format(os);
	os << "COLOR[";
	_valeur1->format(os);
	os << ";";
	_valeur2->format(os);
	os << ";";
	_valeur3->format(os);
	os << ";";
	_valeur4->format(os);
	os << "]";
}

ExpressionSP CouleurRVBA::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP rouge = _valeur1->clone(clones_instances);
	ExpressionSP vert = _valeur2->clone(clones_instances);
	ExpressionSP bleu = _valeur3->clone(clones_instances);
	ExpressionSP alpha = _valeur4->clone(clones_instances);
	CouleurRVBASP clone(new CouleurRVBA());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_valeur1=rouge;
	clone->_valeur2=vert;
	clone->_valeur3=bleu;
	clone->_valeur4=alpha;
	return clone;
}

bool CouleurRVBA::operator==(const CouleurRVBA& couleur)
{
	NombreSP rouge = std::dynamic_pointer_cast<Nombre>(_valeur1);
	TRACEUR_TEST(rouge);
	NombreSP couleur_rouge = std::dynamic_pointer_cast<Nombre>(couleur._valeur1);
	TRACEUR_TEST(couleur_rouge);
	if(rouge->entier()!=couleur_rouge->entier())
		return false;
	NombreSP vert = std::dynamic_pointer_cast<Nombre>(_valeur2);
	TRACEUR_TEST(vert);
	NombreSP couleur_vert = std::dynamic_pointer_cast<Nombre>(couleur._valeur2);
	TRACEUR_TEST(couleur_vert);
	if(vert->entier()!=couleur_vert->entier())
		return false;
	NombreSP bleu = std::dynamic_pointer_cast<Nombre>(_valeur3);
	TRACEUR_TEST(bleu);
	NombreSP couleur_bleu = std::dynamic_pointer_cast<Nombre>(couleur._valeur3);
	TRACEUR_TEST(couleur_bleu);
	if(bleu->entier()!=couleur_bleu->entier())
		return false;
	NombreSP alpha = std::dynamic_pointer_cast<Nombre>(_valeur4);
	TRACEUR_TEST(alpha);
	NombreSP couleur_alpha = std::dynamic_pointer_cast<Nombre>(couleur._valeur4);
	TRACEUR_TEST(couleur_alpha);
	if(alpha->entier()!=couleur_alpha->entier())
		return false;
	return true;
}

bool CouleurRVBA::operator==(const CouleurHSVA& couleur)
{
	CouleurRVBA couleur_rvba(couleur);
	return (*this)==couleur_rvba;
}

bool CouleurRVBA::operator!=(const CouleurRVBA& couleur)
{
	return not ((*this)==couleur);
}

bool CouleurRVBA::operator!=(const CouleurHSVA& couleur)
{
	CouleurRVBA couleur_rvba(couleur);
	return not ((*this)==couleur_rvba);
}

void CouleurRVBA::memes_types_que(const CouleurRVBA& couleur)
{
	NombreSP rouge = std::dynamic_pointer_cast<Nombre>(_valeur1);
	TRACEUR_TEST(rouge);
	if(couleur._valeur1->est_de_nature<Entier>())
		_valeur1 = NombreSP(new Entier(rouge->entier()));
	else
		_valeur1 = NombreSP(new Reel(rouge->reel()));
	NombreSP vert = std::dynamic_pointer_cast<Nombre>(_valeur2);
	TRACEUR_TEST(vert);
	if(couleur._valeur2->est_de_nature<Entier>())
		_valeur2 = NombreSP(new Entier(vert->entier()));
	else
		_valeur2 = NombreSP(new Reel(vert->reel()));
	NombreSP bleu = std::dynamic_pointer_cast<Nombre>(_valeur3);
	TRACEUR_TEST(bleu);
	if(couleur._valeur3->est_de_nature<Entier>())
		_valeur3 = NombreSP(new Entier(bleu->entier()));
	else
		_valeur3 = NombreSP(new Reel(bleu->reel()));
	NombreSP alpha = std::dynamic_pointer_cast<Nombre>(_valeur4);
	TRACEUR_TEST(alpha);
	if(couleur._valeur4->est_de_nature<Entier>())
		_valeur4 = NombreSP(new Entier(alpha->entier()));
	else
		_valeur4 = NombreSP(new Reel(alpha->reel()));
}

CouleurRVBASP CouleurRVBA::augmente_teinte(NombreSP valeur)
{
	CouleurHSVA hsva(*this);
	CouleurHSVASP hsva_resultat = hsva.augmente_teinte(valeur);
	CouleurRVBASP rvba(new CouleurRVBA(*hsva_resultat));
	rvba->memes_types_que(*this);
	return rvba;
}

CouleurRVBASP CouleurRVBA::augmente_saturation(NombreSP valeur)
{
	CouleurHSVA hsva(*this);
	CouleurHSVASP hsva_resultat = hsva.augmente_saturation(valeur);
	CouleurRVBASP rvba(new CouleurRVBA(*hsva_resultat));
	rvba->memes_types_que(*this);
	return rvba;
}

CouleurRVBASP CouleurRVBA::diminue_saturation(NombreSP valeur)
{
	CouleurHSVA hsva(*this);
	CouleurHSVASP hsva_resultat = hsva.diminue_saturation(valeur);
	CouleurRVBASP rvba(new CouleurRVBA(*hsva_resultat));
	rvba->memes_types_que(*this);
	return rvba;
}

CouleurRVBASP CouleurRVBA::augmente_valeur(NombreSP valeur)
{
	CouleurHSVA hsva(*this);
	CouleurHSVASP hsva_resultat = hsva.augmente_valeur(valeur);
	CouleurRVBASP rvba(new CouleurRVBA(*hsva_resultat));
	rvba->memes_types_que(*this);
	return rvba;
}

CouleurRVBASP CouleurRVBA::diminue_valeur(NombreSP valeur)
{
	CouleurHSVA hsva(*this);
	CouleurHSVASP hsva_resultat = hsva.diminue_valeur(valeur);
	CouleurRVBASP rvba(new CouleurRVBA(*hsva_resultat));
	rvba->memes_types_que(*this);
	return rvba;
}

CouleurRVBASP CouleurRVBA::augmente_alpha(NombreSP valeur)
{
	CouleurRVBASP rvba(new CouleurRVBA(*this));
	NombreSP alpha = std::dynamic_pointer_cast<Nombre>(_valeur4);
	TRACEUR_TEST(alpha);
	rvba->_valeur4 = alpha+valeur;
	return rvba;
}

CouleurRVBASP CouleurRVBA::sur(const CouleurRVBASP couleur)
{
	static Reel _0(0);
	static Reel _255(255);
	NombreSP rouge = std::dynamic_pointer_cast<Nombre>(_valeur1);
	TRACEUR_TEST(rouge);
	NombreSP couleur_rouge = std::dynamic_pointer_cast<Nombre>(couleur->_valeur1);
	TRACEUR_TEST(couleur_rouge);
	NombreSP vert = std::dynamic_pointer_cast<Nombre>(_valeur2);
	TRACEUR_TEST(vert);
	NombreSP couleur_vert = std::dynamic_pointer_cast<Nombre>(couleur->_valeur2);
	TRACEUR_TEST(couleur_vert);
	NombreSP bleu = std::dynamic_pointer_cast<Nombre>(_valeur3);
	TRACEUR_TEST(bleu);
	NombreSP couleur_bleu = std::dynamic_pointer_cast<Nombre>(couleur->_valeur3);
	TRACEUR_TEST(couleur_bleu);
	NombreSP alpha = std::dynamic_pointer_cast<Nombre>(_valeur4);
	TRACEUR_TEST(alpha);
	NombreSP couleur_alpha = std::dynamic_pointer_cast<Nombre>(couleur->_valeur4);
	TRACEUR_TEST(couleur_alpha);
	Reel alpha_ = alpha->reel_borne(_0,_255);
	Reel couleur_alpha_ = couleur_alpha->reel_borne(_0,_255);
	if(couleur_alpha_ == _0)
	{
		CouleurRVBASP resultat(new CouleurRVBA(*this));
		return resultat;
	}
	Reel resultat_intermediaire_alpha_ = (_255*(alpha_+couleur_alpha_)-alpha_*couleur_alpha_)/_255;
	NombreSP resultat_alpha(new Reel(resultat_intermediaire_alpha_));
	
	Reel rouge_ = rouge->reel_borne(_0,_255);
	Reel couleur_rouge_ = couleur_rouge->reel_borne(_0,_255);
	NombreSP resultat_rouge(new Reel((_255*rouge_+(couleur_rouge_-rouge_)*couleur_alpha_)/_255));
	
	Reel vert_ = vert->reel_borne(_0,_255);
	Reel couleur_vert_ = couleur_vert->reel_borne(_0,_255);
	NombreSP resultat_vert(new Reel((_255*vert_+(couleur_vert_-vert_)*couleur_alpha_)/_255));
	
	Reel bleu_ = bleu->reel_borne(_0,_255);
	Reel couleur_bleu_ = couleur_bleu->reel_borne(_0,_255);
	NombreSP resultat_bleu(new Reel((_255*bleu_+(couleur_bleu_-bleu_)*couleur_alpha_)/_255));

	CouleurRVBASP resultat(new CouleurRVBA(resultat_rouge,resultat_vert,resultat_bleu,resultat_alpha));
	resultat->memes_types_que(*couleur);
	return resultat;
}

CouleurRVBASP CouleurRVBA::sur(const CouleurHSVASP couleur)
{
	CouleurRVBASP rvba(new CouleurRVBA(*couleur));
	return sur(rvba);
}

CouleurRVBASP CouleurRVBA::gris_equivalent()
{
	static Reel _0(0);
	static Reel _11(11);
	static Reel _30(30);
	static Reel _59(59);
	static Reel _100(100);
	static Reel _255(255);
	NombreSP rouge = std::dynamic_pointer_cast<Nombre>(_valeur1);
	TRACEUR_TEST(rouge);
	NombreSP vert = std::dynamic_pointer_cast<Nombre>(_valeur2);
	TRACEUR_TEST(vert);
	NombreSP bleu = std::dynamic_pointer_cast<Nombre>(_valeur3);
	TRACEUR_TEST(bleu);
	NombreSP alpha = std::dynamic_pointer_cast<Nombre>(_valeur4);
	TRACEUR_TEST(alpha);
	
	Reel rouge_ = rouge->reel_borne(_0,_255);
	Reel vert_ = vert->reel_borne(_0,_255);
	Reel bleu_ = bleu->reel_borne(_0,_255);
	NombreSP gris(new Reel((_30*rouge_+_59*vert_+_11*bleu_)/_100));

	CouleurRVBASP resultat(new CouleurRVBA(gris,gris,gris,alpha));
	resultat->memes_types_que(*this);
	return resultat;
}

ExpressionSP CouleurRVBA::rouge()
{
	return _valeur1;
}

ExpressionSP CouleurRVBA::vert()
{
	return _valeur2;
}

ExpressionSP CouleurRVBA::bleu()
{
	return _valeur3;
}

ExpressionSP CouleurRVBA::alpha()
{
	return _valeur4;
}

ExpressionSP CouleurRVBA::teinte()
{
	CouleurHSVA couleur(*this);
	return couleur._valeur1;
}

ExpressionSP CouleurRVBA::saturation()
{
	CouleurHSVA couleur(*this);
	return couleur._valeur2;
}

ExpressionSP CouleurRVBA::valeur()
{
	CouleurHSVA couleur(*this);
	return couleur._valeur3;
}
NombreSP CouleurRVBA::operator-(CouleurHSVA& couleur)
{
	CouleurHSVA moi(*this);
	return moi-couleur;
}

NombreSP CouleurRVBA::operator-(CouleurRVBA& couleur)
{
	CouleurHSVA moi(*this);
	CouleurHSVA hsva(couleur);
	return moi-hsva;
}


using namespace Traceur::Donnees::Expression::Couleur;

	Noir::Noir()
:CouleurRVBA(Nombre::zero(),Nombre::zero(),Nombre::zero(),Nombre::max_canal_couleur())
{
}
	Gris::Gris()
:CouleurRVBA(Nombre::med_canal_couleur(),Nombre::med_canal_couleur(),Nombre::med_canal_couleur(),Nombre::max_canal_couleur())
{
}
	Blanc::Blanc()
:CouleurRVBA(Nombre::max_canal_couleur(),Nombre::max_canal_couleur(),Nombre::max_canal_couleur(),Nombre::max_canal_couleur())
{
}
	Rouge::Rouge()
:CouleurRVBA(Nombre::max_canal_couleur(),Nombre::zero(),Nombre::zero(),Nombre::max_canal_couleur())
{
}
	Vert::Vert()
:CouleurRVBA(Nombre::zero(),Nombre::max_canal_couleur(),Nombre::zero(),Nombre::max_canal_couleur())
{
}
	Bleu::Bleu()
:CouleurRVBA(Nombre::zero(),Nombre::zero(),Nombre::max_canal_couleur(),Nombre::max_canal_couleur())
{
}
	Magenta::Magenta()
:CouleurRVBA(Nombre::max_canal_couleur(),Nombre::zero(),Nombre::max_canal_couleur(),Nombre::max_canal_couleur())
{
}
	Cyan::Cyan()
:CouleurRVBA(Nombre::zero(),Nombre::max_canal_couleur(),Nombre::max_canal_couleur(),Nombre::max_canal_couleur())
{
}
	Jaune::Jaune()
:CouleurRVBA(Nombre::max_canal_couleur(),Nombre::max_canal_couleur(),Nombre::zero(),Nombre::max_canal_couleur())
{
}
	Transparent::Transparent()
:CouleurRVBA(Nombre::zero(),Nombre::zero(),Nombre::zero(),Nombre::zero())
{
}

Entier::TypeInterne valeur_numerique(char msb, char lsb)
{
	if(msb>'9') msb -= ('A'-'9');
	if(lsb>'9') lsb -= ('A'-'9');
	return (msb - '0')*16+(lsb - '0');
}

Hexa::Hexa(std::string hexa)
{
	Entier::TypeInterne rouge = valeur_numerique(hexa[0],hexa[1]);
	Entier::TypeInterne vert  = valeur_numerique(hexa[2],hexa[3]);
	Entier::TypeInterne bleu  = valeur_numerique(hexa[4],hexa[5]);
	_type = Traceur::Donnees::Type::Color::SP();
	_valeur1 = NombreSP(new Entier(rouge));
	_valeur2 = NombreSP(new Entier(vert));
	_valeur3 = NombreSP(new Entier(bleu));
	_valeur4 = Nombre::max_canal_couleur();
}
