#include <src/donnees/expressions/expressiondegres.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteur.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_DEGRES(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = DegresSP(new EvaluateurOperateurUnaire<Degres,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<Degres,Entier> : public Degres
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							EntierSP valeur_valeur=std::static_pointer_cast<Entier>(expr_valeur);
							// utiliser deg signifie qu'on a donne un nombre en degres, et qu'on veut le nombre en radians
							SECURISE_VALEUR(ReelSP(new Reel(Entier::radians(*valeur_valeur))));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<Degres,Reel> : public Degres
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ReelSP valeur_valeur=std::static_pointer_cast<Reel>(expr_valeur);
							// utiliser deg signifie qu'on a donne un nombre en degres, et qu'on veut le nombre en radians
							SECURISE_VALEUR(ReelSP(new Reel(Reel::radians(*valeur_valeur))));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Degres::Degres(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(DegresSP(new EvaluateurOperateurUnaire<Degres,Expression>()))
{

	EVALUATEUR_DEGRES(Entier);
	EVALUATEUR_DEGRES(Reel  );
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionPostfixe(position,"deg",valeur);
}

ValeurSP Degres::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	Degres resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP Degres::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void Degres::format(std::ostream& os) const
{
	os << "(";
	_valeur->format(os);
	os << ") deg";
}

ExpressionSP Degres::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	DegresSP clone(new Degres(valeur));
	return clone;
}
