#include <src/donnees/expressions/expressiondifference.h>

#include <src/donnees/types/types.h>

#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionchaine.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteur.h>
#include <src/donnees/expressions/expressionbooleenvrai.h>
#include <src/donnees/expressions/expressionbooleenpeutetre.h>
#include <src/donnees/expressions/expressionbooleenfaux.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/expressions/expressioncouleurhsva.h>

#include <src/donnees/assembleur/systemecoordonnees.h>
#include <src/donnees/assembleur/assembleur.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_DIFFERENCE(type1,type2) \
	if(gauche->est_de_nature<type1>() and droite->est_de_nature<type2>()) { _evaluateur = DifferenceSP(new EvaluateurOperateurBinaire<type1,Difference,type2>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurBinaire<Entier,Difference,Entier> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							EntierSP valeur_gauche=std::static_pointer_cast<Entier>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)!=(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Reel,Difference,Entier> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ReelSP valeur_gauche=std::static_pointer_cast<Reel>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)!=(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Entier,Difference,Reel> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							EntierSP valeur_gauche=std::static_pointer_cast<Entier>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ReelSP valeur_droite=std::static_pointer_cast<Reel>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)!=(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Reel,Difference,Reel> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ReelSP valeur_gauche=std::static_pointer_cast<Reel>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ReelSP valeur_droite=std::static_pointer_cast<Reel>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)!=(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Point,Difference,Point> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							PointSP valeur_gauche=std::static_pointer_cast<Point>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							PointSP valeur_droite=std::static_pointer_cast<Point>(expr_droite);
							try
							{
								Traceur::Donnees::Assembleur::SystemeCoordonnees coords = _assembleur->coordonnees(Valeur::comparaison_hors_image());
								SECURISE_VALEUR(Booleen::SP(not coords.egalite_points(valeur_gauche,valeur_droite)));
							}
							catch(ValeurSP valeur)
							{
								return valeur;
							}
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Vecteur,Difference,Vecteur> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							VecteurSP valeur_gauche=std::static_pointer_cast<Vecteur>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							VecteurSP valeur_droite=std::static_pointer_cast<Vecteur>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)!=(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurRVBA,Difference,CouleurRVBA> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurRVBASP valeur_gauche=std::static_pointer_cast<CouleurRVBA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							CouleurRVBASP valeur_droite=std::static_pointer_cast<CouleurRVBA>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)!=(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurRVBA,Difference,CouleurHSVA> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurRVBASP valeur_gauche=std::static_pointer_cast<CouleurRVBA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							CouleurHSVASP valeur_droite=std::static_pointer_cast<CouleurHSVA>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)!=(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurHSVA,Difference,CouleurRVBA> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurHSVASP valeur_gauche=std::static_pointer_cast<CouleurHSVA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							CouleurRVBASP valeur_droite=std::static_pointer_cast<CouleurRVBA>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)!=(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurHSVA,Difference,CouleurHSVA> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurHSVASP valeur_gauche=std::static_pointer_cast<CouleurHSVA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							CouleurHSVASP valeur_droite=std::static_pointer_cast<CouleurHSVA>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)!=(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Chaine,Difference,Chaine> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ChaineSP valeur_gauche=std::static_pointer_cast<Chaine>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ChaineSP valeur_droite=std::static_pointer_cast<Chaine>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)!=(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenVrai,Difference,BooleenVrai> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenFaux::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenVrai,Difference,BooleenPeutetre> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenVrai::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenVrai,Difference,BooleenFaux> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenVrai::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenPeutetre,Difference,BooleenVrai> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenVrai::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenPeutetre,Difference,BooleenPeutetre> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenFaux::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenPeutetre,Difference,BooleenFaux> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenVrai::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenFaux,Difference,BooleenVrai> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenVrai::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenFaux,Difference,BooleenPeutetre> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenVrai::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenFaux,Difference,BooleenFaux> : public Difference
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenFaux::SP()));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Difference::Difference(ExpressionSP gauche, ExpressionSP droite, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneBinaire(gauche,droite), _evaluateur(DifferenceSP(new EvaluateurOperateurBinaire<Expression,Difference,Expression>()))
{
	EVALUATEUR_DIFFERENCE(Entier         ,Entier         );
	EVALUATEUR_DIFFERENCE(Reel           ,Entier         );
	EVALUATEUR_DIFFERENCE(Entier         ,Reel           );
	EVALUATEUR_DIFFERENCE(Reel           ,Reel           );
	EVALUATEUR_DIFFERENCE(Chaine         ,Chaine         );
	EVALUATEUR_DIFFERENCE(BooleenVrai    ,BooleenVrai    );
	EVALUATEUR_DIFFERENCE(BooleenVrai    ,BooleenPeutetre);
	EVALUATEUR_DIFFERENCE(BooleenVrai    ,BooleenFaux    );
	EVALUATEUR_DIFFERENCE(BooleenPeutetre,BooleenVrai    );
	EVALUATEUR_DIFFERENCE(BooleenPeutetre,BooleenPeutetre);
	EVALUATEUR_DIFFERENCE(BooleenPeutetre,BooleenFaux    );
	EVALUATEUR_DIFFERENCE(BooleenFaux    ,BooleenVrai    );
	EVALUATEUR_DIFFERENCE(BooleenFaux    ,BooleenPeutetre);
	EVALUATEUR_DIFFERENCE(BooleenFaux    ,BooleenFaux    );
	EVALUATEUR_DIFFERENCE(Point          ,Point          );
	EVALUATEUR_DIFFERENCE(Vecteur        ,Vecteur        );
	EVALUATEUR_DIFFERENCE(CouleurRVBA    ,CouleurRVBA    );
	EVALUATEUR_DIFFERENCE(CouleurRVBA    ,CouleurHSVA    );
	EVALUATEUR_DIFFERENCE(CouleurHSVA    ,CouleurRVBA    );
	EVALUATEUR_DIFFERENCE(CouleurHSVA    ,CouleurHSVA    );
	// calcul du type de retour
	if(_type)
		return;
	if(gauche->est_de_type<Traceur::Donnees::Type::Integer>() and droite->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Real>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Number>() and droite->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Number>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Point>() and droite->est_de_type<Traceur::Donnees::Type::Point>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Vector>() and droite->est_de_type<Traceur::Donnees::Type::Vector>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Color>() and droite->est_de_type<Traceur::Donnees::Type::Color>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::String>() and droite->est_de_type<Traceur::Donnees::Type::String>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Boolean>() and droite->est_de_type<Traceur::Donnees::Type::Boolean>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionInfixe(position,gauche,"<>",droite);
}

ValeurSP Difference::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_gauche = _gauche->evaluation(assembleur); if(not valeur_gauche->a_une_valeur_retour()) return valeur_gauche;
	ValeurSP valeur_droite = _droite->evaluation(assembleur); if(not valeur_droite->a_une_valeur_retour()) return valeur_droite;
	Difference resultat(*valeur_gauche,*valeur_droite);
	return resultat.evaluation(valeur_gauche,valeur_droite,assembleur);
}

ValeurSP Difference::evaluation(const ValeurSP gauche, const ValeurSP droite, Traceur::Donnees::Assembleur::AssembleurSP& assembleur) const
{
	_evaluateur->_assembleur=assembleur;
	return _evaluateur->evaluation(gauche,droite);
}

ValeurSP Difference::evaluation(const ValeurSP gauche, const ValeurSP droite) const
{
	throw;
}

void Difference::format(std::ostream& os) const
{
	//_type->format(os);
	os << "((";
	_gauche->format(os);
	os << ")<>(";
	_droite->format(os);
	os << "))";
}

ExpressionSP Difference::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP gauche = _gauche->clone(clones_instances);
	ExpressionSP droite = _droite->clone(clones_instances);
	DifferenceSP clone(new Difference());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_gauche=gauche;
	clone->_droite=droite;
	return clone;
}
