#include <limits.h>

#include <src/donnees/expressions/expressiondivision.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteur.h>
#include <src/donnees/expressions/expressionvecteurpolaire.h>
#include <src/donnees/expressions/expressionchaine.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/expressions/expressioncouleurhsva.h>
#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_DIVISION(type1,type2) \
	if(gauche->est_de_nature<type1>() and droite->est_de_nature<type2>()) { _evaluateur = DivisionSP(new EvaluateurOperateurBinaire<type1,Division,type2>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici

			template<>
				class EvaluateurOperateurBinaire<Entier,Division,Entier> : public Division
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							EntierSP valeur_gauche=std::static_pointer_cast<Entier>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							if((*valeur_droite)==Entier(0))
								return Valeur::division_par_zero();
							else if(((*valeur_gauche)==Entier(INT_MIN)) and ((*valeur_droite)==Entier(-1)))
								return Valeur::depassement_capacite();
							else
							{ SECURISE_VALEUR(EntierSP(new Entier((*valeur_gauche)/(*valeur_droite)))); }
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Reel,Division,Entier> : public Division
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ReelSP valeur_gauche=std::static_pointer_cast<Reel>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							if((*valeur_droite)==Entier(0))
								return Valeur::division_par_zero();
							else
							{ SECURISE_VALEUR(ReelSP(new Reel((*valeur_gauche)/(*valeur_droite)))); }
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Entier,Division,Reel> : public Division
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							EntierSP valeur_gauche=std::static_pointer_cast<Entier>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ReelSP valeur_droite=std::static_pointer_cast<Reel>(expr_droite);
							if((*valeur_droite)==Reel(0))
								return Valeur::division_par_zero();
							else
							{ SECURISE_VALEUR(ReelSP(new Reel((*valeur_gauche)/(*valeur_droite)))); }
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Reel,Division,Reel> : public Division
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ReelSP valeur_gauche=std::static_pointer_cast<Reel>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ReelSP valeur_droite=std::static_pointer_cast<Reel>(expr_droite);
							if((*valeur_droite)==Reel(0))
								return Valeur::division_par_zero();
							else
							{ SECURISE_VALEUR(ReelSP(new Reel((*valeur_gauche)/(*valeur_droite)))); }
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Vecteur,Division,Entier> : public Division
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							VecteurSP valeur_gauche=std::static_pointer_cast<Vecteur>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							if((*valeur_droite)==Entier(0))
								return Valeur::division_par_zero();
							else
							{ SECURISE_VALEUR(VecteurSP(new Vecteur((*valeur_gauche)/(valeur_droite)))); }
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Vecteur,Division,Reel> : public Division
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							VecteurSP valeur_gauche=std::static_pointer_cast<Vecteur>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ReelSP valeur_droite=std::static_pointer_cast<Reel>(expr_droite);
							if((*valeur_droite)==Reel(0))
								return Valeur::division_par_zero();
							else
							{ SECURISE_VALEUR(VecteurSP(new Vecteur((*valeur_gauche)/(valeur_droite)))); }
						}
				};
			template<>
				class EvaluateurOperateurBinaire<VecteurPolaire,Division,Entier> : public Division
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							VecteurPolaireSP valeur_gauche=std::static_pointer_cast<VecteurPolaire>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							if((*valeur_droite)==Entier(0))
								return Valeur::division_par_zero();
							else
							{ SECURISE_VALEUR(VecteurPolaireSP(new VecteurPolaire((*valeur_gauche)/(valeur_droite)))); }
						}
				};
			template<>
				class EvaluateurOperateurBinaire<VecteurPolaire,Division,Reel> : public Division
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							VecteurPolaireSP valeur_gauche=std::static_pointer_cast<VecteurPolaire>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ReelSP valeur_droite=std::static_pointer_cast<Reel>(expr_droite);
							if((*valeur_droite)==Reel(0))
								return Valeur::division_par_zero();
							else
							{ SECURISE_VALEUR(VecteurPolaireSP(new VecteurPolaire((*valeur_gauche)/(valeur_droite)))); }
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Vecteur,Division,Vecteur> : public Division
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							VecteurSP valeur_gauche=std::static_pointer_cast<Vecteur>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							VecteurSP valeur_droite=std::static_pointer_cast<Vecteur>(expr_droite);
							if(valeur_droite->est_le_vecteur_nul())
								return Valeur::projection_sur_vecteur_nul();
							else
							{ SECURISE_VALEUR(VecteurSP(new Vecteur(valeur_gauche->projection(*valeur_droite)))); }
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Chaine,Division,Entier> : public Division
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ChaineSP valeur_gauche=std::static_pointer_cast<Chaine>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							if(((*valeur_droite)<(Entier(0))) or ((*valeur_droite)>=((*valeur_gauche).longueur())))
								return Valeur::hors_domaine();
							else
							{ SECURISE_VALEUR(EntierSP(new Entier((*valeur_gauche)/(*valeur_droite)))); }

						}
				};
			template<>
				class EvaluateurOperateurBinaire<Chaine,Division,Chaine> : public Division
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ChaineSP valeur_gauche=std::static_pointer_cast<Chaine>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ChaineSP valeur_droite=std::static_pointer_cast<Chaine>(expr_droite);
							SECURISE_VALEUR(ChaineSP(new Chaine((*valeur_gauche)/(*valeur_droite))));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurRVBA,Division,Nombre> : public Division
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurRVBASP valeur_gauche=std::static_pointer_cast<CouleurRVBA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							NombreSP valeur_droite=std::static_pointer_cast<Nombre>(expr_droite);
							SECURISE_VALEUR(valeur_gauche->diminue_valeur(valeur_droite));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurHSVA,Division,Nombre> : public Division
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurHSVASP valeur_gauche=std::static_pointer_cast<CouleurHSVA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							NombreSP valeur_droite=std::static_pointer_cast<Nombre>(expr_droite);
							SECURISE_VALEUR(valeur_gauche->diminue_valeur(valeur_droite));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurRVBA,Division,CouleurRVBA> : public Division
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurRVBASP valeur_gauche=std::static_pointer_cast<CouleurRVBA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							CouleurRVBASP valeur_droite=std::static_pointer_cast<CouleurRVBA>(expr_droite);
							SECURISE_VALEUR(valeur_gauche->sur(valeur_droite));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurRVBA,Division,CouleurHSVA> : public Division
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurRVBASP valeur_gauche=std::static_pointer_cast<CouleurRVBA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							CouleurHSVASP valeur_droite=std::static_pointer_cast<CouleurHSVA>(expr_droite);
							SECURISE_VALEUR(valeur_gauche->sur(valeur_droite));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurHSVA,Division,CouleurRVBA> : public Division
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurHSVASP valeur_gauche=std::static_pointer_cast<CouleurHSVA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							CouleurRVBASP valeur_droite=std::static_pointer_cast<CouleurRVBA>(expr_droite);
							SECURISE_VALEUR(valeur_gauche->sur(valeur_droite));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurHSVA,Division,CouleurHSVA> : public Division
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurHSVASP valeur_gauche=std::static_pointer_cast<CouleurHSVA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							CouleurHSVASP valeur_droite=std::static_pointer_cast<CouleurHSVA>(expr_droite);
							SECURISE_VALEUR(valeur_gauche->sur(valeur_droite));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Division::Division(ExpressionSP gauche, ExpressionSP droite, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneBinaire(gauche,droite), _evaluateur(DivisionSP(new EvaluateurOperateurBinaire<Expression,Division,Expression>()))
{

	EVALUATEUR_DIVISION(Entier        ,Entier     );
	EVALUATEUR_DIVISION(Reel          ,Entier     );
	EVALUATEUR_DIVISION(Entier        ,Reel       );
	EVALUATEUR_DIVISION(Reel          ,Reel       );
	EVALUATEUR_DIVISION(Vecteur       ,Entier     );
	EVALUATEUR_DIVISION(VecteurPolaire,Entier     );
	EVALUATEUR_DIVISION(Vecteur       ,Reel       );
	EVALUATEUR_DIVISION(VecteurPolaire,Reel       );
	EVALUATEUR_DIVISION(Vecteur       ,Vecteur    );
	EVALUATEUR_DIVISION(Chaine        ,Entier     );
	EVALUATEUR_DIVISION(Chaine        ,Chaine     );
	EVALUATEUR_DIVISION(CouleurRVBA   ,Nombre     );
	EVALUATEUR_DIVISION(CouleurHSVA   ,Nombre     );
	EVALUATEUR_DIVISION(CouleurRVBA   ,CouleurRVBA);
	EVALUATEUR_DIVISION(CouleurRVBA   ,CouleurHSVA);
	EVALUATEUR_DIVISION(CouleurHSVA   ,CouleurRVBA);
	EVALUATEUR_DIVISION(CouleurHSVA   ,CouleurHSVA);
	// calcul du type de retour
	if(_type)
		return;
	if(gauche->est_de_type<Traceur::Donnees::Type::Integer>() and droite->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Integer::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Real>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Number>() and droite->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Number>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Vector>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Vector::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Vector>() and droite->est_de_type<Traceur::Donnees::Type::Vector>())
	{
		_type = Traceur::Donnees::Type::Vector::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::String>() and droite->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Integer::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::String>() and droite->est_de_type<Traceur::Donnees::Type::String>())
	{
		_type = Traceur::Donnees::Type::String::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Color>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Color::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Color>() and droite->est_de_type<Traceur::Donnees::Type::Color>())
	{
		_type = Traceur::Donnees::Type::Color::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionInfixe(position,gauche,"/",droite);
}

ValeurSP Division::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_gauche = _gauche->evaluation(assembleur); if(not valeur_gauche->a_une_valeur_retour()) return valeur_gauche;
	ValeurSP valeur_droite = _droite->evaluation(assembleur); if(not valeur_droite->a_une_valeur_retour()) return valeur_droite;
	Division resultat(*valeur_gauche,*valeur_droite);
	return resultat.evaluation(valeur_gauche,valeur_droite);
}

ValeurSP Division::evaluation(const ValeurSP gauche, const ValeurSP droite) const
{
	return _evaluateur->evaluation(gauche,droite);
}

void Division::format(std::ostream& os) const
{
	//_type->format(os);
	os << "(";
	_gauche->format(os);
	os << ")/(";
	_droite->format(os);
	os << ")";
}
ExpressionSP Division::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP gauche = _gauche->clone(clones_instances);
	ExpressionSP droite = _droite->clone(clones_instances);
	DivisionSP clone(new Division());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_gauche=gauche;
	clone->_droite=droite;
	return clone;
}
