#include <src/donnees/expressions/expressionegalite.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionchaine.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteur.h>
#include <src/donnees/expressions/expressionbooleenvrai.h>
#include <src/donnees/expressions/expressionbooleenpeutetre.h>
#include <src/donnees/expressions/expressionbooleenfaux.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/expressions/expressioncouleurhsva.h>

#include <src/donnees/assembleur/systemecoordonnees.h>
#include <src/donnees/assembleur/assembleur.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_EGALITE(type1,type2) \
	if(gauche->est_de_nature<type1>() and droite->est_de_nature<type2>()) { _evaluateur = EgaliteSP(new EvaluateurOperateurBinaire<type1,Egalite,type2>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurBinaire<Entier,Egalite,Entier> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							EntierSP valeur_gauche=std::static_pointer_cast<Entier>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)==(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Reel,Egalite,Entier> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ReelSP valeur_gauche=std::static_pointer_cast<Reel>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)==(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Entier,Egalite,Reel> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							EntierSP valeur_gauche=std::static_pointer_cast<Entier>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ReelSP valeur_droite=std::static_pointer_cast<Reel>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)==(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Reel,Egalite,Reel> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ReelSP valeur_gauche=std::static_pointer_cast<Reel>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ReelSP valeur_droite=std::static_pointer_cast<Reel>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)==(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Point,Egalite,Point> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							PointSP valeur_gauche=std::static_pointer_cast<Point>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							PointSP valeur_droite=std::static_pointer_cast<Point>(expr_droite);
							try
							{
								Traceur::Donnees::Assembleur::SystemeCoordonnees coords = _assembleur->coordonnees(Valeur::comparaison_hors_image());
								SECURISE_VALEUR(Booleen::SP(coords.egalite_points(valeur_gauche,valeur_droite)));
							}
							catch(ValeurSP valeur)
							{
								return valeur;
							}
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Vecteur,Egalite,Vecteur> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							VecteurSP valeur_gauche=std::static_pointer_cast<Vecteur>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							VecteurSP valeur_droite=std::static_pointer_cast<Vecteur>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)==(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Chaine,Egalite,Chaine> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ChaineSP valeur_gauche=std::static_pointer_cast<Chaine>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ChaineSP valeur_droite=std::static_pointer_cast<Chaine>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)==(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurRVBA,Egalite,CouleurRVBA> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurRVBASP valeur_gauche=std::static_pointer_cast<CouleurRVBA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							CouleurRVBASP valeur_droite=std::static_pointer_cast<CouleurRVBA>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)==(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurRVBA,Egalite,CouleurHSVA> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurRVBASP valeur_gauche=std::static_pointer_cast<CouleurRVBA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							CouleurHSVASP valeur_droite=std::static_pointer_cast<CouleurHSVA>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)==(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurHSVA,Egalite,CouleurRVBA> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurHSVASP valeur_gauche=std::static_pointer_cast<CouleurHSVA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							CouleurRVBASP valeur_droite=std::static_pointer_cast<CouleurRVBA>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)==(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurHSVA,Egalite,CouleurHSVA> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurHSVASP valeur_gauche=std::static_pointer_cast<CouleurHSVA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							CouleurHSVASP valeur_droite=std::static_pointer_cast<CouleurHSVA>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)==(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenVrai,Egalite,BooleenVrai> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenVrai::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenVrai,Egalite,BooleenPeutetre> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenFaux::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenVrai,Egalite,BooleenFaux> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenFaux::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenPeutetre,Egalite,BooleenVrai> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenFaux::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenPeutetre,Egalite,BooleenPeutetre> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenVrai::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenPeutetre,Egalite,BooleenFaux> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenFaux::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenFaux,Egalite,BooleenVrai> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenFaux::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenFaux,Egalite,BooleenPeutetre> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenFaux::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenFaux,Egalite,BooleenFaux> : public Egalite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenVrai::SP()));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Egalite::Egalite(ExpressionSP gauche, ExpressionSP droite, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneBinaire(gauche,droite), _evaluateur(EgaliteSP(new EvaluateurOperateurBinaire<Expression,Egalite,Expression>()))
{

	EVALUATEUR_EGALITE(Entier         ,Entier         );
	EVALUATEUR_EGALITE(Reel           ,Entier         );
	EVALUATEUR_EGALITE(Entier         ,Reel           );
	EVALUATEUR_EGALITE(Reel           ,Reel           );
	EVALUATEUR_EGALITE(Chaine         ,Chaine         );
	EVALUATEUR_EGALITE(BooleenVrai    ,BooleenVrai    );
	EVALUATEUR_EGALITE(BooleenVrai    ,BooleenPeutetre);
	EVALUATEUR_EGALITE(BooleenVrai    ,BooleenFaux    );
	EVALUATEUR_EGALITE(BooleenPeutetre,BooleenVrai    );
	EVALUATEUR_EGALITE(BooleenPeutetre,BooleenPeutetre);
	EVALUATEUR_EGALITE(BooleenPeutetre,BooleenFaux    );
	EVALUATEUR_EGALITE(BooleenFaux    ,BooleenVrai    );
	EVALUATEUR_EGALITE(BooleenFaux    ,BooleenPeutetre);
	EVALUATEUR_EGALITE(BooleenFaux    ,BooleenFaux    );
	EVALUATEUR_EGALITE(Point          ,Point          );
	EVALUATEUR_EGALITE(Vecteur        ,Vecteur        );
	EVALUATEUR_EGALITE(CouleurRVBA    ,CouleurRVBA    );
	EVALUATEUR_EGALITE(CouleurRVBA    ,CouleurHSVA    );
	EVALUATEUR_EGALITE(CouleurHSVA    ,CouleurRVBA    );
	EVALUATEUR_EGALITE(CouleurHSVA    ,CouleurHSVA    );
	// calcul du type de retour
	if(_type)
		return;
	if(gauche->est_de_type<Traceur::Donnees::Type::Integer>() and droite->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Real>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Number>() and droite->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Number>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Point>() and droite->est_de_type<Traceur::Donnees::Type::Point>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Vector>() and droite->est_de_type<Traceur::Donnees::Type::Vector>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::String>() and droite->est_de_type<Traceur::Donnees::Type::String>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Color>() and droite->est_de_type<Traceur::Donnees::Type::Color>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Boolean>() and droite->est_de_type<Traceur::Donnees::Type::Boolean>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionInfixe(position,gauche,"=",droite);
}

ValeurSP Egalite::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_gauche = _gauche->evaluation(assembleur); if(not valeur_gauche->a_une_valeur_retour()) return valeur_gauche;
	ValeurSP valeur_droite = _droite->evaluation(assembleur); if(not valeur_droite->a_une_valeur_retour()) return valeur_droite;
	Egalite resultat(*valeur_gauche,*valeur_droite);
	return resultat.evaluation(valeur_gauche,valeur_droite,assembleur);
}

ValeurSP Egalite::evaluation(const ValeurSP gauche, const ValeurSP droite, Traceur::Donnees::Assembleur::AssembleurSP& assembleur) const
{
	_evaluateur->_assembleur = assembleur;
	return _evaluateur->evaluation(gauche,droite);
}

ValeurSP Egalite::evaluation(const ValeurSP gauche, const ValeurSP droite) const
{
	throw;
}

void Egalite::format(std::ostream& os) const
{
	//_type->format(os);
	os << "((";
	_gauche->format(os);
	os << ")=(";
	_droite->format(os);
	os << "))";
}

ExpressionSP Egalite::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP gauche = _gauche->clone(clones_instances);
	ExpressionSP droite = _droite->clone(clones_instances);
	EgaliteSP clone(new Egalite());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_gauche=gauche;
	clone->_droite=droite;
	return clone;
}
