#include <src/donnees/expressions/expressionellipse.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionnombre.h>
#include <src/donnees/expressions/expressionvecteur.h>
#include <src/donnees/expressions/expressionvecteurpolaire.h>
#include <src/donnees/assembleur/point.h>
#include <src/donnees/assembleur/drawellipse.h>
#include <src/donnees/assembleur/image.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_ELLIPSE(type1,type2,type3) \
	if(centre->est_de_nature<type1>() and grand_axe->est_de_nature<type2>() and petit_axe->est_de_nature<type3>()) { _evaluateur = EllipseSP(new EvaluateurOperateurTernaire<Ellipse,type1,type2,type3>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurTernaire<Ellipse,Point,Nombre,Nombre> : public Ellipse
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3) const
						{
							ExpressionSP expr_valeur1(*valeur1);
							PointSP valeur_valeur1=std::static_pointer_cast<Point>(expr_valeur1);
							ExpressionSP expr_valeur2(*valeur2);
							NombreSP valeur_valeur2=std::static_pointer_cast<Nombre>(expr_valeur2);
							ExpressionSP expr_valeur3(*valeur3);
							NombreSP valeur_valeur3=std::static_pointer_cast<Nombre>(expr_valeur3);
							SECURISE_VALEUR(EllipseSP(new Ellipse(valeur_valeur1,valeur_valeur2,valeur_valeur3,_options)));
						}
				};
			template<>
				class EvaluateurOperateurTernaire<Ellipse,Point,Vecteur,Nombre> : public Ellipse
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3) const
						{
							ExpressionSP expr_valeur1(*valeur1);
							PointSP valeur_valeur1=std::static_pointer_cast<Point>(expr_valeur1);
							ExpressionSP expr_valeur2(*valeur2);
							VecteurSP valeur_valeur2=std::static_pointer_cast<Vecteur>(expr_valeur2);
							ExpressionSP expr_valeur3(*valeur3);
							NombreSP valeur_valeur3=std::static_pointer_cast<Nombre>(expr_valeur3);
							SECURISE_VALEUR(EllipseSP(new Ellipse(valeur_valeur1,valeur_valeur2,valeur_valeur3,_options)));
						}
				};
			template<>
				class EvaluateurOperateurTernaire<Ellipse,Point,VecteurPolaire,Nombre> : public Ellipse
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3) const
						{
							ExpressionSP expr_valeur1(*valeur1);
							PointSP valeur_valeur1=std::static_pointer_cast<Point>(expr_valeur1);
							ExpressionSP expr_valeur2(*valeur2);
							VecteurPolaireSP valeur_valeur2=std::static_pointer_cast<VecteurPolaire>(expr_valeur2);
							ExpressionSP expr_valeur3(*valeur3);
							NombreSP valeur_valeur3=std::static_pointer_cast<Nombre>(expr_valeur3);
							SECURISE_VALEUR(EllipseSP(new Ellipse(valeur_valeur1,valeur_valeur2,valeur_valeur3,_options)));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Ellipse::Ellipse(ExpressionSP centre, ExpressionSP grand_axe, ExpressionSP petit_axe, OptionsSurfaceSP options, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneTernaire(centre,grand_axe,petit_axe), _evaluateur(EllipseSP(new EvaluateurOperateurTernaire<Ellipse,Expression,Expression,Expression>())), _options(options)
{
	EVALUATEUR_ELLIPSE(Point,Nombre        ,Nombre);
	EVALUATEUR_ELLIPSE(Point,Vecteur       ,Nombre);
	EVALUATEUR_ELLIPSE(Point,VecteurPolaire,Nombre);
	// calcul du type de retour
	if(_type)
		return;
	if(centre->est_de_type<Traceur::Donnees::Type::Point>())
	{
		if(grand_axe->est_de_type<Traceur::Donnees::Type::Number>() and petit_axe->est_de_type<Traceur::Donnees::Type::Number>())
		{
			_type = Traceur::Donnees::Type::Ellipse::SP();
			return;
		}
		if(grand_axe->est_de_type<Traceur::Donnees::Type::Vector>() and petit_axe->est_de_type<Traceur::Donnees::Type::Number>())
		{
			_type = Traceur::Donnees::Type::Ellipse::SP();
			return;
		}
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionTrois(position,"ELLIPSE",centre,grand_axe,petit_axe);
}

ValeurSP Ellipse::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur1 = _valeur1->evaluation(assembleur); if(not valeur_valeur1->a_une_valeur_retour()) return valeur_valeur1;
	ValeurSP valeur_valeur2 = _valeur2->evaluation(assembleur); if(not valeur_valeur2->a_une_valeur_retour()) return valeur_valeur2;
	ValeurSP valeur_valeur3 = _valeur3->evaluation(assembleur); if(not valeur_valeur3->a_une_valeur_retour()) return valeur_valeur3;
	try
	{
		OptionsSurfaceSP options = _options->evaluation(assembleur);
		Ellipse resultat(*valeur_valeur1,*valeur_valeur2,*valeur_valeur3,options);
		return resultat.evaluation(valeur_valeur1,valeur_valeur2,valeur_valeur3,options);
	} catch(ValeurSP valeur) { return valeur; }
}

ValeurSP Ellipse::evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3 , const OptionsSurfaceSP options) const
{
	_evaluateur->_options=options;
	return _evaluateur->evaluation(valeur1,valeur2,valeur3);
}

ValeurSP Ellipse::evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3) const
{
	return _evaluateur->evaluation(valeur1,valeur2,valeur3);
}

void Ellipse::format(std::ostream& os) const
{
	//_type->format(os);
	os << "ELLIPSE[" << _valeur1 << ";" << _valeur2 << ";" << _valeur3 << _options << "]";
}

ExpressionSP Ellipse::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur1 = _valeur1->clone(clones_instances);
	ExpressionSP valeur2 = _valeur2->clone(clones_instances);
	ExpressionSP valeur3 = _valeur3->clone(clones_instances);
	EllipseSP clone(new Ellipse());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_valeur1=valeur1;
	clone->_valeur2=valeur2;
	clone->_valeur3=valeur3;
	clone->_options = _options->clone(clones_instances);
	return clone;
}

Traceur::Donnees::Assembleur::InstructionSP Ellipse::draw(Traceur::Donnees::Assembleur::Assembleur& assembleur)
{
	PointSP expr_valeur1 = std::static_pointer_cast<Point>(_valeur1);
	Traceur::Donnees::Assembleur::Point centre(expr_valeur1,assembleur._pile.top()->_systeme_coordonnees);
	Traceur::Donnees::Assembleur::Point axe_horizontal;
	Traceur::Donnees::Assembleur::Point axe_vertical;
	if(_valeur2->est_de_nature<Nombre>())
	{
		NombreSP expr_valeur2 = std::static_pointer_cast<Nombre>(_valeur2);
		VecteurSP vecteur_axe_horizontal(new Vecteur(_valeur2,Entier::zero()));
		PointSP point_axe_horizontal(new Point(expr_valeur1->translation(*vecteur_axe_horizontal)));
		axe_horizontal=Traceur::Donnees::Assembleur::Point(point_axe_horizontal,assembleur._pile.top()->_systeme_coordonnees);
		NombreSP expr_valeur3 = std::static_pointer_cast<Nombre>(_valeur3);
		VecteurSP vecteur_axe_vertical(new Vecteur(Entier::zero(),_valeur3));
		PointSP point_axe_vertical(new Point(expr_valeur1->translation(*vecteur_axe_vertical)));
		axe_vertical=Traceur::Donnees::Assembleur::Point(point_axe_vertical,assembleur._pile.top()->_systeme_coordonnees);
	}
	else if(_valeur2->est_de_nature<Vecteur>())
	{
		VecteurSP expr_valeur2 = std::dynamic_pointer_cast<Vecteur>(_valeur2);
		PointSP point2(new Point(expr_valeur1->translation(*expr_valeur2)));
		axe_horizontal=Traceur::Donnees::Assembleur::Point(point2,assembleur._pile.top()->_systeme_coordonnees);
		NombreSP expr_valeur3 = std::dynamic_pointer_cast<Nombre>(_valeur3);
		VecteurSP vecteur_orthogonal(new VecteurPolaire(expr_valeur3,expr_valeur2->rotation(Nombre::pi_sur_deux()).argument()));
		PointSP point3(new Point(expr_valeur1->translation(*vecteur_orthogonal)));
		axe_vertical=Traceur::Donnees::Assembleur::Point(point3,assembleur._pile.top()->_systeme_coordonnees);
	}
	else { throw; }
	Traceur::Donnees::Assembleur::InstructionSP drawellipse(new Traceur::Donnees::Assembleur::DrawEllipse(centre,axe_horizontal,axe_vertical,*_options));
	return drawellipse;
}
