#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionchaine.h>
#include <cstdlib>
#include <ctime>
using namespace Traceur::Donnees::Expression;

Entier Entier::entier_borne(Entier min, Entier max)
{
	if((*this)<min)
		return min;
	if((*this)>max)
		return max;
	return *this;
}

Entier Entier::operator+()
{
	return Entier(_entier);
}
Entier Entier::operator-()
{
	return Entier(-_entier);
}
Entier Entier::operator+(Entier entier)
{
	return Entier(_entier+entier._entier);
}
Reel Entier::operator+(Reel reel)
{
	Reel moi(*this);
	return Reel(moi._reel+reel._reel);
}
Entier Entier::operator-(Entier entier)
{
	return Entier(_entier-entier._entier);
}
Reel Entier::operator-(Reel reel)
{
	Reel moi(*this);
	return Reel(moi._reel-reel._reel);
}
Entier Entier::operator*(Entier entier)
{
	return Entier(_entier*entier._entier);
}
Reel Entier::operator*(Reel reel)
{
	Reel moi(*this);
	return Reel(moi._reel*reel._reel);
}
Entier Entier::operator/(Entier entier)
{
	return Entier(_entier/entier._entier);
}
Reel Entier::operator/(Reel reel)
{
	Reel moi(*this);
	return Reel(moi._reel/reel._reel);
}
Entier Entier::operator%(Entier entier)
{
	return Entier(_entier%entier._entier);
}
Reel Entier::operator%(Reel reel)
{
	Reel moi(*this);
	return moi%reel;
}
Entier::TypeInterne Entier::exposant(Entier::TypeInterne exp)
{
	if(not exp)
		return 1;
	if(exp%2)
		return _entier*exposant(exp-1);
	TypeInterne res = exposant(exp/2);
	return res*res;
}

Entier Entier::operator^(Entier entier)
{
	return Entier(exposant(entier._entier));
}

Reel Entier::operator^(Reel reel)
{
	Reel moi(*this);
	return moi^reel;
}
bool Entier::operator==(Entier entier)
{
	return (_entier==entier._entier);
}
bool Entier::operator==(Reel reel)
{
	Reel moi(*this);
	return moi==reel;
}
bool Entier::operator!=(Entier entier)
{
	return (_entier!=entier._entier);
}
bool Entier::operator!=(Reel reel)
{
	Reel moi(*this);
	return moi!=reel;
}
bool Entier::operator<(Entier entier)
{
	return (_entier<entier._entier);
}
bool Entier::operator<(Reel reel)
{
	Reel moi(*this);
	return moi<reel;
}
bool Entier::operator>(Entier entier)
{
	return (_entier>entier._entier);
}
bool Entier::operator>(Reel reel)
{
	Reel moi(*this);
	return moi>reel;
}
bool Entier::operator<=(Entier entier)
{
	return (_entier<=entier._entier);
}
bool Entier::operator<=(Reel reel)
{
	Reel moi(*this);
	return moi<=reel;
}
bool Entier::operator>=(Entier entier)
{
	return (_entier>=entier._entier);
}
bool Entier::operator>=(Reel reel)
{
	Reel moi(*this);
	return moi>=reel;
}
Reel Entier::cosinus(Entier& entier)
{
	Reel reel(entier);
	return Reel(::cos(reel._reel));
}
Reel Entier::sinus(Entier& entier)
{
	Reel reel(entier);
	return Reel(::sin(reel._reel));
}
Reel Entier::tangente(Entier& entier)
{
	Reel reel(entier);
	return Reel(::tan(reel._reel));
}
Reel Entier::arccosinus(Entier& entier)
{
	Reel reel(entier);
	return Reel(::acos(reel._reel));
}
Reel Entier::arcsinus(Entier& entier)
{
	Reel reel(entier);
	return Reel(::asin(reel._reel));
}
Reel Entier::arctangente(Entier& entier)
{
	Reel reel(entier);
	return Reel(::atan(reel._reel));
}
Reel Entier::exponentielle(Entier& entier)
{
	Reel reel(entier);
	return Reel(::exp(reel._reel));
}
Reel Entier::logarithme_naturel(Entier& entier)
{
	Reel reel(entier);
	return Reel(::log(reel._reel));
}
Reel Entier::racine_carree(Entier& entier)
{
	Reel reel(entier);
	return Reel(::sqrt(reel._reel));
}
Entier Entier::valeur_absolue(Entier& entier)
{
	if(entier._entier>=0)
		return entier;
	return Entier(-entier._entier);
}

Entier Entier::partie_entiere(Entier& entier)
{
	return entier;
}

Entier Entier::partie_decimale(Entier& entier)
{
	return Entier(0);
}

struct InitialiseAleatoire
{
	InitialiseAleatoire()
	{
		::srand(::time(NULL));
	};
};

EntierSP Entier::aleatoire()
{
	static InitialiseAleatoire initialisation;
	EntierSP valeur(new Entier(::rand()));
	return valeur;
}

Entier Entier::pgcd(Entier gauche, Entier droite)
{
	if(droite==Entier(0))
		return gauche;
	return Entier::pgcd(droite,gauche%droite);
}

Entier Entier::ppcm(Entier gauche, Entier droite)
{
	Entier pgcd = Entier::pgcd(gauche,droite);
	if(pgcd==Entier(0))
	{
		return Entier(0);
	}
	return gauche*droite/pgcd;
}

Reel Entier::degres(Entier& entier)
{
	Reel reel(entier);
	return Reel((reel._reel*180)/M_PI);
}
Reel Entier::radians(Entier& entier)
{
	Reel reel(entier);
	return Reel((reel._reel*M_PI)/180.);
}
Entier::Entier(const Reel& reel)
	:_entier(static_cast<Entier::TypeInterne>(::floor(reel._reel+0.5)))
{
	_type = Traceur::Donnees::Type::Integer::SP();
}

Entier::Entier(const Chaine& chaine)
{
	_type = Traceur::Donnees::Type::Integer::SP();
	std::istringstream iss;
	iss.str(chaine._chaine);
	iss >> _entier;
}

Traceur::Donnees::Instruction::ValeurSP Entier::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	return Traceur::Donnees::Instruction::ValeurSP(new Traceur::Donnees::Instruction::Valeur(EntierSP(new Entier(*this))));
}

void Entier::format(std::ostream& os) const
{
	os << _entier;
}
