#ifndef _DONNEES_EXPRESSIONS_ENTIER_H_
#define _DONNEES_EXPRESSIONS_ENTIER_H_


#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Expression { DECL_SHARED_PTR(Entier); } } }

#include <src/donnees/expressions/expressionnombre.h>
#include <src/donnees/expressions/expressionbooleenvrai.h>
#include <src/donnees/expressions/expressionbooleenfaux.h>
#include <src/donnees/types/typeinteger.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Assembleur { class Assembleur; class Point; class DrawEllipse; namespace Image { class Reelle; class Frequencielle; class Noyau; } class SystemeCoordonnees; namespace Couleur { class Reelle; } }
		namespace Expression
		{
			class Reel;
			class Chaine;
			class CouleurRVBA;
			class OptionsLigne;
			class Entier: public Nombre
			{
				friend class Reel;
				friend class Chaine;
				friend class CouleurRVBA;
				friend class OptionsLigne;
				friend class Traceur::Donnees::Assembleur::Assembleur;
				friend class Traceur::Donnees::Assembleur::Couleur::Reelle;
				friend class Traceur::Donnees::Assembleur::Image::Reelle;
				friend class Traceur::Donnees::Assembleur::Image::Frequencielle;
				friend class Traceur::Donnees::Assembleur::Image::Noyau;
				friend class Traceur::Donnees::Assembleur::SystemeCoordonnees;
				friend class Traceur::Donnees::Assembleur::Point;
				friend class Traceur::Donnees::Assembleur::DrawEllipse;
				public:
					typedef long int TypeInterne;
					Entier(Entier::TypeInterne entier)
						:_entier(entier) { _type=Traceur::Donnees::Type::Integer::SP(); };
					Entier(const Reel& reel);
					Entier(const Chaine& chaine);
					static EntierSP zero() { static EntierSP constante_zero(new Entier(0)); return constante_zero; };
					static EntierSP un() { static EntierSP constante_un(new Entier(1)); return constante_un; };
					static EntierSP deux() { static EntierSP constante_deux(new Entier(2)); return constante_deux; };
					static EntierSP quatre() { static EntierSP constante_quatre(new Entier(4)); return constante_quatre; };
					static EntierSP quarante_cinq() { static EntierSP constante_quarante_cinq(new Entier(45)); return constante_quarante_cinq; };
					static EntierSP quatre_vingt_dix() { static EntierSP constante_quatre_vingt_dix(new Entier(90)); return constante_quatre_vingt_dix; };
					static EntierSP cent() { static EntierSP constante_cent(new Entier(100)); return constante_cent; };
					static EntierSP cent_quatre_vingts() { static EntierSP constante_cent_quatre_vingts(new Entier(180)); return constante_cent_quatre_vingts; };
					static EntierSP trois_cent_soixante() { static EntierSP constante_trois_cent_soixante(new Entier(360)); return constante_trois_cent_soixante; };
					static EntierSP dix_mille() { static EntierSP constante_dix_mille(new Entier(10000)); return constante_dix_mille; };
					static EntierSP med_canal_couleur() { static EntierSP constante_med_canal_couleur(new Entier(127)); return constante_med_canal_couleur; };
					static EntierSP max_canal_couleur() { static EntierSP constante_max_canal_couleur(new Entier(255)); return constante_max_canal_couleur; };
					Entier entier_borne(Entier min, Entier max);
					virtual Traceur::Donnees::Instruction::ValeurSP evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur);
					virtual void format(std::ostream& os) const;
					virtual ~Entier() {};
					Entier operator+();
					Entier operator-();
					Entier operator+(Entier entier);
					Entier operator-(Entier entier);
					Entier operator*(Entier entier);
					Entier operator/(Entier entier);
					Entier operator%(Entier entier);
					Entier operator^(Entier entier);
					bool operator==(Entier entier);
					bool operator!=(Entier entier);
					bool operator<(Entier entier);
					bool operator>(Entier entier);
					bool operator<=(Entier entier);
					bool operator>=(Entier entier);
					static Reel cosinus(Entier& entier);
					static Reel sinus(Entier& entier);
					static Reel tangente(Entier& entier);
					static Reel arccosinus(Entier& entier);
					static Reel arcsinus(Entier& entier);
					static Reel arctangente(Entier& entier);
					static Reel exponentielle(Entier& entier);
					static Reel logarithme_naturel(Entier& entier);
					static Reel racine_carree(Entier& entier);
					static Entier valeur_absolue(Entier& entier);
					static Entier partie_entiere(Entier& entier);
					static Entier partie_decimale(Entier& entier);
					static EntierSP aleatoire();
					static Entier pgcd(Entier gauche, Entier droite);
					static Entier ppcm(Entier gauche, Entier droite);
					static Reel degres(Entier& entier);
					static Reel radians(Entier& entier);
					Reel operator+(Reel reel);
					Reel operator-(Reel reel);
					Reel operator*(Reel reel);
					Reel operator/(Reel reel);
					Reel operator%(Reel reel);
					Reel operator^(Reel reel);
					bool operator==(Reel reel);
					bool operator!=(Reel reel);
					bool operator<(Reel reel);
					bool operator>(Reel reel);
					bool operator<=(Reel reel);
					bool operator>=(Reel reel);
				private:
					Entier::TypeInterne _entier;
					TypeInterne exposant(TypeInterne exp);
			};
		}
	}
}

#endif
