#include <src/donnees/expressions/expressionetlogique.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionbooleenvrai.h>
#include <src/donnees/expressions/expressionbooleenfaux.h>
#include <src/donnees/expressions/expressionbooleenpeutetre.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_ETLOGIQUE(type1,type2) \
	if(gauche->est_de_nature<type1>() and droite->est_de_nature<type2>()) { _evaluateur = EtLogiqueSP(new EvaluateurOperateurBinaire<type1,EtLogique,type2>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurBinaire<BooleenVrai,EtLogique,BooleenVrai> : public EtLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenVrai::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenVrai,EtLogique,BooleenPeutetre> : public EtLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenPeutetre::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenVrai,EtLogique,BooleenFaux> : public EtLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenFaux::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenPeutetre,EtLogique,BooleenVrai> : public EtLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenPeutetre::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenPeutetre,EtLogique,BooleenPeutetre> : public EtLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenPeutetre::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenPeutetre,EtLogique,BooleenFaux> : public EtLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenFaux::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenFaux,EtLogique,BooleenVrai> : public EtLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenFaux::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenFaux,EtLogique,BooleenPeutetre> : public EtLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenFaux::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenFaux,EtLogique,BooleenFaux> : public EtLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenFaux::SP()));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	EtLogique::EtLogique(ExpressionSP gauche, ExpressionSP droite, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneBinaire(gauche,droite), _evaluateur(EtLogiqueSP(new EvaluateurOperateurBinaire<Expression,EtLogique,Expression>()))
{

	EVALUATEUR_ETLOGIQUE(BooleenVrai    ,BooleenVrai    );
	EVALUATEUR_ETLOGIQUE(BooleenVrai    ,BooleenPeutetre);
	EVALUATEUR_ETLOGIQUE(BooleenVrai    ,BooleenFaux    );
	EVALUATEUR_ETLOGIQUE(BooleenPeutetre,BooleenVrai    );
	EVALUATEUR_ETLOGIQUE(BooleenPeutetre,BooleenPeutetre);
	EVALUATEUR_ETLOGIQUE(BooleenPeutetre,BooleenFaux    );
	EVALUATEUR_ETLOGIQUE(BooleenFaux    ,BooleenVrai    );
	EVALUATEUR_ETLOGIQUE(BooleenFaux    ,BooleenPeutetre);
	EVALUATEUR_ETLOGIQUE(BooleenFaux    ,BooleenFaux    );
	// calcul du type de retour
	if(_type)
		return;
	if(gauche->est_de_type<Traceur::Donnees::Type::Boolean>() and droite->est_de_type<Traceur::Donnees::Type::Boolean>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionInfixe(position,gauche,"and",droite);
}

ValeurSP EtLogique::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_gauche = _gauche->evaluation(assembleur); if(not valeur_gauche->a_une_valeur_retour()) return valeur_gauche;
	ValeurSP valeur_droite = _droite->evaluation(assembleur); if(not valeur_droite->a_une_valeur_retour()) return valeur_droite;
	EtLogique resultat(*valeur_gauche,*valeur_droite);
	return resultat.evaluation(valeur_gauche,valeur_droite);
}

ValeurSP EtLogique::evaluation(const ValeurSP gauche, const ValeurSP droite) const
{
	return _evaluateur->evaluation(gauche,droite);
}

void EtLogique::format(std::ostream& os) const
{
	os << "(";
	_gauche->format(os);
	os << " and ";
	_droite->format(os);
	os << ")";
}

ExpressionSP EtLogique::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP gauche = _gauche->clone(clones_instances);
	ExpressionSP droite = _droite->clone(clones_instances);
	EtLogiqueSP clone(new EtLogique());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_gauche=gauche;
	clone->_droite=droite;
	return clone;
}
