#ifndef _DONNEES_EXPRESSION_EVALUATEURS_H_
#define _DONNEES_EXPRESSION_EVALUATEURS_H_

#include <src/global/global.h>
#include <src/donnees/instructions/valeur.h>
#include <src/donnees/expressions/expressionerreur.h>
#include <src/donnees/expressions/expressionvide.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			/* Pour etre evaluable, un operateur doit definir:
				.h:
					public:
						virtual Traceur::Donnees::Instruction::ValeurSP evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur);
					private:
						OperateurSP _evaluateur;
						virtual Traceur::Donnees::Instruction::ValeurSP evaluation(const ValueSP valeur ...) const;
					protected:
						Operateur() {};
				.cpp:
#include <src/donnees/expressions/expressionevaluateur.h>
namespace Traceur
{
	namespace Donnees
	{
		namepace Expression
		{
			// Specialisations
			template<>
			class EvaluateurOperateur<Operateur, TypeValeur ...> : public Operateur
			{
				public:
					virtual Traceur::Donnees::Instruction::ValeurSP evaluation(const ValueSP valeur ...) const
					{
						// Code specifique
					}
			};
		}
	}
}

#define EVALUATEUR_OPERATEUR(typevaleur,...) \
	if(champ->est_de_nature<typevaleur>() ...) { _evaluateur = OperateurSP(new EvaluateurOperateurBinaire<Operateur, TypeValeur ...>()); }

Operateur::Operateur(const ExpressionSP champ ...)
	:_evaluateur(OperateurSP(new EvaluateurOperateurBinaire<Operateur, TypeValeur ...>())),_champ(champ) ...
{
	EVALUATEUR_OPERATEUR(TypeValeur ...);
}

Traceur::Donnees::Instruction::ValeurSP Operateur::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	Traceur::Donnees::Instruction::ValeurSP valeur_champ(new Traceur::Donnees::Instruction::Valeur(*_champ));
	...
	Operateur resultat(*valeur_champ ...);
	return resultat.evaluation(valeur_champ ...);
}

Traceur::Donnees::Instruction::ValeurSP Operateur::evaluation(const Traceur::Donnees::Instruction::ValeurSP champ ...) const
{
	return _evaluateur->evaluation(champ ...);
}
			*/
			template<typename OperateurUnaire, typename TypeValeur>
				class EvaluateurOperateurUnaire : public OperateurUnaire
			{
				public:
					virtual Traceur::Donnees::Instruction::ValeurSP evaluation(const Traceur::Donnees::Instruction::ValeurSP valeur) const
					{
						return Traceur::Donnees::Instruction::ValeurSP(new Traceur::Donnees::Instruction::Valeur(SHARED_PTR(OperateurUnaire)(new OperateurUnaire(*valeur))));
					}
			};

			template<typename TypeValeurGauche, typename OperateurBinaire, typename TypeValeurDroite>
				class EvaluateurOperateurBinaire : public OperateurBinaire
			{
				public:
					virtual Traceur::Donnees::Instruction::ValeurSP evaluation(const Traceur::Donnees::Instruction::ValeurSP gauche,
							const Traceur::Donnees::Instruction::ValeurSP droite) const
					{
						return Traceur::Donnees::Instruction::ValeurSP(new Traceur::Donnees::Instruction::Valeur(SHARED_PTR(OperateurBinaire)(new OperateurBinaire(*gauche,*droite))));
					}
			};

			template<typename OperateurTernaire, typename TypeValeur1, typename TypeValeur2, typename TypeValeur3>
				class EvaluateurOperateurTernaire : public OperateurTernaire
			{
				public:
					virtual Traceur::Donnees::Instruction::ValeurSP evaluation(const Traceur::Donnees::Instruction::ValeurSP valeur1,
							const Traceur::Donnees::Instruction::ValeurSP valeur2,
							const Traceur::Donnees::Instruction::ValeurSP valeur3) const
					{
						return Traceur::Donnees::Instruction::ValeurSP(new Traceur::Donnees::Instruction::Valeur(SHARED_PTR(OperateurTernaire)(new OperateurTernaire(*valeur1,*valeur2,*valeur3))));
					}
			};

			template<typename OperateurQuaternaire, typename TypeValeur1, typename TypeValeur2, typename TypeValeur3, typename TypeValeur4>
				class EvaluateurOperateurQuaternaire : public OperateurQuaternaire
			{
				public:
					virtual Traceur::Donnees::Instruction::ValeurSP evaluation(const Traceur::Donnees::Instruction::ValeurSP valeur1,
							const Traceur::Donnees::Instruction::ValeurSP valeur2,
							const Traceur::Donnees::Instruction::ValeurSP valeur3,
							const Traceur::Donnees::Instruction::ValeurSP valeur4) const
					{
						return Traceur::Donnees::Instruction::ValeurSP(new Traceur::Donnees::Instruction::Valeur(SHARED_PTR(OperateurQuaternaire)(new OperateurQuaternaire(*valeur1,*valeur2,*valeur3,*valeur4))));
					}
			};
		}
	}
}
#endif
