#include <src/donnees/expressions/expressionexponentielle.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_EXPONENTIELLE(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ExponentielleSP(new EvaluateurOperateurUnaire<Exponentielle,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<Exponentielle,Entier> : public Exponentielle
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ReelSP valeur_valeur=ReelSP(new Reel(*(std::static_pointer_cast<Entier>(expr_valeur))));
							SECURISE_VALEUR(ReelSP(new Reel(Reel::exponentielle(*valeur_valeur))));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<Exponentielle,Reel> : public Exponentielle
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ReelSP valeur_valeur=std::static_pointer_cast<Reel>(expr_valeur);
							SECURISE_VALEUR(ReelSP(new Reel(Reel::exponentielle(*valeur_valeur))));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Exponentielle::Exponentielle(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ExponentielleSP(new EvaluateurOperateurUnaire<Exponentielle,Expression>()))
{

	EVALUATEUR_EXPONENTIELLE(Entier);
	EVALUATEUR_EXPONENTIELLE(Reel  );
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionPrefixe(position,"exp",valeur);
}

ValeurSP Exponentielle::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	Exponentielle resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP Exponentielle::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void Exponentielle::format(std::ostream& os) const
{
	os << "exp (";
	_valeur->format(os);
	os << ")";
}

ExpressionSP Exponentielle::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ExponentielleSP clone(new Exponentielle(valeur));
	return clone;
}
