#include <src/donnees/expressions/expressionexposant.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionvecteur.h>
#include <src/donnees/expressions/expressionvecteurpolaire.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/expressions/expressioncouleurhsva.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_EXPOSANT(type1,type2) \
	if(gauche->est_de_nature<type1>() and droite->est_de_nature<type2>()) { _evaluateur = ExposantSP(new EvaluateurOperateurBinaire<type1,Exposant,type2>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici

			template<>
				class EvaluateurOperateurBinaire<Entier,Exposant,Entier> : public Exposant
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							EntierSP valeur_gauche=std::static_pointer_cast<Entier>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							if((*valeur_droite)<Entier(0))
								return Valeur::hors_domaine();
							else
							{ SECURISE_VALEUR(EntierSP(new Entier((*valeur_gauche)^(*valeur_droite)))); }
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Reel,Exposant,Entier> : public Exposant
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ReelSP valeur_gauche=std::static_pointer_cast<Reel>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							if(((*valeur_gauche)==Reel(0)) and ((*valeur_droite)<Entier(0)))
								return Valeur::hors_domaine();
							else
							{ SECURISE_VALEUR(ReelSP(new Reel((*valeur_gauche)^(*valeur_droite)))); }
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Entier,Exposant,Reel> : public Exposant
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ReelSP valeur_gauche=ReelSP(new Reel(*(std::static_pointer_cast<Entier>(expr_gauche))));
							ExpressionSP expr_droite(*droite);
							ReelSP valeur_droite=std::static_pointer_cast<Reel>(expr_droite);
							if(Reel::exposantinvalide(*valeur_gauche,*valeur_droite))
								return Valeur::hors_domaine();
							else
							{ SECURISE_VALEUR(ReelSP(new Reel((*valeur_gauche)^(*valeur_droite)))); }
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Reel,Exposant,Reel> : public Exposant
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ReelSP valeur_gauche=std::static_pointer_cast<Reel>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ReelSP valeur_droite=std::static_pointer_cast<Reel>(expr_droite);
							if(Reel::exposantinvalide(*valeur_gauche,*valeur_droite))
								return Valeur::hors_domaine();
							else
							{ SECURISE_VALEUR(ReelSP(new Reel((*valeur_gauche)^(*valeur_droite)))); }
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Vecteur,Exposant,Nombre> : public Exposant
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							VecteurSP valeur_gauche=std::static_pointer_cast<Vecteur>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							NombreSP valeur_droite=std::static_pointer_cast<Nombre>(expr_droite);
							SECURISE_VALEUR(VecteurSP(new Vecteur(valeur_gauche->rotation(valeur_droite))));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<VecteurPolaire,Exposant,Nombre> : public Exposant
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							VecteurPolaireSP valeur_gauche=std::static_pointer_cast<VecteurPolaire>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							NombreSP valeur_droite=std::static_pointer_cast<Nombre>(expr_droite);
							SECURISE_VALEUR(VecteurPolaireSP(new VecteurPolaire(valeur_gauche->rotation(valeur_droite))));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Vecteur,Exposant,Vecteur> : public Exposant
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							VecteurSP valeur_gauche=std::static_pointer_cast<Vecteur>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							VecteurSP valeur_droite=std::static_pointer_cast<Vecteur>(expr_droite);
							SECURISE_VALEUR((*valeur_gauche)^(*valeur_droite));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurRVBA,Exposant,Nombre> : public Exposant
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurRVBASP valeur_gauche=std::static_pointer_cast<CouleurRVBA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							NombreSP valeur_droite=std::static_pointer_cast<Nombre>(expr_droite);
							SECURISE_VALEUR(valeur_gauche->augmente_teinte(valeur_droite));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurHSVA,Exposant,Nombre> : public Exposant
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurHSVASP valeur_gauche=std::static_pointer_cast<CouleurHSVA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							NombreSP valeur_droite=std::static_pointer_cast<Nombre>(expr_droite);
							SECURISE_VALEUR(valeur_gauche->augmente_teinte(valeur_droite));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Exposant::Exposant(ExpressionSP gauche, ExpressionSP droite, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneBinaire(gauche,droite), _evaluateur(ExposantSP(new EvaluateurOperateurBinaire<Expression,Exposant,Expression>()))
{

	EVALUATEUR_EXPOSANT(Entier        ,Entier );
	EVALUATEUR_EXPOSANT(Reel          ,Entier );
	EVALUATEUR_EXPOSANT(Entier        ,Reel   );
	EVALUATEUR_EXPOSANT(Reel          ,Reel   );
	EVALUATEUR_EXPOSANT(Vecteur       ,Nombre );
	EVALUATEUR_EXPOSANT(VecteurPolaire,Nombre );
	EVALUATEUR_EXPOSANT(Vecteur       ,Vecteur);
	EVALUATEUR_EXPOSANT(CouleurRVBA   ,Nombre );
	EVALUATEUR_EXPOSANT(CouleurHSVA   ,Nombre );
	// calcul du type de retour
	if(_type)
		return;
	if(gauche->est_de_type<Traceur::Donnees::Type::Integer>() and droite->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Integer::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Real>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Number>() and droite->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Number>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Vector>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Vector::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Vector>() and droite->est_de_type<Traceur::Donnees::Type::Vector>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Color>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Color::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionInfixe(position,gauche,"^",droite);
}

ValeurSP Exposant::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_gauche = _gauche->evaluation(assembleur); if(not valeur_gauche->a_une_valeur_retour()) return valeur_gauche;
	ValeurSP valeur_droite = _droite->evaluation(assembleur); if(not valeur_droite->a_une_valeur_retour()) return valeur_droite;
	Exposant resultat(*valeur_gauche,*valeur_droite);
	return resultat.evaluation(valeur_gauche,valeur_droite);
}

ValeurSP Exposant::evaluation(const ValeurSP gauche, const ValeurSP droite) const
{
	return _evaluateur->evaluation(gauche,droite);
}

void Exposant::format(std::ostream& os) const
{
	//_type->format(os);
	os << "(";
	_gauche->format(os);
	os << ")^(";
	_droite->format(os);
	os << ")";
}
ExpressionSP Exposant::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP gauche = _gauche->clone(clones_instances);
	ExpressionSP droite = _droite->clone(clones_instances);
	ExposantSP clone(new Exposant());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_gauche=gauche;
	clone->_droite=droite;
	return clone;
}
