#include <src/donnees/expressions/expressiongomme.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionevaluateur.h>

#include <src/donnees/assembleur/draweraser.h>

#define EVALUATEUR_GOMME(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = GommeSP(new EvaluateurOperateurUnaire<Gomme,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<Gomme,Point> : public Gomme
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur)
						{
							ExpressionSP expr_valeur(*valeur);
							PointSP valeur_valeur=std::static_pointer_cast<Point>(expr_valeur);
							SECURISE_VALEUR(GommeSP(new Gomme(valeur_valeur)));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Gomme::Gomme(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(GommeSP(new EvaluateurOperateurUnaire<Gomme,Expression>()))
{

	EVALUATEUR_GOMME(Point);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Point>())
	{
		_type = Traceur::Donnees::Type::Eraser::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionUn(position,"ERASER",valeur);
}

ValeurSP Gomme::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	Gomme resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP Gomme::evaluation(const ValeurSP valeur)
{
	return _evaluateur->evaluation(valeur);
}

void Gomme::format(std::ostream& os) const
{
	os << "ERASER[";
	_valeur->format(os);
	os << "]";
}

ExpressionSP Gomme::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	GommeSP clone(new Gomme(valeur));
	return clone;
}

Traceur::Donnees::Assembleur::InstructionSP Gomme::draw(Traceur::Donnees::Assembleur::Assembleur& assembleur)
{
	PointSP expr_valeur = std::static_pointer_cast<Point>(_valeur);
	Traceur::Donnees::Assembleur::Point point(expr_valeur,assembleur._pile.top()->_systeme_coordonnees);
	Traceur::Donnees::Assembleur::InstructionSP draweraser(new Traceur::Donnees::Assembleur::DrawEraser(point));
	return draweraser;
}
