#include <src/donnees/expressions/expressionidentite.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionvecteur.h>
#include <src/donnees/expressions/expressionvecteurpolaire.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_IDENTITE(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = IdentiteSP(new EvaluateurOperateurUnaire<Identite,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<Identite,Entier> : public Identite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							EntierSP valeur_valeur=std::static_pointer_cast<Entier>(expr_valeur);
							SECURISE_VALEUR(EntierSP(new Entier(+(*valeur_valeur))));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<Identite,Reel> : public Identite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ReelSP valeur_valeur=std::static_pointer_cast<Reel>(expr_valeur);
							SECURISE_VALEUR(ReelSP(new Reel(+(*valeur_valeur))));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<Identite,Vecteur> : public Identite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							VecteurSP valeur_valeur=std::static_pointer_cast<Vecteur>(expr_valeur);
							SECURISE_VALEUR(VecteurSP(new Vecteur(+(*valeur_valeur))));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<Identite,VecteurPolaire> : public Identite
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							VecteurPolaireSP valeur_valeur=std::static_pointer_cast<VecteurPolaire>(expr_valeur);
							SECURISE_VALEUR(VecteurPolaireSP(new VecteurPolaire(+(*valeur_valeur))));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Identite::Identite(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(IdentiteSP(new EvaluateurOperateurUnaire<Identite,Expression>()))
{

	EVALUATEUR_IDENTITE(Entier        );
	EVALUATEUR_IDENTITE(Reel          );
	EVALUATEUR_IDENTITE(Vecteur       );
	EVALUATEUR_IDENTITE(VecteurPolaire);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Integer::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Vector>())
	{
		_type = Traceur::Donnees::Type::Vector::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionPrefixe(position,"+",valeur);
}

ValeurSP Identite::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	Identite resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP Identite::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void Identite::format(std::ostream& os) const
{
	os << "+(";
	_valeur->format(os);
	os << ")";
}

ExpressionSP Identite::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	IdentiteSP clone(new Identite(valeur));
	return clone;
}
