#include <src/donnees/expressions/expressionimage.h>

#include <src/donnees/types/types.h>
#include <src/donnees/assembleur/assembleur.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteur.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/expressions/expressionimageevaluee.h>

using namespace Traceur::Donnees::Expression;

Image::Image(ExpressionSP taille_x, ExpressionSP taille_y, SystemeCoordonneesSP coordonnees, ExpressionSP couleur_fond,
		Traceur::Donnees::Instruction::BlocSP contenu,Traceur::Donnees::Source::PositionSP position)
: _taille_x(taille_x), _taille_y(taille_y), _coordonnees(coordonnees), _couleur_fond(couleur_fond), _contenu(contenu)
{
	if(not coordonnees)
		_coordonnees = SystemeCoordonneesSP(new SystemeGeometrique());
	if(not couleur_fond)
		_couleur_fond = CouleurRVBASP(new Couleur::Transparent());
	// calcul du type de retour
	if(_taille_x->est_de_type<Traceur::Donnees::Type::Integer>() and _taille_y->est_de_type<Traceur::Donnees::Type::Integer>() and _couleur_fond->est_de_type<Traceur::Donnees::Type::Color>())
	{
		_type = Traceur::Donnees::Type::Image::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectImage(position,_taille_x,_taille_y,_coordonnees,_couleur_fond);
}

Traceur::Donnees::Instruction::ValeurSP Image::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	Traceur::Donnees::Instruction::ValeurSP taille_x = _taille_x->evaluation(assembleur); if(not taille_x->a_une_valeur_retour()) return taille_x;
	Traceur::Donnees::Instruction::ValeurSP taille_y = _taille_y->evaluation(assembleur); if(not taille_y->a_une_valeur_retour()) return taille_y;
	SystemeCoordonneesSP coordonnees;
	try
	{
		coordonnees = _coordonnees->evaluation(assembleur);
		Traceur::Donnees::Instruction::ValeurSP couleur_fond = _couleur_fond->evaluation(assembleur); if(not couleur_fond->a_une_valeur_retour()) return couleur_fond;
		ExpressionSP expr_taille_x = *taille_x;
		ExpressionSP expr_taille_y = *taille_y;
		ExpressionSP expr_couleur_fond = *couleur_fond;
		assembleur->image(expr_taille_x,expr_taille_y,coordonnees,expr_couleur_fond);
	} catch(Traceur::Donnees::Instruction::ValeurSP valeur) { return valeur; }
	Traceur::Donnees::Instruction::InstructionInterruptionSP interruption = _contenu->evaluation(assembleur);
	if(interruption)
	{
		auto action = interruption->comportement_image();
		switch(action.first)
		{
			case Traceur::Donnees::Instruction::InstructionInterruption::Action::SORTIE:
			{
				Traceur::Donnees::Assembleur::Image::ImageSP image = assembleur->image();
				Traceur::Donnees::Expression::ImageEvalueeSP image_evaluee(new Traceur::Donnees::Expression::ImageEvaluee(image));
				Traceur::Donnees::Instruction::ValeurSP valeur(new Traceur::Donnees::Instruction::Valeur(image_evaluee));
				if(action.second)
				{
					valeur->_interruption=action.second;
				}
				return valeur;
			}
			default:
				break;
		}
	}
	Traceur::Donnees::Assembleur::Image::ImageSP image = assembleur->image();
	Traceur::Donnees::Expression::ImageEvalueeSP image_evaluee(new Traceur::Donnees::Expression::ImageEvaluee(image));
	Traceur::Donnees::Instruction::ValeurSP valeur(new Traceur::Donnees::Instruction::Valeur(image_evaluee));
	return valeur;
}

void Image::format(std::ostream& os) const
{
	os << "IMAGE[";
	_taille_x->format(os);
	os <<  ";" ;
	_taille_y->format(os);
	os << "] ";
	_coordonnees->format(os);
	os << " ";
	_couleur_fond->format(os);
	os << std::endl << "{" << std::endl;
	_contenu->format(os);
	os << "}" << std::endl;
}

void Image::edition_liens(Traceur::Donnees::Instruction::InstructionSP instruction)
{
	_taille_x->edition_liens(instruction);
	_taille_y->edition_liens(instruction);
	_coordonnees->edition_liens(instruction);
	_couleur_fond->edition_liens(instruction);
	_contenu->edition_liens(instruction->parent(),instruction);
}

void Image::definition_fonction()
{
	_taille_x->definition_fonction();
	_taille_y->definition_fonction();
	_coordonnees->definition_fonction();
	_couleur_fond->definition_fonction();
	_contenu->definition_fonction();
}

ExpressionSP Image::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP taille_x = _taille_x->clone(clones_instances);
	ExpressionSP taille_y = _taille_y->clone(clones_instances);
	SystemeCoordonneesSP coordonnees = _coordonnees->clone(clones_instances);
	ExpressionSP couleur_fond = _couleur_fond->clone(clones_instances);
	Traceur::Donnees::Instruction::BlocSP contenu = _contenu->clone(clones_instances);
	ImageSP clone(new Image(taille_x,taille_y,coordonnees,couleur_fond,contenu));
	return clone;
}

ExpressionSP Image::simplification()
{
	_taille_x->simplification();
	_taille_y->simplification();
	_coordonnees->simplification();
	_couleur_fond->simplification();
	_contenu->simplification();
	return this->shared_from_this();
}
