#include <src/donnees/expressions/expressioninferieurouegal.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionchaine.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_INFERIEUROUEGAL(type1,type2) \
	if(gauche->est_de_nature<type1>() and droite->est_de_nature<type2>()) { _evaluateur = InferieurOuEgalSP(new EvaluateurOperateurBinaire<type1,InferieurOuEgal,type2>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurBinaire<Entier,InferieurOuEgal,Entier> : public InferieurOuEgal
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							EntierSP valeur_gauche=std::static_pointer_cast<Entier>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)<=(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Reel,InferieurOuEgal,Entier> : public InferieurOuEgal
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ReelSP valeur_gauche=std::static_pointer_cast<Reel>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)<=(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Entier,InferieurOuEgal,Reel> : public InferieurOuEgal
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							EntierSP valeur_gauche=std::static_pointer_cast<Entier>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ReelSP valeur_droite=std::static_pointer_cast<Reel>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)<=(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Reel,InferieurOuEgal,Reel> : public InferieurOuEgal
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ReelSP valeur_gauche=std::static_pointer_cast<Reel>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ReelSP valeur_droite=std::static_pointer_cast<Reel>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)<=(*valeur_droite)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Chaine,InferieurOuEgal,Chaine> : public InferieurOuEgal
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ChaineSP valeur_gauche=std::static_pointer_cast<Chaine>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ChaineSP valeur_droite=std::static_pointer_cast<Chaine>(expr_droite);
							SECURISE_VALEUR(Booleen::SP((*valeur_gauche)<=(*valeur_droite)));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	InferieurOuEgal::InferieurOuEgal(ExpressionSP gauche, ExpressionSP droite, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneBinaire(gauche,droite), _evaluateur(InferieurOuEgalSP(new EvaluateurOperateurBinaire<Expression,InferieurOuEgal,Expression>()))
{

	EVALUATEUR_INFERIEUROUEGAL(Entier         ,Entier         );
	EVALUATEUR_INFERIEUROUEGAL(Reel           ,Entier         );
	EVALUATEUR_INFERIEUROUEGAL(Entier         ,Reel           );
	EVALUATEUR_INFERIEUROUEGAL(Reel           ,Reel           );
	EVALUATEUR_INFERIEUROUEGAL(Chaine         ,Chaine         );
	// calcul du type de retour
	if(_type)
		return;
	if(gauche->est_de_type<Traceur::Donnees::Type::Integer>() and droite->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Real>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Number>() and droite->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Number>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::String>() and droite->est_de_type<Traceur::Donnees::Type::String>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionInfixe(position,gauche,"<=",droite);
}

ValeurSP InferieurOuEgal::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_gauche = _gauche->evaluation(assembleur); if(not valeur_gauche->a_une_valeur_retour()) return valeur_gauche;
	ValeurSP valeur_droite = _droite->evaluation(assembleur); if(not valeur_droite->a_une_valeur_retour()) return valeur_droite;
	InferieurOuEgal resultat(*valeur_gauche,*valeur_droite);
	return resultat.evaluation(valeur_gauche,valeur_droite);
}

ValeurSP InferieurOuEgal::evaluation(const ValeurSP gauche, const ValeurSP droite) const
{
	return _evaluateur->evaluation(gauche,droite);
}

void InferieurOuEgal::format(std::ostream& os) const
{
	//_type->format(os);
	os << "(";
	_gauche->format(os);
	os << ")<=(";
	_droite->format(os);
	os << ")";
}

ExpressionSP InferieurOuEgal::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP gauche = _gauche->clone(clones_instances);
	ExpressionSP droite = _droite->clone(clones_instances);
	InferieurOuEgalSP clone(new InferieurOuEgal());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_gauche=gauche;
	clone->_droite=droite;
	return clone;
}
