#include <src/donnees/expressions/expressionligne.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionnombre.h>
#include <src/donnees/expressions/expressionvecteur.h>
#include <src/donnees/expressions/expressionvecteurpolaire.h>
#include <src/donnees/assembleur/point.h>
#include <src/donnees/assembleur/drawline.h>
#include <src/donnees/assembleur/image.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_LIGNE(type1,type2) \
	if(gauche->est_de_nature<type1>() and droite->est_de_nature<type2>()) { _evaluateur = LigneSP(new EvaluateurOperateurBinaire<type1,Ligne,type2>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurBinaire<Point,Ligne,Point> : public Ligne
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							PointSP valeur_gauche=std::static_pointer_cast<Point>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							PointSP valeur_droite=std::static_pointer_cast<Point>(expr_droite);
							SECURISE_VALEUR(LigneSP(new Ligne(valeur_gauche,valeur_droite,_options)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Point,Ligne,Vecteur> : public Ligne
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							PointSP valeur_gauche=std::static_pointer_cast<Point>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							VecteurSP valeur_droite=std::static_pointer_cast<Vecteur>(expr_droite);
							SECURISE_VALEUR(LigneSP(new Ligne(valeur_gauche,valeur_droite,_options)));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Point,Ligne,VecteurPolaire> : public Ligne
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							PointSP valeur_gauche=std::static_pointer_cast<Point>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							VecteurPolaireSP valeur_droite=std::static_pointer_cast<VecteurPolaire>(expr_droite);
							SECURISE_VALEUR(LigneSP(new Ligne(valeur_gauche,valeur_droite,_options)));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Ligne::Ligne(ExpressionSP gauche, ExpressionSP droite, OptionsLigneSP options, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneBinaire(gauche,droite), _evaluateur(LigneSP(new EvaluateurOperateurBinaire<Expression,Ligne,Expression>())), _options(options)
{
	EVALUATEUR_LIGNE(Point,Point         );
	EVALUATEUR_LIGNE(Point,Vecteur       );
	EVALUATEUR_LIGNE(Point,VecteurPolaire);
	// calcul du type de retour
	if(_type)
		return;
	if(gauche->est_de_type<Traceur::Donnees::Type::Point>())
	{
		if(droite->est_de_type<Traceur::Donnees::Type::Point>())
		{
			_type = Traceur::Donnees::Type::Line::SP();
			return;
		}
		if(droite->est_de_type<Traceur::Donnees::Type::Vector>())
		{
			_type = Traceur::Donnees::Type::Line::SP();
			return;
		}
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionDeux(position,"LINE",gauche,droite);
}

ValeurSP Ligne::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_gauche = _gauche->evaluation(assembleur); if(not valeur_gauche->a_une_valeur_retour()) return valeur_gauche;
	ValeurSP valeur_droite = _droite->evaluation(assembleur); if(not valeur_droite->a_une_valeur_retour()) return valeur_droite;
	try
	{
		OptionsLigneSP options = _options->evaluation(assembleur);
		Ligne resultat(*valeur_gauche,*valeur_droite,options);
		return resultat.evaluation(valeur_gauche,valeur_droite,options);
	} catch(ValeurSP valeur) { return valeur; }
}

ValeurSP Ligne::evaluation(const ValeurSP gauche, const ValeurSP droite, const OptionsLigneSP options) const
{
	_evaluateur->_options=options;
	return _evaluateur->evaluation(gauche,droite);
}

ValeurSP Ligne::evaluation(const ValeurSP gauche, const ValeurSP droite) const
{
	return _evaluateur->evaluation(gauche,droite);
}

void Ligne::format(std::ostream& os) const
{
	//_type->format(os);
	os << "LINE[";
	_gauche->format(os);
	os << ";";
	_droite->format(os);
	_options->format(os);
	os << "]";
}

ExpressionSP Ligne::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP gauche = _gauche->clone(clones_instances);
	ExpressionSP droite = _droite->clone(clones_instances);
	LigneSP clone(new Ligne());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_gauche=gauche;
	clone->_droite=droite;
	clone->_options = _options->clone(clones_instances);
	return clone;
}

Traceur::Donnees::Assembleur::InstructionSP Ligne::draw(Traceur::Donnees::Assembleur::Assembleur& assembleur)
{
	PointSP expr_gauche = std::static_pointer_cast<Point>(_gauche);
	Traceur::Donnees::Assembleur::Point gauche(expr_gauche,assembleur._pile.top()->_systeme_coordonnees);
	PointSP expr_point;
	if(_droite->est_de_nature<Point>())
	{
		expr_point = std::static_pointer_cast<Point>(_droite);
	}
	if(_droite->est_de_nature<Vecteur>())
	{
		VecteurSP expr_droite = std::dynamic_pointer_cast<Vecteur>(_droite);
		expr_point = PointSP(new Point(expr_gauche->translation(*expr_droite)));
	}
	Traceur::Donnees::Assembleur::Point droite(expr_point,assembleur._pile.top()->_systeme_coordonnees);
	Traceur::Donnees::Assembleur::InstructionSP drawline(new Traceur::Donnees::Assembleur::DrawLine(gauche,droite,*_options));
	return drawline;
}
