#include <src/donnees/expressions/expressionlogarithmenaturel.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_LOGARITHMENATUREL(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = LogarithmeNaturelSP(new EvaluateurOperateurUnaire<LogarithmeNaturel,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<LogarithmeNaturel,Entier> : public LogarithmeNaturel
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							EntierSP valeur_valeur=std::static_pointer_cast<Entier>(expr_valeur);
							if((*valeur_valeur)<=Entier(0))
								return Valeur::hors_domaine();
							else
							{ SECURISE_VALEUR(ReelSP(new Reel(Entier::logarithme_naturel(*valeur_valeur)))); }
						}
				};
			template<>
				class EvaluateurOperateurUnaire<LogarithmeNaturel,Reel> : public LogarithmeNaturel
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ReelSP valeur_valeur=std::static_pointer_cast<Reel>(expr_valeur);
							if((*valeur_valeur)<=Reel(0))
								return Valeur::hors_domaine();
							else
							{ SECURISE_VALEUR(ReelSP(new Reel(Reel::logarithme_naturel(*valeur_valeur)))); }
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	LogarithmeNaturel::LogarithmeNaturel(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(LogarithmeNaturelSP(new EvaluateurOperateurUnaire<LogarithmeNaturel,Expression>()))
{

	EVALUATEUR_LOGARITHMENATUREL(Entier);
	EVALUATEUR_LOGARITHMENATUREL(Reel  );
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionPrefixe(position,"ln",valeur);
}

ValeurSP LogarithmeNaturel::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	LogarithmeNaturel resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP LogarithmeNaturel::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void LogarithmeNaturel::format(std::ostream& os) const
{
	os << "ln (";
	_valeur->format(os);
	os << ")";
}

ExpressionSP LogarithmeNaturel::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	LogarithmeNaturelSP clone(new LogarithmeNaturel(valeur));
	return clone;
}
