#include <src/donnees/expressions/expressionmaximum.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionchaine.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_MAXIMUM(type1,type2) \
	if(gauche->est_de_nature<type1>() and droite->est_de_nature<type2>()) { _evaluateur = MaximumSP(new EvaluateurOperateurBinaire<type1,Maximum,type2>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurBinaire<Entier,Maximum,Entier> : public Maximum
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							EntierSP valeur_gauche=std::static_pointer_cast<Entier>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							if((*valeur_gauche)<(*valeur_droite))
								return droite;
							else
								return gauche;
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Reel,Maximum,Entier> : public Maximum
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ReelSP valeur_gauche=std::static_pointer_cast<Reel>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							if((*valeur_gauche)<(*valeur_droite))
								return droite;
							else
								return gauche;
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Entier,Maximum,Reel> : public Maximum
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							EntierSP valeur_gauche=std::static_pointer_cast<Entier>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ReelSP valeur_droite=std::static_pointer_cast<Reel>(expr_droite);
							if((*valeur_gauche)<(*valeur_droite))
								return droite;
							else
								return gauche;
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Reel,Maximum,Reel> : public Maximum
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ReelSP valeur_gauche=std::static_pointer_cast<Reel>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ReelSP valeur_droite=std::static_pointer_cast<Reel>(expr_droite);
							if((*valeur_gauche)<(*valeur_droite))
								return droite;
							else
								return gauche;
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Chaine,Maximum,Chaine> : public Maximum
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ChaineSP valeur_gauche=std::static_pointer_cast<Chaine>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ChaineSP valeur_droite=std::static_pointer_cast<Chaine>(expr_droite);
							if((*valeur_gauche)<(*valeur_droite))
								return droite;
							else
								return gauche;
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Maximum::Maximum(ExpressionSP gauche, ExpressionSP droite, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneBinaire(gauche,droite), _evaluateur(MaximumSP(new EvaluateurOperateurBinaire<Expression,Maximum,Expression>()))
{

	EVALUATEUR_MAXIMUM(Entier,Entier);
	EVALUATEUR_MAXIMUM(Reel  ,Entier);
	EVALUATEUR_MAXIMUM(Entier,Reel  );
	EVALUATEUR_MAXIMUM(Reel  ,Reel  );
	EVALUATEUR_MAXIMUM(Chaine,Chaine);
	// calcul du type de retour
	if(_type)
		return;
	if(gauche->est_de_type<Traceur::Donnees::Type::Integer>() and droite->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Integer::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Real>() and droite->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Number>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::String>() and droite->est_de_type<Traceur::Donnees::Type::String>())
	{
		_type = Traceur::Donnees::Type::String::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionInfixe(position,gauche,"max",droite);
}

ValeurSP Maximum::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_gauche = _gauche->evaluation(assembleur); if(not valeur_gauche->a_une_valeur_retour()) return valeur_gauche;
	ValeurSP valeur_droite = _droite->evaluation(assembleur); if(not valeur_droite->a_une_valeur_retour()) return valeur_droite;
	Maximum resultat(*valeur_gauche,*valeur_droite);
	return resultat.evaluation(valeur_gauche,valeur_droite);
}

ValeurSP Maximum::evaluation(const ValeurSP gauche, const ValeurSP droite) const
{
	return _evaluateur->evaluation(gauche,droite);
}

void Maximum::format(std::ostream& os) const
{
	//_type->format(os);
	os << "(";
	_gauche->format(os);
	os << ") max (";
	_droite->format(os);
	os << ")";
}

ExpressionSP Maximum::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP gauche = _gauche->clone(clones_instances);
	ExpressionSP droite = _droite->clone(clones_instances);
	MaximumSP clone(new Maximum());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_gauche=gauche;
	clone->_droite=droite;
	return clone;
}

