#include <src/donnees/expressions/expressionnombre.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <cmath>
using namespace Traceur::Donnees::Expression;

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			NombreSP cosinus(NombreSP valeur)
			{
				EntierSP entier = std::dynamic_pointer_cast<Entier>(valeur);
				if(entier)
				{
					Reel reel(*entier);
					NombreSP nombre(new Reel(Reel::cosinus(reel)));
					return nombre;
				}
				ReelSP reel = std::dynamic_pointer_cast<Reel>(valeur);
				NombreSP nombre(new Reel(Reel::cosinus(*reel)));
				return nombre;
			}

			NombreSP sinus(NombreSP valeur)
			{
				EntierSP entier = std::dynamic_pointer_cast<Entier>(valeur);
				if(entier)
				{
					Reel reel(*entier);
					NombreSP nombre(new Reel(Reel::sinus(reel)));
					return nombre;
				}
				ReelSP reel = std::dynamic_pointer_cast<Reel>(valeur);
				NombreSP nombre(new Reel(Reel::sinus(*reel)));
				return nombre;
			}

			NombreSP partie_entiere(NombreSP valeur)
			{
				EntierSP entier = std::dynamic_pointer_cast<Entier>(valeur);
				if(entier)
				{
					NombreSP nombre(new Entier(Entier::partie_entiere(*entier)));
					return nombre;
				}
				ReelSP reel = std::dynamic_pointer_cast<Reel>(valeur);
				NombreSP nombre(new Reel(Reel::partie_entiere(*reel)));
				return nombre;
			}

			NombreSP operator+(NombreSP gauche, NombreSP droite)
			{
				EntierSP entier_gauche = std::dynamic_pointer_cast<Entier>(gauche);
				EntierSP entier_droite = std::dynamic_pointer_cast<Entier>(droite);
				if(entier_gauche and entier_droite)
				{
					NombreSP nombre(new Entier((*entier_gauche)+(*entier_droite)));
					return nombre;
				}
				ReelSP reel_gauche = std::dynamic_pointer_cast<Reel>(gauche);
				if(reel_gauche and entier_droite)
				{
					NombreSP nombre(new Reel((*reel_gauche)+(*entier_droite)));
					return nombre;
				}
				ReelSP reel_droite = std::dynamic_pointer_cast<Reel>(droite);
				if(entier_gauche and reel_droite)
				{
					NombreSP nombre(new Reel((*entier_gauche)+(*reel_droite)));
					return nombre;
				}
				NombreSP nombre(new Reel((*reel_gauche)+(*reel_droite)));
				return nombre;
			}

			NombreSP operator-(NombreSP gauche, NombreSP droite)
			{
				EntierSP entier_gauche = std::dynamic_pointer_cast<Entier>(gauche);
				EntierSP entier_droite = std::dynamic_pointer_cast<Entier>(droite);
				if(entier_gauche and entier_droite)
				{
					NombreSP nombre(new Entier((*entier_gauche)-(*entier_droite)));
					return nombre;
				}
				ReelSP reel_gauche = std::dynamic_pointer_cast<Reel>(gauche);
				if(reel_gauche and entier_droite)
				{
					NombreSP nombre(new Reel((*reel_gauche)-(*entier_droite)));
					return nombre;
				}
				ReelSP reel_droite = std::dynamic_pointer_cast<Reel>(droite);
				if(entier_gauche and reel_droite)
				{
					NombreSP nombre(new Reel((*entier_gauche)-(*reel_droite)));
					return nombre;
				}
				NombreSP nombre(new Reel((*reel_gauche)-(*reel_droite)));
				return nombre;
			}

			NombreSP operator-(NombreSP valeur)
			{
				EntierSP entier = std::dynamic_pointer_cast<Entier>(valeur);
				if(entier)
				{
					NombreSP nombre(new Entier(-(*entier)));
					return nombre;
				}
				ReelSP reel = std::dynamic_pointer_cast<Reel>(valeur);
				NombreSP nombre(new Reel(-(*reel)));
				return nombre;
			}

			NombreSP operator*(NombreSP gauche, NombreSP droite)
			{
				EntierSP entier_gauche = std::dynamic_pointer_cast<Entier>(gauche);
				EntierSP entier_droite = std::dynamic_pointer_cast<Entier>(droite);
				if(entier_gauche and entier_droite)
				{
					NombreSP nombre(new Entier((*entier_gauche)*(*entier_droite)));
					return nombre;
				}
				ReelSP reel_gauche = std::dynamic_pointer_cast<Reel>(gauche);
				if(reel_gauche and entier_droite)
				{
					NombreSP nombre(new Reel((*reel_gauche)*(*entier_droite)));
					return nombre;
				}
				ReelSP reel_droite = std::dynamic_pointer_cast<Reel>(droite);
				if(entier_gauche and reel_droite)
				{
					NombreSP nombre(new Reel((*entier_gauche)*(*reel_droite)));
					return nombre;
				}
				NombreSP nombre(new Reel((*reel_gauche)*(*reel_droite)));
				return nombre;
			}

			NombreSP operator/(NombreSP gauche, NombreSP droite)
			{
				EntierSP entier_gauche = std::dynamic_pointer_cast<Entier>(gauche);
				EntierSP entier_droite = std::dynamic_pointer_cast<Entier>(droite);
				if(entier_gauche and entier_droite)
				{
					NombreSP nombre(new Entier((*entier_gauche)/(*entier_droite)));
					return nombre;
				}
				ReelSP reel_gauche = std::dynamic_pointer_cast<Reel>(gauche);
				if(reel_gauche and entier_droite)
				{
					NombreSP nombre(new Reel((*reel_gauche)/(*entier_droite)));
					return nombre;
				}
				ReelSP reel_droite = std::dynamic_pointer_cast<Reel>(droite);
				if(entier_gauche and reel_droite)
				{
					NombreSP nombre(new Reel((*entier_gauche)/(*reel_droite)));
					return nombre;
				}
				NombreSP nombre(new Reel((*reel_gauche)/(*reel_droite)));
				return nombre;
			}

			NombreSP operator%(NombreSP gauche, NombreSP droite)
			{
				EntierSP entier_gauche = std::dynamic_pointer_cast<Entier>(gauche);
				EntierSP entier_droite = std::dynamic_pointer_cast<Entier>(droite);
				if(entier_gauche and entier_droite)
				{
					NombreSP nombre(new Entier((*entier_gauche)%(*entier_droite)));
					return nombre;
				}
				ReelSP reel_gauche = std::dynamic_pointer_cast<Reel>(gauche);
				if(reel_gauche and entier_droite)
				{
					NombreSP nombre(new Reel((*reel_gauche)%(*entier_droite)));
					return nombre;
				}
				ReelSP reel_droite = std::dynamic_pointer_cast<Reel>(droite);
				if(entier_gauche and reel_droite)
				{
					NombreSP nombre(new Reel((*entier_gauche)%(*reel_droite)));
					return nombre;
				}
				NombreSP nombre(new Reel((*reel_gauche)%(*reel_droite)));
				return nombre;
			}

			bool operator==(NombreSP gauche, NombreSP droite)
			{
				EntierSP entier_gauche = std::dynamic_pointer_cast<Entier>(gauche);
				EntierSP entier_droite = std::dynamic_pointer_cast<Entier>(droite);
				if(entier_gauche and entier_droite)
				{
					return (*entier_gauche)==(*entier_droite);
				}
				ReelSP reel_gauche = std::dynamic_pointer_cast<Reel>(gauche);
				if(reel_gauche and entier_droite)
				{
					return (*reel_gauche)==(*entier_droite);
				}
				ReelSP reel_droite = std::dynamic_pointer_cast<Reel>(droite);
				if(entier_gauche and reel_droite)
				{
					return (*entier_gauche)==(*reel_droite);
				}
				return (*reel_gauche)==(*reel_droite);
			}
			bool operator!=(NombreSP gauche, NombreSP droite)
			{
				EntierSP entier_gauche = std::dynamic_pointer_cast<Entier>(gauche);
				EntierSP entier_droite = std::dynamic_pointer_cast<Entier>(droite);
				if(entier_gauche and entier_droite)
				{
					return (*entier_gauche)!=(*entier_droite);
				}
				ReelSP reel_gauche = std::dynamic_pointer_cast<Reel>(gauche);
				if(reel_gauche and entier_droite)
				{
					return (*reel_gauche)!=(*entier_droite);
				}
				ReelSP reel_droite = std::dynamic_pointer_cast<Reel>(droite);
				if(entier_gauche and reel_droite)
				{
					return (*entier_gauche)!=(*reel_droite);
				}
				return (*reel_gauche)!=(*reel_droite);
			}

			bool operator<(NombreSP gauche, NombreSP droite)
			{
				EntierSP entier_gauche = std::dynamic_pointer_cast<Entier>(gauche);
				EntierSP entier_droite = std::dynamic_pointer_cast<Entier>(droite);
				if(entier_gauche and entier_droite)
				{
					return (*entier_gauche)<(*entier_droite);
				}
				ReelSP reel_gauche = std::dynamic_pointer_cast<Reel>(gauche);
				if(reel_gauche and entier_droite)
				{
					return (*reel_gauche)<(*entier_droite);
				}
				ReelSP reel_droite = std::dynamic_pointer_cast<Reel>(droite);
				if(entier_gauche and reel_droite)
				{
					return (*entier_gauche)<(*reel_droite);
				}
				return (*reel_gauche)<(*reel_droite);
			}

			bool operator>(NombreSP gauche, NombreSP droite)
			{
				EntierSP entier_gauche = std::dynamic_pointer_cast<Entier>(gauche);
				EntierSP entier_droite = std::dynamic_pointer_cast<Entier>(droite);
				if(entier_gauche and entier_droite)
				{
					return (*entier_gauche)>(*entier_droite);
				}
				ReelSP reel_gauche = std::dynamic_pointer_cast<Reel>(gauche);
				if(reel_gauche and entier_droite)
				{
					return (*reel_gauche)>(*entier_droite);
				}
				ReelSP reel_droite = std::dynamic_pointer_cast<Reel>(droite);
				if(entier_gauche and reel_droite)
				{
					return (*entier_gauche)>(*reel_droite);
				}
				return (*reel_gauche)>(*reel_droite);
			}

			bool operator<=(NombreSP gauche, NombreSP droite)
			{
				EntierSP entier_gauche = std::dynamic_pointer_cast<Entier>(gauche);
				EntierSP entier_droite = std::dynamic_pointer_cast<Entier>(droite);
				if(entier_gauche and entier_droite)
				{
					return (*entier_gauche)<=(*entier_droite);
				}
				ReelSP reel_gauche = std::dynamic_pointer_cast<Reel>(gauche);
				if(reel_gauche and entier_droite)
				{
					return (*reel_gauche)<=(*entier_droite);
				}
				ReelSP reel_droite = std::dynamic_pointer_cast<Reel>(droite);
				if(entier_gauche and reel_droite)
				{
					return (*entier_gauche)<=(*reel_droite);
				}
				return (*reel_gauche)<=(*reel_droite);
			}

			bool operator>=(NombreSP gauche, NombreSP droite)
			{
				EntierSP entier_gauche = std::dynamic_pointer_cast<Entier>(gauche);
				EntierSP entier_droite = std::dynamic_pointer_cast<Entier>(droite);
				if(entier_gauche and entier_droite)
				{
					return (*entier_gauche)>=(*entier_droite);
				}
				ReelSP reel_gauche = std::dynamic_pointer_cast<Reel>(gauche);
				if(reel_gauche and entier_droite)
				{
					return (*reel_gauche)>=(*entier_droite);
				}
				ReelSP reel_droite = std::dynamic_pointer_cast<Reel>(droite);
				if(entier_gauche and reel_droite)
				{
					return (*entier_gauche)>=(*reel_droite);
				}
				return (*reel_gauche)>=(*reel_droite);
			}

			bool est_nul(NombreSP valeur)
			{
				EntierSP entier = std::dynamic_pointer_cast<Entier>(valeur);
				if(entier)
				{
					return (*entier)==Entier(0);
				}
				ReelSP reel = std::dynamic_pointer_cast<Reel>(valeur);
				return (*reel)==Reel(0);
			}

			NombreSP racine_carree(NombreSP valeur)
			{
				EntierSP entier = std::dynamic_pointer_cast<Entier>(valeur);
				if(entier)
				{
					Reel reel(*entier);
					NombreSP nombre(new Reel(Reel::racine_carree(reel)));
					return nombre;
				}
				ReelSP reel = std::dynamic_pointer_cast<Reel>(valeur);
				NombreSP nombre(new Reel(Reel::racine_carree(*reel)));
				return nombre;
			}

			NombreSP valeur_absolue(NombreSP valeur)
			{
				EntierSP entier = std::dynamic_pointer_cast<Entier>(valeur);
				if(entier)
				{
					NombreSP nombre(new Entier(Entier::valeur_absolue(*entier)));
					return nombre;
				}
				ReelSP reel = std::dynamic_pointer_cast<Reel>(valeur);
				NombreSP nombre(new Reel(Reel::valeur_absolue(*reel)));
				return nombre;
			}

			NombreSP argument(NombreSP gauche, NombreSP droite)
			{
				EntierSP entier_gauche = std::dynamic_pointer_cast<Entier>(gauche);
				EntierSP entier_droite = std::dynamic_pointer_cast<Entier>(droite);
				ReelSP reel_gauche = std::dynamic_pointer_cast<Reel>(gauche);
				ReelSP reel_droite = std::dynamic_pointer_cast<Reel>(droite);
				Reel valeur_gauche(0);
				Reel valeur_droite(0);
				if(entier_gauche)
				{
					valeur_gauche = Reel(*entier_gauche);
				}
				else
				{
					valeur_gauche = *reel_gauche;
				}
				if(entier_droite)
				{
					valeur_droite = Reel(*entier_droite);
				}
				else
				{
					valeur_droite = *reel_droite;
				}
				NombreSP nombre(new Reel(Reel::argument(valeur_gauche,valeur_droite)));
				return nombre;
			}
		}
	}
}
NombreSP Nombre::zero()
{
	static NombreSP zero = Entier::zero();
	return zero;
}
NombreSP Nombre::un()
{
	static NombreSP un = Entier::un();
	return un;
}
NombreSP Nombre::deux()
{
	static NombreSP deux = Entier::deux();
	return deux;
}
NombreSP Nombre::max_canal_couleur()
{
	static NombreSP max_canal_couleur = Entier::max_canal_couleur();
	return max_canal_couleur;
}
NombreSP Nombre::med_canal_couleur()
{
	static NombreSP med_canal_couleur = Entier::med_canal_couleur();
	return med_canal_couleur;
}
NombreSP Nombre::pi_sur_deux()
{
	static NombreSP pi_sur_deux = Reel::pi_sur_deux();
	return pi_sur_deux;
}

Reel Nombre::reel()
{
	Reel *reel = dynamic_cast<Reel*>(this);
	if(reel)
	{
		Reel copie = *reel;
		return copie;
	}
	Reel conversion(*(dynamic_cast<Entier*>(this)));
	return conversion;
}

Entier Nombre::entier()
{
	Entier *entier = dynamic_cast<Entier*>(this);
	if(entier)
	{
		Entier copie = *entier;
		return copie;
	}
	Entier conversion(*(dynamic_cast<Reel*>(this)));
	return conversion;
}

Reel Nombre::reel_borne(const Reel& inf, const Reel& sup)
{
	Reel reel = this->reel();
	if(reel<inf)
		return inf;
	if(reel>sup)
		return sup;
	return reel;
}

Entier Nombre::entier_borne(const Entier& inf, const Entier& sup)
{
	Entier entier = this->entier();
	if(entier<inf)
		return inf;
	if(entier>sup)
		return sup;
	return entier;
}
