#include <src/donnees/expressions/expressionnonlogique.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionbooleenvrai.h>
#include <src/donnees/expressions/expressionbooleenfaux.h>
#include <src/donnees/expressions/expressionbooleenpeutetre.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_NONLOGIQUE(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = NonLogiqueSP(new EvaluateurOperateurUnaire<NonLogique,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<NonLogique,BooleenVrai> : public NonLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							return ValeurSP(new Valeur(BooleenFaux::SP()));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<NonLogique,BooleenPeutetre> : public NonLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							return ValeurSP(new Valeur(BooleenPeutetre::SP()));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<NonLogique,BooleenFaux> : public NonLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							return ValeurSP(new Valeur(BooleenVrai::SP()));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	NonLogique::NonLogique(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(NonLogiqueSP(new EvaluateurOperateurUnaire<NonLogique,Expression>()))
{

	EVALUATEUR_NONLOGIQUE(BooleenVrai    );
	EVALUATEUR_NONLOGIQUE(BooleenPeutetre);
	EVALUATEUR_NONLOGIQUE(BooleenFaux    );
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Boolean>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionPrefixe(position,"not",valeur);
}

ValeurSP NonLogique::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	NonLogique resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP NonLogique::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void NonLogique::format(std::ostream& os) const
{
	os << "not (";
	_valeur->format(os);
	os << ")";
}

ExpressionSP NonLogique::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	NonLogiqueSP clone(new NonLogique(valeur));
	return clone;
}
