#include <src/donnees/expressions/expressionnoyau.h>

#include <src/donnees/types/types.h>
#include <src/donnees/assembleur/assembleur.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteur.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/expressions/expressionimageevaluee.h>

using namespace Traceur::Donnees::Expression;

Noyau::Noyau(ExpressionSP taille_x, ExpressionSP taille_y, SystemeCoordonneesSP coordonnees,
	Traceur::Donnees::Instruction::BlocSP contenu, Traceur::Donnees::Source::PositionSP position)
: _taille_x(taille_x), _taille_y(taille_y), _coordonnees(coordonnees), _contenu(contenu)
{
	// calcul du type de retour
	if(not coordonnees)
		_coordonnees = SystemeCoordonneesSP(new SystemeGeometrique());
	if(_taille_x->est_de_type<Traceur::Donnees::Type::Integer>() and _taille_y->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Kernel::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectNoyau(position,_taille_x,_taille_y);
}

Traceur::Donnees::Instruction::ValeurSP Noyau::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	Traceur::Donnees::Instruction::ValeurSP taille_x = _taille_x->evaluation(assembleur); if(not taille_x->a_une_valeur_retour()) return taille_x;
	Traceur::Donnees::Instruction::ValeurSP taille_y = _taille_y->evaluation(assembleur); if(not taille_y->a_une_valeur_retour()) return taille_y;
	SystemeCoordonneesSP coordonnees;
	try
	{
		coordonnees = _coordonnees->evaluation(assembleur);
		ExpressionSP expr_taille_x = *taille_x;
		ExpressionSP expr_taille_y = *taille_y;
		assembleur->noyau(expr_taille_x,expr_taille_y,coordonnees);
	} catch(Traceur::Donnees::Instruction::ValeurSP valeur) { return valeur; }
	Traceur::Donnees::Instruction::InstructionInterruptionSP interruption = _contenu->evaluation(assembleur);
	if(interruption)
	{
		auto action = interruption->comportement_image();
		switch(action.first)
		{
			case Traceur::Donnees::Instruction::InstructionInterruption::Action::SORTIE:
			{
				Traceur::Donnees::Assembleur::Image::ImageSP image = assembleur->image();
				Traceur::Donnees::Expression::ImageEvalueeSP image_evaluee(new Traceur::Donnees::Expression::ImageEvaluee(image));
				Traceur::Donnees::Instruction::ValeurSP valeur(new Traceur::Donnees::Instruction::Valeur(image_evaluee));
				if(action.second)
				{
					valeur->_interruption=action.second;
				}
				return valeur;
			}
			default:
				break;
		}
	}
	Traceur::Donnees::Assembleur::Image::ImageSP image = assembleur->image();
	Traceur::Donnees::Expression::ImageEvalueeSP image_evaluee(new Traceur::Donnees::Expression::ImageEvaluee(image));
	Traceur::Donnees::Instruction::ValeurSP valeur(new Traceur::Donnees::Instruction::Valeur(image_evaluee));
	return valeur;
}

void Noyau::format(std::ostream& os) const
{
	os << "KERNEL[";
	_taille_x->format(os);
	os <<  ";" ;
	_taille_y->format(os);
	os << "] ";
	_coordonnees->format(os);
	os << std::endl << "{" << std::endl;
	_contenu->format(os);
	os << "}" << std::endl;
}

void Noyau::edition_liens(Traceur::Donnees::Instruction::InstructionSP instruction)
{
	_taille_x->edition_liens(instruction);
	_taille_y->edition_liens(instruction);
	_coordonnees->edition_liens(instruction);
	_contenu->edition_liens(instruction->parent(),instruction);
}

void Noyau::definition_fonction()
{
	_taille_x->definition_fonction();
	_taille_y->definition_fonction();
	_coordonnees->definition_fonction();
	_contenu->definition_fonction();
}

ExpressionSP Noyau::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP taille_x = _taille_x->clone(clones_instances);
	ExpressionSP taille_y = _taille_y->clone(clones_instances);
	SystemeCoordonneesSP coordonnees = _coordonnees->clone(clones_instances);
	Traceur::Donnees::Instruction::BlocSP contenu = _contenu->clone(clones_instances);
	NoyauSP clone(new Noyau(taille_x,taille_y,coordonnees,contenu));
	return clone;
}

ExpressionSP Noyau::simplification()
{
	_taille_x->simplification();
	_taille_y->simplification();
	_coordonnees->simplification();
	_contenu->simplification();
	return this->shared_from_this();
}
