#include <src/donnees/expressions/expressionouexclusiflogique.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionbooleenvrai.h>
#include <src/donnees/expressions/expressionbooleenfaux.h>
#include <src/donnees/expressions/expressionbooleenpeutetre.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_OUEXCLUSIFLOGIQUE(type1,type2) \
	if(gauche->est_de_nature<type1>() and droite->est_de_nature<type2>()) { _evaluateur = OuExclusifLogiqueSP(new EvaluateurOperateurBinaire<type1,OuExclusifLogique,type2>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurBinaire<BooleenVrai,OuExclusifLogique,BooleenVrai> : public OuExclusifLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenFaux::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenVrai,OuExclusifLogique,BooleenPeutetre> : public OuExclusifLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenPeutetre::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenVrai,OuExclusifLogique,BooleenFaux> : public OuExclusifLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenVrai::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenPeutetre,OuExclusifLogique,BooleenVrai> : public OuExclusifLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenPeutetre::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenPeutetre,OuExclusifLogique,BooleenPeutetre> : public OuExclusifLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenPeutetre::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenPeutetre,OuExclusifLogique,BooleenFaux> : public OuExclusifLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenPeutetre::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenFaux,OuExclusifLogique,BooleenVrai> : public OuExclusifLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenVrai::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenFaux,OuExclusifLogique,BooleenPeutetre> : public OuExclusifLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenPeutetre::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenFaux,OuExclusifLogique,BooleenFaux> : public OuExclusifLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenFaux::SP()));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	OuExclusifLogique::OuExclusifLogique(ExpressionSP gauche, ExpressionSP droite, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneBinaire(gauche,droite), _evaluateur(OuExclusifLogiqueSP(new EvaluateurOperateurBinaire<Expression,OuExclusifLogique,Expression>()))
{

	EVALUATEUR_OUEXCLUSIFLOGIQUE(BooleenVrai    ,BooleenVrai    );
	EVALUATEUR_OUEXCLUSIFLOGIQUE(BooleenVrai    ,BooleenPeutetre);
	EVALUATEUR_OUEXCLUSIFLOGIQUE(BooleenVrai    ,BooleenFaux    );
	EVALUATEUR_OUEXCLUSIFLOGIQUE(BooleenPeutetre,BooleenVrai    );
	EVALUATEUR_OUEXCLUSIFLOGIQUE(BooleenPeutetre,BooleenPeutetre);
	EVALUATEUR_OUEXCLUSIFLOGIQUE(BooleenPeutetre,BooleenFaux    );
	EVALUATEUR_OUEXCLUSIFLOGIQUE(BooleenFaux    ,BooleenVrai    );
	EVALUATEUR_OUEXCLUSIFLOGIQUE(BooleenFaux    ,BooleenPeutetre);
	EVALUATEUR_OUEXCLUSIFLOGIQUE(BooleenFaux    ,BooleenFaux    );
	// calcul du type de retour
	if(_type)
		return;
	if(gauche->est_de_type<Traceur::Donnees::Type::Boolean>() and droite->est_de_type<Traceur::Donnees::Type::Boolean>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionInfixe(position,gauche,"xor",droite);
}

ValeurSP OuExclusifLogique::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_gauche = _gauche->evaluation(assembleur); if(not valeur_gauche->a_une_valeur_retour()) return valeur_gauche;
	ValeurSP valeur_droite = _droite->evaluation(assembleur); if(not valeur_droite->a_une_valeur_retour()) return valeur_droite;
	OuExclusifLogique resultat(*valeur_gauche,*valeur_droite);
	return resultat.evaluation(valeur_gauche,valeur_droite);
}

ValeurSP OuExclusifLogique::evaluation(const ValeurSP gauche, const ValeurSP droite) const
{
	return _evaluateur->evaluation(gauche,droite);
}

void OuExclusifLogique::format(std::ostream& os) const
{
	os << "(";
	_gauche->format(os);
	os << " xor ";
	_droite->format(os);
	os << ")";
}

ExpressionSP OuExclusifLogique::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP gauche = _gauche->clone(clones_instances);
	ExpressionSP droite = _droite->clone(clones_instances);
	OuExclusifLogiqueSP clone(new OuExclusifLogique());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_gauche=gauche;
	clone->_droite=droite;
	return clone;
}
