#include <src/donnees/expressions/expressionoulogique.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionbooleenvrai.h>
#include <src/donnees/expressions/expressionbooleenfaux.h>
#include <src/donnees/expressions/expressionbooleenpeutetre.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_OULOGIQUE(type1,type2) \
	if(gauche->est_de_nature<type1>() and droite->est_de_nature<type2>()) { _evaluateur = OuLogiqueSP(new EvaluateurOperateurBinaire<type1,OuLogique,type2>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurBinaire<BooleenVrai,OuLogique,BooleenVrai> : public OuLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenVrai::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenVrai,OuLogique,BooleenPeutetre> : public OuLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenVrai::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenVrai,OuLogique,BooleenFaux> : public OuLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenVrai::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenPeutetre,OuLogique,BooleenVrai> : public OuLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenVrai::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenPeutetre,OuLogique,BooleenPeutetre> : public OuLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenPeutetre::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenPeutetre,OuLogique,BooleenFaux> : public OuLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenPeutetre::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenFaux,OuLogique,BooleenVrai> : public OuLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenVrai::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenFaux,OuLogique,BooleenPeutetre> : public OuLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenPeutetre::SP()));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<BooleenFaux,OuLogique,BooleenFaux> : public OuLogique
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							return ValeurSP(new Valeur(BooleenFaux::SP()));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	OuLogique::OuLogique(ExpressionSP gauche, ExpressionSP droite, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneBinaire(gauche,droite), _evaluateur(OuLogiqueSP(new EvaluateurOperateurBinaire<Expression,OuLogique,Expression>()))
{

	EVALUATEUR_OULOGIQUE(BooleenVrai    ,BooleenVrai    );
	EVALUATEUR_OULOGIQUE(BooleenVrai    ,BooleenPeutetre);
	EVALUATEUR_OULOGIQUE(BooleenVrai    ,BooleenFaux    );
	EVALUATEUR_OULOGIQUE(BooleenPeutetre,BooleenVrai    );
	EVALUATEUR_OULOGIQUE(BooleenPeutetre,BooleenPeutetre);
	EVALUATEUR_OULOGIQUE(BooleenPeutetre,BooleenFaux    );
	EVALUATEUR_OULOGIQUE(BooleenFaux    ,BooleenVrai    );
	EVALUATEUR_OULOGIQUE(BooleenFaux    ,BooleenPeutetre);
	EVALUATEUR_OULOGIQUE(BooleenFaux    ,BooleenFaux    );
	// calcul du type de retour
	if(_type)
		return;
	if(gauche->est_de_type<Traceur::Donnees::Type::Boolean>() and droite->est_de_type<Traceur::Donnees::Type::Boolean>())
	{
		_type = Traceur::Donnees::Type::Boolean::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionInfixe(position,gauche,"or",droite);
}

ValeurSP OuLogique::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_gauche = _gauche->evaluation(assembleur); if(not valeur_gauche->a_une_valeur_retour()) return valeur_gauche;
	ValeurSP valeur_droite = _droite->evaluation(assembleur); if(not valeur_droite->a_une_valeur_retour()) return valeur_droite;
	OuLogique resultat(*valeur_gauche,*valeur_droite);
	return resultat.evaluation(valeur_gauche,valeur_droite);
}

ValeurSP OuLogique::evaluation(const ValeurSP gauche, const ValeurSP droite) const
{
	return _evaluateur->evaluation(gauche,droite);
}

void OuLogique::format(std::ostream& os) const
{
	os << "(";
	_gauche->format(os);
	os << " or ";
	_droite->format(os);
	os << ")";
}

ExpressionSP OuLogique::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP gauche = _gauche->clone(clones_instances);
	ExpressionSP droite = _droite->clone(clones_instances);
	OuLogiqueSP clone(new OuLogique());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_gauche=gauche;
	clone->_droite=droite;
	return clone;
}
