#include <src/donnees/expressions/expressionpartiedecimale.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_PARTIEDECIMALE(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = PartieDecimaleSP(new EvaluateurOperateurUnaire<PartieDecimale,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<PartieDecimale,Entier> : public PartieDecimale
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							EntierSP valeur_valeur=std::static_pointer_cast<Entier>(expr_valeur);
							SECURISE_VALEUR(EntierSP(new Entier(Entier::partie_decimale(*valeur_valeur))));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<PartieDecimale,Reel> : public PartieDecimale
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ReelSP valeur_valeur=std::static_pointer_cast<Reel>(expr_valeur);
							SECURISE_VALEUR(ReelSP(new Reel(Reel::partie_decimale(*valeur_valeur))));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	PartieDecimale::PartieDecimale(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(PartieDecimaleSP(new EvaluateurOperateurUnaire<PartieDecimale,Expression>()))
{

	EVALUATEUR_PARTIEDECIMALE(Entier);
	EVALUATEUR_PARTIEDECIMALE(Reel  );
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Integer::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionPrefixe(position,"frac",valeur);
}

ValeurSP PartieDecimale::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	PartieDecimale resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP PartieDecimale::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void PartieDecimale::format(std::ostream& os) const
{
	os << "frac (";
	_valeur->format(os);
	os << ")";
}

ExpressionSP PartieDecimale::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	PartieDecimaleSP clone(new PartieDecimale(valeur));
	return clone;
}
