#include <src/donnees/expressions/expressionpartieentiere.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_PARTIEENTIERE(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = PartieEntiereSP(new EvaluateurOperateurUnaire<PartieEntiere,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<PartieEntiere,Entier> : public PartieEntiere
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							EntierSP valeur_valeur=std::static_pointer_cast<Entier>(expr_valeur);
							SECURISE_VALEUR(EntierSP(new Entier(Entier::partie_entiere(*valeur_valeur))));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<PartieEntiere,Reel> : public PartieEntiere
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ReelSP valeur_valeur=std::static_pointer_cast<Reel>(expr_valeur);
							SECURISE_VALEUR(ReelSP(new Reel(Reel::partie_entiere(*valeur_valeur))));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	PartieEntiere::PartieEntiere(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(PartieEntiereSP(new EvaluateurOperateurUnaire<PartieEntiere,Expression>()))
{

	EVALUATEUR_PARTIEENTIERE(Entier);
	EVALUATEUR_PARTIEENTIERE(Reel  );
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Integer::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionPrefixe(position,"int",valeur);
}

ValeurSP PartieEntiere::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	PartieEntiere resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP PartieEntiere::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void PartieEntiere::format(std::ostream& os) const
{
	os << "int (";
	_valeur->format(os);
	os << ")";
}

ExpressionSP PartieEntiere::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	PartieEntiereSP clone(new PartieEntiere(valeur));
	return clone;
}
