#include <src/donnees/expressions/expressionpoint.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionnombre.h>
#include <src/donnees/expressions/expressionvecteur.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/assembleur/point.h>
#include <src/donnees/assembleur/drawpoint.h>
#include <src/donnees/assembleur/image.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_POINT(type1,type2) \
	if(gauche->est_de_nature<type1>() and droite->est_de_nature<type2>()) { _evaluateur = PointSP(new EvaluateurOperateurBinaire<type1,Point,type2>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurBinaire<Nombre,Point,Nombre> : public Point
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							NombreSP valeur_gauche=std::static_pointer_cast<Nombre>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							NombreSP valeur_droite=std::static_pointer_cast<Nombre>(expr_droite);
							SECURISE_VALEUR(PointSP(new Point(valeur_gauche,valeur_droite)));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Point::Point(ExpressionSP gauche, ExpressionSP droite, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneBinaire(gauche,droite), _evaluateur(PointSP(new EvaluateurOperateurBinaire<Expression,Point,Expression>()))
{

	EVALUATEUR_POINT(Nombre,Nombre);
	// calcul du type de retour
	if(_type)
		return;
	if(gauche->est_de_type<Traceur::Donnees::Type::Number>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Point::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionDeux(position,"POINT",gauche,droite);
}

ValeurSP Point::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_gauche = _gauche->evaluation(assembleur); if(not valeur_gauche->a_une_valeur_retour()) return valeur_gauche;
	ValeurSP valeur_droite = _droite->evaluation(assembleur); if(not valeur_droite->a_une_valeur_retour()) return valeur_droite;
	Point resultat(*valeur_gauche,*valeur_droite);
	return resultat.evaluation(valeur_gauche,valeur_droite);
}

ValeurSP Point::evaluation(const ValeurSP gauche, const ValeurSP droite) const
{
	return _evaluateur->evaluation(gauche,droite);
}

void Point::format(std::ostream& os) const
{
	//_type->format(os);
	os << "POINT[";
	_gauche->format(os);
	os << ";";
	_droite->format(os);
	os << "]";
}

ExpressionSP Point::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP gauche = _gauche->clone(clones_instances);
	ExpressionSP droite = _droite->clone(clones_instances);
	PointSP clone(new Point());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_gauche=gauche;
	clone->_droite=droite;
	return clone;
}

Point::Point(const Vecteur& vecteur)
{
	_type = Traceur::Donnees::Type::Point::SP();
	NombreSP abscisse;
	NombreSP ordonnee;
	vecteur.en_cartesien(abscisse,ordonnee);
	_gauche = abscisse;
	_droite = ordonnee;
}

void Point::en_cartesien(NombreSP& abscisse, NombreSP& ordonnee) const
{
	abscisse = std::dynamic_pointer_cast<Nombre>(_gauche);
	TRACEUR_TEST(abscisse);
	ordonnee = std::dynamic_pointer_cast<Nombre>(_droite);
	TRACEUR_TEST(ordonnee);
}

void Point::en_polaire(NombreSP& module, NombreSP& argument) const
{
	NombreSP abscisse = std::dynamic_pointer_cast<Nombre>(_gauche);
	TRACEUR_TEST(abscisse);
	NombreSP ordonnee = std::dynamic_pointer_cast<Nombre>(_droite);
	TRACEUR_TEST(ordonnee);
	module = racine_carree(abscisse*abscisse+ordonnee*ordonnee);
	argument = Traceur::Donnees::Expression::argument(abscisse,ordonnee);
}
Point Point::translation(Vecteur& vecteur) const
{
	Vecteur cartesien(vecteur.en_cartesien());
	NombreSP abscisse_depart;
	NombreSP ordonnee_depart;
	this->en_cartesien(abscisse_depart,ordonnee_depart);
	NombreSP abscisse_translation = std::dynamic_pointer_cast<Nombre>(cartesien._gauche);
	TRACEUR_TEST(abscisse_translation);
	NombreSP ordonnee_translation = std::dynamic_pointer_cast<Nombre>(cartesien._droite);
	TRACEUR_TEST(ordonnee_translation);
	Point point_translate(abscisse_depart+abscisse_translation,ordonnee_depart+ordonnee_translation);
	return point_translate;
}

Point Point::translation_opposee(Vecteur& vecteur) const
{
	Vecteur cartesien(vecteur.en_cartesien());
	NombreSP abscisse_depart;
	NombreSP ordonnee_depart;
	this->en_cartesien(abscisse_depart,ordonnee_depart);
	NombreSP abscisse_translation = std::dynamic_pointer_cast<Nombre>(cartesien._gauche);
	TRACEUR_TEST(abscisse_translation);
	NombreSP ordonnee_translation = std::dynamic_pointer_cast<Nombre>(cartesien._droite);
	TRACEUR_TEST(ordonnee_translation);
	Point point_translate(abscisse_depart-abscisse_translation,ordonnee_depart-ordonnee_translation);
	return point_translate;
}

Vecteur Point::decalage(Point& point) const
{
	NombreSP abscisse_depart;
	NombreSP ordonnee_depart;
	point.en_cartesien(abscisse_depart,ordonnee_depart);
	NombreSP abscisse_arrivee;
	NombreSP ordonnee_arrivee;
	this->en_cartesien(abscisse_arrivee,ordonnee_arrivee);
	Vecteur vecteur(abscisse_arrivee-abscisse_depart,ordonnee_arrivee-ordonnee_depart);
	return vecteur;
}

NombreSP Point::abscisse() const
{
	NombreSP abscisse;
	NombreSP ordonnee;
	this->en_cartesien(abscisse,ordonnee);
	return abscisse;
}

NombreSP Point::ordonnee() const
{
	NombreSP abscisse;
	NombreSP ordonnee;
	this->en_cartesien(abscisse,ordonnee);
	return ordonnee;
}

NombreSP Point::module() const
{
	NombreSP module;
	NombreSP argument;
	this->en_polaire(module,argument);
	return module;
}

NombreSP Point::argument() const
{
	NombreSP module;
	NombreSP argument;
	this->en_polaire(module,argument);
	return argument;
}

CouleurRVBASP Point::couleur(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	Traceur::Donnees::Assembleur::Image::ImageSP image = assembleur->trace_image();
	PointSP moi = std::static_pointer_cast<Point>(this->shared_from_this());
	Traceur::Donnees::Assembleur::Point point(moi,image->_systeme_coordonnees);
	CouleurRVBASP couleur(new CouleurRVBA(image->couleur(point)));
	return couleur;
}

Traceur::Donnees::Assembleur::InstructionSP Point::draw(Traceur::Donnees::Assembleur::Assembleur& assembleur)
{
	PointSP moi = std::static_pointer_cast<Point>(this->shared_from_this());
	Traceur::Donnees::Assembleur::Point point(moi,assembleur._pile.top()->_systeme_coordonnees);
	Traceur::Donnees::Assembleur::InstructionSP drawpoint(new Traceur::Donnees::Assembleur::DrawPoint(point));
	return drawpoint;
}
