#include <src/donnees/expressions/expressionppcm.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_PPCM(type1,type2) \
	if(gauche->est_de_nature<type1>() and droite->est_de_nature<type2>()) { _evaluateur = PpcmSP(new EvaluateurOperateurBinaire<type1,Ppcm,type2>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurBinaire<Entier,Ppcm,Entier> : public Ppcm
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							EntierSP valeur_gauche=std::static_pointer_cast<Entier>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							SECURISE_VALEUR(EntierSP(new Entier(Entier::ppcm(*valeur_gauche,*valeur_droite))));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Ppcm::Ppcm(ExpressionSP gauche, ExpressionSP droite, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneBinaire(gauche,droite), _evaluateur(PpcmSP(new EvaluateurOperateurBinaire<Expression,Ppcm,Expression>()))
{
	EVALUATEUR_PPCM(Entier,Entier);
	// calcul du type de retour
	if(_type)
		return;
	if(gauche->est_de_type<Traceur::Donnees::Type::Integer>() and droite->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Integer::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionInfixe(position,gauche,"lcm",droite);
}

ValeurSP Ppcm::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_gauche = _gauche->evaluation(assembleur); if(not valeur_gauche->a_une_valeur_retour()) return valeur_gauche;
	ValeurSP valeur_droite = _droite->evaluation(assembleur); if(not valeur_droite->a_une_valeur_retour()) return valeur_droite;
	Ppcm resultat(*valeur_gauche,*valeur_droite);
	return resultat.evaluation(valeur_gauche,valeur_droite);
}

ValeurSP Ppcm::evaluation(const ValeurSP gauche, const ValeurSP droite) const
{
	return _evaluateur->evaluation(gauche,droite);
}

void Ppcm::format(std::ostream& os) const
{
	//_type->format(os);
	os << "(";
	_gauche->format(os);
	os << ") lcm (";
	_droite->format(os);
	os << ")";
}

ExpressionSP Ppcm::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP gauche = _gauche->clone(clones_instances);
	ExpressionSP droite = _droite->clone(clones_instances);
	PpcmSP clone(new Ppcm());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_gauche=gauche;
	clone->_droite=droite;
	return clone;
}

