#include <src/donnees/expressions/expressionracinecarree.h>

#include <src/donnees/types/types.h>

#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_RACINECARREE(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = RacineCarreeSP(new EvaluateurOperateurUnaire<RacineCarree,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<RacineCarree,Entier> : public RacineCarree
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							EntierSP valeur_valeur=std::static_pointer_cast<Entier>(expr_valeur);
							if((*valeur_valeur)<Entier(0))
								return Valeur::hors_domaine();
							else
							{ SECURISE_VALEUR(ReelSP(new Reel(Entier::racine_carree(*valeur_valeur)))); }
						}
				};
			template<>
				class EvaluateurOperateurUnaire<RacineCarree,Reel> : public RacineCarree
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ReelSP valeur_valeur=std::static_pointer_cast<Reel>(expr_valeur);
							if((*valeur_valeur)<Reel(0))
								return Valeur::hors_domaine();
							else
							{ SECURISE_VALEUR(ReelSP(new Reel(Reel::racine_carree(*valeur_valeur)))); }
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	RacineCarree::RacineCarree(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(RacineCarreeSP(new EvaluateurOperateurUnaire<RacineCarree,Expression>()))
{

	EVALUATEUR_RACINECARREE(Entier);
	EVALUATEUR_RACINECARREE(Reel  );
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionPrefixe(position,"sqrt",valeur);
}

ValeurSP RacineCarree::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	RacineCarree resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP RacineCarree::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void RacineCarree::format(std::ostream& os) const
{
	os << "sqrt (";
	_valeur->format(os);
	os << ")";
}

ExpressionSP RacineCarree::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	RacineCarreeSP clone(new RacineCarree(valeur));
	return clone;
}
