#include <src/donnees/expressions/expressionradians.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteur.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_RADIANS(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = RadiansSP(new EvaluateurOperateurUnaire<Radians,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<Radians,Entier> : public Radians
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							EntierSP valeur_valeur=std::static_pointer_cast<Entier>(expr_valeur);
							// utilser rad signifie qu'on a donne un nombre en radians, et qu'on veut le nombre en degres
							SECURISE_VALEUR(ReelSP(new Reel(Entier::degres(*valeur_valeur))));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<Radians,Reel> : public Radians
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ReelSP valeur_valeur=std::static_pointer_cast<Reel>(expr_valeur);
							// utilser rad signifie qu'on a donne un nombre en radians, et qu'on veut le nombre en degres
							SECURISE_VALEUR(ReelSP(new Reel(Reel::degres(*valeur_valeur))));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Radians::Radians(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(RadiansSP(new EvaluateurOperateurUnaire<Radians,Expression>()))
{

	EVALUATEUR_RADIANS(Entier);
	EVALUATEUR_RADIANS(Reel  );
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionPostfixe(position,"rad",valeur);
}

ValeurSP Radians::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	Radians resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP Radians::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void Radians::format(std::ostream& os) const
{
	os << "(";
	_valeur->format(os);
	os << ") rad";
}

ExpressionSP Radians::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	RadiansSP clone(new Radians(valeur));
	return clone;
}
