#include <src/donnees/expressions/expressionrectangle.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteur.h>
#include <src/donnees/expressions/expressionvecteurpolaire.h>
#include <src/donnees/expressions/expressionvide.h>
#include <src/donnees/assembleur/point.h>
#include <src/donnees/assembleur/drawrectangle.h>
#include <src/donnees/assembleur/image.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_RECTANGLE(type1,type2,type3) \
	if(sommet1->est_de_nature<type1>() and sommet2->est_de_nature<type2>() and sommet3->est_de_nature<type3>()) { _evaluateur = RectangleSP(new EvaluateurOperateurTernaire<Rectangle,type1,type2,type3>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurTernaire<Rectangle,Point,Point,Vide> : public Rectangle
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3) const
						{
							ExpressionSP expr_valeur1(*valeur1);
							PointSP valeur_valeur1=std::static_pointer_cast<Point>(expr_valeur1);
							ExpressionSP expr_valeur2(*valeur2);
							PointSP valeur_valeur2=std::static_pointer_cast<Point>(expr_valeur2);
							SECURISE_VALEUR(RectangleSP(new Rectangle(valeur_valeur1,valeur_valeur2,_options)));
						}
				};
			template<>
				class EvaluateurOperateurTernaire<Rectangle,Point,Vecteur,Vide> : public Rectangle
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3) const
						{
							ExpressionSP expr_valeur1(*valeur1);
							PointSP valeur_valeur1=std::static_pointer_cast<Point>(expr_valeur1);
							ExpressionSP expr_valeur2(*valeur2);
							VecteurSP valeur_valeur2=std::static_pointer_cast<Vecteur>(expr_valeur2);
							SECURISE_VALEUR(RectangleSP(new Rectangle(valeur_valeur1,valeur_valeur2,_options)));
						}
				};
			template<>
				class EvaluateurOperateurTernaire<Rectangle,Point,VecteurPolaire,Vide> : public Rectangle
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3) const
						{
							ExpressionSP expr_valeur1(*valeur1);
							PointSP valeur_valeur1=std::static_pointer_cast<Point>(expr_valeur1);
							ExpressionSP expr_valeur2(*valeur2);
							VecteurPolaireSP valeur_valeur2=std::static_pointer_cast<VecteurPolaire>(expr_valeur2);
							SECURISE_VALEUR(RectangleSP(new Rectangle(valeur_valeur1,valeur_valeur2,_options)));
						}
				};
			template<>
				class EvaluateurOperateurTernaire<Rectangle,Point,Point,Nombre> : public Rectangle
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3) const
						{
							ExpressionSP expr_valeur1(*valeur1);
							PointSP valeur_valeur1=std::static_pointer_cast<Point>(expr_valeur1);
							ExpressionSP expr_valeur2(*valeur2);
							PointSP valeur_valeur2=std::static_pointer_cast<Point>(expr_valeur2);
							ExpressionSP expr_valeur3(*valeur3);
							NombreSP valeur_valeur3=std::static_pointer_cast<Nombre>(expr_valeur3);
							SECURISE_VALEUR(RectangleSP(new Rectangle(valeur_valeur1,valeur_valeur2,valeur_valeur3,_options)));
						}
				};
			template<>
				class EvaluateurOperateurTernaire<Rectangle,Point,Vecteur,Nombre> : public Rectangle
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3) const
						{
							ExpressionSP expr_valeur1(*valeur1);
							PointSP valeur_valeur1=std::static_pointer_cast<Point>(expr_valeur1);
							ExpressionSP expr_valeur2(*valeur2);
							VecteurSP valeur_valeur2=std::static_pointer_cast<Vecteur>(expr_valeur2);
							ExpressionSP expr_valeur3(*valeur3);
							NombreSP valeur_valeur3=std::static_pointer_cast<Nombre>(expr_valeur3);
							SECURISE_VALEUR(RectangleSP(new Rectangle(valeur_valeur1,valeur_valeur2,valeur_valeur3,_options)));
						}
				};
			template<>
				class EvaluateurOperateurTernaire<Rectangle,Point,VecteurPolaire,Nombre> : public Rectangle
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3) const
						{
							ExpressionSP expr_valeur1(*valeur1);
							PointSP valeur_valeur1=std::static_pointer_cast<Point>(expr_valeur1);
							ExpressionSP expr_valeur2(*valeur2);
							VecteurPolaireSP valeur_valeur2=std::static_pointer_cast<VecteurPolaire>(expr_valeur2);
							ExpressionSP expr_valeur3(*valeur3);
							NombreSP valeur_valeur3=std::static_pointer_cast<Nombre>(expr_valeur3);
							SECURISE_VALEUR(RectangleSP(new Rectangle(valeur_valeur1,valeur_valeur2,valeur_valeur3,_options)));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Rectangle::Rectangle(ExpressionSP sommet1, ExpressionSP sommet2, OptionsSurfaceSP options, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneTernaire(sommet1,sommet2,ExpressionSP(new Vide())), _evaluateur(RectangleSP(new EvaluateurOperateurTernaire<Rectangle,Expression,Expression,Expression>())), _options(options)
{
	ExpressionSP sommet3 = _valeur3;

	EVALUATEUR_RECTANGLE(Point,Point         ,Vide);
	EVALUATEUR_RECTANGLE(Point,Vecteur       ,Vide);
	EVALUATEUR_RECTANGLE(Point,VecteurPolaire,Vide);
	// calcul du type de retour

	if(_type)
		return;
	if(sommet1->est_de_type<Traceur::Donnees::Type::Point>())
	{
		if(sommet2->est_de_type<Traceur::Donnees::Type::Point>())
		{
			_type = Traceur::Donnees::Type::Rectangle::SP();
			return;
		}
		if(sommet2->est_de_type<Traceur::Donnees::Type::Vector>())
		{
			_type = Traceur::Donnees::Type::Rectangle::SP();
			return;
		}
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionDeux(position,"RECTANGLE",sommet1,sommet2);
}

	Rectangle::Rectangle(ExpressionSP sommet1, ExpressionSP sommet2, ExpressionSP sommet3, OptionsSurfaceSP options, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneTernaire(sommet1,sommet2,sommet3), _evaluateur(RectangleSP(new EvaluateurOperateurTernaire<Rectangle,Expression,Expression,Expression>())), _options(options)
{
	EVALUATEUR_RECTANGLE(Point,Point         ,Nombre);
	EVALUATEUR_RECTANGLE(Point,Vecteur       ,Nombre);
	EVALUATEUR_RECTANGLE(Point,VecteurPolaire,Nombre);
	EVALUATEUR_RECTANGLE(Point,Point         ,Vide  );
	EVALUATEUR_RECTANGLE(Point,Vecteur       ,Vide  );
	EVALUATEUR_RECTANGLE(Point,VecteurPolaire,Vide  );
	// calcul du type de retour
	if(_type)
		return;
	if(sommet1->est_de_type<Traceur::Donnees::Type::Point>() and (sommet3->est_de_type<Traceur::Donnees::Type::Number>() or sommet3->est_de_type<Traceur::Donnees::Type::Void>()))
	{
		if(sommet2->est_de_type<Traceur::Donnees::Type::Point>())
		{
			_type = Traceur::Donnees::Type::Rectangle::SP();
			return;
		}
		if(sommet2->est_de_type<Traceur::Donnees::Type::Vector>())
		{
			_type = Traceur::Donnees::Type::Rectangle::SP();
			return;
		}
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionTrois(position,"RECTANGLE",sommet1,sommet2,sommet3);
}

ValeurSP Rectangle::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur1 = _valeur1->evaluation(assembleur); if(not valeur_valeur1->a_une_valeur_retour()) return valeur_valeur1;
	ValeurSP valeur_valeur2 = _valeur2->evaluation(assembleur); if(not valeur_valeur2->a_une_valeur_retour()) return valeur_valeur2;
	ValeurSP valeur_valeur3 = _valeur3->evaluation(assembleur); if(not valeur_valeur3->a_une_valeur_retour()) return valeur_valeur3;
	try
	{
		OptionsSurfaceSP options = _options->evaluation(assembleur);
		Rectangle resultat(*valeur_valeur1,*valeur_valeur2,*valeur_valeur3,options);
		return resultat.evaluation(valeur_valeur1,valeur_valeur2,valeur_valeur3,options);
	} catch(ValeurSP valeur) { return valeur; }
}

ValeurSP Rectangle::evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3 , const OptionsSurfaceSP options) const
{
	_evaluateur->_options=options;
	return _evaluateur->evaluation(valeur1,valeur2,valeur3);
}

ValeurSP Rectangle::evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3) const
{
	return _evaluateur->evaluation(valeur1,valeur2,valeur3);
}

void Rectangle::format(std::ostream& os) const
{
	//_type->format(os);
	os << "RECTANGLE[";
	_valeur1->format(os);
	os << ";";
	_valeur2->format(os);
	if(not _valeur3->est_de_nature<Vide>())
	{
		os << ";";
		_valeur3->format(os);
	}
	_options->format(os);
	os << "]";
}

ExpressionSP Rectangle::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur1 = _valeur1->clone(clones_instances);
	ExpressionSP valeur2 = _valeur2->clone(clones_instances);
	ExpressionSP valeur3 = _valeur3->clone(clones_instances);
	RectangleSP clone(new Rectangle());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_valeur1=valeur1;
	clone->_valeur2=valeur2;
	clone->_valeur3=valeur3;
	clone->_options = _options->clone(clones_instances);
	return clone;
}

Traceur::Donnees::Assembleur::InstructionSP Rectangle::draw(Traceur::Donnees::Assembleur::Assembleur& assembleur)
{
	PointSP expr_valeur1 = std::static_pointer_cast<Point>(_valeur1);
	Traceur::Donnees::Assembleur::Point sommet1(expr_valeur1,assembleur._pile.top()->_systeme_coordonnees);
	Traceur::Donnees::Assembleur::Point sommet2;
	Traceur::Donnees::Assembleur::Point sommet3;
	Traceur::Donnees::Assembleur::Point sommet4;
	if(_valeur3->est_de_nature<Vide>())
	{
		if(_valeur2->est_de_nature<Point>())
		{
			PointSP expr_valeur2 = std::static_pointer_cast<Point>(_valeur2);
			sommet4=Traceur::Donnees::Assembleur::Point(expr_valeur2,assembleur._pile.top()->_systeme_coordonnees);
		}
		else if(_valeur2->est_de_nature<Vecteur>())
		{
			VecteurSP expr_valeur2 = std::dynamic_pointer_cast<Vecteur>(_valeur2);
			PointSP point4(new Point(expr_valeur1->translation(*expr_valeur2)));
			sommet4=Traceur::Donnees::Assembleur::Point(point4,assembleur._pile.top()->_systeme_coordonnees);
		}
		else { throw; }
		sommet2=Traceur::Donnees::Assembleur::Point(sommet1._x,sommet4._y);
		sommet3=Traceur::Donnees::Assembleur::Point(sommet4._x,sommet1._y);
	}
	else
	{
		VecteurSP decalage;
		PointSP point3;
		if(_valeur2->est_de_nature<Point>())
		{
			PointSP expr_valeur2 = std::static_pointer_cast<Point>(_valeur2);
			sommet3=Traceur::Donnees::Assembleur::Point(expr_valeur2,assembleur._pile.top()->_systeme_coordonnees);
			decalage = VecteurSP(new Vecteur(expr_valeur2->decalage(*expr_valeur1)));
			point3 = expr_valeur2;
		}
		else if(_valeur2->est_de_nature<Vecteur>())
		{
			VecteurSP expr_valeur2 = std::dynamic_pointer_cast<Vecteur>(_valeur2);
			point3 = PointSP(new Point(expr_valeur1->translation(*expr_valeur2)));
			sommet3=Traceur::Donnees::Assembleur::Point(point3,assembleur._pile.top()->_systeme_coordonnees);
			decalage = expr_valeur2;
		}
		else { throw; }
		NombreSP expr_valeur3 = std::dynamic_pointer_cast<Nombre>(_valeur3);
		decalage = VecteurSP(new VecteurPolaire(expr_valeur3,(decalage->rotation(Nombre::pi_sur_deux()).argument())));
		Traceur::Donnees::Assembleur::Point point_decalage(PointSP(new Point(*decalage)),assembleur._pile.top()->_systeme_coordonnees);
		sommet2=sommet1.translate(point_decalage);
		sommet4=sommet3.translate(point_decalage);
	}
	Traceur::Donnees::Assembleur::InstructionSP drawrectangle(new Traceur::Donnees::Assembleur::DrawRectangle(sommet1,sommet2,sommet3,sommet4,*_options));
	return drawrectangle;
}
