#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionchaine.h>
#include <cmath>
using namespace Traceur::Donnees::Expression;

Reel Reel::operator+()
{
	return Reel(_reel);
}
Reel Reel::operator-()
{
	return Reel(-_reel);
}
Reel Reel::operator+(Reel reel) const
{
	return Reel(_reel+reel._reel);
}
Reel Reel::operator+(Entier entier) const
{
	Reel reel(entier);
	return Reel(_reel+reel._reel);
}
Reel Reel::operator-(Reel reel) const
{
	return Reel(_reel-reel._reel);
}
Reel Reel::operator-(Entier entier) const
{
	Reel reel(entier);
	return Reel(_reel-reel._reel);
}
Reel Reel::operator*(Reel reel) const
{
	return Reel(_reel*reel._reel);
}
Reel Reel::operator*(Entier entier) const
{
	Reel reel(entier);
	return Reel(_reel*reel._reel);
}
Reel Reel::operator/(Reel reel) const
{
	return Reel(_reel/reel._reel);
}
Reel Reel::operator/(Entier entier) const
{
	Reel reel(entier);
	return Reel(_reel/reel._reel);
}
Reel Reel::operator%(Reel reel) const
{
	Reel::TypeInterne intpart;
	::modf(_reel/::fabs(reel._reel),&intpart);
	return Reel(_reel-intpart);
}
Reel Reel::operator%(Entier entier) const
{
	Reel reel(entier);
	return (*this)%reel;
}
Reel Reel::operator^(Reel reel) const
{
	return Reel(::pow(_reel,reel._reel));
}

bool Reel::exposantinvalide(Reel gauche, Reel droite)
{
	Reel::TypeInterne valeur=::pow(gauche._reel,droite._reel);
	return isnan(valeur) or isinf(valeur);
}

Reel::TypeInterne Reel::exposant(Entier::TypeInterne exp) const
{
	if(not exp)
		return 1.;
	if(exp%2)
		return _reel*exposant(exp-1);
	TypeInterne res = exposant(exp/2);
	return res*res;
}

Reel Reel::operator^(Entier entier) const
{
	if(entier<Entier(0))
		return Reel(1./(exposant(-entier._entier)));
	return Reel(exposant(entier._entier));
}
bool Reel::operator==(Reel reel) const
{
	return (::fabs(_reel-reel._reel)<PRECISION);
}
bool Reel::operator==(Entier entier) const
{
	Reel reel(entier);
	return (::fabs(_reel-reel._reel)<PRECISION);
}
bool Reel::operator!=(Reel reel)
{
	return (::fabs(_reel-reel._reel)>=PRECISION);
}
bool Reel::operator!=(Entier entier)
{
	Reel reel(entier);
	return (::fabs(_reel-reel._reel)>=PRECISION);
}
bool Reel::operator<(Reel reel)
{
	return (_reel<reel._reel);
}
bool Reel::operator<(Entier entier)
{
	Reel reel(entier);
	return (_reel<reel._reel);
}
bool Reel::operator>(Reel reel)
{
	return (_reel>reel._reel);
}
bool Reel::operator>(Entier entier)
{
	Reel reel(entier);
	return (_reel>reel._reel);
}
bool Reel::operator<=(Reel reel)
{
	return ((_reel-PRECISION/2)<=reel._reel);
}
bool Reel::operator<=(Entier entier)
{
	Reel reel(entier);
	return ((_reel-PRECISION/2)<=reel._reel);
}
bool Reel::operator>=(Reel reel)
{
	return ((_reel+PRECISION/2)>=reel._reel);
}
bool Reel::operator>=(Entier entier)
{
	Reel reel(entier);
	return ((_reel+PRECISION/2)>=reel._reel);
}
Reel Reel::cosinus(Reel& reel)
{
	return Reel(::cos(reel._reel));
}

Reel Reel::sinus(Reel& reel)
{
	return Reel(::sin(reel._reel));
}

Reel Reel::tangente(Reel& reel)
{
	return Reel(::tan(reel._reel));
}

Reel Reel::arccosinus(Reel& reel)
{
	return Reel(::acos(reel._reel));
}

Reel Reel::arcsinus(Reel& reel)
{
	return Reel(::asin(reel._reel));
}

Reel Reel::arctangente(Reel& reel)
{
	return Reel(::atan(reel._reel));
}

Reel Reel::exponentielle(Reel& reel)
{
	return Reel(::exp(reel._reel));
}

Reel Reel::logarithme_naturel(Reel& reel)
{
	return Reel(::log(reel._reel));
}

Reel Reel::racine_carree(Reel& reel)
{
	return Reel(::sqrt(reel._reel));
}

Reel Reel::valeur_absolue(Reel& reel)
{
	if(reel._reel>0)
		return reel;
	return Reel(-(reel._reel));
}

Reel Reel::partie_entiere(Reel& reel)
{
	Reel::TypeInterne intpart;
	::modf(reel._reel,&intpart);
	return Reel(intpart);
}

Reel Reel::partie_decimale(Reel& reel)
{
	Reel::TypeInterne intpart;
	return Reel(::modf(reel._reel,&intpart));
}

Reel Reel::argument(Reel& gauche, Reel& droite)
{
	return Reel(::atan2(droite._reel,gauche._reel));
}
Reel Reel::degres(Reel& radians)
{
	return Reel((radians._reel*180)/M_PI);
}

Reel Reel::radians(Reel& degres)
{
	return Reel((degres._reel*M_PI)/180);
}

Reel::Reel(const Entier& entier)
	:_reel(static_cast<Reel::TypeInterne>(entier._entier))
{
	_type = Traceur::Donnees::Type::Real::SP();
}

Reel::Reel(const Chaine& chaine)
{
	_type = Traceur::Donnees::Type::Real::SP();
	std::istringstream iss;
	iss.str(chaine._chaine);
	iss >> _reel;
}

Traceur::Donnees::Instruction::ValeurSP Reel::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	return Traceur::Donnees::Instruction::ValeurSP(new Traceur::Donnees::Instruction::Valeur(ReelSP(new Reel(*this))));
}

void Reel::format(std::ostream& os) const
{
	std::ostringstream oss;
	oss << _reel;
	os << oss.str();
	if(oss.str().find('.')==std::string::npos)
		os << ".";
}

ReelSP Reel::pi_sur_deux()
{
	static ReelSP pi_sur_deux = ReelSP(new Reel(M_PI_2));
	return pi_sur_deux;
}

ReelSP Reel::pi()
{
	static ReelSP pi = ReelSP(new Reel(M_PI));
	return pi;
}
