#ifndef _DONNEES_EXPRESSIONS_REEL_H_
#define _DONNEES_EXPRESSIONS_REEL_H_

#include <cmath>

#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Expression { DECL_SHARED_PTR(Reel); } } }

#include <src/donnees/expressions/expressionnombre.h>
#include <src/donnees/expressions/expressionbooleenvrai.h>
#include <src/donnees/expressions/expressionbooleenfaux.h>
#include <src/donnees/types/typereal.h>
#include <src/donnees/expressions/expressionentier.h>

#define PRECISION 1e-8

namespace Traceur
{
	namespace Donnees
	{
		namespace Assembleur { namespace Couleur { class Couleur; class Frequencielle; } }
		namespace Expression
		{
			class Chaine;
			class Reel: public Nombre
			{
				friend class Entier;
				friend class Chaine;
				friend class Traceur::Donnees::Assembleur::Couleur::Couleur;
				friend class Traceur::Donnees::Assembleur::Couleur::Frequencielle;
				public:
					typedef double TypeInterne;
					Reel(Reel::TypeInterne reel)
						:_reel(reel) { _type=Traceur::Donnees::Type::Real::SP(); };
					Reel(const Entier& entier);
					Reel(const Chaine& chaine);
					static ReelSP pi_sur_deux();
					static ReelSP pi();

					virtual ~Reel() {};

					virtual Traceur::Donnees::Instruction::ValeurSP evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur);
					virtual void format(std::ostream& os) const;

					Reel operator+();
					Reel operator-();
					Reel operator+(Reel reel) const;
					Reel operator-(Reel reel) const;
					Reel operator*(Reel reel) const;
					Reel operator/(Reel reel) const;
					Reel operator%(Reel reel) const;
					Reel operator^(Reel reel) const;
					static bool exposantinvalide(Reel gauche, Reel droite);
					bool operator==(Reel reel) const;
					bool operator!=(Reel reel);
					bool operator<(Reel reel);
					bool operator>(Reel reel);
					bool operator<=(Reel reel);
					bool operator>=(Reel reel);
					static Reel cosinus(Reel& reel);
					static Reel sinus(Reel& reel);
					static Reel tangente(Reel& reel);
					static Reel arccosinus(Reel& reel);
					static Reel arcsinus(Reel& reel);
					static Reel arctangente(Reel& reel);
					static Reel exponentielle(Reel& reel);
					static Reel logarithme_naturel(Reel& reel);
					static Reel racine_carree(Reel& reel);
					static Reel valeur_absolue(Reel& reel);
					static Reel partie_entiere(Reel& reel);
					static Reel partie_decimale(Reel& reel);
					static Reel argument(Reel& gauche, Reel& droite);
					static Reel degres(Reel& radians);
					static Reel radians(Reel& degres);
					Reel operator+(Entier entier) const;
					Reel operator-(Entier entier) const;
					Reel operator*(Entier entier) const;
					Reel operator/(Entier entier) const;
					Reel operator%(Entier entier) const;
					Reel operator^(Entier entier) const;
					bool operator==(Entier entier) const;
					bool operator!=(Entier entier);
					bool operator<(Entier entier);
					bool operator>(Entier entier);
					bool operator<=(Entier entier);
					bool operator>=(Entier entier);
				private:
					Reel::TypeInterne _reel;
					Reel::TypeInterne exposant(Entier::TypeInterne exp) const;
			};
		}
	}
}

#endif
