#include <src/donnees/expressions/expressionremplissage.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionevaluateur.h>

#include <src/donnees/assembleur/drawfill.h>

#define EVALUATEUR_REMPLISSAGE(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = RemplissageSP(new EvaluateurOperateurUnaire<Remplissage,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<Remplissage,Point> : public Remplissage
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur)
						{
							ExpressionSP expr_valeur(*valeur);
							PointSP valeur_valeur=std::static_pointer_cast<Point>(expr_valeur);
							SECURISE_VALEUR(RemplissageSP(new Remplissage(valeur_valeur,_options)));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Remplissage::Remplissage(ExpressionSP valeur, OptionsSurfaceSP options, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(RemplissageSP(new EvaluateurOperateurUnaire<Remplissage,Expression>())), _options(options)
{

	EVALUATEUR_REMPLISSAGE(Point);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Point>())
	{
		_type = Traceur::Donnees::Type::Fill::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionUn(position,"SOLID",valeur);
}

ValeurSP Remplissage::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	try
	{
		OptionsSurfaceSP options = _options->evaluation(assembleur);
		Remplissage resultat(*valeur_valeur,options);
		return resultat.evaluation(valeur_valeur,options);
	} catch(ValeurSP valeur) { return valeur; }
}

ValeurSP Remplissage::evaluation(const ValeurSP valeur, const OptionsSurfaceSP options)
{
	_evaluateur->_options=options;
	return _evaluateur->evaluation(valeur);
}

ValeurSP Remplissage::evaluation(const ValeurSP valeur)
{
	return _evaluateur->evaluation(valeur);
}

void Remplissage::format(std::ostream& os) const
{
	os << "FILL[";
	_valeur->format(os);
	_options->format(os);
	os << "]";
}

ExpressionSP Remplissage::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	RemplissageSP clone(new Remplissage());
	clone->_valeur = valeur;
	clone->_options = _options->clone(clones_instances);
	return clone;
}

Traceur::Donnees::Assembleur::InstructionSP Remplissage::draw(Traceur::Donnees::Assembleur::Assembleur& assembleur)
{
	PointSP expr_valeur = std::static_pointer_cast<Point>(_valeur);
	Traceur::Donnees::Assembleur::Point point(expr_valeur,assembleur._pile.top()->_systeme_coordonnees);
	Traceur::Donnees::Assembleur::InstructionSP drawfill(new Traceur::Donnees::Assembleur::DrawFill(point,*_options));
	return drawfill;
}
