#include <src/donnees/expressions/expressionreste.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionvecteur.h>
#include <src/donnees/expressions/expressionvecteurpolaire.h>
#include <src/donnees/expressions/expressionchaine.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/expressions/expressioncouleurhsva.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_RESTE(type1,type2) \
	if(gauche->est_de_nature<type1>() and droite->est_de_nature<type2>()) { _evaluateur = ResteSP(new EvaluateurOperateurBinaire<type1,Reste,type2>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici

			template<>
				class EvaluateurOperateurBinaire<Entier,Reste,Entier> : public Reste
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							EntierSP valeur_gauche=std::static_pointer_cast<Entier>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							if((*valeur_droite)==Entier(0))
								return Valeur::reste_par_zero();
							else
							{ SECURISE_VALEUR(EntierSP(new Entier((*valeur_gauche)%(*valeur_droite)))); }
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Reel,Reste,Entier> : public Reste
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ReelSP valeur_gauche=std::static_pointer_cast<Reel>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							if((*valeur_droite)==Reel(0))
								return Valeur::reste_par_zero();
							else
							{ SECURISE_VALEUR(ReelSP(new Reel((*valeur_gauche)%(*valeur_droite)))); }
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Entier,Reste,Reel> : public Reste
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							EntierSP valeur_gauche=std::static_pointer_cast<Entier>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ReelSP valeur_droite=std::static_pointer_cast<Reel>(expr_droite);
							if((*valeur_droite)==Reel(0))
								return Valeur::reste_par_zero();
							else
							{ SECURISE_VALEUR(ReelSP(new Reel((*valeur_gauche)%(*valeur_droite)))); }
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Reel,Reste,Reel> : public Reste
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ReelSP valeur_gauche=std::static_pointer_cast<Reel>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ReelSP valeur_droite=std::static_pointer_cast<Reel>(expr_droite);
							if((*valeur_droite)==Reel(0))
								return Valeur::reste_par_zero();
							else
							{ SECURISE_VALEUR(ReelSP(new Reel((*valeur_gauche)%(*valeur_droite)))); }
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Vecteur,Reste,Vecteur> : public Reste
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							VecteurSP valeur_gauche=std::static_pointer_cast<Vecteur>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							VecteurSP valeur_droite=std::static_pointer_cast<Vecteur>(expr_droite);
							if(valeur_droite->est_le_vecteur_nul())
								return Valeur::symetrie_par_vecteur_nul();
							else
							{ SECURISE_VALEUR(VecteurSP(new Vecteur(valeur_gauche->symetrie(*valeur_droite)))); }
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Chaine,Reste,Chaine> : public Reste
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ChaineSP valeur_gauche=std::static_pointer_cast<Chaine>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ChaineSP valeur_droite=std::static_pointer_cast<Chaine>(expr_droite);
							SECURISE_VALEUR(ChaineSP(new Chaine((*valeur_gauche)%(*valeur_droite))));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurRVBA,Reste,Nombre> : public Reste
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurRVBASP valeur_gauche=std::static_pointer_cast<CouleurRVBA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							NombreSP valeur_droite=std::static_pointer_cast<Nombre>(expr_droite);
							SECURISE_VALEUR(valeur_gauche->augmente_alpha(valeur_droite));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurHSVA,Reste,Nombre> : public Reste
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurHSVASP valeur_gauche=std::static_pointer_cast<CouleurHSVA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							NombreSP valeur_droite=std::static_pointer_cast<Nombre>(expr_droite);
							SECURISE_VALEUR(valeur_gauche->augmente_alpha(valeur_droite));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Reste::Reste(ExpressionSP gauche, ExpressionSP droite, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneBinaire(gauche,droite), _evaluateur(ResteSP(new EvaluateurOperateurBinaire<Expression,Reste,Expression>()))
{

	EVALUATEUR_RESTE(Entier     ,Entier );
	EVALUATEUR_RESTE(Reel       ,Entier );
	EVALUATEUR_RESTE(Entier     ,Reel   );
	EVALUATEUR_RESTE(Reel       ,Reel   );
	EVALUATEUR_RESTE(Vecteur    ,Vecteur);
	EVALUATEUR_RESTE(Chaine     ,Chaine );
	EVALUATEUR_RESTE(CouleurRVBA,Nombre );
	EVALUATEUR_RESTE(CouleurHSVA,Nombre );
	// calcul du type de retour
	if(_type)
		return;
	if(gauche->est_de_type<Traceur::Donnees::Type::Integer>() and droite->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Integer::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Real>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Number>() and droite->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Number>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Vector>() and droite->est_de_type<Traceur::Donnees::Type::Vector>())
	{
		_type = Traceur::Donnees::Type::Vector::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::String>() and droite->est_de_type<Traceur::Donnees::Type::String>())
	{
		_type = Traceur::Donnees::Type::String::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Color>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Color::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionInfixe(position,gauche,"%",droite);
}

ValeurSP Reste::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_gauche = _gauche->evaluation(assembleur); if(not valeur_gauche->a_une_valeur_retour()) return valeur_gauche;
	ValeurSP valeur_droite = _droite->evaluation(assembleur); if(not valeur_droite->a_une_valeur_retour()) return valeur_droite;
	Reste resultat(*valeur_gauche,*valeur_droite);
	return resultat.evaluation(valeur_gauche,valeur_droite);
}

ValeurSP Reste::evaluation(const ValeurSP gauche, const ValeurSP droite) const
{
	return _evaluateur->evaluation(gauche,droite);
}

void Reste::format(std::ostream& os) const
{
	//_type->format(os);
	os << "(";
	_gauche->format(os);
	os << ")%(";
	_droite->format(os);
	os << ")";
}
ExpressionSP Reste::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP gauche = _gauche->clone(clones_instances);
	ExpressionSP droite = _droite->clone(clones_instances);
	ResteSP clone(new Reste());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_gauche=gauche;
	clone->_droite=droite;
	return clone;
}
