#include <src/donnees/expressions/expressionsinus.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteur.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_SINUS(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = SinusSP(new EvaluateurOperateurUnaire<Sinus,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<Sinus,Entier> : public Sinus
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							EntierSP valeur_valeur=std::static_pointer_cast<Entier>(expr_valeur);
							SECURISE_VALEUR(ReelSP(new Reel(Entier::sinus(*valeur_valeur))));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<Sinus,Reel> : public Sinus
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ReelSP valeur_valeur=std::static_pointer_cast<Reel>(expr_valeur);
							SECURISE_VALEUR(ReelSP(new Reel(Reel::sinus(*valeur_valeur))));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<Sinus,Point> : public Sinus
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							PointSP valeur_valeur=std::static_pointer_cast<Point>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->ordonnee());
						}
				};
			template<>
				class EvaluateurOperateurUnaire<Sinus,Vecteur> : public Sinus
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							VecteurSP valeur_valeur=std::static_pointer_cast<Vecteur>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->ordonnee());
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Sinus::Sinus(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(SinusSP(new EvaluateurOperateurUnaire<Sinus,Expression>()))
{

	EVALUATEUR_SINUS(Entier );
	EVALUATEUR_SINUS(Reel   );
	EVALUATEUR_SINUS(Point  );
	EVALUATEUR_SINUS(Vecteur);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Point>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Vector>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionPrefixe(position,"sin",valeur);
}

ValeurSP Sinus::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	Sinus resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP Sinus::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void Sinus::format(std::ostream& os) const
{
	os << "sin (";
	_valeur->format(os);
	os << ")";
}

ExpressionSP Sinus::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	SinusSP clone(new Sinus(valeur));
	return clone;
}
