#include <src/donnees/expressions/expressionsolide.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionevaluateur.h>

#include <src/donnees/assembleur/drawsolid.h>

#define EVALUATEUR_SOLIDE(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = SolideSP(new EvaluateurOperateurUnaire<Solide,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<Solide,Point> : public Solide
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur)
						{
							ExpressionSP expr_valeur(*valeur);
							PointSP valeur_valeur=std::static_pointer_cast<Point>(expr_valeur);
							SECURISE_VALEUR(SolideSP(new Solide(valeur_valeur)));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Solide::Solide(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(SolideSP(new EvaluateurOperateurUnaire<Solide,Expression>()))
{

	EVALUATEUR_SOLIDE(Point);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Point>())
	{
		_type = Traceur::Donnees::Type::Solid::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionUn(position,"SOLID",valeur);
}

ValeurSP Solide::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	Solide resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP Solide::evaluation(const ValeurSP valeur)
{
	return _evaluateur->evaluation(valeur);
}

void Solide::format(std::ostream& os) const
{
	os << "SOLID[";
	_valeur->format(os);
	os << "]";
}

ExpressionSP Solide::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	SolideSP clone(new Solide(valeur));
	return clone;
}

Traceur::Donnees::Assembleur::InstructionSP Solide::draw(Traceur::Donnees::Assembleur::Assembleur& assembleur)
{
	PointSP expr_valeur = std::static_pointer_cast<Point>(_valeur);
	Traceur::Donnees::Assembleur::Point point(expr_valeur,assembleur._pile.top()->_systeme_coordonnees);
	Traceur::Donnees::Assembleur::InstructionSP drawsolid(new Traceur::Donnees::Assembleur::DrawSolid(point));
	return drawsolid;
}
