#include <src/donnees/expressions/expressionsoustraction.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteur.h>
#include <src/donnees/expressions/expressionvecteurpolaire.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/expressions/expressioncouleurhsva.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_SOUSTRACTION(type1,type2) \
	if(gauche->est_de_nature<type1>() and droite->est_de_nature<type2>()) { _evaluateur = SoustractionSP(new EvaluateurOperateurBinaire<type1,Soustraction,type2>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurBinaire<Entier,Soustraction,Entier> : public Soustraction
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							EntierSP valeur_gauche=std::static_pointer_cast<Entier>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							SECURISE_VALEUR(EntierSP(new Entier((*valeur_gauche)-(*valeur_droite))));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Reel,Soustraction,Entier> : public Soustraction
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ReelSP valeur_gauche=std::static_pointer_cast<Reel>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							EntierSP valeur_droite=std::static_pointer_cast<Entier>(expr_droite);
							SECURISE_VALEUR(ReelSP(new Reel((*valeur_gauche)-(*valeur_droite))));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Entier,Soustraction,Reel> : public Soustraction
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							EntierSP valeur_gauche = std::dynamic_pointer_cast<Entier>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ReelSP valeur_droite=std::static_pointer_cast<Reel>(expr_droite);
							SECURISE_VALEUR(ReelSP(new Reel((*valeur_gauche)-(*valeur_droite))));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Reel,Soustraction,Reel> : public Soustraction
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							ReelSP valeur_gauche=std::static_pointer_cast<Reel>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							ReelSP valeur_droite=std::static_pointer_cast<Reel>(expr_droite);
							SECURISE_VALEUR(ReelSP(new Reel((*valeur_gauche)-(*valeur_droite))));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Point,Soustraction,Point> : public Soustraction
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							PointSP valeur_gauche=std::static_pointer_cast<Point>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							PointSP valeur_droite=std::static_pointer_cast<Point>(expr_droite);
							SECURISE_VALEUR(VecteurSP(new Vecteur(valeur_gauche->decalage(*valeur_droite))));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Point,Soustraction,Vecteur> : public Soustraction
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							PointSP valeur_gauche=std::static_pointer_cast<Point>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							VecteurSP valeur_droite=std::static_pointer_cast<Vecteur>(expr_droite);
							SECURISE_VALEUR(PointSP(new Point(valeur_gauche->translation_opposee(*valeur_droite))));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<Vecteur,Soustraction,Vecteur> : public Soustraction
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							VecteurSP valeur_gauche=std::static_pointer_cast<Vecteur>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							VecteurSP valeur_droite=std::static_pointer_cast<Vecteur>(expr_droite);
							SECURISE_VALEUR(VecteurSP(new Vecteur((*valeur_gauche)-(*valeur_droite))));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurRVBA,Soustraction,Nombre> : public Soustraction
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurRVBASP valeur_gauche=std::static_pointer_cast<CouleurRVBA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							NombreSP valeur_droite=std::static_pointer_cast<Nombre>(expr_droite);
							SECURISE_VALEUR(valeur_gauche->diminue_saturation(valeur_droite));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurHSVA,Soustraction,Nombre> : public Soustraction
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurHSVASP valeur_gauche=std::static_pointer_cast<CouleurHSVA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							NombreSP valeur_droite=std::static_pointer_cast<Nombre>(expr_droite);
							SECURISE_VALEUR(valeur_gauche->diminue_saturation(valeur_droite));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurRVBA,Soustraction,CouleurRVBA> : public Soustraction
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurRVBASP valeur_gauche=std::static_pointer_cast<CouleurRVBA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							CouleurRVBASP valeur_droite=std::static_pointer_cast<CouleurRVBA>(expr_droite);
							SECURISE_VALEUR((*valeur_gauche)-(*valeur_droite));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurRVBA,Soustraction,CouleurHSVA> : public Soustraction
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurRVBASP valeur_gauche=std::static_pointer_cast<CouleurRVBA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							CouleurHSVASP valeur_droite=std::static_pointer_cast<CouleurHSVA>(expr_droite);
							SECURISE_VALEUR((*valeur_gauche)-(*valeur_droite));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurHSVA,Soustraction,CouleurRVBA> : public Soustraction
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurHSVASP valeur_gauche=std::static_pointer_cast<CouleurHSVA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							CouleurRVBASP valeur_droite=std::static_pointer_cast<CouleurRVBA>(expr_droite);
							SECURISE_VALEUR((*valeur_gauche)-(*valeur_droite));
						}
				};
			template<>
				class EvaluateurOperateurBinaire<CouleurHSVA,Soustraction,CouleurHSVA> : public Soustraction
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							CouleurHSVASP valeur_gauche=std::static_pointer_cast<CouleurHSVA>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							CouleurHSVASP valeur_droite=std::static_pointer_cast<CouleurHSVA>(expr_droite);
							SECURISE_VALEUR((*valeur_gauche)-(*valeur_droite));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Soustraction::Soustraction(ExpressionSP gauche, ExpressionSP droite, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneBinaire(gauche,droite), _evaluateur(SoustractionSP(new EvaluateurOperateurBinaire<Expression,Soustraction,Expression>()))
{

	EVALUATEUR_SOUSTRACTION(Entier     ,Entier     );
	EVALUATEUR_SOUSTRACTION(Reel       ,Entier     );
	EVALUATEUR_SOUSTRACTION(Entier     ,Reel       );
	EVALUATEUR_SOUSTRACTION(Reel       ,Reel       );
	EVALUATEUR_SOUSTRACTION(Vecteur    ,Vecteur    );
	EVALUATEUR_SOUSTRACTION(Point      ,Point      );
	EVALUATEUR_SOUSTRACTION(Point      ,Vecteur    );
	EVALUATEUR_SOUSTRACTION(CouleurRVBA,Nombre     );
	EVALUATEUR_SOUSTRACTION(CouleurHSVA,Nombre     );
	EVALUATEUR_SOUSTRACTION(CouleurRVBA,CouleurRVBA);
	EVALUATEUR_SOUSTRACTION(CouleurRVBA,CouleurHSVA);
	EVALUATEUR_SOUSTRACTION(CouleurHSVA,CouleurRVBA);
	EVALUATEUR_SOUSTRACTION(CouleurHSVA,CouleurHSVA);
	// calcul du type de retour
	if(_type)
		return;
	if(gauche->est_de_type<Traceur::Donnees::Type::Integer>() and droite->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Integer::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Real>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Number>() and droite->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Number>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Vector>() and droite->est_de_type<Traceur::Donnees::Type::Vector>())
	{
		_type = Traceur::Donnees::Type::Vector::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Point>() and droite->est_de_type<Traceur::Donnees::Type::Vector>())
	{
		_type = Traceur::Donnees::Type::Point::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Point>() and droite->est_de_type<Traceur::Donnees::Type::Point>())
	{
		_type = Traceur::Donnees::Type::Vector::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Color>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Color::SP();
		return;
	}
	if(gauche->est_de_type<Traceur::Donnees::Type::Color>() and droite->est_de_type<Traceur::Donnees::Type::Color>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionInfixe(position,gauche,"-",droite);
}

ValeurSP Soustraction::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_gauche = _gauche->evaluation(assembleur); if(not valeur_gauche->a_une_valeur_retour()) return valeur_gauche;
	ValeurSP valeur_droite = _droite->evaluation(assembleur); if(not valeur_droite->a_une_valeur_retour()) return valeur_droite;
	Soustraction resultat(*valeur_gauche,*valeur_droite);
	return resultat.evaluation(valeur_gauche,valeur_droite);
}

ValeurSP Soustraction::evaluation(const ValeurSP gauche, const ValeurSP droite) const
{
	return _evaluateur->evaluation(gauche,droite);
}

void Soustraction::format(std::ostream& os) const
{
	//_type->format(os);
	os << "(";
	_gauche->format(os);
	os << ")-(";
	_droite->format(os);
	os << ")";
}

ExpressionSP Soustraction::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP gauche = _gauche->clone(clones_instances);
	ExpressionSP droite = _droite->clone(clones_instances);
	SoustractionSP clone(new Soustraction());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_gauche=gauche;
	clone->_droite=droite;
	return clone;
}
