#include <src/donnees/expressions/expressiontangente.h>

#include <src/donnees/types/types.h>

#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#include <cmath>
#define EVALUATEUR_TANGENTE(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = TangenteSP(new EvaluateurOperateurUnaire<Tangente,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<Tangente,Entier> : public Tangente
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							EntierSP valeur_valeur=std::static_pointer_cast<Entier>(expr_valeur);
							Reel test(((*valeur_valeur)-Reel(M_PI/2))/Reel(M_PI));
							if(Reel::partie_decimale(test)==Reel(0))
								return Valeur::hors_domaine();
							else
								{ SECURISE_VALEUR(ReelSP(new Reel(Entier::tangente(*valeur_valeur)))); }
						}
				};
			template<>
				class EvaluateurOperateurUnaire<Tangente,Reel> : public Tangente
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ReelSP valeur_valeur=std::static_pointer_cast<Reel>(expr_valeur);
							Reel test(((*valeur_valeur)-Reel(M_PI/2))/Reel(M_PI));
							if(Reel::partie_decimale(test)==Reel(0))
								return Valeur::hors_domaine();
							else
								{ SECURISE_VALEUR(ReelSP(new Reel(Reel::tangente(*valeur_valeur)))); }
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Tangente::Tangente(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(TangenteSP(new EvaluateurOperateurUnaire<Tangente,Expression>()))
{

	EVALUATEUR_TANGENTE(Entier);
	EVALUATEUR_TANGENTE(Reel  );
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionPrefixe(position,"tan",valeur);
}

ValeurSP Tangente::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	Tangente resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP Tangente::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void Tangente::format(std::ostream& os) const
{
	os << "tan (";
	_valeur->format(os);
	os << ")";
}

ExpressionSP Tangente::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	TangenteSP clone(new Tangente(valeur));
	return clone;
}
