#include <src/donnees/expressions/expressiontriangle.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteur.h>
#include <src/donnees/expressions/expressionvecteurpolaire.h>
#include <src/donnees/assembleur/point.h>
#include <src/donnees/assembleur/drawtriangle.h>
#include <src/donnees/assembleur/image.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_TRIANGLE(type1,type2,type3) \
	if(sommet1->est_de_nature<type1>() and sommet2->est_de_nature<type2>() and sommet3->est_de_nature<type3>()) { _evaluateur = TriangleSP(new EvaluateurOperateurTernaire<Triangle,type1,type2,type3>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurTernaire<Triangle,Point,Point,Point> : public Triangle
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3) const
						{
							ExpressionSP expr_valeur1(*valeur1);
							PointSP valeur_valeur1=std::static_pointer_cast<Point>(expr_valeur1);
							ExpressionSP expr_valeur2(*valeur2);
							PointSP valeur_valeur2=std::static_pointer_cast<Point>(expr_valeur2);
							ExpressionSP expr_valeur3(*valeur3);
							PointSP valeur_valeur3=std::static_pointer_cast<Point>(expr_valeur3);
							SECURISE_VALEUR(TriangleSP(new Triangle(valeur_valeur1,valeur_valeur2,valeur_valeur3,_options)));
						}
				};
			template<>
				class EvaluateurOperateurTernaire<Triangle,Point,Vecteur,Vecteur> : public Triangle
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3) const
						{
							ExpressionSP expr_valeur1(*valeur1);
							PointSP valeur_valeur1=std::static_pointer_cast<Point>(expr_valeur1);
							ExpressionSP expr_valeur2(*valeur2);
							VecteurSP valeur_valeur2=std::static_pointer_cast<Vecteur>(expr_valeur2);
							ExpressionSP expr_valeur3(*valeur3);
							VecteurSP valeur_valeur3=std::static_pointer_cast<Vecteur>(expr_valeur3);
							SECURISE_VALEUR(TriangleSP(new Triangle(valeur_valeur1,valeur_valeur2,valeur_valeur3,_options)));
						}
				};
			template<>
				class EvaluateurOperateurTernaire<Triangle,Point,Vecteur,VecteurPolaire> : public Triangle
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3) const
						{
							ExpressionSP expr_valeur1(*valeur1);
							PointSP valeur_valeur1=std::static_pointer_cast<Point>(expr_valeur1);
							ExpressionSP expr_valeur2(*valeur2);
							VecteurSP valeur_valeur2=std::static_pointer_cast<Vecteur>(expr_valeur2);
							ExpressionSP expr_valeur3(*valeur3);
							VecteurPolaireSP valeur_valeur3=std::static_pointer_cast<VecteurPolaire>(expr_valeur3);
							SECURISE_VALEUR(TriangleSP(new Triangle(valeur_valeur1,valeur_valeur2,valeur_valeur3,_options)));
						}
				};
			template<>
				class EvaluateurOperateurTernaire<Triangle,Point,VecteurPolaire,Vecteur> : public Triangle
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3) const
						{
							ExpressionSP expr_valeur1(*valeur1);
							PointSP valeur_valeur1=std::static_pointer_cast<Point>(expr_valeur1);
							ExpressionSP expr_valeur2(*valeur2);
							VecteurPolaireSP valeur_valeur2=std::static_pointer_cast<VecteurPolaire>(expr_valeur2);
							ExpressionSP expr_valeur3(*valeur3);
							VecteurSP valeur_valeur3=std::static_pointer_cast<Vecteur>(expr_valeur3);
							SECURISE_VALEUR(TriangleSP(new Triangle(valeur_valeur1,valeur_valeur2,valeur_valeur3,_options)));
						}
				};
			template<>
				class EvaluateurOperateurTernaire<Triangle,Point,VecteurPolaire,VecteurPolaire> : public Triangle
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3) const
						{
							ExpressionSP expr_valeur1(*valeur1);
							PointSP valeur_valeur1=std::static_pointer_cast<Point>(expr_valeur1);
							ExpressionSP expr_valeur2(*valeur2);
							VecteurPolaireSP valeur_valeur2=std::static_pointer_cast<VecteurPolaire>(expr_valeur2);
							ExpressionSP expr_valeur3(*valeur3);
							VecteurPolaireSP valeur_valeur3=std::static_pointer_cast<VecteurPolaire>(expr_valeur3);
							SECURISE_VALEUR(TriangleSP(new Triangle(valeur_valeur1,valeur_valeur2,valeur_valeur3,_options)));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Triangle::Triangle(ExpressionSP sommet1, ExpressionSP sommet2, ExpressionSP sommet3, OptionsSurfaceSP options, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneTernaire(sommet1,sommet2,sommet3), _evaluateur(TriangleSP(new EvaluateurOperateurTernaire<Triangle,Expression,Expression,Expression>())), _options(options)
{
	EVALUATEUR_TRIANGLE(Point,Point         ,Point         );
	EVALUATEUR_TRIANGLE(Point,Vecteur       ,Vecteur       );
	EVALUATEUR_TRIANGLE(Point,VecteurPolaire,Vecteur       );
	EVALUATEUR_TRIANGLE(Point,Vecteur       ,VecteurPolaire);
	EVALUATEUR_TRIANGLE(Point,VecteurPolaire,VecteurPolaire);
	// calcul du type de retour
	if(_type)
		return;
	if(sommet1->est_de_type<Traceur::Donnees::Type::Point>())
	{
		if(sommet2->est_de_type<Traceur::Donnees::Type::Point>() and sommet3->est_de_type<Traceur::Donnees::Type::Point>())
		{
			_type = Traceur::Donnees::Type::Triangle::SP();
			return;
		}
		if(sommet2->est_de_type<Traceur::Donnees::Type::Vector>() and sommet3->est_de_type<Traceur::Donnees::Type::Vector>())
		{
			_type = Traceur::Donnees::Type::Triangle::SP();
			return;
		}
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionTrois(position,"TRIANGLE",sommet1,sommet2,sommet3);
}

ValeurSP Triangle::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur1 = _valeur1->evaluation(assembleur); if(not valeur_valeur1->a_une_valeur_retour()) return valeur_valeur1;
	ValeurSP valeur_valeur2 = _valeur2->evaluation(assembleur); if(not valeur_valeur2->a_une_valeur_retour()) return valeur_valeur2;
	ValeurSP valeur_valeur3 = _valeur3->evaluation(assembleur); if(not valeur_valeur3->a_une_valeur_retour()) return valeur_valeur3;
	try
	{
		OptionsSurfaceSP options = _options->evaluation(assembleur);
		Triangle resultat(*valeur_valeur1,*valeur_valeur2,*valeur_valeur3,options);
		return resultat.evaluation(valeur_valeur1,valeur_valeur2,valeur_valeur3,options);
	} catch(ValeurSP valeur) { return valeur; }
}

ValeurSP Triangle::evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3 , const OptionsSurfaceSP options) const
{
	_evaluateur->_options=options;
	return _evaluateur->evaluation(valeur1,valeur2,valeur3);
}

ValeurSP Triangle::evaluation(const ValeurSP valeur1, const ValeurSP valeur2, const ValeurSP valeur3) const
{
	return _evaluateur->evaluation(valeur1,valeur2,valeur3);
}

void Triangle::format(std::ostream& os) const
{
	//_type->format(os);
	os << "TRIANGLE[";
	_valeur1->format(os);
	os << ";";
	_valeur2->format(os);
	os << ";";
	_valeur3->format(os);
	_options->format(os);
	os << "]";
}

ExpressionSP Triangle::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur1 = _valeur1->clone(clones_instances);
	ExpressionSP valeur2 = _valeur2->clone(clones_instances);
	ExpressionSP valeur3 = _valeur3->clone(clones_instances);
	TriangleSP clone(new Triangle());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_valeur1=valeur1;
	clone->_valeur2=valeur2;
	clone->_valeur3=valeur3;
	clone->_options = _options->clone(clones_instances);
	return clone;
}

Traceur::Donnees::Assembleur::InstructionSP Triangle::draw(Traceur::Donnees::Assembleur::Assembleur& assembleur)
{
	PointSP expr_valeur1 = std::static_pointer_cast<Point>(_valeur1);
	Traceur::Donnees::Assembleur::Point sommet1(expr_valeur1,assembleur._pile.top()->_systeme_coordonnees);
	Traceur::Donnees::Assembleur::Point sommet2;
	Traceur::Donnees::Assembleur::Point sommet3;
	if(_valeur2->est_de_nature<Point>() and _valeur3->est_de_nature<Point>())
	{
		PointSP expr_valeur2 = std::static_pointer_cast<Point>(_valeur2);
		sommet2=Traceur::Donnees::Assembleur::Point(expr_valeur2,assembleur._pile.top()->_systeme_coordonnees);
		PointSP expr_valeur3 = std::static_pointer_cast<Point>(_valeur3);
		sommet3=Traceur::Donnees::Assembleur::Point(expr_valeur3,assembleur._pile.top()->_systeme_coordonnees);
	}
	else if(_valeur2->est_de_nature<Vecteur>() and _valeur3->est_de_nature<Vecteur>())
	{
		VecteurSP expr_valeur2 = std::dynamic_pointer_cast<Vecteur>(_valeur2);
		PointSP point2(new Point(expr_valeur1->translation(*expr_valeur2)));
		sommet2=Traceur::Donnees::Assembleur::Point(point2,assembleur._pile.top()->_systeme_coordonnees);
		VecteurSP expr_valeur3 = std::dynamic_pointer_cast<Vecteur>(_valeur3);
		PointSP point3(new Point(expr_valeur1->translation(*expr_valeur3)));
		sommet3=Traceur::Donnees::Assembleur::Point(point3,assembleur._pile.top()->_systeme_coordonnees);
	}
	else { throw; }
	Traceur::Donnees::Assembleur::InstructionSP drawtriangle(new Traceur::Donnees::Assembleur::DrawTriangle(sommet1,sommet2,sommet3,*_options));
	return drawtriangle;
}
