#include <src/donnees/expressions/expressionvaleurabsolue.h>

#include <src/donnees/types/types.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionreel.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteur.h>
#include <src/donnees/expressions/expressionchaine.h>
#include <src/donnees/expressions/expressioncouleurrvba.h>
#include <src/donnees/expressions/expressioncouleurhsva.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_VALEURABSOLUE(type) \
	if(valeur->est_de_nature<type>()) { _evaluateur = ValeurAbsolueSP(new EvaluateurOperateurUnaire<ValeurAbsolue,type>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurUnaire<ValeurAbsolue,Entier> : public ValeurAbsolue
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							EntierSP valeur_valeur=std::static_pointer_cast<Entier>(expr_valeur);
							SECURISE_VALEUR(EntierSP(new Entier(Entier::valeur_absolue(*valeur_valeur))));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ValeurAbsolue,Reel> : public ValeurAbsolue
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ReelSP valeur_valeur=std::static_pointer_cast<Reel>(expr_valeur);
							SECURISE_VALEUR(ReelSP(new Reel(Reel::valeur_absolue(*valeur_valeur))));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ValeurAbsolue,Point> : public ValeurAbsolue
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							PointSP valeur_valeur=std::static_pointer_cast<Point>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->module());
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ValeurAbsolue,Vecteur> : public ValeurAbsolue
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							VecteurSP valeur_valeur=std::static_pointer_cast<Vecteur>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->module());
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ValeurAbsolue,Chaine> : public ValeurAbsolue
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							ChaineSP valeur_valeur=std::static_pointer_cast<Chaine>(expr_valeur);
							SECURISE_VALEUR(EntierSP(new Entier(valeur_valeur->longueur())));
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ValeurAbsolue,CouleurRVBA> : public ValeurAbsolue
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							CouleurRVBASP valeur_valeur=std::static_pointer_cast<CouleurRVBA>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->gris_equivalent());
						}
				};
			template<>
				class EvaluateurOperateurUnaire<ValeurAbsolue,CouleurHSVA> : public ValeurAbsolue
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP valeur) const
						{
							ExpressionSP expr_valeur(*valeur);
							CouleurHSVASP valeur_valeur=std::static_pointer_cast<CouleurHSVA>(expr_valeur);
							SECURISE_VALEUR(valeur_valeur->gris_equivalent());
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	ValeurAbsolue::ValeurAbsolue(ExpressionSP valeur, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneUnaire(valeur), _evaluateur(ValeurAbsolueSP(new EvaluateurOperateurUnaire<ValeurAbsolue,Expression>()))
{

	EVALUATEUR_VALEURABSOLUE(Entier     );
	EVALUATEUR_VALEURABSOLUE(Reel       );
	EVALUATEUR_VALEURABSOLUE(Point      );
	EVALUATEUR_VALEURABSOLUE(Vecteur    );
	EVALUATEUR_VALEURABSOLUE(Chaine     );
	EVALUATEUR_VALEURABSOLUE(CouleurRVBA);
	EVALUATEUR_VALEURABSOLUE(CouleurHSVA);
	// calcul du type de retour
	if(_type)
		return;
	if(valeur->est_de_type<Traceur::Donnees::Type::Integer>())
	{
		_type = Traceur::Donnees::Type::Integer::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Real>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Point>())
	{
		_type = Traceur::Donnees::Type::Real::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Vector>())
	{
		_type = Traceur::Donnees::Type::Number::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::String>())
	{
		_type = Traceur::Donnees::Type::Integer::SP();
		return;
	}
	if(valeur->est_de_type<Traceur::Donnees::Type::Color>())
	{
		_type = Traceur::Donnees::Type::Color::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionAutour(position,"|",valeur,"|");
}

ValeurSP ValeurAbsolue::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_valeur = _valeur->evaluation(assembleur); if(not valeur_valeur->a_une_valeur_retour()) return valeur_valeur;
	ValeurAbsolue resultat(*valeur_valeur);
	return resultat.evaluation(valeur_valeur);
}

ValeurSP ValeurAbsolue::evaluation(const ValeurSP valeur) const
{
	return _evaluateur->evaluation(valeur);
}

void ValeurAbsolue::format(std::ostream& os) const
{
	os << "|";
	_valeur->format(os);
	os << "|";
}

ExpressionSP ValeurAbsolue::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP valeur = _valeur->clone(clones_instances);
	ValeurAbsolueSP clone(new ValeurAbsolue(valeur));
	return clone;
}
