#include <src/donnees/expressions/expressionvecteur.h>

#include <src/donnees/types/types.h>

#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressionnombre.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteurpolaire.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_VECTEUR(type1,type2) \
	if(gauche->est_de_nature<type1>() and droite->est_de_nature<type2>()) { _evaluateur = VecteurSP(new EvaluateurOperateurBinaire<type1,Vecteur,type2>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurBinaire<Nombre,Vecteur,Nombre> : public Vecteur
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							NombreSP valeur_gauche=std::static_pointer_cast<Nombre>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							NombreSP valeur_droite=std::static_pointer_cast<Nombre>(expr_droite);
							SECURISE_VALEUR(VecteurSP(new Vecteur(valeur_gauche,valeur_droite)));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	Vecteur::Vecteur(ExpressionSP gauche, ExpressionSP droite, Traceur::Donnees::Source::PositionSP position)
:AppelFonctionInterneBinaire(gauche,droite), _evaluateur(VecteurSP(new EvaluateurOperateurBinaire<Expression,Vecteur,Expression>()))
{

	EVALUATEUR_VECTEUR(Nombre,Nombre);
	// calcul du type de retour
	if(_type)
		return;
	if(gauche->est_de_type<Traceur::Donnees::Type::Number>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Vector::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionDeux(position,"VECTOR",gauche,droite);
}

ValeurSP Vecteur::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_gauche = _gauche->evaluation(assembleur); if(not valeur_gauche->a_une_valeur_retour()) return valeur_gauche;
	ValeurSP valeur_droite = _droite->evaluation(assembleur); if(not valeur_droite->a_une_valeur_retour()) return valeur_droite;
	Vecteur resultat(*valeur_gauche,*valeur_droite);
	return resultat.evaluation(valeur_gauche,valeur_droite);
}

ValeurSP Vecteur::evaluation(const ValeurSP gauche, const ValeurSP droite) const
{
	return _evaluateur->evaluation(gauche,droite);
}

void Vecteur::format(std::ostream& os) const
{
	//_type->format(os);
	os << "VECTOR[";
	_gauche->format(os);
	os << ";";
	_droite->format(os);
	os << "]";
}

ExpressionSP Vecteur::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP gauche = _gauche->clone(clones_instances);
	ExpressionSP droite = _droite->clone(clones_instances);
	VecteurSP clone(new Vecteur());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_gauche=gauche;
	clone->_droite=droite;
	return clone;
}

Vecteur::Vecteur(const VecteurPolaire& vecteur_polaire)
{
	NombreSP module=std::dynamic_pointer_cast<Nombre>(vecteur_polaire._gauche);
	TRACEUR_TEST(module);
	NombreSP argument=std::dynamic_pointer_cast<Nombre>(vecteur_polaire._droite);
	TRACEUR_TEST(argument);
	_type = Traceur::Donnees::Type::Vector::SP();
	_gauche = module*cosinus(argument);
	_droite = module*sinus(argument);
}

Vecteur::Vecteur(const Point& point)
{
	NombreSP abscisse=std::dynamic_pointer_cast<Nombre>(point._gauche);
	TRACEUR_TEST(abscisse);
	NombreSP ordonnee=std::dynamic_pointer_cast<Nombre>(point._droite);
	TRACEUR_TEST(ordonnee);
	_type = Traceur::Donnees::Type::Vector::SP();
	_gauche = abscisse;
	_droite = ordonnee;
}

void Vecteur::en_cartesien(NombreSP& abscisse, NombreSP& ordonnee) const
{
	Vecteur vecteur(this->en_cartesien());
	abscisse = std::dynamic_pointer_cast<Nombre>(vecteur._gauche);
	TRACEUR_TEST(abscisse);
	ordonnee = std::dynamic_pointer_cast<Nombre>(vecteur._droite);
	TRACEUR_TEST(ordonnee);
}

void Vecteur::en_polaire(NombreSP& module, NombreSP& argument) const
{
	VecteurPolaire vecteur(this->en_polaire());
	module = std::dynamic_pointer_cast<Nombre>(vecteur._gauche);
	TRACEUR_TEST(module);
	argument = std::dynamic_pointer_cast<Nombre>(vecteur._droite);
	TRACEUR_TEST(argument);
}

Vecteur Vecteur::operator+(Vecteur& vecteur) const
{
	NombreSP abscisse_gauche;
	NombreSP ordonnee_gauche;
	this->en_cartesien(abscisse_gauche,ordonnee_gauche);
	NombreSP abscisse_droite;
	NombreSP ordonnee_droite;
	vecteur.en_cartesien(abscisse_droite,ordonnee_droite);
	Vecteur somme(abscisse_gauche+abscisse_droite,ordonnee_gauche+ordonnee_droite);
	return somme;
}

Vecteur Vecteur::operator-(Vecteur& vecteur) const
{
	NombreSP abscisse_gauche;
	NombreSP ordonnee_gauche;
	this->en_cartesien(abscisse_gauche,ordonnee_gauche);
	NombreSP abscisse_droite;
	NombreSP ordonnee_droite;
	vecteur.en_cartesien(abscisse_droite,ordonnee_droite);
	Vecteur difference(abscisse_gauche-abscisse_droite,ordonnee_gauche-ordonnee_droite);
	return difference;
}

Vecteur Vecteur::operator-() const
{
	NombreSP abscisse_gauche;
	NombreSP ordonnee_gauche;
	this->en_cartesien(abscisse_gauche,ordonnee_gauche);
	Vecteur oppose(-abscisse_gauche,-ordonnee_gauche);
	return oppose;
}

NombreSP Vecteur::operator*(Vecteur& vecteur) const
{
	NombreSP abscisse_gauche;
	NombreSP ordonnee_gauche;
	this->en_cartesien(abscisse_gauche,ordonnee_gauche);
	NombreSP abscisse_droite;
	NombreSP ordonnee_droite;
	vecteur.en_cartesien(abscisse_droite,ordonnee_droite);
	NombreSP scalaire = (abscisse_gauche*abscisse_droite)+(ordonnee_gauche*ordonnee_droite);
	return scalaire;
}

NombreSP Vecteur::operator^(Vecteur& vecteur) const
{
	NombreSP abscisse_gauche;
	NombreSP ordonnee_gauche;
	this->en_cartesien(abscisse_gauche,ordonnee_gauche);
	NombreSP abscisse_droite;
	NombreSP ordonnee_droite;
	vecteur.en_cartesien(abscisse_droite,ordonnee_droite);
	NombreSP normale_produit_vectoriel = (abscisse_gauche*ordonnee_droite)-(abscisse_droite*ordonnee_gauche);
	return normale_produit_vectoriel;
}

Vecteur Vecteur::projection(Vecteur& vecteur) const
{
	NombreSP abscisse_gauche;
	NombreSP ordonnee_gauche;
	this->en_cartesien(abscisse_gauche,ordonnee_gauche);
	NombreSP abscisse_droite;
	NombreSP ordonnee_droite;
	vecteur.en_cartesien(abscisse_droite,ordonnee_droite);
	NombreSP facteur = abscisse_gauche*abscisse_droite+ordonnee_gauche*ordonnee_droite;
	NombreSP carre_norme = abscisse_droite*abscisse_droite+ordonnee_droite*ordonnee_droite;
	Vecteur projete((facteur*abscisse_droite)/carre_norme,(facteur*ordonnee_droite)/carre_norme);
	return projete;
}

Vecteur Vecteur::symetrie(Vecteur& vecteur) const
{
	NombreSP abscisse_gauche;
	NombreSP ordonnee_gauche;
	this->en_cartesien(abscisse_gauche,ordonnee_gauche);
	NombreSP abscisse_droite;
	NombreSP ordonnee_droite;
	vecteur.en_cartesien(abscisse_droite,ordonnee_droite);
	NombreSP carre_norme = (abscisse_gauche*abscisse_gauche+ordonnee_gauche*ordonnee_gauche)*(abscisse_droite*abscisse_droite+ordonnee_droite*ordonnee_droite);
	NombreSP facteur = Nombre::deux()*(abscisse_gauche*abscisse_droite+ordonnee_gauche*ordonnee_droite);
	NombreSP cossym = ((facteur*(abscisse_gauche*abscisse_droite+ordonnee_gauche*ordonnee_droite))/carre_norme)-Nombre::un();
	NombreSP sinsym = (facteur*(abscisse_gauche*ordonnee_droite-abscisse_droite*ordonnee_gauche))/carre_norme;
	Vecteur symetrique(cossym*abscisse_gauche-sinsym*ordonnee_gauche,sinsym*abscisse_gauche+cossym*ordonnee_gauche);
	return symetrique;
}

Vecteur Vecteur::rotation(NombreSP angle) const
{
	NombreSP abscisse_gauche;
	NombreSP ordonnee_gauche;
	this->en_cartesien(abscisse_gauche,ordonnee_gauche);
	NombreSP cosangle = cosinus(angle);
	NombreSP sinangle = sinus(angle);
	Vecteur tourne(cosangle*abscisse_gauche-sinangle*ordonnee_gauche,sinangle*abscisse_gauche+cosangle*ordonnee_gauche);
	return tourne;
}

Vecteur Vecteur::operator*(NombreSP nombre) const
{
	NombreSP abscisse_gauche;
	NombreSP ordonnee_gauche;
	this->en_cartesien(abscisse_gauche,ordonnee_gauche);
	Vecteur produit(abscisse_gauche*nombre,ordonnee_gauche*nombre);
	return produit;
}

Vecteur Vecteur::operator/(NombreSP nombre) const
{
	NombreSP abscisse_gauche;
	NombreSP ordonnee_gauche;
	this->en_cartesien(abscisse_gauche,ordonnee_gauche);
	Vecteur produit(abscisse_gauche/nombre,ordonnee_gauche/nombre);
	return produit;
}

VecteurPolaire Vecteur::en_polaire() const
{
	return VecteurPolaire(*this);
}

bool Vecteur::est_le_vecteur_nul() const
{
	NombreSP abscisse_gauche;
	NombreSP ordonnee_gauche;
	this->en_cartesien(abscisse_gauche,ordonnee_gauche);
	return est_nul(abscisse_gauche) and est_nul(ordonnee_gauche);
}

bool Vecteur::operator==(Vecteur& vecteur) const
{
	NombreSP abscisse_gauche;
	NombreSP ordonnee_gauche;
	this->en_cartesien(abscisse_gauche,ordonnee_gauche);
	NombreSP abscisse_droite;
	NombreSP ordonnee_droite;
	vecteur.en_cartesien(abscisse_droite,ordonnee_droite);
	return (abscisse_gauche==abscisse_droite) and (ordonnee_gauche==ordonnee_droite);
}

bool Vecteur::operator!=(Vecteur& vecteur) const
{
	NombreSP abscisse_gauche;
	NombreSP ordonnee_gauche;
	this->en_cartesien(abscisse_gauche,ordonnee_gauche);
	NombreSP abscisse_droite;
	NombreSP ordonnee_droite;
	vecteur.en_cartesien(abscisse_droite,ordonnee_droite);
	return (abscisse_gauche!=abscisse_droite) or (ordonnee_gauche!=ordonnee_droite);
}

NombreSP Vecteur::abscisse() const
{
	NombreSP abscisse;
	NombreSP ordonnee;
	this->en_cartesien(abscisse,ordonnee);
	return abscisse;
}

NombreSP Vecteur::ordonnee() const
{
	NombreSP abscisse;
	NombreSP ordonnee;
	this->en_cartesien(abscisse,ordonnee);
	return ordonnee;
}

NombreSP Vecteur::module() const
{
	NombreSP module;
	NombreSP argument;
	this->en_polaire(module,argument);
	return module;
}

NombreSP Vecteur::argument() const
{
	NombreSP module;
	NombreSP argument;
	this->en_polaire(module,argument);
	return argument;
}
