#include <src/donnees/expressions/expressionvecteurpolaire.h>

#include <src/donnees/types/types.h>

#include <src/donnees/expressions/expressionnombre.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteur.h>

#include <src/donnees/expressions/expressionevaluateur.h>

#define EVALUATEUR_VECTEURPOLAIRE(type1,type2) \
	if(gauche->est_de_nature<type1>() and droite->est_de_nature<type2>()) { _evaluateur = VecteurPolaireSP(new EvaluateurOperateurBinaire<type1,VecteurPolaire,type2>()); }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			using namespace Traceur::Donnees::Instruction; // a titre exceptionnel ici
			template<>
				class EvaluateurOperateurBinaire<Nombre,VecteurPolaire,Nombre> : public VecteurPolaire
				{
					public:
						virtual ValeurSP evaluation(const ValeurSP gauche, const ValeurSP droite) const
						{
							ExpressionSP expr_gauche(*gauche);
							NombreSP valeur_gauche=std::static_pointer_cast<Nombre>(expr_gauche);
							ExpressionSP expr_droite(*droite);
							NombreSP valeur_droite=std::static_pointer_cast<Nombre>(expr_droite);
							SECURISE_VALEUR(VecteurPolaireSP(new VecteurPolaire(valeur_gauche,valeur_droite)));
						}
				};
		}
	}
}
using namespace Traceur::Donnees::Expression;
	VecteurPolaire::VecteurPolaire(ExpressionSP gauche, ExpressionSP droite, Traceur::Donnees::Source::PositionSP position)
:Vecteur(gauche,droite,VecteurSP(new EvaluateurOperateurBinaire<Expression,VecteurPolaire,Expression>()))
{

	EVALUATEUR_VECTEURPOLAIRE(Nombre,Nombre);
	// calcul du type de retour
	if(_type)
		return;
	if(gauche->est_de_type<Traceur::Donnees::Type::Number>() and droite->est_de_type<Traceur::Donnees::Type::Number>())
	{
		_type = Traceur::Donnees::Type::Vector::SP();
		return;
	}
	throw Traceur::Exception::TypeIncorrectAppelFonctionConversionDeux(position,"POLAR",gauche,droite);
}

VecteurPolaire::VecteurPolaire(const Vecteur& vecteur)
{
	NombreSP abscisse=std::dynamic_pointer_cast<Nombre>(vecteur._gauche);
	TRACEUR_TEST(abscisse);
	NombreSP ordonnee=std::dynamic_pointer_cast<Nombre>(vecteur._droite);
	TRACEUR_TEST(ordonnee);
	_type = Traceur::Donnees::Type::Vector::SP();
	_gauche = racine_carree(abscisse*abscisse+ordonnee*ordonnee);
	_droite = Traceur::Donnees::Expression::argument(abscisse,ordonnee);
}

ValeurSP VecteurPolaire::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_gauche = _gauche->evaluation(assembleur); if(not valeur_gauche->a_une_valeur_retour()) return valeur_gauche;
	ValeurSP valeur_droite = _droite->evaluation(assembleur); if(not valeur_droite->a_une_valeur_retour()) return valeur_droite;
	VecteurPolaire resultat(*valeur_gauche,*valeur_droite);
	return resultat.evaluation(valeur_gauche,valeur_droite);
}

ValeurSP VecteurPolaire::evaluation(const ValeurSP gauche, const ValeurSP droite) const
{
	return _evaluateur->evaluation(gauche,droite);
}

void VecteurPolaire::format(std::ostream& os) const
{
	//_type->format(os);
	os << "POLAR[";
	_gauche->format(os);
	os << ";";
	_droite->format(os);
	os << "]";
}

ExpressionSP VecteurPolaire::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	ExpressionSP gauche = _gauche->clone(clones_instances);
	ExpressionSP droite = _droite->clone(clones_instances);
	VecteurPolaireSP clone(new VecteurPolaire());
	Traceur::Donnees::Type::TypeSP type = _type->clone(clones_instances);
	clone->_type=type;
	clone->_gauche=gauche;
	clone->_droite=droite;
	return clone;
}

VecteurPolaire VecteurPolaire::operator-() const
{
	NombreSP module = std::dynamic_pointer_cast<Nombre>(_gauche);
	TRACEUR_TEST(module);
	VecteurPolaire oppose(-module,_droite);
	return oppose;
}

VecteurPolaire VecteurPolaire::operator*(NombreSP nombre) const
{
	NombreSP module = std::dynamic_pointer_cast<Nombre>(_gauche);
	TRACEUR_TEST(module);
	VecteurPolaire produit(module*nombre,_droite);
	return produit;
}

VecteurPolaire VecteurPolaire::operator/(NombreSP nombre) const
{
	NombreSP module = std::dynamic_pointer_cast<Nombre>(_gauche);
	TRACEUR_TEST(module);
	VecteurPolaire produit(module/nombre,_droite);
	return produit;
}

bool VecteurPolaire::est_le_vecteur_nul() const
{
	NombreSP module = std::dynamic_pointer_cast<Nombre>(_gauche);
	TRACEUR_TEST(module);
	return est_nul(module);
}

VecteurPolaire VecteurPolaire::rotation(NombreSP& angle) const
{
	NombreSP argument = std::dynamic_pointer_cast<Nombre>(_droite);
	TRACEUR_TEST(argument);
	VecteurPolaire tourne(_gauche,argument+angle);
	return tourne;
}
