#ifndef _DONNEES_EXPRESSIONS_OPTIONSLIGNE_H_
#define _DONNEES_EXPRESSIONS_OPTIONSLIGNE_H_

#include <iostream>

#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Expression { DECL_SHARED_PTR(OptionsLigne); } } }
#include <src/exceptions/exceptions.h>
#include <src/donnees/sources/position.h>
#include <src/donnees/expressions/expression.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/types/typeinteger.h>
#include <src/donnees/assembleur/optionsligne.h>

namespace Traceur
{
	namespace Exception
	{
		class OptionLigneIncorrecte : public Compilation
		{
			public:
				OptionLigneIncorrecte(Traceur::Donnees::Source::PositionSP& position, std::string nom_option)
					:Compilation(position,std::string("Incompatible line option ")+nom_option+std::string(".")) {};
				virtual ~OptionLigneIncorrecte() {};
		};
		class TypeIncorrectTirets : public Compilation
		{
			public:
				TypeIncorrectTirets(Traceur::Donnees::Source::PositionSP& position, Traceur::Donnees::Expression::ExpressionSP gauche, Traceur::Donnees::Expression::ExpressionSP droite)
					:Compilation(position,std::string("Invalid parameter type in DASHED[")+std::string(*(gauche->type()))+(droite?std::string(";")+std::string(*(droite->type())):std::string(""))+("] line option.")) {};
				virtual ~TypeIncorrectTirets() {};
		};
	}
}

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			class OptionsLigne : public std::enable_shared_from_this<OptionsLigne>
			{
				friend struct OptionLigneSansDebut;
				friend struct OptionLigneSansFin;
				friend struct OptionLigneSolide;
				friend struct OptionLigneGomme;
				friend struct OptionLigneTirets;
				friend struct OptionLignePointilles;
				public:
					OptionsLigne()
						:_sans_debut(false), _sans_fin(false), _solide(false), _gomme(false), _discontinue(false) {};
					virtual ~OptionsLigne() {};
					void format(std::ostream& os)
					{
						if(_sans_debut or _sans_fin or _solide or _gomme or _discontinue)
							os << ";";
						if(_sans_debut)
						{
							os << "NOSTART";
							if(_sans_fin or _solide or _gomme or _discontinue)
								os << ",";
						}
						if(_sans_fin)
						{
							os << "NOEND";
							if(_solide or _gomme or _discontinue)
								os << ",";
						}
						if(_solide)
						{
							os << "SOLID";
							if(_gomme or _discontinue)
								os << ",";
						}
						if(_gomme)
						{
							os << "ERASER";
							if(_discontinue)
								os << ",";
						}
						if(_discontinue)
						{
							os << "DASHED[";
							_ecart_trace->format(os);
							if(_ecart_non_trace)
							{
								os << ";";
								_ecart_non_trace->format(os);
							}
							os << "]";
						}
					};
					void edition_liens(Traceur::Donnees::Instruction::InstructionSP instruction)
					{
						if(_discontinue)
						{
							_ecart_trace->edition_liens(instruction);
							if(_ecart_non_trace)
								_ecart_non_trace->edition_liens(instruction);
						}
					};
					void definition_fonction()
					{
						if(_discontinue)
						{
							_ecart_trace->definition_fonction();
							if(_ecart_non_trace)
								_ecart_non_trace->definition_fonction();
						}
					};
					virtual OptionsLigneSP simplification()
					{
						if(_discontinue)
						{
							_ecart_trace->simplification();
							if(_ecart_non_trace)
								_ecart_non_trace->simplification();
						}
						return this->shared_from_this();
					};
					OptionsLigneSP clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
					{
						return this->shared_from_this();
					};
					OptionsLigneSP evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
					{
						Traceur::Donnees::Instruction::ValeurSP ecart_trace;
						Traceur::Donnees::Instruction::ValeurSP ecart_non_trace;
						if(_discontinue)
						{
							ecart_trace = _ecart_trace->evaluation(assembleur); if(not ecart_trace->a_une_valeur_retour()) throw ecart_trace;
							if(_ecart_non_trace)
							{
								ecart_non_trace = _ecart_non_trace->evaluation(assembleur); if(not ecart_non_trace->a_une_valeur_retour()) throw ecart_non_trace;
							}
						}
						ExpressionSP expr_ecart_trace;
						ExpressionSP expr_ecart_non_trace;
						if(ecart_trace)
							expr_ecart_trace = *ecart_trace;
						if(ecart_non_trace)
							expr_ecart_non_trace = *ecart_non_trace;
						OptionsLigneSP options(new OptionsLigne(_sans_debut,_sans_fin,_solide,_gomme,_discontinue,expr_ecart_trace,expr_ecart_non_trace));
						return options;
					};
					operator Traceur::Donnees::Assembleur::OptionsLigne ()
					{
						Traceur::Donnees::Assembleur::Coordonnee ecart_trace = 0;
						Traceur::Donnees::Assembleur::Coordonnee ecart_non_trace = 0;
						if(_discontinue)
						{
							NombreSP expr_ecart_trace = std::dynamic_pointer_cast<Nombre>(_ecart_trace);
							ecart_trace = expr_ecart_trace->entier()._entier;
							if(ecart_trace<1)
								throw Traceur::Donnees::Instruction::Valeur::hors_domaine();
							if(_ecart_non_trace)
							{
								NombreSP expr_ecart_non_trace = std::dynamic_pointer_cast<Nombre>(_ecart_non_trace);
								ecart_non_trace = expr_ecart_non_trace->entier()._entier;
								if(ecart_non_trace<1)
									throw Traceur::Donnees::Instruction::Valeur::hors_domaine();
							}
						}
						return Traceur::Donnees::Assembleur::OptionsLigne(_sans_debut,_sans_fin,_solide,_gomme,_discontinue,ecart_trace,ecart_non_trace);
					};
				private:
					OptionsLigne(bool sans_debut, bool sans_fin, bool solide, bool gomme, bool discontinue, ExpressionSP ecart_trace, ExpressionSP ecart_non_trace)
						:_sans_debut(sans_debut), _sans_fin(sans_fin), _solide(solide), _gomme(gomme),
						 _discontinue(discontinue), _ecart_trace(ecart_trace), _ecart_non_trace(ecart_non_trace) {};
					bool _sans_debut;
					bool _sans_fin;
					bool _solide;
					bool _gomme;
					bool _discontinue;
					ExpressionSP _ecart_trace;
					ExpressionSP _ecart_non_trace;
			};
			struct OptionLigneSimple
			{
				virtual void option(OptionsLigne *options, Traceur::Donnees::Source::PositionSP& position) = 0;
				virtual ~OptionLigneSimple() {};
			};

			struct OptionLigneSansDebut : public OptionLigneSimple
			{
				virtual void option(OptionsLigne *options, Traceur::Donnees::Source::PositionSP& position)
				{
					if(options->_sans_debut)
						throw Traceur::Exception::OptionLigneIncorrecte(position,"NOSTART");
					options->_sans_debut=true;
				};
				virtual ~OptionLigneSansDebut() {};
			};

			struct OptionLigneSansFin : public OptionLigneSimple
			{
				virtual void option(OptionsLigne *options, Traceur::Donnees::Source::PositionSP& position)
				{
					if(options->_sans_fin)
						throw Traceur::Exception::OptionLigneIncorrecte(position,"NOEND");
					options->_sans_fin=true;
				};
				virtual ~OptionLigneSansFin() {};
			};

			struct OptionLigneSolide : public OptionLigneSimple
			{
				virtual void option(OptionsLigne *options, Traceur::Donnees::Source::PositionSP& position)
				{
					if(options->_solide or options->_gomme)
						throw Traceur::Exception::OptionLigneIncorrecte(position,"SOLID");
					options->_solide=true;
				};
				virtual ~OptionLigneSolide() {};
			};

			struct OptionLigneGomme : public OptionLigneSimple
			{
				virtual void option(OptionsLigne *options, Traceur::Donnees::Source::PositionSP& position)
				{
					if(options->_solide or options->_gomme)
						throw Traceur::Exception::OptionLigneIncorrecte(position,"ERASER");
					options->_gomme=true;
				};
				virtual ~OptionLigneGomme() {};
			};

			struct OptionLigneTirets : public OptionLigneSimple
			{
				OptionLigneTirets(ExpressionSP ecart)
					:_ecartTrace(ecart) { };
				OptionLigneTirets(ExpressionSP ecart_trace, ExpressionSP ecart_non_trace)
					:_ecartTrace(ecart_trace),_ecartNonTrace(ecart_non_trace) { };
				ExpressionSP _ecartTrace;
				ExpressionSP _ecartNonTrace;

				virtual ~OptionLigneTirets() {};
				virtual void option(OptionsLigne *options, Traceur::Donnees::Source::PositionSP& position)
				{
					if(options->_discontinue)
					{
						if(_ecartNonTrace)
							throw Traceur::Exception::OptionLigneIncorrecte(position,"DASHED[...;...]");
						throw Traceur::Exception::OptionLigneIncorrecte(position,"DASHED[...]");
					}
					if(not _ecartTrace->est_de_type<Traceur::Donnees::Type::Integer>())
						throw Traceur::Exception::TypeIncorrectTirets(position,_ecartTrace,_ecartNonTrace);
					if(_ecartNonTrace and (not _ecartNonTrace->est_de_type<Traceur::Donnees::Type::Integer>()))
						throw Traceur::Exception::TypeIncorrectTirets(position,_ecartTrace,_ecartNonTrace);
					ecarts(options);
				};
				void ecarts(OptionsLigne *options)
				{
					options->_discontinue=true;
					options->_ecart_trace = _ecartTrace;
					if(_ecartNonTrace)
					{
						options->_ecart_non_trace = _ecartNonTrace;
					}
					else
					{
						options->_ecart_non_trace = _ecartTrace;
					}
				};
			};

			struct OptionLignePointilles : public OptionLigneTirets
			{
				OptionLignePointilles()
					:OptionLigneTirets(ExpressionSP(new Entier(1))) {};
				virtual ~OptionLignePointilles() {};
				virtual void option(OptionsLigne *options, Traceur::Donnees::Source::PositionSP& position)
				{
					if(options->_discontinue)
						throw Traceur::Exception::OptionLigneIncorrecte(position,"DOTTED");
					ecarts(options);
				};
			};

		}
	}
}
#endif
