#ifndef _DONNEES_EXPRESSIONS_OPTIONSSURFACE_H_
#define _DONNEES_EXPRESSIONS_OPTIONSSURFACE_H_

#include <iostream>

#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Expression { DECL_SHARED_PTR(OptionsSurface); } } }
#include <src/exceptions/exceptions.h>
#include <src/donnees/sources/position.h>
#include <src/donnees/expressions/expression.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/types/typeinteger.h>
#include <src/donnees/assembleur/optionssurface.h>

namespace Traceur
{
	namespace Exception
	{
		class OptionSurfaceIncorrecte : public Compilation
		{
			public:
				OptionSurfaceIncorrecte(Traceur::Donnees::Source::PositionSP& position, std::string nom_option)
					:Compilation(position,std::string("Incompatible surface option ")+nom_option+std::string(".")) {};
				virtual ~OptionSurfaceIncorrecte() {};
		};
	}
}

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression
		{
			class OptionsSurface : public std::enable_shared_from_this<OptionsSurface>
			{
				friend struct OptionSurfaceRemplissage;
				friend struct OptionSurfaceSolide;
				friend struct OptionSurfaceGomme;
				public:
					OptionsSurface()
						:_remplissage(false), _solide(false), _gomme(false) {};
					virtual ~OptionsSurface() {};
					void format(std::ostream& os)
					{
						if(_remplissage or _solide or _gomme)
							os << ";";
						if(_remplissage)
						{
							os << "FILL";
							if(_solide or _gomme)
								os << ",";
						}
						if(_solide)
						{
							os << "SOLID";
							if(_gomme)
								os << ",";
						}
						if(_gomme)
							os << "ERASER";
					};
					OptionsSurfaceSP clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
					{
						return this->shared_from_this();
					};
					OptionsSurfaceSP evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
					{
						return this->shared_from_this();
					};
					operator Traceur::Donnees::Assembleur::OptionsSurface ()
					{
						return Traceur::Donnees::Assembleur::OptionsSurface(_remplissage,_solide,_gomme);
					};
				private:
					bool _remplissage;
					bool _solide;
					bool _gomme;
			};
			struct OptionSurfaceSimple
			{
				virtual void option(OptionsSurface *options, Traceur::Donnees::Source::PositionSP& position) = 0;
				virtual ~OptionSurfaceSimple() {};
			};

			struct OptionSurfaceRemplissage : public OptionSurfaceSimple
			{
				virtual void option(OptionsSurface *options, Traceur::Donnees::Source::PositionSP& position)
				{
					if(options->_remplissage)
						throw Traceur::Exception::OptionSurfaceIncorrecte(position,"FILL");
					options->_remplissage=true;
				};
				virtual ~OptionSurfaceRemplissage() {};
			};

			struct OptionSurfaceSolide : public OptionSurfaceSimple
			{
				virtual void option(OptionsSurface *options, Traceur::Donnees::Source::PositionSP& position)
				{
					if(options->_solide or options->_gomme)
						throw Traceur::Exception::OptionSurfaceIncorrecte(position,"SOLID");
					options->_solide=true;
				};
				virtual ~OptionSurfaceSolide() {};
			};

			struct OptionSurfaceGomme : public OptionSurfaceSimple
			{
				virtual void option(OptionsSurface *options, Traceur::Donnees::Source::PositionSP& position)
				{
					if(options->_solide or options->_gomme)
						throw Traceur::Exception::OptionSurfaceIncorrecte(position,"ERASER");
					options->_gomme=true;
				};
				virtual ~OptionSurfaceGomme() {};
			};

		}
	}
}
#endif
