#ifndef _DONNEES_EXPRESSIONS_SYSTEME_COORDONNEES_H_
#define _DONNEES_EXPRESSIONS_SYSTEME_COORDONNEES_H_

#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Expression { DECL_SHARED_PTR(Image); } } }

#include <src/donnees/instructions/valeur.h>
#include <src/donnees/expressions/expression.h>
#include <src/donnees/instructions/bloc.h>
#include <src/donnees/types/typenumber.h>
#include <src/donnees/types/typepoint.h>
#include <src/donnees/expressions/expressionappelfonction.h>
#include <src/donnees/expressions/expressionpoint.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Assembleur { class SystemeCoordonnees; }
		namespace Expression
		{
			DECL_SHARED_PTR(SystemeCoordonnees);
			class SystemeCoordonnees
			{
				public:
					SystemeCoordonnees() = default;
					virtual ~SystemeCoordonnees() {};
					virtual operator std::string () = 0;
					virtual void format(std::ostream& os) const = 0;
					virtual SystemeCoordonneesSP clone(Traceur::Donnees::Global::ClonesInstances& clones_instances) = 0;
					virtual void definition_fonction() { };
					virtual void edition_liens(const Traceur::Donnees::Instruction::InstructionSP instruction) { };
					virtual void simplification() {};
					virtual SystemeCoordonneesSP evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur) = 0;
			};

			class SystemeGeometrique : public SystemeCoordonnees
			{
				public:
					virtual ~SystemeGeometrique() {};
					virtual operator std::string() { return "GEOMETRY"; };
					virtual void format(std::ostream& os) const { os << "GEOMETRY" ; }
					virtual SystemeCoordonneesSP clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
					{
						SystemeCoordonneesSP clone(new SystemeGeometrique());
						return clone;
					};
					virtual SystemeCoordonneesSP evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
					{
					//	Traceur::Donnees::Instruction::ValeurSP xsiz = _xsiz->evaluation(assembleur); if(not xsiz->a_une_valeur_retour()) throw xsiz;
						SystemeCoordonneesSP coordonnees(new SystemeGeometrique());
						return coordonnees;
					};
			};

			class SystemeStandard : public SystemeCoordonnees
			{
				public:
					virtual ~SystemeStandard() {};
					virtual operator std::string() { return "STANDARD"; };
					virtual void format(std::ostream& os) const { os << "STANDARD" ; }
					virtual SystemeCoordonneesSP clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
					{
						SystemeCoordonneesSP clone(new SystemeStandard());
						return clone;
					};
					virtual SystemeCoordonneesSP evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
					{
						SystemeCoordonneesSP coordonnees(new SystemeStandard());
						return coordonnees;
					};
			};

			DECL_SHARED_PTR(SystemeMathematique);
			class SystemeMathematique : public SystemeCoordonnees
			{
				friend class Traceur::Donnees::Assembleur::SystemeCoordonnees;
				public:
					SystemeMathematique(const ExpressionSP& point_min, const ExpressionSP& point_max,Traceur::Donnees::Source::PositionSP position = Traceur::Donnees::Source::PositionSP())
						:_point_min(point_min), _point_max(point_max)
					{
						if(point_min->est_de_type<Traceur::Donnees::Type::Point>() and point_max->est_de_type<Traceur::Donnees::Type::Point>())
						{
							return;
						}
						throw Traceur::Exception::TypeIncorrectAppelFonctionConversionDeux(position,"MATHEMATICAL",point_min,point_max);
					}
					SystemeMathematique(const ExpressionSP& xmin, const ExpressionSP& ymin, const ExpressionSP& xmax, const ExpressionSP& ymax,Traceur::Donnees::Source::PositionSP position = Traceur::Donnees::Source::PositionSP())
					{
						if(xmin->est_de_type<Traceur::Donnees::Type::Number>() and ymin->est_de_type<Traceur::Donnees::Type::Number>()
							and xmax->est_de_type<Traceur::Donnees::Type::Number>() and ymax->est_de_type<Traceur::Donnees::Type::Number>())
						{
							_point_min = ExpressionSP(new Point(xmin,ymin));
							_point_max = ExpressionSP(new Point(xmax,ymax));
							return;
						}
						throw Traceur::Exception::TypeIncorrectAppelFonctionConversionQuatre(position,"MATHEMATICAL",xmin,ymin,xmax,ymax);
					}
					virtual ~SystemeMathematique() {};
					virtual operator std::string()
					{
						return std::string("MATHEMATICAL[")+std::string(*(_point_min->type()))+std::string(";")+
							std::string(*(_point_max->type()))+std::string("]");
					};
					virtual void format(std::ostream& os) const
					{
						os << "MATHEMATICAL[" ;
						_point_min->format(os);
						os << ";" ;
						_point_max->format(os);
						os << "]" ;
					}
					virtual SystemeCoordonneesSP clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
					{
						ExpressionSP point_min = _point_min->clone(clones_instances);
						ExpressionSP point_max = _point_max->clone(clones_instances);
						SystemeCoordonneesSP clone(new SystemeMathematique(point_min,point_max));
						return clone;
					};
					virtual void definition_fonction()
					{
						_point_min->definition_fonction();
						_point_max->definition_fonction();
					};
					virtual void edition_liens(const Traceur::Donnees::Instruction::InstructionSP instruction)
					{
						_point_min->edition_liens(instruction);
						_point_max->edition_liens(instruction);
					};
					virtual SystemeCoordonneesSP evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
					{
						Traceur::Donnees::Instruction::ValeurSP point_min = _point_min->evaluation(assembleur); if(not point_min->a_une_valeur_retour()) throw point_min;
						Traceur::Donnees::Instruction::ValeurSP point_max = _point_max->evaluation(assembleur); if(not point_max->a_une_valeur_retour()) throw point_max;
						SystemeCoordonneesSP coordonnees(new SystemeMathematique(*point_min,*point_max));
						return coordonnees;
					};
					virtual void simplification()
					{
						_point_min->simplification();
						_point_max->simplification();
					};
				private:
					ExpressionSP _point_min;
					ExpressionSP _point_max;
			};
		}
	}
}
#endif
