#include <src/donnees/fonctions/declarations.h>
#include <src/donnees/fonctions/instancerecursive.h>
#include <src/donnees/fonctions/instancenormale.h>

#include <functional>

using namespace Traceur::Donnees::Fonction;

namespace Traceur
{
	namespace Donnees
	{
		namespace Fonction
		{
			struct AfficheDeclaration
			{
				AfficheDeclaration(std::ostream& os) :_os(os) {};
				void operator() (std::pair<PrototypeSP,InstanceSP> declaration)
				{
					declaration.first->format(_os);
#ifdef TRACEUR_DEBUG
					_os << ": " ;
					declaration.second->format(_os);
#endif
					_os << std::endl;
				};
				private:
				std::ostream& _os;
			};
			struct AfficheSousMap
			{
				AfficheSousMap(std::ostream& os) :_os(os) {};
				void operator() (std::pair<Prototype::Clef,std::map<PrototypeSP,InstanceSP,Traceur::Global::SPinferieurSP<Prototype> > > sous_map)
				{
					_os <<"Key : ("<< std::get<1>(sous_map.first) <<") "<< std::get<0>(sous_map.first) <<" ("<< std::get<2>(sous_map.first) <<")"<< std::endl;
					std::for_each(sous_map.second.cbegin(),sous_map.second.cend(),AfficheDeclaration(_os));
				};
				private:
				std::ostream& _os;
			};
			struct DispDeclaration
			{
				DispDeclaration(std::ostream& os) :_os(os) {};
				void operator() (std::pair<PrototypeSP,InstanceSP> declaration)
				{
					if(declaration.second and declaration.second->valeur_dynamique())
					{
						_os << "func ";
						declaration.first->format(_os);
						_os << " is" << std::endl;
						declaration.second->valeur_dynamique()->format_disp(_os);
						_os << std::endl;
					}
				};
				private:
				std::ostream& _os;
			};
			struct DispSousMap
			{
				DispSousMap(std::ostream& os) :_os(os) {};
				void operator() (std::pair<Prototype::Clef,std::map<PrototypeSP,InstanceSP,Traceur::Global::SPinferieurSP<Prototype> > > sous_map)
				{
					std::for_each(sous_map.second.cbegin(),sous_map.second.cend(),DispDeclaration(_os));
				};
				private:
				std::ostream& _os;
			};
		}
	}
}

void Declarations::disp(std::ostream& os) const
{
	os << "// --- stack dump: declarations ------------------------------" << std::endl;
	std::for_each(_declarations.cbegin(),_declarations.cend(),DispSousMap(os));
}

void Declarations::format(std::ostream& os) const
{
	os << "Declarations" << std::endl;
	std::for_each(_declarations.cbegin(),_declarations.cend(),AfficheSousMap(os));
	os << "End of declarations" << std::endl;
}

bool Declarations::ajoute_declaration(const PrototypeSP& prototype, const InstanceSP& instance)
{
	auto res = _declarations[prototype->clef()].insert(std::pair<PrototypeSP,InstanceSP>(prototype,instance));
	return not res.second;
}

InstanceSP Declarations::recherche_stricte_locale(const PrototypeSP& prototype) 
{
	auto sous_map = _declarations.find(prototype->clef());
	if(sous_map==_declarations.end())
		return InstanceSP();
	for(auto declaration = sous_map->second.begin() ; declaration!=sous_map->second.end() ; ++declaration)
	{
		if(Traceur::Global::SPegalSP<Prototype>()(declaration->first,prototype))
			return declaration->second;
	}
	return InstanceSP();
}

Declarations::RechercheDeclarations Declarations::recherche_large_locale(const PrototypeSP& prototype)
{
	Declarations::RechercheDeclarations resultat;
	auto sous_map = _declarations.find(prototype->clef());
	if(sous_map==_declarations.end())
		return resultat;
	for(auto iter = sous_map->second.begin() ; iter != sous_map->second.end() ; ++iter)
	{
		// teste si le prototype d'appel est compatilble avec la definition
		Traceur::Donnees::Type::Compatibilite compatibilite = (iter->first)->compatible(prototype,false);
		if(compatibilite)
		{
			resultat[iter->first]=std::pair<InstanceSP,Traceur::Donnees::Type::Compatibilite>(iter->second,compatibilite);
		}
	}
	return resultat;
}

InstanceSP Declarations::recherche_definition(const PrototypeSP& prototype)
{
	InstanceSP instance = recherche_stricte_locale(prototype);
	if(instance)
		return instance;
	DeclarationsSP parent=_parent.lock();
	return parent->recherche_definition(prototype);
}

InstanceSP Declarations::recherche_lien(const PrototypeSP& prototype)
{
	InstanceSP instance = recherche_stricte_locale(prototype);
	if(instance)
		return instance;
	DeclarationsSP parent=_parent.lock();
	return parent->recherche_lien(prototype);
}

Declarations::RechercheDeclarations Declarations::recherche_appel(const PrototypeSP& prototype)
{
	Declarations::RechercheDeclarations instances = recherche_large_locale(prototype);
	DeclarationsSP parent=_parent.lock();
	Declarations::RechercheDeclarations instances_parentes = parent->recherche_appel(prototype);
	instances.insert(instances_parentes.begin(),instances_parentes.end());
	return instances;
}

InstanceSP Declarations::definition(const PrototypeSP& prototype, const FonctionSP& fonction, bool& definition_locale, Traceur::Donnees::Source::PositionSP position)
{
	InstanceSP instance = recherche_definition(prototype);
	if(instance)
	{
		definition_locale=false;
		if(std::dynamic_pointer_cast<InstanceRecursive>(instance))
			
			throw Traceur::Exception::RedefinitionFonctionRecursiveInterdite(position,prototype);
		return instance;
	}
	else
	{
		definition_locale = true;
		InstanceSP nouvelle_instance(new InstanceNormale());
		ajoute_declaration(prototype,nouvelle_instance);
		return nouvelle_instance;
	}
}

InstanceSP Declarations::lien(const PrototypeSP& prototype, Traceur::Donnees::Source::PositionSP position)
{
	InstanceSP instance = recherche_lien(prototype);
	if(not instance)
		throw Traceur::Exception::FonctionNonTrouvee(position,"Imaginary function creation",prototype);
	if(std::dynamic_pointer_cast<InstanceRecursive>(instance))
		throw Traceur::Exception::FonctionImaginaireSurFonctionRecursiveInterdite(position,prototype);
	return instance;
}

InstanceSP Declarations::appel(const PrototypeSP& prototype, Traceur::Donnees::Source::PositionSP position)
{
	Declarations::RechercheDeclarations fonctions_compatibles = recherche_appel(prototype);
	return tri("Function call",prototype,fonctions_compatibles,position);
}

InstanceSP Declarations::tri(const std::string& action, const PrototypeSP& prototype, const Declarations::RechercheDeclarations& fonctions_compatibles, Traceur::Donnees::Source::PositionSP position)
{
	std::map<size_t,std::map<size_t,CategorieDeclarations> > recherche;
	for(auto iter = fonctions_compatibles.begin() ; iter != fonctions_compatibles.end() ; ++iter)
	{
		auto test = recherche[iter->second.second.transtypage_parametres()][iter->second.second.transtypage_retour()]
			.insert( std::pair<PrototypeSP,InstanceSP>(iter->first,iter->second.first)).second;
		TRACEUR_TEST(test);
	}
	if(recherche.empty())
		throw Traceur::Exception::FonctionNonTrouvee(position,action,prototype);
	
	auto candidates = (recherche.begin()->second).begin()->second;
	if(candidates.size()>1)
		throw Traceur::Exception::TropDeFonctionTrouvees(position,action,prototype,candidates);
	return candidates.begin()->second;
}

namespace Traceur
{
	namespace Donnees
	{
		namespace Fonction
		{
			struct CloneDeclaration
			{
				CloneDeclaration(std::map<PrototypeSP,InstanceSP,Traceur::Global::SPinferieurSP<Prototype> >& declarations,
					Traceur::Donnees::Global::ClonesInstances& clones_instances)
					:_declarations(declarations),_clones_instances(clones_instances) {};
				void operator() (std::pair<PrototypeSP,InstanceSP> declaration)
				{
					PrototypeSP prototype = declaration.first->clone(_clones_instances);
					InstanceSP instance = declaration.second->clone(_clones_instances);
					_declarations[prototype]=instance;
				};
				private:
				std::map<PrototypeSP,InstanceSP,Traceur::Global::SPinferieurSP<Prototype> >& _declarations;
				Traceur::Donnees::Global::ClonesInstances& _clones_instances;
			};
			struct CloneSousMap
			{
				CloneSousMap(DeclarationsSP clone, Traceur::Donnees::Global::ClonesInstances& clones_instances)
					:_clone(clone),_clones_instances(clones_instances) {};
				void operator() (std::pair<Prototype::Clef,std::map<PrototypeSP,InstanceSP,Traceur::Global::SPinferieurSP<Prototype> > > sous_map)
				{
					std::for_each(sous_map.second.begin(),sous_map.second.end(),
						CloneDeclaration(_clone->_declarations[sous_map.first],_clones_instances));
				};
				private:
				DeclarationsSP _clone;
				Traceur::Donnees::Global::ClonesInstances& _clones_instances;
			};
		}
	}
}

void Declarations::clone_outil(DeclarationsSP clone, Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	std::for_each(_declarations.begin(),_declarations.end(),CloneSousMap(clone,clones_instances));
}

DeclarationsSP Declarations::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	DeclarationsSP parent = clones_instances[_parent.lock().get()];
	if(not parent)
	{
		parent = _parent.lock();
	}
	DeclarationsSP clone(new Declarations(parent));
	clone_outil(clone,clones_instances);
	clones_instances.ajout(this,clone);
	return clone;
}

bool recursive_categorie_outil(std::pair<PrototypeSP,InstanceSP> declaration)
{
	return static_cast<bool>(std::dynamic_pointer_cast<InstanceRecursive>(declaration.second));
}

namespace Traceur
{
	namespace Donnees
	{
		namespace Fonction
		{
			 struct Recursive_outil
			 {
				bool operator()(std::pair<Prototype::Clef,Declarations::CategorieDeclarations> categorie)
				{
					return std::find_if(categorie.second.begin(),categorie.second.end(),recursive_categorie_outil) != categorie.second.end();
				}
			};
		}
	}
}

bool Declarations::recursive() const
{
	return std::find_if(_declarations.begin(),_declarations.end(),Recursive_outil()) != _declarations.end();
}
