#ifndef _DONNEES_FONCTIONS_DECLARATIONS_H_
#define _DONNEES_FONCTIONS_DECLARATIONS_H_
#include <map>
#include <iostream>
#include <algorithm>

#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Fonction { DECL_SHARED_PTR(Declarations); DECL_WEAK_PTR(Declarations); } } }

#include <src/donnees/global/clonesinstances.h>
#include <src/donnees/fonctions/fonctionidentifiant.h>
#include <src/donnees/fonctions/instance.h>
#include <src/donnees/types/type.h>
#include <src/exceptions/exceptions.h>

namespace Traceur
{
	namespace Exception { class TropDeFonctionTrouvees; }
	namespace Donnees
	{
		namespace Instruction { class For; }
		namespace Fonction
		{
			class FonctionReelle;
			struct CreationLienParametre;
			class DeclarationsIsolees;
			class DeclarationsRestreintes;
			struct CloneSousMap;
			struct Recursive_outil;
			class Declarations
			{
				friend class FonctionReelle;
				friend struct CreationLienParametre;
				friend class DeclarationsRestreintes;
				friend class DeclarationsIsolees;
				friend class CloneSousMap;
				friend struct Recursive_outil;
				friend class Traceur::Donnees::Instruction::For;
				friend struct Traceur::Exception::TropDeFonctionTrouvees;
				public:
					Declarations(DeclarationsSP parent) :_parent(parent) {};
					virtual void format(std::ostream& os) const;
					virtual ~Declarations() {};
					InstanceSP definition(const PrototypeSP& prototype, const FonctionSP& fonction, bool& definition_locale, Traceur::Donnees::Source::PositionSP position);
					InstanceSP lien(const PrototypeSP& prototype, Traceur::Donnees::Source::PositionSP position);
					InstanceSP appel(const PrototypeSP& prototype, Traceur::Donnees::Source::PositionSP position);
					virtual DeclarationsSP clone(Traceur::Donnees::Global::ClonesInstances& clones_instances);
					bool recursive() const;
					void disp(std::ostream& os) const;
				private:
					typedef std::map<PrototypeSP, InstanceSP, Traceur::Global::SPinferieurSP<Prototype> > CategorieDeclarations;
					typedef std::map<PrototypeSP, std::pair<InstanceSP,Traceur::Donnees::Type::Compatibilite>,
						Traceur::Global::SPinferieurSP<Prototype> > RechercheDeclarations;
					InstanceSP tri(const std::string& action, const PrototypeSP& prototype, const Declarations::RechercheDeclarations& fonctions_compatibles, Traceur::Donnees::Source::PositionSP position);
				protected:
					virtual InstanceSP recherche_definition(const PrototypeSP& prototype);
					virtual InstanceSP recherche_lien(const PrototypeSP& prototype);
					virtual Declarations::RechercheDeclarations recherche_appel(const PrototypeSP& prototype);
					void clone_outil(DeclarationsSP clone, Traceur::Donnees::Global::ClonesInstances& clones_instances);
					// renvoie un SP vide si la recherche n'aboutit pas
					InstanceSP recherche_stricte_locale(const PrototypeSP& prototype);
					// renvoie une map vide si la recherche n'aboutit pas
					RechercheDeclarations recherche_large_locale(const PrototypeSP& prototype);
					// renvoie true si la declaration existait deja
					bool ajoute_declaration(const PrototypeSP& prototype, const InstanceSP& instance);
				private:
					std::map<Prototype::Clef, CategorieDeclarations > _declarations;
					DeclarationsWP _parent;
			};
		}
	}
}

namespace Traceur
{
	namespace Exception
	{
		class FonctionNonTrouvee: public Compilation
		{
			public:
				FonctionNonTrouvee(const Traceur::Donnees::Source::PositionSP position, const std::string& action, const Traceur::Donnees::Fonction::PrototypeSP& prototype)
					:Compilation(position,action+std::string(" ")+std::string(*prototype)+std::string(" does not match any function.")) {};
				virtual ~FonctionNonTrouvee() {};
		};
		class TropDeFonctionTrouvees: public Compilation
		{
			struct AjoutePrototype
			{
				AjoutePrototype(std::string& message)
					:_message(message){};
				void operator() (const std::pair<Traceur::Donnees::Fonction::PrototypeSP, Traceur::Donnees::Fonction::InstanceSP>& candidate)
				{
					std::ostringstream oss;
					oss << std::endl << "- " << std::string(*(candidate.first)) ;
					_message += oss.str();
				};
				private:
				std::string& _message;
			};
			public:
				TropDeFonctionTrouvees(const Traceur::Donnees::Source::PositionSP position, const std::string& action, const Traceur::Donnees::Fonction::PrototypeSP& prototype,
					const Traceur::Donnees::Fonction::Declarations::CategorieDeclarations& candidates)
					:Compilation(position,action+std::string(" ")+std::string(*prototype)+std::string(" matches several functions:"))
					{
						std::for_each(candidates.begin(),candidates.end(),TropDeFonctionTrouvees::AjoutePrototype(_message));
					};
				virtual ~TropDeFonctionTrouvees() {};
		};
		class RedefinitionFonctionRecursiveInterdite: public Compilation
		{
			public:
				RedefinitionFonctionRecursiveInterdite(const Traceur::Donnees::Source::PositionSP position, const Traceur::Donnees::Fonction::PrototypeSP& prototype)
					:Compilation(position,std::string("Redefinition of recursive function ")+std::string(*prototype)+std::string(" not allowed inside its definition.")) {};
				virtual ~RedefinitionFonctionRecursiveInterdite() {};
		};
		class FonctionImaginaireSurFonctionRecursiveInterdite: public Compilation
		{
			public:
				FonctionImaginaireSurFonctionRecursiveInterdite(const Traceur::Donnees::Source::PositionSP position, const Traceur::Donnees::Fonction::PrototypeSP& prototype)
					:Compilation(position,std::string("Imaginary function ")+std::string(*prototype)+std::string(" linked to a recursive function not allowed inside its definition.")) {};
				virtual ~FonctionImaginaireSurFonctionRecursiveInterdite() {};
		};
	}
}
#endif
