#include <src/donnees/fonctions/fonctionimaginaire.h>
#include <src/donnees/expressions/expressionfonctionimaginaire.h>
#include <src/donnees/instructions/instructionreturn.h>
#include <src/donnees/fonctions/fonctionreelle.h>
#include <src/donnees/fonctions/instancenormale.h>
#include <src/donnees/sources/position.h>
using namespace Traceur::Donnees::Fonction;

FonctionImaginaire::FonctionImaginaire(const Traceur::Donnees::Expression::ExpressionSP& expression, const Traceur::Donnees::Source::PositionSP& position, const Traceur::Donnees::Type::TypeSP& type_attendu)
{
	_position=position;
	Traceur::Donnees::Expression::FonctionImaginaireSP fonction_imaginaire = std::dynamic_pointer_cast<Traceur::Donnees::Expression::FonctionImaginaire>(expression);
#ifdef TRACEUR_DEBUG
		if(fonction_imaginaire and type_attendu)
		{
			TRACEUR_TRACE << "Lien special de parametre pour la fonction imaginaire " << fonction_imaginaire << " ? Type attendu: " << type_attendu << ". Type fonction trouve: " << fonction_imaginaire << std::endl;
		}
#endif
	if(fonction_imaginaire and ((not type_attendu) or fonction_imaginaire->_type->compatible(type_attendu)))
	{	// lien special des parametres
		_prototype = fonction_imaginaire->_fonction->prototype();
		FonctionImaginaireSP fonction(new FonctionImaginaire(_prototype,fonction_imaginaire->_fonction,position));
		_valeur_statique = fonction;

		InstanceSP instance(new InstanceNormale());
		_instance = instance;
		_instance->valeur_dynamique(fonction);
	}
	else
	{
		Traceur::Donnees::Fonction::PrototypeSP prototype(new Traceur::Donnees::Fonction::Prototype(expression->_type,
					Traceur::Donnees::Fonction::Prototype::ListeParametres(),
					Traceur::Donnees::Fonction::IDENTIFIANTVIDE(),
					Traceur::Donnees::Fonction::Prototype::ListeParametres()));
		_prototype = prototype;

		Traceur::Donnees::Instruction::InstructionSP instruction(new Traceur::Donnees::Instruction::Return(expression));
		FonctionSP fonction(new FonctionReelle(prototype,instruction,position));
		_valeur_statique = fonction;

		InstanceSP instance(new InstanceNormale());
		_instance = instance;
		_instance->valeur_dynamique(fonction);
	}
}

FonctionImaginaire::FonctionImaginaire(const PrototypeSP& prototype, const FonctionSP& fonction, Traceur::Donnees::Source::PositionSP position)
:Fonction(prototype,position), _valeur_statique(fonction)
{
	InstanceSP instance(new InstanceNormale());
	_instance = instance;
}

void FonctionImaginaire::disp(std::ostream& os) const
{
	throw;
}

void FonctionImaginaire::format_disp(std::ostream& os) const
{
	os << "func " ;
	if(_valeur_statique and _valeur_statique->recursive())
		os << "rec ";
	_prototype->format(os);
	os << " linked to " << std::endl;
	if(_valeur_statique)
	{
		_valeur_statique->format_disp(os);
	}
	else
	{
		_instance->valeur_dynamique()->format_disp(os);
	}
}

void FonctionImaginaire::format(std::ostream& os) const
{
	os << "func " ;
	if(_valeur_statique and _valeur_statique->recursive())
		os << "rec ";
	_prototype->format(os);
#ifdef TRACEUR_DEBUG
	if(_instance)
	{
		os << ": ";
		_instance->format(os);
	}
	os << std::endl;
#endif
	if(_valeur_statique)
	{
#ifdef TRACEUR_DEBUG
		os << "Redirected to " ;
#else
		os << " := do" << std::endl;
#endif
		_valeur_statique->format(os);
#ifndef TRACEUR_DEBUG
		os << "end" << std::endl;
#endif
	}
#ifdef TRACEUR_DEBUG
	else
	{
		os << "No static function" << std::endl;
	}
#endif
}

void FonctionImaginaire::edition_liens(Traceur::Donnees::Instruction::BlocSP bloc)
{
	if(_valeur_statique)
	{
		TRACEUR_TEST(_instance);
		_valeur_statique->edition_liens(bloc);
	}
	else
	{
		_instance = bloc->_declarations->lien(_prototype,_position);
		TRACEUR_TEST(_instance);
	}
}

void FonctionImaginaire::definition_fonction()
{
	if(_valeur_statique)
	{
		Traceur::Donnees::Global::ClonesInstances clones_instances;
		FonctionSP clone = _valeur_statique->clone(clones_instances);
#ifdef TRACEUR_DEBUG
		TRACEUR_TRACE; format(std::cerr); std::cerr << std::endl;
#endif
		clone->definition_fonction();
		_instance->valeur_dynamique(clone);
#ifdef TRACEUR_DEBUG
		TRACEUR_TRACE << clone.get() << std::endl;
		TRACEUR_TRACE << clone << std::endl;
		TRACEUR_TRACE; format(std::cerr); std::cerr << std::endl;
#endif
	}
}

Traceur::Donnees::Instruction::InstructionInterruptionSP FonctionImaginaire::appel(
	ListeParametres gauche, ListeParametres droite, Traceur::Donnees::Assembleur::AssembleurSP& assembleur,
						Traceur::Donnees::Expression::AppelFonctionUtilisateurSP appel)
{
	FonctionSP fonction = _instance->valeur_appel();
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE << "appel fonction imaginaire> " << _instance << std::endl;
#endif
	//TRACEUR_TEST(fonction);
	if(fonction)
		return fonction->appel(gauche,droite,assembleur,appel);
	return _valeur_statique->appel(gauche,droite,assembleur,appel); // cas des fonctions imaginaires pures : elles n'ont pas de valeur dynamique
}

Traceur::Donnees::Expression::ExpressionSP FonctionImaginaire::substitution(ListeParametres gauche, ListeParametres droite)
{
	if(not _valeur_statique)
	{
		return Traceur::Donnees::Expression::ExpressionSP();
	}
	definition_fonction();
	FonctionSP valeur_dynamique = _instance->valeur_appel();
	if(not valeur_dynamique)
	{
		return Traceur::Donnees::Expression::ExpressionSP();
	}
	return valeur_dynamique->substitution(gauche,droite);
}

FonctionSP FonctionImaginaire::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	FonctionSP clone_ptetre = clones_instances[this];
	if(clone_ptetre)
		return clone_ptetre;
	PrototypeSP prototype = _prototype->clone(clones_instances);
	FonctionImaginaireSP clone(new FonctionImaginaire(prototype,_position));
	clones_instances.ajout(this,clone);
	if(_valeur_statique)
	{
		FonctionSP fonction = _valeur_statique->clone(clones_instances);
		clone->_valeur_statique=fonction;
	}
	InstanceSP instance = _instance->clone(clones_instances);
	clone->_instance=instance;
	return clone;
}

#ifdef TRACEUR_DEBUG
void debug_fonctionimaginaire(Traceur::Donnees::Fonction::Fonction *fi)
{
	fi->format(std::cout);
}
void debug_instance(Traceur::Donnees::Fonction::Instance *i)
{
	i->format(std::cout);
}
#endif
