#include <src/donnees/fonctions/fonctionprototype.h>
#include <algorithm>
#include <functional>
#include <set>

using namespace Traceur::Donnees::Fonction;

	Prototype::Prototype(Traceur::Donnees::Type::TypeSP type_retour, ListeParametres parametres_gauche, Traceur::Donnees::Fonction::Identifiant nom_fonction, ListeParametres parametres_droite)
:_type_retour(type_retour),_parametres_gauche(parametres_gauche),_nom_fonction(nom_fonction),_parametres_droite(parametres_droite)
{
}

void Prototype::format(std::ostream& os) const
{
	_type_retour->format(os);
	os << " " ;
	format(_parametres_gauche,os);
	os << _nom_fonction ;
	format(_parametres_droite,os);
}

Traceur::Donnees::Type::Compatibilite Prototype::compatible(const PrototypeSP prototype, bool test_complet)
{
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE << "Compatibilite prototypes de " << prototype << " envers " << this->shared_from_this() << std::endl;
#endif
	if(test_complet)
	{
		if(clef()!=prototype->clef())
			return Traceur::Donnees::Type::NonCompatible();
	}

	Traceur::Donnees::Type::TypeSP type_this(new Traceur::Donnees::Type::Fonction(this->shared_from_this()));
	Traceur::Donnees::Type::TypeSP type_prototype(new Traceur::Donnees::Type::Fonction(prototype));
	return (type_this)->compatible_prototype(type_prototype);
}

void separateur(const Traceur::Donnees::Fonction::PrototypeSP& prototype, std::ostream& os)
{
	os << ", " ;
	prototype->format(os);
}

void Prototype::format(ListeParametres liste_parametres, std::ostream& os) const
{
	if(!liste_parametres.empty())
	{
		os << "(" ;
		auto iterateur = liste_parametres.cbegin();
		(*iterateur)->format(os);
		++iterateur;
		std::for_each(iterateur,liste_parametres.cend(),bind(separateur,std::placeholders::_1,std::ref(os)));
		os << ")" ;
	}
}


struct EgaliteType
{
	bool operator()(const PrototypeSP& p1, const PrototypeSP& p2)
	{
		Traceur::Donnees::Type::TypeSP type_p1(new Traceur::Donnees::Type::Fonction(p1));
		Traceur::Donnees::Type::TypeSP type_p2(new Traceur::Donnees::Type::Fonction(p2));
		return Traceur::Global::SPegalSP<Traceur::Donnees::Type::Type>()(type_p1,type_p2);
	};
};

bool Prototype::operator==(const Prototype& prototype) const
{
	if(clef()!=prototype.clef())
		return false;

	if(not Traceur::Global::SPegalSP<Traceur::Donnees::Type::Type>()(prototype._type_retour,_type_retour))
		return false;

	if(not std::equal(_parametres_gauche.cbegin(),_parametres_gauche.cend(),prototype._parametres_gauche.cbegin(),EgaliteType()))
		return false;

	if(not std::equal(_parametres_droite.cbegin(),_parametres_droite.cend(),prototype._parametres_droite.cbegin(),EgaliteType()))
		return false;

	return true;
}

bool Prototype::operator<(const Prototype& prototype) const
{
	if(_nom_fonction<prototype._nom_fonction)
		return true;
	if(prototype._nom_fonction<_nom_fonction)
		return false;

	if(Traceur::Global::SPinferieurSP<Traceur::Donnees::Type::Type>()(_type_retour,prototype._type_retour))
		return true;
	if(Traceur::Global::SPinferieurSP<Traceur::Donnees::Type::Type>()(prototype._type_retour,_type_retour))
		return false;

	if(std::lexicographical_compare(_parametres_gauche.cbegin(),_parametres_gauche.cend(),
				prototype._parametres_gauche.cbegin(),prototype._parametres_gauche.cend(),
				Traceur::Global::SPinferieurSP<Prototype>()))
		return true;
	if(std::lexicographical_compare(prototype._parametres_gauche.cbegin(),prototype._parametres_gauche.cend(),
				_parametres_gauche.cbegin(),_parametres_gauche.cend(),
				Traceur::Global::SPinferieurSP<Prototype>()))
		return false;

	if(std::lexicographical_compare(_parametres_droite.cbegin(),_parametres_droite.cend(),
				prototype._parametres_droite.cbegin(),prototype._parametres_droite.cend(),
				Traceur::Global::SPinferieurSP<Prototype>()))
		return true;
	if(std::lexicographical_compare(prototype._parametres_droite.cbegin(),prototype._parametres_droite.cend(),
				_parametres_droite.cbegin(),_parametres_droite.cend(),
				Traceur::Global::SPinferieurSP<Prototype>()))
		return false;

	return false;
}

PrototypeSP Prototype::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	return this->shared_from_this();
}
