#ifndef _DONNEES_FONCTIONS_PROTOTYPE_H_
#define _DONNEES_FONCTIONS_PROTOTYPE_H_
#include <vector>
#include <iostream>

#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Fonction { DECL_SHARED_PTR(Prototype); } } }

#include <src/donnees/fonctions/fonctionidentifiant.h>
#include <src/donnees/global/clonesinstances.h>
#include <src/donnees/types/typefonction.h>
#include <src/exceptions/exceptions.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Type { class SPconversionDePrototypeVersType; }
		namespace Instruction { class For; }
		namespace Expression { class AppelFonctionUtilisateur; }
		namespace Fonction
		{
			class VerificationNomParametre;
			class FonctionReelle;
			class FonctionImaginaire;
			class Declarations;
			struct DispSousMap;
			struct AfficheSousMap;
			struct CloneSousMap;
			struct Recursive_outil;
			class Prototype : public std::enable_shared_from_this<Prototype>
			{
				friend class Traceur::Donnees::Type::Fonction;
				friend class Traceur::Donnees::Fonction::FonctionReelle;
				friend class Traceur::Donnees::Fonction::FonctionImaginaire;
				friend class Traceur::Donnees::Type::SPconversionDePrototypeVersType;
				friend class Traceur::Donnees::Instruction::For;
				friend class Traceur::Donnees::Expression::AppelFonctionUtilisateur;
				friend class Declarations;
				friend class VerificationNomParametre;
				friend struct AfficheSousMap;
				friend struct DispSousMap;
				friend struct CloneSousMap;
				friend struct Recursive_outil;
				public:
					typedef std::vector<PrototypeSP> ListeParametres;
					Prototype(Traceur::Donnees::Type::TypeSP type_retour, ListeParametres parametres_gauche, Traceur::Donnees::Fonction::Identifiant nom_fonction, ListeParametres parametres_droite);
					// retourne la distance entre les prototypes s'ils sont compatibles
					Traceur::Donnees::Type::Compatibilite compatible(const PrototypeSP prototype, bool test_complet = true);
					operator std::string () const
					{
						std::ostringstream oss;
						format(oss);
						return oss.str();
					};
					PrototypeSP clone(Traceur::Donnees::Global::ClonesInstances& clones_instances);
					void format(std::ostream& os) const;
					bool operator<(const Prototype&) const;
					bool operator==(const Prototype&) const;
					virtual ~Prototype() {};
				private:
					typedef std::tuple<Identifiant,size_t,size_t> Clef;
					Clef clef() const
					{
						return Clef(_nom_fonction,_parametres_gauche.size(),_parametres_droite.size());
					}
					Traceur::Donnees::Type::TypeSP _type_retour;
					ListeParametres _parametres_gauche;
					Traceur::Donnees::Fonction::Identifiant _nom_fonction;
					ListeParametres _parametres_droite;
					void format(ListeParametres liste_parametres, std::ostream& os) const;
			};
		}
	}
}

namespace Traceur
{
namespace Exception
{
	class ParametreAmbigu : public Compilation
	{
		public:
			ParametreAmbigu(Traceur::Donnees::Source::PositionSP position, Traceur::Donnees::Fonction::PrototypeSP parametre, Traceur::Donnees::Fonction::PrototypeSP prototype)
				:Compilation(position,std::string("Parameter ")+std::string(*parametre)+
					std::string(" is ambiguous in prototype ")+std::string(*prototype)+
					std::string(".")) {};
			virtual ~ParametreAmbigu() {};
	};
}
}
#endif
