#ifndef _DONNEES_FONCTIONS_FONCTIONREELLE_H_
#define _DONNEES_FONCTIONS_FONCTIONREELLE_H_
#include <vector>

#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Fonction { DECL_SHARED_PTR(FonctionReelle); DECL_WEAK_PTR(FonctionReelle); } } }

#include <src/donnees/fonctions/fonction.h>
#include <src/donnees/fonctions/fonctionprototype.h>
#include <src/donnees/instructions/instruction.h>
#include <src/exceptions/exceptions.h>
#include <src/donnees/expressions/expressionappelfonctionutilisateur.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression { DECL_WEAK_PTR(AppelFonctionUtilisateur); }
		namespace Fonction
		{
			struct CreationLienParametre;
			class FonctionReelle : public Fonction, public std::enable_shared_from_this<FonctionReelle>
			{
				friend struct CreationLienParametre;
				public:
					enum class Type { NORMALE, RECURSIVE };
					FonctionReelle(const PrototypeSP& prototype, const Traceur::Donnees::Instruction::InstructionSP& corps, const Traceur::Donnees::Source::PositionSP& position);
					virtual void edition_liens(Traceur::Donnees::Instruction::BlocSP bloc);
					virtual void definition_fonction();
					virtual FonctionSP clone(Traceur::Donnees::Global::ClonesInstances& clones_instances);
					virtual Traceur::Donnees::Instruction::InstructionInterruptionSP appel(
						ListeParametres gauche, ListeParametres droite, Traceur::Donnees::Assembleur::AssembleurSP& assembleur,
						Traceur::Donnees::Expression::AppelFonctionUtilisateurSP appel);
					virtual Traceur::Donnees::Expression::ExpressionSP substitution(ListeParametres gauche, ListeParametres droite);
					virtual void format_disp(std::ostream& os) const;
					virtual void disp(std::ostream& os) const;
					virtual void format(std::ostream& os) const;
					virtual ~FonctionReelle() {};
				protected:
					Type _type;
					Traceur::Donnees::Instruction::BlocSP _corps;
					typedef std::vector<InstanceSP> LiensParametres;
					LiensParametres _parametres_gauche;
					LiensParametres _parametres_droite;
					Traceur::Donnees::Expression::AppelFonctionUtilisateurWP _appel;
				private:
					virtual bool recursive() const;
					FonctionReelle(const Type type, const PrototypeSP& prototype, const Traceur::Donnees::Source::PositionSP& position)
						:Fonction(prototype,position),_type(type) {};
					typedef std::vector<FonctionSP> PileParametres;
					void empile_parametres(Traceur::Donnees::Fonction::Fonction::ListeParametres gauche, Traceur::Donnees::Fonction::Fonction::ListeParametres droite, PileParametres& sauvegarde_gauche, PileParametres& sauvegarde_droite);
					void depile_parametres(PileParametres& sauvegarde_gauche, PileParametres& sauvegarde_droite);
			};

			class FonctionReelleRecursive : public FonctionReelle
			{
				public:
					FonctionReelleRecursive(const PrototypeSP& prototype, const Traceur::Donnees::Instruction::InstructionSP& corps, const Traceur::Donnees::Source::PositionSP& position)
						:FonctionReelle(prototype,corps,position) { _type = Type::RECURSIVE; _corps->mode_isole(); }
					virtual ~FonctionReelleRecursive() {};
			};
		}
	}
}

namespace Traceur
{
	namespace Exception
	{
		class Retour_manquant : public Compilation
		{
			public:
				Retour_manquant(const Traceur::Donnees::Source::PositionSP position, const Traceur::Donnees::Fonction::PrototypeSP& prototype)
					:Compilation(position,std::string("Missing return instruction in function ")+
						std::string(*prototype)+
						std::string(".")) {};
				virtual ~Retour_manquant() {};
				
		};
	}
}
#endif
