#ifndef _DONNEES_GLOBAL_CACHEOBJETS_H_
#define _DONNEES_GLOBAL_CACHEOBJETS_H_
#include <map>

#include <src/global/global.h>

#include <src/donnees/expressions/expressionentier.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression { DECL_SHARED_PTR(Expression); }
		namespace Global
		{
			template<typename TypeObjet>
			class CacheObjet
			{
				public:
					typedef TypeObjet Type;
					CacheObjet() = default;
					// le pointeur passe a cette methode DOIT AVOIR ETE ALLOUE PAR NEW
					SHARED_PTR(TypeObjet) operator()(TypeObjet *objet)
					{
						auto iter = _cache.find(*objet);
						if(iter != _cache.end())
						{
							delete objet;
							return iter->second;
						}
						SHARED_PTR(TypeObjet) pointeur(objet);
						_cache[*objet]=pointeur;
						return pointeur;
					};
				private:
					struct comparaison { bool operator() (TypeObjet inf, TypeObjet sup) { return inf<sup; }; };
					std::map<TypeObjet,SHARED_PTR(TypeObjet), comparaison > _cache;
			};

			class CacheObjetGenerique
			{
				public:
					CacheObjetGenerique() = default;
					Traceur::Donnees::Expression::ExpressionSP operator()(Traceur::Donnees::Expression::Expression* objet)
					{
						Traceur::Donnees::Expression::Entier *entier = dynamic_cast<Traceur::Donnees::Expression::Entier*>(objet);
						if(entier)
						{
							return Traceur::Donnees::Expression::ExpressionSP(_cache_entiers(entier));
						}
						Traceur::Donnees::Expression::ExpressionSP expression(objet);
						return expression;
					}
				private:
					CacheObjet<Traceur::Donnees::Expression::Entier> _cache_entiers;
			};

		}
	}
}
#endif
