#ifndef _DONNEES_GLOBAL_CLONESINSTANCES_H_
#define _DONNEES_GLOBAL_CLONESINSTANCES_H_
#include <map>

#include <src/global/global.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Fonction
		{
			DECL_SHARED_PTR(Instance);
			DECL_SHARED_PTR(Fonction);
			DECL_SHARED_PTR(Declarations);
		}
		namespace Instruction
		{
			DECL_SHARED_PTR(Bloc);
			DECL_SHARED_PTR(Instruction);
		}
		namespace Expression
		{
			DECL_SHARED_PTR(Expression);
		}
		namespace Global
		{
			class ClonesInstances
			{
				public:
					ClonesInstances() = default;
					~ClonesInstances() {};

					Traceur::Donnees::Fonction::InstanceSP operator[] (Traceur::Donnees::Fonction::Instance* instance)
					{
						auto clone_iter = _instances.find(instance);
						if(clone_iter==_instances.end())
						{
							return Traceur::Donnees::Fonction::InstanceSP();
						}
						return clone_iter->second;
					};
					void ajout(Traceur::Donnees::Fonction::Instance* original, Traceur::Donnees::Fonction::InstanceSP clone)
					{
						_instances[original]=clone;
					};

					Traceur::Donnees::Fonction::FonctionSP operator[] (Traceur::Donnees::Fonction::Fonction* fonction)
					{
						auto clone_iter = _fonctions.find(fonction);
						if(clone_iter==_fonctions.end())
						{
							return Traceur::Donnees::Fonction::FonctionSP();
						}
						return clone_iter->second;
					};
					void ajout(Traceur::Donnees::Fonction::Fonction* original, Traceur::Donnees::Fonction::FonctionSP clone)
					{
						_fonctions[original]=clone;
					};

					Traceur::Donnees::Fonction::DeclarationsSP operator[] (Traceur::Donnees::Fonction::Declarations* declarations)
					{
						auto clone_iter = _declarations.find(declarations);
						if(clone_iter==_declarations.end())
						{
							return Traceur::Donnees::Fonction::DeclarationsSP();
						}
						return clone_iter->second;
					};
					void ajout(Traceur::Donnees::Fonction::Declarations* original, Traceur::Donnees::Fonction::DeclarationsSP clone)
					{
						_declarations[original]=clone;
					};

					Traceur::Donnees::Instruction::BlocSP operator[] (Traceur::Donnees::Instruction::Bloc* bloc)
					{
						auto clone_iter = _blocs.find(bloc);
						if(clone_iter==_blocs.end())
						{
							return Traceur::Donnees::Instruction::BlocSP();
						}
						return clone_iter->second;
					};
					void ajout(Traceur::Donnees::Instruction::Bloc* original, Traceur::Donnees::Instruction::BlocSP clone)
					{
						_blocs[original]=clone;
					};

					Traceur::Donnees::Instruction::InstructionSP operator[] (Traceur::Donnees::Instruction::Instruction* instruction)
					{
						auto clone_iter = _instructions.find(instruction);
						if(clone_iter==_instructions.end())
						{
							return Traceur::Donnees::Instruction::InstructionSP();
						}
						return clone_iter->second;
					};
					void ajout(Traceur::Donnees::Instruction::Instruction* original, Traceur::Donnees::Instruction::InstructionSP clone)
					{
						_instructions[original]=clone;
					};

					Traceur::Donnees::Expression::ExpressionSP operator[] (Traceur::Donnees::Expression::Expression* expression)
					{
						auto clone_iter = _expressions.find(expression);
						if(clone_iter==_expressions.end())
						{
							return Traceur::Donnees::Expression::ExpressionSP();
						}
						return clone_iter->second;
					};
					void ajout(Traceur::Donnees::Expression::Expression* original, Traceur::Donnees::Expression::ExpressionSP clone)
					{
						_expressions[original]=clone;
					};
				private:
					std::map<Traceur::Donnees::Fonction::Instance*,Traceur::Donnees::Fonction::InstanceSP> _instances;
					std::map<Traceur::Donnees::Fonction::Fonction*,Traceur::Donnees::Fonction::FonctionSP> _fonctions;
					std::map<Traceur::Donnees::Fonction::Declarations*,Traceur::Donnees::Fonction::DeclarationsSP> _declarations;
					std::map<Traceur::Donnees::Instruction::Bloc*,Traceur::Donnees::Instruction::BlocSP> _blocs;
					std::map<Traceur::Donnees::Instruction::Instruction*,Traceur::Donnees::Instruction::InstructionSP> _instructions;
					std::map<Traceur::Donnees::Expression::Expression*,Traceur::Donnees::Expression::ExpressionSP> _expressions;
			};
		}
	}
}
#endif
