#ifndef _DONNEES_INSTRUCTIONS_BLOC_H_
#define _DONNEES_INSTRUCTIONS_BLOC_H_
#include <list>
#include <iostream>

#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Instruction { DECL_SHARED_PTR(Bloc); DECL_WEAK_PTR(Bloc); } } }

#include <src/donnees/global/clonesinstances.h>
#include <src/donnees/fonctions/declarations.h>
#include <src/donnees/instructions/instruction.h>
#include <src/donnees/types/type.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Fonction { class FonctionReelle; class FonctionImaginaire; struct CreationLienParametre; }
		namespace Expression { class AppelFonctionUtilisateur; DECL_SHARED_PTR(Expression); }
		namespace Assembleur { DECL_SHARED_PTR(Assembleur); }
		namespace Instruction
		{
			DECL_SHARED_PTR(InstructionInterruption);
			class Let;
			class For;
			class Bloc: public std::enable_shared_from_this<Bloc>
			{
				friend class Traceur::Donnees::Fonction::FonctionReelle;
				friend class Traceur::Donnees::Fonction::FonctionImaginaire;
				friend class Traceur::Donnees::Expression::AppelFonctionUtilisateur;
				friend struct Traceur::Donnees::Fonction::CreationLienParametre;
				friend class Let;
				friend class For;
				public:
					Bloc() = default;
					void mode_restreint() { if(_type!=Type::ISOLE) _type = Type::RESTREINT; };
					void mode_isole() { _type = Type::ISOLE; }
					void ajoute(InstructionSP instruction);
					bool teste_retour_fonction(const Traceur::Donnees::Type::TypeSP& type) const;
					void edition_liens(BlocSP parent = BlocSP(), InstructionSP instruction = InstructionSP());
					void definition_fonction();
					void simplification();
					BlocSP clone(Traceur::Donnees::Global::ClonesInstances& clones_instances);
					InstructionInterruptionSP evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur);
					void disp(std::ostream& os) const;
					void format(std::ostream& os) const;
					virtual ~Bloc() {};
				private:
					Traceur::Donnees::Expression::ExpressionSP commence_par_return();
					enum class Type {NORMAL, RESTREINT, ISOLE};
					void cree_declarations(BlocSP parent);
					Type _type;
					std::list<InstructionSP> _instructions;
					Traceur::Donnees::Fonction::DeclarationsSP _declarations;
					InstructionWP _parent;
					Traceur::Donnees::Fonction::FonctionWP _fonction;
			};
		}
	}
}
#ifdef TRACEUR_DEBUG
void debug_bloc(Traceur::Donnees::Instruction::BlocSP bloc);
#endif
#endif
