#ifndef _DONNEES_INSTRUCTIONS_INSTRUCTION_H_
#define _DONNEES_INSTRUCTIONS_INSTRUCTION_H_
#include <list>
#include <iostream>

#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Instruction { DECL_SHARED_PTR(Instruction); DECL_WEAK_PTR(Instruction); } } }

#include <src/donnees/global/clonesinstances.h>
#include <src/donnees/instructions/bloc.h>
#include <src/donnees/sources/position.h>
#include <src/exceptions/exceptions.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Assembleur { DECL_SHARED_PTR(Assembleur); }
		namespace Expression { class AppelFonctionUtilisateur; class FonctionImaginaire; class Image; class Noyau; }
		namespace Instruction
		{
			DECL_SHARED_PTR(InstructionInterruption);
			class Bloc;//test
			class Instruction: public std::enable_shared_from_this<Instruction>
			{
				friend class Traceur::Donnees::Expression::AppelFonctionUtilisateur;
				friend class Traceur::Donnees::Expression::FonctionImaginaire;
				friend class Traceur::Donnees::Expression::Image;
				friend class Traceur::Donnees::Expression::Noyau;
				friend class Bloc;//test
				public:
					Instruction(const Traceur::Donnees::Source::PositionSP position) :_position(position) {};
					virtual void edition_liens(BlocSP bloc) { _parent=bloc;	};
					virtual void definition_fonction() = 0;
					virtual InstructionSP clone(Traceur::Donnees::Global::ClonesInstances& clones_instances) = 0;
					virtual InstructionInterruptionSP evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur) = 0;
					virtual void simplification() { };
					//virtual void optimisation()// = 0;		 // TODO => virtual InstructionSP optimisation(Optimiseur);
					virtual void disp(std::ostream& os) const;
					virtual void format(std::ostream& os) const = 0;
					virtual bool teste_retour_fonction(const Traceur::Donnees::Type::TypeSP& type) const =0;
					virtual ~Instruction() {};
				protected:
					BlocWP _parent;
					Traceur::Donnees::Source::PositionSP _position;
				private:
					BlocSP parent() { return _parent.lock(); };
					Traceur::Donnees::Source::PositionSP position() { return _position; }; //test
			};
			// si trop de copies avec la std::list, il faudra voir pour en faire une a la mano avec des concatenations/insertions en O(1)
			typedef std::list<InstructionSP> ListeInstructions;
		}
	}
}

namespace Traceur
{
	namespace Exception
	{
		class TypeInvalide : public Compilation
		{
			public:
			TypeInvalide(Traceur::Donnees::Source::PositionSP position, Traceur::Donnees::Type::TypeSP type, std::string instruction)
				:Compilation(position,std::string("Type ")+std::string(*type)+
					std::string(" invalid for instruction ")+instruction+std::string(".")) {};
		};
	}
}
#endif
