#include <src/donnees/instructions/instructionconv.h>
#include <src/donnees/assembleur/assembleur.h>
#include <src/donnees/instructions/valeur.h>
#include <src/donnees/types/typekernel.h>

using namespace Traceur::Donnees::Instruction;

Conv::Conv(const Traceur::Donnees::Expression::ExpressionSP& expression, const Traceur::Donnees::Source::PositionSP& position)
:Instruction(position), _expression(expression)
{
	if(_expression->est_de_type<Traceur::Donnees::Type::Kernel>())
		return;
	throw Traceur::Exception::TypeInvalide(_position,_expression->type(),"conv");
}

void Conv::edition_liens(BlocSP bloc)
{
	Instruction::edition_liens(bloc);
	_expression->edition_liens(this->shared_from_this());
}

InstructionInterruptionSP Conv::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur = _expression->evaluation(assembleur);
	if(valeur->a_une_valeur_retour())
	{
		Traceur::Donnees::Expression::ExpressionSP expr_valeur = *valeur;
		assembleur->conv(expr_valeur);
		return InstructionInterruptionSP();
	}
	InstructionInterruptionSP interruption = *valeur;
	interruption->ajoute_position(_position);
	return interruption;
}

void Conv::disp(std::ostream& os) const
{
	Instruction::disp(os);
	format(os);
}

void Conv::format(std::ostream& os) const
{
	os << "conv ";
	_expression->format(os);
	os << std::endl;
}

void Conv::definition_fonction()
{
	_expression->definition_fonction();
}

InstructionSP Conv::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	BlocSP parent = clones_instances[_parent.lock().get()];
	if(not parent)
		parent = _parent.lock();
	ConvSP clone(new Conv(_position));
	clones_instances.ajout(this,clone);
	Traceur::Donnees::Expression::ExpressionSP clone_expression = _expression->clone(clones_instances);
	clone->_expression = clone_expression;
	clone->_parent = parent;
	return clone;
}
