#include <src/donnees/instructions/instructiondisp.h>
#include <src/donnees/assembleur/assembleur.h>
#include <src/donnees/instructions/valeur.h>
using namespace Traceur::Donnees::Instruction;

void Disp::edition_liens(BlocSP bloc)
{
	Instruction::edition_liens(bloc);
	if(_expression)
		_expression->edition_liens(this->shared_from_this());
}

InstructionInterruptionSP Disp::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	if(_expression)
	{
		ValeurSP valeur = _expression->evaluation(assembleur);
		if(valeur->a_une_valeur_retour())
		{
			Traceur::Donnees::Expression::ExpressionSP expr_valeur = *valeur;
			assembleur->disp(_expression,expr_valeur);
			return InstructionInterruptionSP();
		}
		InstructionInterruptionSP interruption = *valeur;
		interruption->ajoute_position(_position);
		return *valeur;
	}
	else
	{
		assembleur->disp();
		std::cerr << "// --- stack dump: begin -------------------------------------" << std::endl;
		disp(std::cerr);
		std::cerr << std::endl << "// --- stack dump: end ---------------------------------------" << std::endl << std::endl << std::endl;
		return InstructionInterruptionSP();
	}
}

void Disp::disp(std::ostream& os) const
{
	Instruction::disp(os);
	format(os);
}

void Disp::format(std::ostream& os) const
{
	os << "disp ";
	if(_expression)
		_expression->format(os);
	os << std::endl;
}

void Disp::definition_fonction()
{
	if(_expression)
		_expression->definition_fonction();
}

InstructionSP Disp::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	BlocSP parent = clones_instances[_parent.lock().get()];
	if(not parent)
		parent = _parent.lock();
	DispSP clone(new Disp(_position));
	clones_instances.ajout(this,clone);
	if(_expression)
	{
		Traceur::Donnees::Expression::ExpressionSP clone_expression = _expression->clone(clones_instances);
		clone->_expression = clone_expression;
	}
	clone->_parent = parent;
	return clone;
}
