#include <src/donnees/instructions/instructiondraw.h>
#include <src/donnees/assembleur/assembleur.h>
#include <src/donnees/instructions/valeur.h>
#include <src/donnees/types/typetool.h>
#include <src/donnees/types/typevoid.h>

using namespace Traceur::Donnees::Instruction;

Draw::Draw(const Traceur::Donnees::Expression::ExpressionSP& expression, const Traceur::Donnees::Source::PositionSP& position)
:Instruction(position), _expression(expression)
{
	if(_expression->est_de_type<Traceur::Donnees::Type::Tool>())
		return;
	if(_expression->est_de_type<Traceur::Donnees::Type::Void>())
		return;
	throw Traceur::Exception::TypeInvalide(_position,_expression->type(),"draw");
}

void Draw::edition_liens(BlocSP bloc)
{
	Instruction::edition_liens(bloc);
	_expression->edition_liens(this->shared_from_this());
}

InstructionInterruptionSP Draw::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur = _expression->evaluation(assembleur);
	if(valeur->a_une_valeur_retour())
	{
		Traceur::Donnees::Expression::ExpressionSP expr_valeur = *valeur;
		try
		{
			assembleur->draw(expr_valeur);
			return InstructionInterruptionSP();
		}
		catch(ValeurSP echec) { valeur=echec; }
	}
	InstructionInterruptionSP interruption = *valeur;
	interruption->ajoute_position(_position);
	return interruption;
}

void Draw::disp(std::ostream& os) const
{
	Instruction::disp(os);
	format(os);
}

void Draw::format(std::ostream& os) const
{
	os << "draw ";
	_expression->format(os);
	os << std::endl;
}

void Draw::definition_fonction()
{
	_expression->definition_fonction();
}

InstructionSP Draw::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	BlocSP parent = clones_instances[_parent.lock().get()];
	if(not parent)
		parent = _parent.lock();
	DrawSP clone(new Draw(_position));
	clones_instances.ajout(this,clone);
	Traceur::Donnees::Expression::ExpressionSP clone_expression = _expression->clone(clones_instances);
	clone->_expression = clone_expression;
	clone->_parent = parent;
	return clone;
}
