#include <src/donnees/instructions/instructiondrawat.h>
#include <src/donnees/assembleur/assembleur.h>
#include <src/donnees/instructions/valeur.h>
#include <src/donnees/types/typenumber.h>
#include <src/donnees/types/typepoint.h>
#include <src/donnees/types/typeimage.h>
#include <src/donnees/types/typekernel.h>

using namespace Traceur::Donnees::Instruction;

DrawAt::DrawAt(const Traceur::Donnees::Expression::ExpressionSP& image, const Traceur::Donnees::Expression::ExpressionSP& point, const Traceur::Donnees::Expression::ExpressionSP& alpha, const Traceur::Donnees::Source::PositionSP& position)
:Instruction(position), _image(image), _point(point), _alpha(alpha)
{
	if(not (_image->est_de_type<Traceur::Donnees::Type::Image>() or _image->est_de_type<Traceur::Donnees::Type::Kernel>()))
		throw Traceur::Exception::TypeInvalide(_position,_image->type(),"draw");
	if(not (_point->est_de_type<Traceur::Donnees::Type::Point>()))
		throw Traceur::Exception::TypeInvalide(_position,_point->type(),"draw");
	if(not (_alpha->est_de_type<Traceur::Donnees::Type::Number>()))
		throw Traceur::Exception::TypeInvalide(_position,_alpha->type(),"draw");
}

void DrawAt::edition_liens(BlocSP bloc)
{
	Instruction::edition_liens(bloc);
	_image->edition_liens(this->shared_from_this());
	_point->edition_liens(this->shared_from_this());
	_alpha->edition_liens(this->shared_from_this());
}

InstructionInterruptionSP DrawAt::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_image = _image->evaluation(assembleur); if(not valeur_image->a_une_valeur_retour()) return *valeur_image;
	ValeurSP valeur_point = _point->evaluation(assembleur); if(not valeur_point->a_une_valeur_retour()) return *valeur_point;
	ValeurSP valeur_alpha = _alpha->evaluation(assembleur); if(not valeur_alpha->a_une_valeur_retour()) return *valeur_alpha;
	Traceur::Donnees::Expression::ExpressionSP expr_image = *valeur_image;
	Traceur::Donnees::Expression::ExpressionSP expr_point = *valeur_point;
	Traceur::Donnees::Expression::ExpressionSP expr_alpha = *valeur_alpha;
	try
	{
		assembleur->drawat(expr_image,expr_point,expr_alpha);
	}
	catch(ValeurSP valeur)
	{
		InstructionInterruptionSP interruption = *valeur;
		interruption->ajoute_position(_position);
		return interruption;
	}
	return InstructionInterruptionSP();
}

void DrawAt::disp(std::ostream& os) const
{
	Instruction::disp(os);
	format(os);
}

void DrawAt::format(std::ostream& os) const
{
	os << "draw ";
	_image->format(os);
	os << " at ";
	_point->format(os);
	os << " by ";
	_alpha->format(os);
	os << std::endl;
}

void DrawAt::definition_fonction()
{
	_image->definition_fonction();
	_point->definition_fonction();
	_alpha->definition_fonction();
}

InstructionSP DrawAt::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	BlocSP parent = clones_instances[_parent.lock().get()];
	if(not parent)
		parent = _parent.lock();
	DrawAtSP clone(new DrawAt(_position));
	clones_instances.ajout(this,clone);
	Traceur::Donnees::Expression::ExpressionSP clone_image = _image->clone(clones_instances);
	Traceur::Donnees::Expression::ExpressionSP clone_point = _point->clone(clones_instances);
	Traceur::Donnees::Expression::ExpressionSP clone_alpha = _alpha->clone(clones_instances);
	clone->_image = clone_image;
	clone->_point = clone_point;
	clone->_alpha = clone_alpha;
	clone->_parent = parent;
	return clone;
}
