#include <src/donnees/instructions/instructionerror.h>
#include <src/donnees/expressions/expressionerreur.h>
#include <src/donnees/assembleur/assembleur.h>
#include <algorithm>
#include <locale>
using namespace Traceur::Donnees::Instruction;
	Error::Error(const Traceur::Donnees::Source::PositionSP& position)
:InstructionInterruption(position)
{
}

bool Error::teste_retour_fonction(const Traceur::Donnees::Type::TypeSP& type) const
{
	return true;
}

Traceur::Donnees::Expression::ExpressionSP Error::expression_equivalente() const
{
	Traceur::Donnees::Expression::ExpressionSP erreur(new Traceur::Donnees::Expression::Erreur(_nom));
	return erreur;
}

void Error::disp(std::ostream& os) const
{
	Instruction::disp(os);
	format(os);
}

void Error::format(std::ostream& os) const
{
	os << "error " << _nom << std::endl;
}

void Error::edition_liens(BlocSP bloc)
{
}

InstructionInterruptionSP Error::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	InstructionInterruptionSP interruption = std::static_pointer_cast<InstructionInterruption>(this->shared_from_this());
	interruption->ajoute_position(_position);
	return interruption;
}

void Error::definition_fonction()
{
}

InstructionSP Error::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	BlocSP parent = clones_instances[_parent.lock().get()];
	if(not parent)
		parent = _parent.lock();
	ErrorSP clone(new Error(_nom,_position));
	clones_instances.ajout(this,clone);
	clone->_parent = parent;
	return clone;
}

char transforme(char c)
{
	if(c=='_')
		return ' ';
	return c;
}

Traceur::Exception::Execution Error::exception()
{
	std::string nom;
	std::transform(_nom.begin(),_nom.end(),std::back_inserter(nom),transforme);
	nom[0]=std::toupper(nom[0]);
	nom +=".";
	return Traceur::Exception::Execution(_position,nom);
}

std::pair<InstructionInterruption::Action,InstructionInterruptionSP> Error::comportement_boucle()
{
	InstructionInterruptionSP interruption(new Error(_nom,_position));
	return std::make_pair(InstructionInterruption::Action::SORTIE,interruption);
}

std::pair<InstructionInterruption::Action,InstructionInterruptionSP> Error::comportement_image()
{
	InstructionInterruptionSP interruption(new Error(_nom,_position));
	return std::make_pair(InstructionInterruption::Action::SORTIE,interruption);
}

ErrorSP Error::condition_peut_etre_trouvee(Traceur::Donnees::Source::PositionSP position)
{
	ErrorSP erreur(new Error("maybe_condition_found"));
	erreur->ajoute_position(position);
	return erreur;
}

ErrorSP Error::boucle_for_infinie_detectee(Traceur::Donnees::Source::PositionSP position)
{
	ErrorSP erreur(new Error("infinite_for_loop_detected"));
	erreur->ajoute_position(position);
	return erreur;
}
