#include <src/donnees/instructions/instructionexit.h>
#include <src/donnees/expressions/expressionvide.h>
#include <src/donnees/assembleur/assembleur.h>
#include <algorithm>
#include <locale>
using namespace Traceur::Donnees::Instruction;

bool Exit::teste_retour_fonction(const Traceur::Donnees::Type::TypeSP& type) const
{
	return true;
}

Traceur::Donnees::Expression::ExpressionSP Exit::expression_equivalente() const
{
	return Traceur::Donnees::Expression::Vide::SP();
}

void Exit::disp(std::ostream& os) const
{
	Instruction::disp(os);
	format(os);
}

void Exit::format(std::ostream& os) const
{
	os << "exit " << _niveau << std::endl;
}

void Exit::edition_liens(BlocSP bloc)
{
}

InstructionInterruptionSP Exit::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	InstructionInterruptionSP interruption = std::static_pointer_cast<InstructionInterruption>(this->shared_from_this());
	return interruption;
}

void Exit::definition_fonction()
{
}

InstructionSP Exit::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	BlocSP parent = clones_instances[_parent.lock().get()];
	if(not parent)
		parent = _parent.lock();
	ExitSP clone(new Exit(_niveau,_position));
	clones_instances.ajout(this,clone);
	clone->_parent = parent;
	return clone;
}

std::pair<InstructionInterruption::Action,InstructionInterruptionSP> Exit::comportement_boucle()
{
	switch(_niveau)
	{
		case 0:
			{
				InstructionInterruptionSP interruption(new Exit(_niveau));
				return std::make_pair(InstructionInterruption::Action::SORTIE,interruption);
			}
		case 1:
			return std::make_pair(InstructionInterruption::Action::SORTIE,InstructionInterruptionSP());
		default:
			{
				InstructionInterruptionSP interruption(new Exit(_niveau-1));
				return std::make_pair(InstructionInterruption::Action::SORTIE,interruption);
			}
	}
}

std::pair<InstructionInterruption::Action,InstructionInterruptionSP> Exit::comportement_image()
{
	InstructionInterruptionSP interruption(new Exit(_niveau));
	return std::make_pair(InstructionInterruption::Action::SORTIE,interruption);
}
