#include <src/donnees/instructions/instructionfor.h>
#include <src/donnees/instructions/instructionerror.h>
#include <src/donnees/instructions/instructionreturn.h>
#include <src/donnees/expressions/expressionentier.h>
#include <src/donnees/expressions/expressioninferieur.h>
#include <src/donnees/expressions/expressioninferieurouegal.h>
#include <src/donnees/expressions/expressionsuperieurouegal.h>
#include <src/donnees/expressions/expressionegalite.h>
#include <src/donnees/expressions/expressionaddition.h>
#include <src/donnees/expressions/expressionpoint.h>
#include <src/donnees/expressions/expressionvecteur.h>
#include <src/donnees/fonctions/instancenormale.h>
#include <src/donnees/fonctions/fonctionreelle.h>
#include <src/donnees/types/typenumber.h>
#include <src/donnees/types/typepoint.h>
#include <src/donnees/types/typevector.h>
#include <algorithm>

using namespace Traceur::Donnees::Instruction;

For::For(const Traceur::Donnees::Fonction::PrototypeSP indice_prototype, const Traceur::Donnees::Expression::ExpressionSP debut,
		const Traceur::Donnees::Expression::ExpressionSP fin, const Traceur::Donnees::Expression::ExpressionSP pas,
		const BlocSP& bloc, const Traceur::Donnees::Source::PositionSP& position)
:InstructionControle(position),_indice_prototype(indice_prototype),_debut(debut),_fin(fin),_pas(pas),_bloc(bloc),_type(Type::INVALIDE)
{
	Traceur::Donnees::Fonction::PrototypeSP prototype_nombre(new Traceur::Donnees::Fonction::Prototype(
				Traceur::Donnees::Type::Number::SP(),
				Traceur::Donnees::Fonction::Prototype::ListeParametres(),
				_indice_prototype->_nom_fonction,
				Traceur::Donnees::Fonction::Prototype::ListeParametres()));
	if(_indice_prototype->compatible(prototype_nombre))
	{
		_type = Type::NOMBRE;
		if(not _pas)
		{
			_pas = Traceur::Donnees::Expression::ExpressionSP(new Traceur::Donnees::Expression::Entier(1));
		}
		Traceur::Donnees::Type::TypeSP type_indice = _indice_prototype->_type_retour;
		if(not type_indice->compatible(_debut->type()))
		{
			throw Traceur::Exception::BorneInvalideBoucleFor(_position,type_indice,"from",_debut->type());
		}
		if(not type_indice->compatible(_fin->type()))
		{
			throw Traceur::Exception::BorneInvalideBoucleFor(_position,type_indice,"to",_fin->type());
		}
		if(pas and (not type_indice->compatible(_pas->type())))
		{
			throw Traceur::Exception::BorneInvalideBoucleFor(_position,type_indice,"by",_pas->type());
		}
	}
	Traceur::Donnees::Fonction::PrototypeSP prototype_point(new Traceur::Donnees::Fonction::Prototype(
				Traceur::Donnees::Type::Point::SP(),
				Traceur::Donnees::Fonction::Prototype::ListeParametres(),
				_indice_prototype->_nom_fonction,
				Traceur::Donnees::Fonction::Prototype::ListeParametres()));
	if(_indice_prototype->compatible(prototype_point))
	{
		_type = Type::POINT;
		if(not _pas)
		{
			Traceur::Donnees::Expression::EntierSP pas_num(new Traceur::Donnees::Expression::Entier(1));// a terme, VECTOR[xscl;yscl]
			_pas = Traceur::Donnees::Expression::ExpressionSP(new Traceur::Donnees::Expression::Vecteur(pas_num,pas_num));
		}
		if(not Traceur::Donnees::Type::Point::SP()->compatible(_debut->type()))
		{
			throw Traceur::Exception::BorneInvalideBoucleFor(_position,Traceur::Donnees::Type::Point::SP(),"from",_debut->type());
		}
		if(not Traceur::Donnees::Type::Point::SP()->compatible(_fin->type()))
		{
			throw Traceur::Exception::BorneInvalideBoucleFor(_position,Traceur::Donnees::Type::Point::SP(),"to",_fin->type());
		}
		if(pas and (not Traceur::Donnees::Type::Vector::SP()->compatible(_pas->type())))
		{
			throw Traceur::Exception::BorneInvalideBoucleFor(_position,Traceur::Donnees::Type::Vector::SP(),"by",_pas->type());
		}
	}
	if(_type==Type::INVALIDE)
		throw Traceur::Exception::PrototypeIndiceIncorrectBoucleFor(_position,_indice_prototype);
}

bool For::teste_retour_fonction(const Traceur::Donnees::Type::TypeSP& type) const
{
	return false;
}

void For::disp(std::ostream& os) const
{
	Instruction::disp(os);
	format(os);
}

void For::format(std::ostream& os) const
{
	os << "for ";
	_indice_prototype->format(os);
	os << " from " ;
	_debut->format(os);
	os << " to " ;
	_fin->format(os);
	os << " by " ;
	_pas->format(os);
#ifdef TRACEUR_DEBUG
	if(_indice_instance)
	{
		os << ": " ;
		_indice_instance->format(os);
	}
#endif
	os << std::endl << "do" << std::endl;
	_bloc->format(os);
	os << "end" << std::endl;
}

void For::edition_liens(BlocSP bloc)
{
	Instruction::edition_liens(bloc);
	_debut->edition_liens(this->shared_from_this());
	_fin->edition_liens(this->shared_from_this());
	_pas->edition_liens(this->shared_from_this());
	_bloc->cree_declarations(bloc);
	_indice_instance = Traceur::Donnees::Fonction::InstanceSP(new Traceur::Donnees::Fonction::InstanceNormale());
	_bloc->_declarations->ajoute_declaration(_indice_prototype,_indice_instance);
	_bloc->edition_liens(bloc,this->shared_from_this());
}


InstructionInterruptionSP For::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_debut = _debut->evaluation(assembleur);
	if(not valeur_debut->a_une_valeur_retour())
	{
		InstructionInterruptionSP interruption = *valeur_debut;
		interruption->ajoute_position(_position);
		return interruption;
	}

	ValeurSP valeur_fin = _fin->evaluation(assembleur);
	if(not valeur_fin->a_une_valeur_retour())
	{
		InstructionInterruptionSP interruption = *valeur_fin;
		interruption->ajoute_position(_position);
		return interruption;
	}

	ValeurSP valeur_pas = _pas->evaluation(assembleur);
	if(not valeur_pas->a_une_valeur_retour())
	{
		InstructionInterruptionSP interruption = *valeur_pas;
		interruption->ajoute_position(_position);
		return interruption;
	}

	Traceur::Donnees::Expression::ExpressionSP indice_debut_generique = *valeur_debut;
	Traceur::Donnees::Expression::ExpressionSP indice_fin_generique = *valeur_fin;
	Traceur::Donnees::Expression::ExpressionSP indice_pas_generique = *valeur_pas;
	Sens sens_externe = Sens::INVALIDE;
	Traceur::Donnees::Expression::NombreSP debut_externe;
	Traceur::Donnees::Expression::NombreSP fin_externe;
	Traceur::Donnees::Expression::NombreSP pas_externe;
	Sens sens_interne = Sens::INVALIDE;
	Traceur::Donnees::Expression::NombreSP debut_interne;
	Traceur::Donnees::Expression::NombreSP fin_interne;
	Traceur::Donnees::Expression::NombreSP pas_interne;

	switch(_type)
	{
		case Type::NOMBRE:
			{
				debut_externe = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(indice_debut_generique);
				TRACEUR_TEST(debut_externe);
				fin_externe = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(indice_fin_generique);
				TRACEUR_TEST(fin_externe);
				pas_externe = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Nombre>(indice_pas_generique);
				TRACEUR_TEST(pas_externe);
				if(pas_externe==Traceur::Donnees::Expression::Nombre::zero())
					return Error::boucle_for_infinie_detectee(_position);
				if(pas_externe<Traceur::Donnees::Expression::Nombre::zero())
					sens_externe = Sens::NEGATIF;
				else
					sens_externe = Sens::POSITIF;
				break;
			}
		case Type::POINT:
			{
				Traceur::Donnees::Expression::PointSP debut = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Point>(indice_debut_generique);
				TRACEUR_TEST(debut);
				debut->en_cartesien(debut_externe,debut_interne);
				Traceur::Donnees::Expression::PointSP fin = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Point>(indice_fin_generique);
				TRACEUR_TEST(fin);
				fin->en_cartesien(fin_externe,fin_interne);
				Traceur::Donnees::Expression::VecteurSP pas = std::dynamic_pointer_cast<Traceur::Donnees::Expression::Vecteur>(indice_pas_generique);
				TRACEUR_TEST(pas);
				pas->en_cartesien(pas_externe,pas_interne);
				if((pas_externe==Traceur::Donnees::Expression::Nombre::zero()) or (pas_interne==Traceur::Donnees::Expression::Nombre::zero()))
					return Error::boucle_for_infinie_detectee(_position);
				if(pas_externe<Traceur::Donnees::Expression::Nombre::zero())
					sens_externe = Sens::NEGATIF;
				else
					sens_externe = Sens::POSITIF;
				if(pas_interne<Traceur::Donnees::Expression::Nombre::zero())
					sens_interne = Sens::NEGATIF;
				else
					sens_interne = Sens::POSITIF;
				break;
			}
		default:
			throw;
	}

	Traceur::Donnees::Expression::NombreSP indice_externe = debut_externe;
	Traceur::Donnees::Expression::NombreSP indice_interne = debut_interne;
	Traceur::Donnees::Expression::ExpressionSP indice_generique;
	switch(sens_interne)
	{
		case Sens::POSITIF:
			if(not (indice_interne<=fin_interne))
				return InstructionInterruptionSP();
			break;
		case Sens::NEGATIF:
			if(not (indice_interne>=fin_interne))
				return InstructionInterruptionSP();
			break;
		default: // type nombre
			break;
	}
	for(;;)
	{
		switch(sens_externe)
		{
			case Sens::POSITIF:
				if(not (indice_externe<=fin_externe))
					return InstructionInterruptionSP();
				break;
			case Sens::NEGATIF:
				if(not (indice_externe>=fin_externe))
					return InstructionInterruptionSP();
				break;
			default: throw;
		}
		switch(_type)
		{
			case Type::NOMBRE:
				indice_generique = indice_externe;
				break;
			case Type::POINT:
				indice_generique = Traceur::Donnees::Expression::PointSP(new Traceur::Donnees::Expression::Point(indice_externe,indice_interne));
				break;
			default:
				throw;
		}

		ReturnSP retour(new Return(indice_generique));
		Traceur::Donnees::Fonction::FonctionReelleSP fonction_indice(new Traceur::Donnees::Fonction::FonctionReelle(_indice_prototype,retour,_position));
		_indice_instance->valeur_dynamique(fonction_indice);

		InstructionInterruptionSP interruption = _bloc->evaluation(assembleur);
		if(interruption)
		{
			std::pair<InstructionInterruption::Action,InstructionInterruptionSP> comportement = interruption->comportement_boucle();
			switch(comportement.first)
			{
				case InstructionInterruption::Action::RIEN:
				case InstructionInterruption::Action::SUIVANT:
					break;
				case InstructionInterruption::Action::SORTIE:
					if(comportement.second)
						comportement.second->ajoute_position(_position);
					return comportement.second;
					break;
			}
		}
		switch(_type)
		{
			case Type::NOMBRE:
				indice_externe = indice_externe+pas_externe;
				break;
			case Type::POINT:
				indice_interne = indice_interne+pas_interne;
				switch(sens_interne)
				{
					case Sens::POSITIF:
						if(not (indice_interne<=fin_interne))
						{
							indice_interne = debut_interne;
							indice_externe = indice_externe+pas_externe;
						}
						break;
					case Sens::NEGATIF:
						if(not (indice_interne>=fin_interne))
						{
							indice_interne = debut_interne;
							indice_externe = indice_externe+pas_externe;
						}
						break;
					default: throw;
				}
				break;
			default:
				throw;
		}
	}
	return InstructionInterruptionSP();
}

void For::definition_fonction()
{
	_debut->definition_fonction();
	_fin->definition_fonction();
	_pas->definition_fonction();
	_bloc->definition_fonction();
}

InstructionSP For::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	BlocSP parent = clones_instances[_parent.lock().get()];
	if(not parent)
		parent=_parent.lock();
	ForSP clone(new For(_position));
	clone->_parent=parent;
	clones_instances.ajout(this,clone);
	clone->_type = _type;
	clone->_indice_prototype = _indice_prototype->clone(clones_instances);
	clone->_indice_instance = _indice_instance->clone(clones_instances);
	clone->_debut = _debut->clone(clones_instances);
	clone->_fin = _fin->clone(clones_instances);
	clone->_pas = _pas->clone(clones_instances);
	clone->_bloc = _bloc->clone(clones_instances);
	clone->_parent = parent;
	return clone;
}
