#ifndef _DONNEES_INSTRUCTIONS_INSTRUCTIONFOR_H_
#define _DONNEES_INSTRUCTIONS_INSTRUCTIONFOR_H_
#include <deque>

#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Instruction { DECL_SHARED_PTR(For); } } }

#include <src/donnees/instructions/instructioncontrole.h>
#include <src/donnees/fonctions/fonctionprototype.h>
#include <src/donnees/expressions/expression.h>
#include <src/exceptions/exceptions.h>
namespace Traceur
{
	namespace Donnees
	{
		namespace Instruction
		{
			class For : public InstructionControle
			{
				public:
					For(const Traceur::Donnees::Fonction::PrototypeSP indice_prototype,
						const Traceur::Donnees::Expression::ExpressionSP debut,
						const Traceur::Donnees::Expression::ExpressionSP fin,
						const Traceur::Donnees::Expression::ExpressionSP pas,
						const BlocSP& bloc,
						const Traceur::Donnees::Source::PositionSP& position = Traceur::Donnees::Source::PositionSP());
					virtual bool teste_retour_fonction(const Traceur::Donnees::Type::TypeSP& type) const;
					virtual void edition_liens(BlocSP bloc);
					virtual void definition_fonction();
					virtual InstructionSP clone(Traceur::Donnees::Global::ClonesInstances& clones_instances);
					virtual InstructionInterruptionSP evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur);
					virtual void disp(std::ostream& os) const;
					virtual void format(std::ostream& os) const;
					virtual ~For() {};
				private:
					For(const Traceur::Donnees::Source::PositionSP& position = Traceur::Donnees::Source::PositionSP()):InstructionControle(position) {};
					Traceur::Donnees::Fonction::PrototypeSP _indice_prototype;
					Traceur::Donnees::Fonction::InstanceSP _indice_instance;
					Traceur::Donnees::Expression::ExpressionSP _debut;
					Traceur::Donnees::Expression::ExpressionSP _fin;
					Traceur::Donnees::Expression::ExpressionSP _pas;
					BlocSP _bloc;
					enum class Type {INVALIDE,NOMBRE,POINT};
					Type _type;
					enum class Sens {INVALIDE,POSITIF,NEGATIF};
			};
		}
	}
}

namespace Traceur
{
	namespace Exception
	{
		class PrototypeIndiceIncorrectBoucleFor : public Compilation
		{
			public:
				PrototypeIndiceIncorrectBoucleFor(Traceur::Donnees::Source::PositionSP position, const Traceur::Donnees::Fonction::PrototypeSP prototype)
					:Compilation(position,std::string("Prototype ")+std::string(*prototype)+std::string(" incorrect in for loop as loop index."))
					{
					};
		};
		class BorneInvalideBoucleFor : public Compilation
		{
			public:
				BorneInvalideBoucleFor(Traceur::Donnees::Source::PositionSP position, const Traceur::Donnees::Type::TypeSP indice, const std::string nom_borne, const Traceur::Donnees::Type::TypeSP borne)
					:Compilation(position,std::string("Incorrect type ")+std::string(*borne)+std::string(" in ")+nom_borne+std::string(" part of for loop. Type compatible with ")+std::string(*indice)+std::string(" expected."))
					{
					};
		};
	}
}
#endif
