#include <src/donnees/instructions/instructionlet.h>
#include <src/donnees/fonctions/fonctionreelle.h>
#include <src/donnees/expressions/expressionfonctionimaginaire.h>
#include <src/donnees/instructions/instructionreturn.h>
using namespace Traceur::Donnees::Instruction;
void Let::disp(std::ostream& os) const
{
	Instruction::disp(os);
	format(os);
}

void Let::format(std::ostream& os) const
{
	os << "let ";
	if(_fonction->_valeur_statique->recursive())
		os << "rec ";
	_fonction->prototype()->format(os);
#ifdef TRACEUR_DEBUG
	if(_fonction->_instance)
	{
		os << ": ";
		_fonction->_instance->format(os);
	}
	os << std::endl;
#else
	os << " := do" << std::endl;
#endif
	_fonction->_valeur_statique->format(os);
#ifndef TRACEUR_DEBUG
	os << "end" << std::endl;
#endif
}

LetFunc::LetFunc(const Traceur::Donnees::Fonction::PrototypeSP& prototype,
		const Traceur::Donnees::Instruction::InstructionSP& instruction,
		const Traceur::Donnees::Source::PositionSP& position)
:Let(position,prototype)
{
	Traceur::Donnees::Type::TypeSP type_retour(new Traceur::Donnees::Type::Fonction(prototype));
	Traceur::Donnees::Fonction::PrototypeSP vrai_prototype(new Traceur::Donnees::Fonction::Prototype(type_retour,
		Traceur::Donnees::Fonction::Prototype::ListeParametres(),
		Traceur::Donnees::Fonction::IDENTIFIANTVIDE(),
		Traceur::Donnees::Fonction::Prototype::ListeParametres()));
	Traceur::Donnees::Fonction::FonctionSP fonction(new Traceur::Donnees::Fonction::FonctionReelle(vrai_prototype,instruction,position));
	_fonction->_valeur_statique = fonction;
	_fonction->_position = position;
}

void LetFunc::disp(std::ostream& os) const
{
	Instruction::disp(os);
	format(os);
}

void LetFunc::format(std::ostream& os) const
{
	os << "let func ";
	_fonction->prototype()->format(os);
#ifdef TRACEUR_DEBUG
	if(_fonction->_instance)
	{
		os << ": ";
		_fonction->_instance->format(os);
	}
	os << std::endl;
#else
	os << " := do" << std::endl;
#endif
	_fonction->_valeur_statique->format(os);
#ifndef TRACEUR_DEBUG
	os << "end" << std::endl;
#endif
}

void Let::edition_liens(BlocSP bloc)
{
	Instruction::edition_liens(bloc);
	_fonction->_valeur_statique->edition_liens(bloc);
	_fonction->_instance = bloc->_declarations->definition(_fonction->prototype(),_fonction->_valeur_statique,_definition_locale,_position);
}

InstructionInterruptionSP Let::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	_fonction->definition_fonction();
	return InstructionInterruptionSP();
}

InstructionInterruptionSP LetFunc::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	Traceur::Donnees::Fonction::Fonction::ListeParametres gauche;
	Traceur::Donnees::Fonction::Fonction::ListeParametres droite;

	Traceur::Donnees::Expression::AppelFonctionUtilisateurSP appel(new Traceur::Donnees::Expression::AppelFonctionUtilisateur(_position,this->shared_from_this()));
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE << this->shared_from_this() << std::endl;
#endif
	InstructionInterruptionSP valeur = _fonction->_valeur_statique->appel(gauche,droite,assembleur,appel);
	ReturnSP retour = std::dynamic_pointer_cast<Return>(valeur);
	if(not retour)
		return valeur;
	Traceur::Donnees::Expression::ExpressionSP expression=retour->expression_equivalente();
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE << expression << std::endl;
#endif
	Traceur::Donnees::Expression::FonctionImaginaireSP fonction = std::dynamic_pointer_cast<Traceur::Donnees::Expression::FonctionImaginaire>(expression);
	TRACEUR_TEST(fonction);
	_fonction->_instance->valeur_dynamique(fonction->_fonction);
	return InstructionInterruptionSP();
}

InstructionSP Let::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	BlocSP parent = clones_instances[_parent.lock().get()];
	if(not parent)
		parent = _parent.lock();
	LetSP clone(new Let(_position,_fonction->prototype()->clone(clones_instances)));
	clones_instances.ajout(this,clone);
	Traceur::Donnees::Fonction::FonctionSP fonction = _fonction->clone(clones_instances);
	Traceur::Donnees::Fonction::FonctionImaginaireSP fonction_imag = std::dynamic_pointer_cast<Traceur::Donnees::Fonction::FonctionImaginaire>(fonction);
	TRACEUR_TEST(fonction_imag);
	clone->_fonction=fonction_imag;
	return clone;
}

InstructionSP LetFunc::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	BlocSP parent = clones_instances[_parent.lock().get()];
	if(not parent)
		parent = _parent.lock();
	LetFuncSP clone(new LetFunc(_fonction->prototype()->clone(clones_instances),_position));
	clones_instances.ajout(this,clone);
	Traceur::Donnees::Fonction::FonctionSP fonction = _fonction->clone(clones_instances);
	Traceur::Donnees::Fonction::FonctionImaginaireSP fonction_imag = std::dynamic_pointer_cast<Traceur::Donnees::Fonction::FonctionImaginaire>(fonction);
	TRACEUR_TEST(fonction_imag);
	clone->_fonction=fonction_imag;
	clone->_parent = parent;
	return clone;
}
