#include <src/donnees/instructions/instructionread.h>
#include <src/donnees/assembleur/assembleur.h>
#include <src/donnees/instructions/valeur.h>
#include <src/donnees/types/typestring.h>

using namespace Traceur::Donnees::Instruction;

Read::Read(const Traceur::Donnees::Expression::ExpressionSP& nom_fichier, const Traceur::Donnees::Source::PositionSP& position)
:Instruction(position), _nom_fichier(nom_fichier)
{
	if(not (_nom_fichier->est_de_type<Traceur::Donnees::Type::String>()))
		throw Traceur::Exception::TypeInvalide(position,_nom_fichier->type(),"read");
}

void Read::edition_liens(BlocSP bloc)
{
	Instruction::edition_liens(bloc);
	_nom_fichier->edition_liens(this->shared_from_this());
}

InstructionInterruptionSP Read::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_nom_fichier = _nom_fichier->evaluation(assembleur);
	if(not valeur_nom_fichier->a_une_valeur_retour())
		return *valeur_nom_fichier;
	Traceur::Donnees::Expression::ExpressionSP expr_nom_fichier = *valeur_nom_fichier;
	try
	{
		assembleur->read(expr_nom_fichier);
	}
	catch(ValeurSP valeur)
	{
		InstructionInterruptionSP interruption = *valeur;
		interruption->ajoute_position(_position);
		return interruption;
	}
	return InstructionInterruptionSP();
}

void Read::disp(std::ostream& os) const
{
	Instruction::disp(os);
	format(os);
}

void Read::format(std::ostream& os) const
{
	os << "read from ";
	_nom_fichier->format(os);
	os << std::endl;
}

void Read::definition_fonction()
{
	_nom_fichier->definition_fonction();
}

InstructionSP Read::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	BlocSP parent = clones_instances[_parent.lock().get()];
	if(not parent)
		parent = _parent.lock();
	ReadSP clone(new Read(_position));
	clones_instances.ajout(this,clone);
	Traceur::Donnees::Expression::ExpressionSP clone_nom_fichier = _nom_fichier->clone(clones_instances);
	clone->_nom_fichier = clone_nom_fichier;
	clone->_parent = parent;
	return clone;
}
