#include <src/donnees/instructions/instructionreturn.h>
#include <src/donnees/assembleur/assembleur.h>
#include <src/donnees/expressions/expressionfonctionimaginaire.h>
#include <src/donnees/expressions/expressionvide.h>
using namespace Traceur::Donnees::Instruction;
	Return::Return(const Traceur::Donnees::Source::PositionSP& position)
:InstructionInterruption(position)
{
	Traceur::Donnees::Expression::ExpressionSP expression(new Traceur::Donnees::Expression::Vide());
	_valeur = expression;
}

Return::Return(const Traceur::Donnees::Fonction::FonctionImaginaireSP& valeur, const Traceur::Donnees::Source::PositionSP& position)
:InstructionInterruption(position)
{
	Traceur::Donnees::Expression::ExpressionSP expression(new Traceur::Donnees::Expression::FonctionImaginaire(valeur));
	_valeur = expression;
}

bool Return::teste_retour_fonction(const Traceur::Donnees::Type::TypeSP& type) const
{
	// le type retourne doit etre compatible avec le type demande
	if(not (type->compatible(_valeur->type())))
	{
		throw Traceur::Exception::Retour_invalide(_position,type,_valeur->type());
	}
	return true;
}

void Return::disp(std::ostream& os) const
{
	Instruction::disp(os);
	format(os);
}

void Return::format(std::ostream& os) const
{
	//bool fonction = false;
	//if(_valeur->est_de_nature<Traceur::Donnees::Expression::FonctionImaginaire>())
	//	fonction=true;
	os << "return ";
	_valeur->format(os);
	os << std::endl;
}

void Return::edition_liens(BlocSP bloc)
{
	Instruction::edition_liens(bloc);
	_valeur->edition_liens(this->shared_from_this());
}

InstructionInterruptionSP Return::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	// la valeur de l'expression est deja dans une instruction return
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE;
#endif
	InstructionInterruptionSP interruption = *(_valeur->evaluation(assembleur));
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE;
	interruption->format(std::cerr);
#endif
	interruption->ajoute_position(_position);
	return interruption;
}

void Return::definition_fonction()
{
	_valeur->definition_fonction();
}

void Return::simplification()
{
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE << "Avant: ";	_valeur->format(std::cerr); std::cerr << std::endl;
#endif
	_valeur = _valeur->simplification();
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE << "Apres: ";	_valeur->format(std::cerr); std::cerr << std::endl;
#endif
}

InstructionSP Return::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	BlocSP parent = clones_instances[_parent.lock().get()];
	if(not parent)
		parent = _parent.lock();
	ReturnSP clone(new Return(_position));
	clones_instances.ajout(this,clone);
	Traceur::Donnees::Expression::ExpressionSP clone_valeur = _valeur->clone(clones_instances);
	clone->_valeur = clone_valeur;
	clone->_parent = parent;
	return clone;
}

std::pair<InstructionInterruption::Action,InstructionInterruptionSP> Return::comportement_boucle()
{
	InstructionInterruptionSP interruption(new Return(_valeur));
	return std::make_pair(InstructionInterruption::Action::SORTIE,interruption);
}

std::pair<InstructionInterruption::Action,InstructionInterruptionSP> Return::comportement_image()
{
	InstructionInterruptionSP interruption(new Return(_valeur));
	return std::make_pair(InstructionInterruption::Action::SORTIE,interruption);
}
