#ifndef _DONNEES_INSTRUCTIONS_INSTRUCTIONRETURN_H_
#define _DONNEES_INSTRUCTIONS_INSTRUCTIONRETURN_H_

#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Instruction { DECL_SHARED_PTR(Return); } } }

#include <src/donnees/instructions/instructioninterruption.h>
#include <src/exceptions/exceptions.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Instruction
		{
			class Valeur;
			class Bloc;
			class Return : public InstructionInterruption
			{
				friend class Valeur;
				friend class Bloc;
				public:
					Return(const Traceur::Donnees::Source::PositionSP& position = Traceur::Donnees::Source::PositionSP());
					Return(const Traceur::Donnees::Expression::ExpressionSP& valeur,
						const Traceur::Donnees::Source::PositionSP& position = Traceur::Donnees::Source::PositionSP())
						:InstructionInterruption(position),_valeur(valeur) {};
					Return(const Traceur::Donnees::Fonction::FonctionImaginaireSP& valeur,
						const Traceur::Donnees::Source::PositionSP& position = Traceur::Donnees::Source::PositionSP());
					virtual bool teste_retour_fonction(const Traceur::Donnees::Type::TypeSP& type) const;
					virtual void edition_liens(BlocSP bloc);
					virtual void definition_fonction();
					virtual void simplification();
					virtual InstructionSP clone(Traceur::Donnees::Global::ClonesInstances& clones_instances);
					virtual InstructionInterruptionSP evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur);
					virtual Traceur::Donnees::Expression::ExpressionSP expression_equivalente() const { return _valeur; };
					virtual std::pair<Action,InstructionInterruptionSP> comportement_boucle();
					virtual std::pair<Action,InstructionInterruptionSP> comportement_image();
					virtual void disp(std::ostream& os) const;
					virtual void format(std::ostream& os) const;
					virtual ~Return() {};
				private:
					Traceur::Donnees::Expression::ExpressionSP _valeur;

			};
		}
	}
}

namespace Traceur
{
	namespace Exception
	{
		class Retour_invalide : public Compilation
		{
			public:
				Retour_invalide(Traceur::Donnees::Source::PositionSP position, const Traceur::Donnees::Type::TypeSP& type_attendu, const Traceur::Donnees::Type::TypeSP& type_recu)
					:Compilation(position,std::string("Return type ")+
						std::string(*type_recu)+
						std::string(" invalid. Type compatible with ")+
						std::string(*type_attendu)+
						std::string(" expected.")) {};
				virtual ~Retour_invalide() {};
		};
	}
}
#endif
