#include <src/donnees/instructions/instructiontry.h>
#include <src/donnees/instructions/instructionerror.h>
#include <src/donnees/types/typeboolean.h>
#include <src/donnees/expressions/expressionbooleenvrai.h>
#include <src/donnees/expressions/expressionbooleenpeutetre.h>
#include <algorithm>

using namespace Traceur::Donnees::Instruction;

Try::Try(BlocSP essai, ClausesCatch clauses_catch,
		const Traceur::Donnees::Source::PositionSP& position)
:InstructionControle(position),_essai(essai),_clauses_catch(clauses_catch)
{
	TRACEUR_TEST(_clauses_catch.size()!=0);
	if(_clauses_catch.back().first.size()==0)
	{
		_clause_totale=_clauses_catch.back().second;
		_clauses_catch.pop_back();
	}
}

bool Try::teste_retour_fonction(const Traceur::Donnees::Type::TypeSP& type) const
{
	return false;
}

struct AfficheCatch
{
	AfficheCatch(std::ostream& os)
	:_os(os) {};
	void operator()(Try::ClauseCatch clause_catch)
	{
		_os << "catch (";
		auto it = clause_catch.first.cbegin();
		_os << *it;
		for(++it ; it!=clause_catch.first.cend() ; ++it)
		{
			_os << ", " << *it;
		}
		_os << ")" << std::endl;
		clause_catch.second->format(_os);
	};
	private:
		std::ostream& _os;
};

void Try::disp(std::ostream& os) const
{
	Instruction::disp(os);
	format(os);
}

void Try::format(std::ostream& os) const
{
	os << "try " << std::endl;
	_essai->format(os);

	std::for_each(_clauses_catch.begin(),_clauses_catch.end(),AfficheCatch(os));
	if(_clause_totale)
	{
		os << "catch" << std::endl;
		_clause_totale->format(os);
	}
	os << "end" << std::endl;
}

struct EditionLiens
{
	EditionLiens(InstructionSP instruction, BlocSP bloc)
		:_instruction(instruction),_bloc(bloc) {};
	void operator() (Try::ClauseCatch clause_catch)
	{
		clause_catch.second->edition_liens(_bloc,_instruction);
	};
	private:
		InstructionSP _instruction;
		BlocSP _bloc;
};

void Try::edition_liens(BlocSP bloc)
{
	Instruction::edition_liens(bloc);
	_essai->edition_liens(bloc,this->shared_from_this());
	std::for_each(_clauses_catch.begin(),_clauses_catch.end(),EditionLiens(this->shared_from_this(),bloc));
	if(_clause_totale)
		_clause_totale->edition_liens(bloc,this->shared_from_this());
}


InstructionInterruptionSP Try::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	InstructionInterruptionSP interruption_essai = _essai->evaluation(assembleur);
	ErrorSP erreur = std::dynamic_pointer_cast<Error>(interruption_essai);
	if(not erreur)
		return interruption_essai;
	Traceur::Donnees::Fonction::Identifiant nom = erreur->_nom;
	for(auto iter=_clauses_catch.begin() ; iter!=_clauses_catch.end() ; ++iter)
	{
		if(std::find(iter->first.begin(),iter->first.end(),nom)!=iter->first.end())
		{
			InstructionInterruptionSP interruption = iter->second->evaluation(assembleur);
			return interruption;
		}
	}
	if(_clause_totale)
	{
		InstructionInterruptionSP interruption = _clause_totale->evaluation(assembleur);
		return interruption;
	}
	return interruption_essai;
}

struct DefinitionFonction
{
	void operator() (Try::ClauseCatch clause_catch)
	{
		clause_catch.second->definition_fonction();
	};
};

void Try::definition_fonction()
{
	_essai->definition_fonction();
	std::for_each(_clauses_catch.begin(),_clauses_catch.end(),DefinitionFonction());
	if(_clause_totale)
		_clause_totale->definition_fonction();
}

struct CloneClauseCatch
{
	CloneClauseCatch(Traceur::Donnees::Global::ClonesInstances& clones_instances)
	:_clones_instances(clones_instances) {};
	Try::ClauseCatch operator() (Try::ClauseCatch clause_catch)
	{
		Try::ClauseCatch clone;
		clone.first = clause_catch.first;
		clone.second = clause_catch.second->clone(_clones_instances);
		return clone;
	};
	private:
		Traceur::Donnees::Global::ClonesInstances& _clones_instances;
};

InstructionSP Try::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	BlocSP parent = clones_instances[_parent.lock().get()];
	if(not parent)
		parent=_parent.lock();
	TrySP clone(new Try(_position));
	clone->_parent=parent;
	clones_instances.ajout(this,clone);
	clone->_essai = _essai->clone(clones_instances);
	std::transform(_clauses_catch.begin(),_clauses_catch.end(),std::back_inserter(clone->_clauses_catch),CloneClauseCatch(clones_instances));
	if(_clause_totale)
		clone->_clause_totale = _clause_totale->clone(clones_instances);
	clone->_parent = parent;
	return clone;
}
