#include <src/donnees/instructions/instructionwhile.h>
#include <src/donnees/instructions/instructionerror.h>
#include <src/donnees/instructions/instructionif.h>
#include <src/donnees/types/typeboolean.h>
#include <src/donnees/expressions/expressionbooleenfaux.h>
#include <src/donnees/expressions/expressionbooleenpeutetre.h>
#include <algorithm>

using namespace Traceur::Donnees::Instruction;

	While::While(const Traceur::Donnees::Expression::ExpressionSP condition, const BlocSP& bloc,const Traceur::Donnees::Source::PositionSP& position)
:InstructionControle(position),_condition(condition),_bloc(bloc)
{
	if(not condition->est_de_type<Traceur::Donnees::Type::Boolean>())
	{
		throw Traceur::Exception::ConditionNonBooleenne(_position,condition->type());
	}
}

bool While::teste_retour_fonction(const Traceur::Donnees::Type::TypeSP& type) const
{
	return false;
}

void While::disp(std::ostream& os) const
{
	Instruction::disp(os);
	format(os);
}

void While::format(std::ostream& os) const
{
	os << "while ";
	_condition->format(os);
	os << std::endl << "do" << std::endl;
	_bloc->format(os);
	os << "end" << std::endl;
}

void While::edition_liens(BlocSP bloc)
{
	Instruction::edition_liens(bloc);
	_condition->edition_liens(this->shared_from_this());
	_bloc->edition_liens(bloc,this->shared_from_this());
}


InstructionInterruptionSP While::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	for(;;)
	{
		ValeurSP condition = _condition->evaluation(assembleur);
		if(not condition->a_une_valeur_retour())
		{
			InstructionInterruptionSP interruption = *condition;
			interruption->ajoute_position(_position);
			return interruption;
		}
		Traceur::Donnees::Expression::ExpressionSP expr_condition = *condition;
		if(expr_condition->est_de_nature<Traceur::Donnees::Expression::BooleenFaux>())
		{
			return InstructionInterruptionSP();	
		}
		if(expr_condition->est_de_nature<Traceur::Donnees::Expression::BooleenPeutetre>())
		{
			return Error::condition_peut_etre_trouvee(_position);
		}
		InstructionInterruptionSP interruption = _bloc->evaluation(assembleur);
		if(interruption)
		{
			std::pair<InstructionInterruption::Action,InstructionInterruptionSP> comportement = interruption->comportement_boucle();
			switch(comportement.first)
			{
				case InstructionInterruption::Action::RIEN:
				case InstructionInterruption::Action::SUIVANT:
					break;
				case InstructionInterruption::Action::SORTIE:
					if(comportement.second)
						comportement.second->ajoute_position(_position);
					return comportement.second;
					break;
			}
		}
	}
	return InstructionInterruptionSP();
}

void While::definition_fonction()
{
	_condition->definition_fonction();
	_bloc->definition_fonction();
}

InstructionSP While::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	BlocSP parent = clones_instances[_parent.lock().get()];
	if(not parent)
		parent=_parent.lock();
	WhileSP clone(new While(_position));
	clone->_parent=parent;
	clones_instances.ajout(this,clone);
	clone->_condition = _condition->clone(clones_instances);
	clone->_bloc = _bloc->clone(clones_instances);
	clone->_parent = parent;
	return clone;
}
