#include <src/donnees/instructions/instructionwith.h>
#include <src/donnees/assembleur/assembleur.h>
#include <src/donnees/instructions/valeur.h>
#include <src/donnees/types/typecolor.h>

using namespace Traceur::Donnees::Instruction;

With::With(const Traceur::Donnees::Expression::ExpressionSP& expression, const Traceur::Donnees::Source::PositionSP& position)
:Instruction(position), _expression(expression)
{
	if(_expression->est_de_type<Traceur::Donnees::Type::Color>())
		return;
	throw Traceur::Exception::TypeInvalide(_position,_expression->type(),"with");
}

With::With(const Traceur::Donnees::Expression::ExpressionSP& expression, const BlocSP& bloc, const Traceur::Donnees::Source::PositionSP& position)
:Instruction(position), _expression(expression), _bloc(bloc)
{
	if(_expression->est_de_type<Traceur::Donnees::Type::Color>())
		return;
	throw Traceur::Exception::TypeInvalide(_position,_expression->type(),"with");
}

void With::edition_liens(BlocSP bloc)
{
	Instruction::edition_liens(bloc);
	_expression->edition_liens(this->shared_from_this());
	if(_bloc)
		_bloc->edition_liens(bloc,this->shared_from_this());
}

InstructionInterruptionSP With::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur = _expression->evaluation(assembleur);
	if(valeur->a_une_valeur_retour())
	{
		if(not _bloc)
		{
			Traceur::Donnees::Expression::ExpressionSP expr_valeur = *valeur;
			assembleur->with(expr_valeur);
			return InstructionInterruptionSP();
		}
		else
		{
			Traceur::Donnees::Assembleur::Couleur::Couleur couleur = assembleur->couleur_courante();
			Traceur::Donnees::Expression::ExpressionSP expr_valeur = *valeur;
			assembleur->with(expr_valeur);
			InstructionInterruptionSP interruption = _bloc->evaluation(assembleur);
			assembleur->with(couleur);
			return interruption;
		}
	}
	return *valeur;
}

void With::disp(std::ostream& os) const
{
	Instruction::disp(os);
	format(os);
}

void With::format(std::ostream& os) const
{
	os << "with ";
	_expression->format(os);
	if(_bloc)
	{
		os << std::endl << "{" << std::endl;
		_bloc->format(os);
		os << "}" << std::endl;
	}
	os << std::endl;
}

void With::definition_fonction()
{
	_expression->definition_fonction();
}

InstructionSP With::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	BlocSP parent = clones_instances[_parent.lock().get()];
	if(not parent)
		parent = _parent.lock();
	WithSP clone(new With(_position));
	clones_instances.ajout(this,clone);
	Traceur::Donnees::Expression::ExpressionSP clone_expression = _expression->clone(clones_instances);
	clone->_expression = clone_expression;
	clone->_parent = parent;
	if(_bloc)
		clone->_bloc = _bloc->clone(clones_instances);
	return clone;
}
