#include <src/donnees/instructions/instructionwrite.h>
#include <src/donnees/assembleur/assembleur.h>
#include <src/donnees/instructions/valeur.h>
#include <src/donnees/types/typestring.h>
#include <src/donnees/types/typeimage.h>

using namespace Traceur::Donnees::Instruction;

Write::Write(const Traceur::Donnees::Expression::ExpressionSP& image, const Traceur::Donnees::Expression::ExpressionSP& nom_fichier, const Traceur::Donnees::Source::PositionSP& position)
:Instruction(position), _image(image), _nom_fichier(nom_fichier)
{
	if(not (_image->est_de_type<Traceur::Donnees::Type::Image>()))
		throw Traceur::Exception::TypeInvalide(_position,_image->type(),"write");
	if(not (_nom_fichier->est_de_type<Traceur::Donnees::Type::String>()))
		throw Traceur::Exception::TypeInvalide(_position,_nom_fichier->type(),"write");
}

void Write::edition_liens(BlocSP bloc)
{
	Instruction::edition_liens(bloc);
	_image->edition_liens(this->shared_from_this());
	_nom_fichier->edition_liens(this->shared_from_this());
}

InstructionInterruptionSP Write::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	ValeurSP valeur_image = _image->evaluation(assembleur);
	if(not valeur_image->a_une_valeur_retour())
		return *valeur_image;
	ValeurSP valeur_nom_fichier = _nom_fichier->evaluation(assembleur);
	if(not valeur_nom_fichier->a_une_valeur_retour())
		return *valeur_nom_fichier;
	Traceur::Donnees::Expression::ExpressionSP expr_image = *valeur_image;
	Traceur::Donnees::Expression::ExpressionSP expr_nom_fichier = *valeur_nom_fichier;
	try
	{
		assembleur->write(expr_image,expr_nom_fichier);
	}
	catch(ValeurSP valeur)
	{
		InstructionInterruptionSP interruption = *valeur;
		interruption->ajoute_position(_position);
		return interruption;
	}
	return InstructionInterruptionSP();
}

void Write::disp(std::ostream& os) const
{
	Instruction::disp(os);
	format(os);
}

void Write::format(std::ostream& os) const
{
	os << "write ";
	_image->format(os);
	os << " to ";
	_nom_fichier->format(os);
	os << std::endl;
}

void Write::definition_fonction()
{
	_image->definition_fonction();
	_nom_fichier->definition_fonction();
}

InstructionSP Write::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	BlocSP parent = clones_instances[_parent.lock().get()];
	if(not parent)
		parent = _parent.lock();
	WriteSP clone(new Write(_position));
	clones_instances.ajout(this,clone);
	Traceur::Donnees::Expression::ExpressionSP clone_image = _image->clone(clones_instances);
	Traceur::Donnees::Expression::ExpressionSP clone_nom_fichier = _nom_fichier->clone(clones_instances);
	clone->_image = clone_image;
	clone->_nom_fichier = clone_nom_fichier;
	clone->_parent = parent;
	return clone;
}
