#include <src/donnees/instructions/valeur.h>
#include <src/donnees/instructions/instructionreturn.h>
#include <src/donnees/instructions/instructionerror.h>
using namespace Traceur::Donnees::Instruction;

Valeur::Valeur(const Traceur::Donnees::Expression::ExpressionSP& expression)
{
	ReturnSP retour(new Return(expression));
	_interruption=retour;
}

Valeur::operator Traceur::Donnees::Expression::ExpressionSP ()
{
	return _interruption->expression_equivalente();
}

Valeur::operator InstructionInterruptionSP ()
{
	return _interruption;
}

bool Valeur::a_une_valeur_retour() const
{
	return static_cast<bool>(std::dynamic_pointer_cast<Return>(_interruption));
}

bool Valeur::est_une_erreur() const
{
	return static_cast<bool>(std::dynamic_pointer_cast<Error>(_interruption));
}

ValeurSP Valeur::erreur(Traceur::Donnees::Fonction::Identifiant nom)
{
	ErrorSP erreur(new Error(nom));
	ValeurSP valeur(new Valeur());
	valeur->_interruption=erreur;
	return valeur;
}

ValeurSP Valeur::depassement_capacite()
{
	return Valeur::erreur("overflow");
}

ValeurSP Valeur::division_par_zero()
{
	return Valeur::erreur("division_by_zero");
}

ValeurSP Valeur::reste_par_zero()
{
	return Valeur::erreur("remainder_by_zero");
}

ValeurSP Valeur::hors_domaine()
{
	return Valeur::erreur("out_of_range");
}

ValeurSP Valeur::conversion_entiere_invalide()
{
	return Valeur::erreur("invalid_integer_conversion");
}

ValeurSP Valeur::conversion_reelle_invalide()
{
	return Valeur::erreur("invalid_real_conversion");
}

ValeurSP Valeur::format_invalide()
{
	return Valeur::erreur("invalid_format");
}

ValeurSP Valeur::projection_sur_vecteur_nul()
{
	return Valeur::erreur("projection_on_nul_vector");
}

ValeurSP Valeur::symetrie_par_vecteur_nul()
{
	return Valeur::erreur("symetry_by_nul_vector");
}

ValeurSP Valeur::image_mathematique_vide()
{
	return Valeur::erreur("empty_mathematical_image");
}

ValeurSP Valeur::comparaison_hors_image()
{
	return Valeur::erreur("point_comparison_outside_image");
}

ValeurSP Valeur::coordonnee_hors_image()
{
	return Valeur::erreur("coordinate_value_outside_image");
}

ValeurSP Valeur::fichier_incorrect()
{
	return Valeur::erreur("invalid_file");
}

ValeurSP Valeur::fichier_non_png()
{
	return Valeur::erreur("non_png_file");
}

ValeurSP Valeur::erreur_png()
{
	return Valeur::erreur("png_error");
}

ValeurSP Valeur::type_image_incorrect()
{
	return Valeur::erreur("invalid_image_type");
}

ValeurSP Valeur::taille_image_incorrecte()
{
	return Valeur::erreur("invalid_image_size");
}

ValeurSP Valeur::taille_noyau_incorrecte()
{
	return Valeur::erreur("invalid_kernel_size");
}

