#ifndef _DONNEES_INSTRUCTIONS_VALEUR_H_
#define _DONNEES_INSTRUCTIONS_VALEUR_H_

#include <src/global/global.h>
namespace Traceur { namespace Donnees { namespace Instruction { DECL_SHARED_PTR(Valeur); } } }

#include <src/donnees/instructions/instructioninterruption.h>

#define SECURISE_VALEUR(expr) try { return Traceur::Donnees::Instruction::Valeur::expression(expr); } catch(Traceur::Donnees::Instruction::ValeurSP erreur) { return erreur; }

namespace Traceur
{
	namespace Donnees
	{
		namespace Expression { class Image; class Noyau; DECL_SHARED_PTR(Expression); class AppelFonctionUtilisateur; }
		namespace Instruction
		{
			class Valeur
			{
				friend class Traceur::Donnees::Expression::AppelFonctionUtilisateur;
				friend class Traceur::Donnees::Expression::Image;
				friend class Traceur::Donnees::Expression::Noyau;
				// Valeur est utilisee par quasiment toutes les classes, donc on place son interface dans le domaine public
				public:
					Valeur() = default;
					// les constructeurs prendront les instructions des valeurs passees en parametre et les ajouteront a la liste
					// les expressions dans les valeurs passees en parametres doivent etre ignorees
					explicit Valeur(const Traceur::Donnees::Expression::ExpressionSP& expression);
					bool a_une_valeur_retour() const;
					bool est_une_erreur() const;
					operator Traceur::Donnees::Expression::ExpressionSP ();
					operator InstructionInterruptionSP ();
					static ValeurSP expression(Traceur::Donnees::Expression::ExpressionSP expression)
					{
						ValeurSP valeur(new Valeur(expression));
						return valeur;
					}
					static ValeurSP erreur(Traceur::Donnees::Fonction::Identifiant nom);
					static ValeurSP depassement_capacite();
					static ValeurSP division_par_zero();
					static ValeurSP reste_par_zero();
					static ValeurSP hors_domaine();
					static ValeurSP conversion_entiere_invalide();
					static ValeurSP conversion_reelle_invalide();
					static ValeurSP format_invalide();
					static ValeurSP projection_sur_vecteur_nul();
					static ValeurSP symetrie_par_vecteur_nul();
					static ValeurSP image_mathematique_vide();
					static ValeurSP comparaison_hors_image();
					static ValeurSP coordonnee_hors_image();
					static ValeurSP fichier_incorrect();
					static ValeurSP fichier_non_png();
					static ValeurSP erreur_png();
					static ValeurSP type_image_incorrect();
					static ValeurSP taille_image_incorrecte();
					static ValeurSP taille_noyau_incorrecte();
					~Valeur() {};
				private:
					InstructionInterruptionSP _interruption;

			};
		}
	}
}
#endif
