#ifndef _DONNEES_TYPES_COMPATIBILITE_H_
#define _DONNEES_TYPES_COMPATIBILITE_H_

#include <iostream>
#include <src/global/global.h>

namespace Traceur { namespace Donnees { namespace Type { DECL_SHARED_PTR(Compatibilite); } } }

namespace Traceur
{
	namespace Donnees
	{
		namespace Type
		{
			class Compatibilite
			{
				public:
					Compatibilite()
						:_compatible(false), _transtypage_retour(0), _transtypage_parametres(0) {};
					~Compatibilite() {};
					operator bool () const
					{
						return _compatible;
					};
					size_t transtypage_retour() const
					{
						return _transtypage_retour;
					};
					size_t transtypage_parametres() const
					{
						return _transtypage_parametres;
					};
					// ajoute les transtypages de parametres
					Compatibilite& operator>>(Compatibilite compatibilite)
					{
						TRACEUR_TEST(compatibilite);
						compatible();
						_transtypage_parametres += compatibilite._transtypage_retour+compatibilite._transtypage_parametres;
						return *this;
					};
					// ajoute les transtypages de retour
					Compatibilite& operator<<(Compatibilite compatibilite)
					{
						TRACEUR_TEST(compatibilite);
						compatible();
						_transtypage_retour += compatibilite._transtypage_retour+compatibilite._transtypage_parametres;
						return *this;
					};
					bool operator< (Compatibilite compatibilite)
					{
						TRACEUR_TEST((*this) or compatibilite);
						return (_transtypage_retour+_transtypage_parametres)<(compatibilite._transtypage_retour+compatibilite._transtypage_parametres);
					};
					void format(std::ostream& os) const
					{
						if(not _compatible)
						{
							os << "Non compatibles" << std::endl;
							return;
						}
						os << "Compatibles: " << _transtypage_retour << "/" << _transtypage_parametres << std::endl;
					};
					friend std::ostream& operator<<(std::ostream& flux, const Compatibilite compatibilite);

				private:
					bool _compatible;
					size_t _transtypage_retour;
					size_t _transtypage_parametres;
				protected:
					void compatible()
					{
						_compatible = true;
					};
					void transtypage_retour()
					{
						_transtypage_retour++;
					};
					
			};
			class Compatible : public Compatibilite
			{
				public:
					enum class Transtypage { SANS, AVEC } ;
					Compatible(Transtypage transtypage)
					{
						compatible();
						if(transtypage==Transtypage::AVEC)
							transtypage_retour();
					}
					~Compatible() {};
			};
			class NonCompatible : public Compatibilite
			{
				public:
					NonCompatible() {};
					~NonCompatible() {};
			};
		}
	}
}

#endif
