#ifndef _DONNEES_TYPES_TYPE_H_
#define _DONNEES_TYPES_TYPE_H_

#include <iostream>
#include <sstream>
#include <string>

#include <src/global/global.h>

namespace Traceur { namespace Donnees { namespace Type { DECL_SHARED_PTR(Type); } } }

#include <src/donnees/global/clonesinstances.h>
#include <src/donnees/types/compatibilite.h>

namespace Traceur
{
	namespace Donnees
	{
		namespace Type
		{
			class Type : public std::enable_shared_from_this<Type>
			{
				public:
					Type() = default;
					virtual ~Type(){}; // mettre = default provoque un memory leak ici
					virtual void format(std::ostream& os) const = 0;
					operator std::string() const
					{
						std::ostringstream oss;
						format(oss);
						return oss.str();
					}
					virtual bool operator==(const Type&) const = 0;
					// renvoie true si type est compatible avec this
					// donc, si type est une specialisation de this dans le cas d'un type atomique
					// ou une generalisation dans le cas d'un type fonction
					virtual bool compatible(const TypeSP& type) = 0;
					// renvoie true si type est compatible avec this. Renvoie aussi la distance entre les deux types.
					// la compatibilite non exacte implique le lien special des parametres
					virtual Compatibilite compatible_prototype(const TypeSP& type, const bool compatibilite_exacte = false)= 0;
					friend std::ostream& operator<<(std::ostream& os, const Type& type);
					virtual bool operator<(const Type&) const = 0;
					TypeSP clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
					{
						return this->shared_from_this();
					};
					virtual bool atomique() const { return false; };
				private:
			};
		}
	}
}

#endif
