#include <functional>
#include <algorithm>
#include <src/donnees/types/typefonction.h>
#include <src/donnees/types/typeatomique.h>
#include <src/donnees/fonctions/fonctionprototype.h>

using namespace Traceur::Donnees::Type;

	Fonction::Fonction(TypeSP type_retour, ListeParametres parametres_gauche, Traceur::Donnees::Fonction::Identifiant nom_fonction, ListeParametres parametres_droite)
:_type_retour(type_retour),_parametres_gauche(parametres_gauche),_nom_fonction(nom_fonction),_parametres_droite(parametres_droite)
{
}

namespace Traceur
{
	namespace Donnees
	{
		namespace Type
		{
			class SPconversionDePrototypeVersType
			{
				public:
					TypeSP operator() (const Traceur::Donnees::Fonction::PrototypeSP& prototype)
					{
						if(prototype->_parametres_gauche.size()==0 and prototype->_parametres_droite.size()==0 and prototype->_type_retour->atomique())
						{
							// Integer fonction(Integer) et pas Integer fonction([Integer parametre])
							return prototype->_type_retour;
						}
						else
						{
							TypeSP type(new Fonction(prototype));
							return type;
						}
					}
			};
		}
	}
}


	Fonction::Fonction(const Traceur::Donnees::Fonction::PrototypeSP& prototype)
:_type_retour(prototype->_type_retour),_nom_fonction(prototype->_nom_fonction)
{
	std::transform(prototype->_parametres_gauche.cbegin(),prototype->_parametres_gauche.cend(),
			std::back_inserter<ListeParametres>(_parametres_gauche),SPconversionDePrototypeVersType());
	std::transform(prototype->_parametres_droite.cbegin(),prototype->_parametres_droite.cend(),
			std::back_inserter<ListeParametres>(_parametres_droite),SPconversionDePrototypeVersType());
}

struct CompatibleType
{
	bool operator() (const TypeSP& t1, const TypeSP& t2)
	{
		return t2->compatible(t1);
	};
};

bool Fonction::compatible(const TypeSP& type)
{
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE << "Compatibilite de types demandee" << std::endl;
#endif
	return type->compatible_prototype(this->shared_from_this(),true);
}

namespace Traceur
{
	namespace Donnees
	{
		namespace Type
		{
struct CompatiblePrototypeType
{
	CompatiblePrototypeType(Compatibilite& compatibilite, const bool compatibilite_exacte)
		:_compatibilite(compatibilite),_compatibilite_exacte(compatibilite_exacte) {};
	bool operator() (const TypeSP& t1, const TypeSP& t2)
	{
		Compatibilite compatibilite = t1->compatible_prototype(t2,_compatibilite_exacte);
		if(compatibilite)
		{
			_compatibilite >> compatibilite;
			return true;
		}
		FonctionSP fonction = std::dynamic_pointer_cast<Fonction>(t1);
#ifdef TRACEUR_DEBUG
		TRACEUR_TRACE << "Non compatibles avec test normal" << std::endl;
#endif
		if((not fonction) or _compatibilite_exacte)
		{
#ifdef TRACEUR_DEBUG
		TRACEUR_TRACE << "Non fonction ou compatibilite exacte demandee" << std::endl;
#endif
			return false;
		}	
#ifdef TRACEUR_DEBUG
		TRACEUR_TRACE << "Test avec fonction - test par lien special" << std::endl;
#endif
		compatibilite = fonction->_type_retour->compatible_prototype(t2,_compatibilite_exacte);
		if(compatibilite)
		{
			_compatibilite >> compatibilite;
			return true;
		}
#ifdef TRACEUR_DEBUG
		TRACEUR_TRACE << "Test lien special faux, prototypes non compatibles" << std::endl;
#endif
		return false;
	};
	private:
		Compatibilite& _compatibilite;
		bool _compatibilite_exacte;
};
		}
	}
}

Compatibilite Fonction::compatible_prototype(const TypeSP& type, const bool compatibilite_exacte)
{
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE << "Compatibilite type fonction de " << type << " envers " << this->shared_from_this() << std::endl;
#endif
	Compatibilite compatibilite;

	FonctionSP fonction = std::dynamic_pointer_cast<Fonction>(type);
	if(not fonction)
		return NonCompatible();

	Compatibilite compatibilite_retour = _type_retour->compatible_prototype(fonction->_type_retour);
	if(not compatibilite_retour)
		return NonCompatible();
	compatibilite << compatibilite_retour;
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE << compatibilite;
#endif

	if(_parametres_gauche.size()!=fonction->_parametres_gauche.size())
		return NonCompatible();
	if(not std::equal(fonction->_parametres_gauche.cbegin(),fonction->_parametres_gauche.cend(),_parametres_gauche.cbegin(),CompatiblePrototypeType(compatibilite,compatibilite_exacte)))
		return NonCompatible();

	if(_parametres_droite.size()!=fonction->_parametres_droite.size())
		return NonCompatible();
	if(not std::equal(fonction->_parametres_droite.cbegin(),fonction->_parametres_droite.cend(),_parametres_droite.cbegin(),CompatiblePrototypeType(compatibilite,compatibilite_exacte)))
		return NonCompatible();
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE << compatibilite;
#endif
	return compatibilite;
}

void Fonction::format(std::ostream& os) const
{
	os << "[" ;
	_type_retour->format(os);
	os << " " ;
	format(_parametres_gauche,os);
	os << _nom_fonction ;
	format(_parametres_droite,os);
	os << "]" ;
}

void separateur(const TypeSP typeSP, std::ostream& os)
{
	os << ", " ;
	typeSP->format(os);
}

bool Fonction::operator==(const Type& type) const
{
	const Fonction *fonction = dynamic_cast<const Fonction*>(&type);
	if(not fonction)
		return false;

	if(not Traceur::Global::SPegalSP<Traceur::Donnees::Type::Type>()(fonction->_type_retour,_type_retour))
		return false;

	if(_parametres_gauche.size()!=fonction->_parametres_gauche.size())
		return false;

	if(_parametres_droite.size()!=fonction->_parametres_droite.size())
		return false;
	
	if(not std::equal(fonction->_parametres_gauche.cbegin(),fonction->_parametres_gauche.cend(),_parametres_gauche.cbegin(),Traceur::Global::SPegalSP<Type>()))
		return false;                                   
                                                                          
	if(not std::equal(fonction->_parametres_droite.cbegin(),fonction->_parametres_droite.cend(),_parametres_droite.cbegin(),Traceur::Global::SPegalSP<Type>()))
		return false;

	return true;
}

bool Fonction::operator<(const Type& type) const
{
	const Fonction *fonction = dynamic_cast<const Fonction*>(&type);
	if(not fonction)
		return false;

	if(Traceur::Global::SPinferieurSP<Type>()(_type_retour,fonction->_type_retour))
		return true;
	if(Traceur::Global::SPinferieurSP<Type>()(fonction->_type_retour,_type_retour))
		return false;

	if(std::lexicographical_compare(_parametres_gauche.cbegin(),_parametres_gauche.cend(),
			fonction->_parametres_gauche.cbegin(),fonction->_parametres_gauche.cend(),
			Traceur::Global::SPinferieurSP<Type>()))
		return true;
	if(std::lexicographical_compare(fonction->_parametres_gauche.cbegin(),fonction->_parametres_gauche.cend(),
			_parametres_gauche.cbegin(),_parametres_gauche.cend(),
			Traceur::Global::SPinferieurSP<Type>()))
		return false;

	if(std::lexicographical_compare(_parametres_droite.cbegin(),_parametres_droite.cend(),
			fonction->_parametres_droite.cbegin(),fonction->_parametres_droite.cend(),
			Traceur::Global::SPinferieurSP<Type>()))
		return true;
	if(std::lexicographical_compare(fonction->_parametres_droite.cbegin(),fonction->_parametres_droite.cend(),
			_parametres_droite.cbegin(),_parametres_droite.cend(),
			Traceur::Global::SPinferieurSP<Type>()))
		return false;

	return false;
}

void Fonction::format(ListeParametres liste_parametres, std::ostream& os) const
{
	if(!liste_parametres.empty())
	{
		os << "(" ;
		auto iterateur = liste_parametres.cbegin();
		(*iterateur)->format(os);
		++iterateur;
		std::for_each(iterateur,liste_parametres.cend(),bind(separateur,std::placeholders::_1,std::ref(os)));
		os << ")" ;
	}
}
