#include <src/png/imagepng.h>
#include <cstdio>
#include <unistd.h>
#include <sys/stat.h>
#include <src/donnees/instructions/valeur.h>
#define PNG_STDIO_SUPPORTED
typedef FILE* png_FILE_p;
#include <png.h>
using namespace Traceur::Png;

DECL_SHARED_PTR(PngBitmap);
struct PngBitmap
{
	PngBitmap(Traceur::Donnees::Assembleur::Coordonnee lignes, Traceur::Donnees::Assembleur::Coordonnee colonnes)
		:lignes(lignes), colonnes(colonnes)
	{
		bitmap = new png_bytep[lignes];
		for(png_int_32 indice=0 ; indice<lignes ; ++indice)
		{
			bitmap[indice] = new png_byte[4*colonnes];
		}
	};

	~PngBitmap()
	{
		for(png_int_32 indice=0 ; indice<lignes ; ++indice)
		{
			delete [] bitmap[indice];
		}
		delete [] bitmap;
	};
	png_int_32 lignes;
	png_int_32 colonnes;
	png_bytepp bitmap;
};



struct PngPtrs
{
	PngPtrs() :fichier(0),write_ptr(0),read_ptr(0),info_ptr(0),bitmap() {};
	~PngPtrs()
	{
		if(fichier and (fichier!=stdout))
			::fclose(fichier);
		if(write_ptr)
		{
			if(info_ptr)
				::png_destroy_write_struct(&write_ptr, &info_ptr);
			else
				::png_destroy_write_struct(&write_ptr, NULL);
		}
		if(read_ptr)
		{
			if(info_ptr)
				::png_destroy_read_struct(&read_ptr, &info_ptr,NULL);
			else
				::png_destroy_read_struct(&read_ptr, NULL, NULL);
		}
	}
	png_FILE_p fichier;
	png_structp write_ptr;
	png_structp read_ptr;
	png_infop info_ptr;
	PngBitmapSP bitmap;
	static void erreur(png_structp png, const char *mesg)
	{
#ifdef TRACEUR_DEBUG
		std::cerr << "erreur PNG: " << mesg << std::endl;
#endif
		throw Traceur::Donnees::Instruction::Valeur::erreur_png();
	}
	static void attention(png_structp png, const char *mesg)
	{
#ifdef TRACEUR_DEBUG
		std::cerr << "attention PNG: " << mesg << std::endl;
#endif
	}
};

void Image::ecriture(Traceur::Donnees::Assembleur::Image::ReelleSP image, Traceur::Donnees::Assembleur::NomFichier nom_fichier)
{
	PngPtrs pngptrs;
	Traceur::Donnees::Assembleur::Coordonnee lignes = image->externe_lignes();
	Traceur::Donnees::Assembleur::Coordonnee colonnes = image->externe_colonnes();

#ifdef TRACEUR_DEBUG
	std::cerr << lignes << "x" << colonnes << std::endl;
#endif
	if(nom_fichier.est_le_terminal())
	{
		pngptrs.fichier=stdout;
	}
	else
	{
		pngptrs.fichier=::fopen(nom_fichier,"wb");
	}
	if(not pngptrs.fichier)
		throw Traceur::Donnees::Instruction::Valeur::fichier_incorrect();
	pngptrs.write_ptr = ::png_create_write_struct(PNG_LIBPNG_VER_STRING,NULL,PngPtrs::erreur,PngPtrs::attention);
	if(not pngptrs.write_ptr)
		throw Traceur::Donnees::Instruction::Valeur::erreur_png();
	pngptrs.info_ptr = ::png_create_info_struct(pngptrs.write_ptr);
	if(not pngptrs.info_ptr)
		throw Traceur::Donnees::Instruction::Valeur::erreur_png();

	::png_init_io(pngptrs.write_ptr,pngptrs.fichier);

	::png_set_IHDR(pngptrs.write_ptr,pngptrs.info_ptr,colonnes,lignes,8,PNG_COLOR_TYPE_RGB_ALPHA,PNG_INTERLACE_ADAM7,PNG_COMPRESSION_TYPE_DEFAULT,PNG_FILTER_TYPE_DEFAULT);
	::png_set_gAMA(pngptrs.write_ptr,pngptrs.info_ptr,1.);
	static std::string commentaire("Image generated by traceur");
	static std::string clef("Source");
	png_text texte;
	texte.compression = PNG_TEXT_COMPRESSION_zTXt;
	texte.key = const_cast<char*>(clef.c_str());
	texte.text = const_cast<char*>(commentaire.c_str());
	texte.text_length = commentaire.size();

	::png_set_text(pngptrs.write_ptr,pngptrs.info_ptr,&texte,1);

	::png_set_user_limits(pngptrs.write_ptr,0x7FFFFFFF,0x7FFFFFFF);

	pngptrs.bitmap = PngBitmapSP(new PngBitmap(lignes,colonnes));

	for(Traceur::Donnees::Assembleur::Coordonnee ligne=0 ; ligne<lignes ; ++ligne)
	{
		for(Traceur::Donnees::Assembleur::Coordonnee colonne=0 ; colonne<colonnes ; ++colonne)
		{
#ifdef TRACEUR_DEBUG
			std::cerr << "(" << ligne << "," << colonne << ")";
#endif
			Traceur::Donnees::Assembleur::Couleur::Reelle couleur = image->externe_pixel(ligne, colonne);
			pngptrs.bitmap->bitmap[ligne][4*colonne  ]=couleur._rouge;
			pngptrs.bitmap->bitmap[ligne][4*colonne+1]=couleur._vert;
			pngptrs.bitmap->bitmap[ligne][4*colonne+2]=couleur._bleu;
			pngptrs.bitmap->bitmap[ligne][4*colonne+3]=couleur._alpha;
		}
#ifdef TRACEUR_DEBUG
		std::cerr << std::endl;
#endif
	}

	::png_set_rows(pngptrs.write_ptr,pngptrs.info_ptr,pngptrs.bitmap->bitmap);
	::png_write_png(pngptrs.write_ptr,pngptrs.info_ptr,PNG_TRANSFORM_IDENTITY,NULL);
}

void Image::lecture(Traceur::Donnees::Assembleur::Image::Reelle& image, Traceur::Donnees::Assembleur::NomFichier nom_fichier)
{
	PngPtrs pngptrs;
	Traceur::Donnees::Assembleur::Coordonnee lignes = image.externe_lignes();
	Traceur::Donnees::Assembleur::Coordonnee colonnes = image.externe_colonnes();

	pngptrs.fichier=::fopen(nom_fichier,"rb");
	if(not pngptrs.fichier)
		throw Traceur::Donnees::Instruction::Valeur::fichier_incorrect();
	pngptrs.read_ptr = ::png_create_read_struct(PNG_LIBPNG_VER_STRING,NULL,PngPtrs::erreur,PngPtrs::attention);
	if(not pngptrs.read_ptr)
		throw Traceur::Donnees::Instruction::Valeur::erreur_png();
	pngptrs.info_ptr = ::png_create_info_struct(pngptrs.read_ptr);
	if(not pngptrs.info_ptr)
		throw Traceur::Donnees::Instruction::Valeur::erreur_png();

	::png_init_io(pngptrs.read_ptr,pngptrs.fichier);

	::png_set_user_limits(pngptrs.read_ptr,0x7FFFFFFF,0x7FFFFFFF);

//	::png_read_png(pngptrs.read_ptr,pngptrs.info_ptr,PNG_TRANSFORM_STRIP_16|PNG_TRANSFORM_GRAY_TO_RGB,NULL);
	::png_read_info(pngptrs.read_ptr,pngptrs.info_ptr);

	png_byte color_type = ::png_get_color_type(pngptrs.read_ptr,pngptrs.info_ptr);
	png_byte bit_depth = ::png_get_bit_depth(pngptrs.read_ptr,pngptrs.info_ptr);

	if(color_type==PNG_COLOR_TYPE_PALETTE)
		::png_set_palette_to_rgb(pngptrs.read_ptr);

	if (color_type == PNG_COLOR_TYPE_GRAY and bit_depth < 8)
		::png_set_expand_gray_1_2_4_to_8(pngptrs.read_ptr);

	if (color_type == PNG_COLOR_TYPE_GRAY or color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
		::png_set_gray_to_rgb(pngptrs.read_ptr);

	if (png_get_valid(pngptrs.read_ptr, pngptrs.info_ptr, PNG_INFO_tRNS))
		::png_set_tRNS_to_alpha(pngptrs.read_ptr);                                                                                                                        
	
	if(bit_depth>8)
		::png_set_strip_16(pngptrs.read_ptr);

	if(not (color_type&PNG_COLOR_MASK_ALPHA))
		::png_set_add_alpha(pngptrs.read_ptr,255,PNG_FILLER_AFTER);

	pngptrs.bitmap = PngBitmapSP(new PngBitmap(::png_get_image_height(pngptrs.read_ptr,pngptrs.info_ptr),::png_get_image_width(pngptrs.read_ptr,pngptrs.info_ptr)));
	::png_read_image(pngptrs.read_ptr,pngptrs.bitmap->bitmap);

	::png_read_end(pngptrs.read_ptr,NULL);

	png_int_32 png_lignes = ::png_get_image_height(pngptrs.read_ptr,pngptrs.info_ptr);
	png_int_32 png_colonnes = ::png_get_image_width(pngptrs.read_ptr,pngptrs.info_ptr);
	if(png_lignes<lignes) lignes = png_lignes;
	if(png_colonnes<colonnes) colonnes = png_colonnes;

	for(Traceur::Donnees::Assembleur::Coordonnee ligne=0 ; ligne<lignes ; ++ligne)
	{
		for(Traceur::Donnees::Assembleur::Coordonnee colonne=0 ; colonne<colonnes ; ++colonne)
		{
			png_byte rouge = pngptrs.bitmap->bitmap[ligne][4*colonne  ];
			png_byte vert  = pngptrs.bitmap->bitmap[ligne][4*colonne+1];
			png_byte bleu  = pngptrs.bitmap->bitmap[ligne][4*colonne+2];
			png_byte alpha = pngptrs.bitmap->bitmap[ligne][4*colonne+3];
			Traceur::Donnees::Assembleur::Couleur::Reelle couleur(rouge,vert,bleu,alpha);
			image.externe_trace(ligne, colonne, couleur);
		}
	}
}

void Image::teste_fichier_ecriture(Traceur::Donnees::Assembleur::NomFichier nom_fichier)
{
	teste_fichier(nom_fichier,W_OK);
}

void Image::teste_fichier_lecture(Traceur::Donnees::Assembleur::NomFichier nom_fichier)
{
	teste_fichier(nom_fichier,R_OK);
	PngPtrs pngptrs;

	pngptrs.fichier=::fopen(nom_fichier,"rb");
	if(not pngptrs.fichier)
		throw Traceur::Donnees::Instruction::Valeur::fichier_incorrect();

//	pngptrs.read_ptr = ::png_create_read_struct(PNG_LIBPNG_VER_STRING,NULL,PngPtrs::erreur,PngPtrs::attention);
//	if(not pngptrs.read_ptr)
//		throw Traceur::Donnees::Instruction::Valeur::erreur_png();
//	pngptrs.info_ptr = ::png_create_info_struct(pngptrs.read_ptr);
//	if(not pngptrs.info_ptr)
//		throw Traceur::Donnees::Instruction::Valeur::erreur_png();

	png_byte buffer[16];
	if(::fread(buffer,1,8,pngptrs.fichier)!=8)
		throw Traceur::Donnees::Instruction::Valeur::fichier_non_png();
	if(::png_sig_cmp(buffer,0,8))
		throw Traceur::Donnees::Instruction::Valeur::fichier_non_png();
//	::png_set_sig_bytes(pngptrs.read_ptr,8);
//
//	::png_init_io(pngptrs.read_ptr,pngptrs.fichier);
//
//	::png_set_user_limits(pngptrs.read_ptr,0x7FFFFFFF,0x7FFFFFFF);

}

void Image::teste_fichier(Traceur::Donnees::Assembleur::NomFichier nom_fichier, int mode)
{
	if(::access(nom_fichier,mode))
	{
		if((mode|W_OK) and (errno==ENOENT))
			return;
		throw Traceur::Donnees::Instruction::Valeur::fichier_incorrect();
	}
	struct stat buf;
	if(::stat(nom_fichier,&buf))
		throw Traceur::Donnees::Instruction::Valeur::fichier_incorrect();
	if(S_ISDIR(buf.st_mode))
		throw Traceur::Donnees::Instruction::Valeur::fichier_incorrect();
}
