#include <cstdio>
#include <iostream>
#include <set>

#include <src/global/global.h>
#include <src/donnees/types/types.h>
#include <src/donnees/fonctions/fonctionprototype.h>
#include <src/donnees/expressions/expressions.h>
#include <src/donnees/expressions/expressionetlogique.h>
#include <src/donnees/expressions/expressionoulogique.h>
#include <src/donnees/expressions/expressionouexclusiflogique.h>
#include <src/donnees/expressions/expressionnonlogique.h>
#include <src/donnees/instructions/instructiondisp.h>
#include <src/donnees/instructions/instructionerror.h>
#include <src/donnees/assembleur/assembleur.h>
#include <src/traceur/lignedecommande.h>
#include <src/png/imagepng.h>

using namespace Traceur::Donnees;

int main(int argc, char *argv[])
{
	Traceur::LigneDeCommande lignedecommande(argc,argv);
	try
	{
#ifdef TRACEUR_DEBUG_ETAPE
		std::cout << "--- Compilation -----------------------------------------------------------------" << std::endl;
#endif
		Instruction::BlocSP bloc;
		bloc = Traceur::Analyseur::Analyseur::analyse(lignedecommande);
#ifdef TRACEUR_DEBUG
		bloc->format(std::cout);
#endif
#ifdef TRACEUR_DEBUG_ETAPE
		std::cout << "--- Edition de liens ------------------------------------------------------------" << std::endl;
#endif
		bloc->edition_liens();
#ifdef TRACEUR_DEBUG
		bloc->format(std::cout);
#endif
#ifdef TRACEUR_DEBUG_ETAPE
		std::cout << "--- Optimisation ----------------------------------------------------------------" << std::endl;
#endif
		bloc->simplification();
#ifdef TRACEUR_DEBUG_ETAPE
		std::cout << "--- Evaluation ------------------------------------------------------------------" << std::endl;
#endif
		Traceur::Donnees::Assembleur::AssembleurSP assembleur(new Traceur::Donnees::Assembleur::Assembleur());
		Traceur::Donnees::Instruction::InstructionInterruptionSP interruption = bloc->evaluation(assembleur);
		if(std::dynamic_pointer_cast<Instruction::Error>(interruption))
		{
			Instruction::ErrorSP error = std::dynamic_pointer_cast<Instruction::Error>(interruption);
			throw error->exception();
		}
#ifdef TRACEUR_DEBUG_ETAPE
		std::cout << "--- Etat apres evaluation -------------------------------------------------------" << std::endl;
#endif
#ifdef TRACEUR_DEBUG
		bloc->format(std::cout);
#endif
#ifdef TRACEUR_DEBUG_ETAPE
		std::cout << "--- Execution -------------------------------------------------------------------" << std::endl;
#endif
		try
		{
			assembleur->execute();
		}
		catch(Traceur::Donnees::Instruction::ValeurSP valeur)
		{
			Traceur::Donnees::Instruction::InstructionInterruptionSP interruption = *valeur;
			Instruction::ErrorSP error = std::dynamic_pointer_cast<Instruction::Error>(interruption);
			throw error->exception();
		}
#ifdef TRACEUR_DEBUG_ETAPE
		std::cout << "--- Sortie ----------------------------------------------------------------------" << std::endl;
#endif
	}
	catch(Traceur::Exception::Generique exception)
	{
		exception.format(std::cerr);
		return 1;
	}
	return 0;
}
